/*!
 * @file fs_worker.h
 * @brief This is the header file for the Fastec Filesystem worker thread.
 *
 * Copyright 2013-2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */
#ifndef FS_WORKER_H_
#define FS_WORKER_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>

#include <map>
#include <memory>
#include <mutex>
#include <string>
#include <vector>

// Windows or Linux?
#if _WIN32
//
// Windows
#include "fas_thread.h"
#else
//
// Linux
#include <fas_thread.h>
#endif
#include "fas_filesystem.h"

namespace FAS_FILESYSTEM
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Our cache entry structure.
typedef struct FS_CACHE_ENTRY
{
    uint64_t    drive_size;
    PARTITION_INFO_VEC  part_info;
} FS_CACHE_ENTRY;

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

//! Message class for FS_ANALYZE message.
class fsAnalyzeMessage: public FAS_THREAD::SyncMessage
{
public:
    fsAnalyzeMessage();
    virtual ~fsAnalyzeMessage();

    std::string dev_path;               // [IN] The device we're interested in.
    bool is_fastec;                     // [OUT] True if dev_path is a Fastec filesystem, false otherwise.
};

//! Message class for FS_APPEND_CAPTURE message.
class fsAppendCaptureMessage: public FAS_THREAD::SyncMessage
{
public:
    fsAppendCaptureMessage();
    virtual ~fsAppendCaptureMessage();

    std::string dev_path;               // [IN] The device we're interested in.
    uint64_t part_lba;                  // [IN] The LBA of the partition we're interested in.
    LAYOUT_CAPTURE_VEC layout;          // [IN] The vector of capture layout information.
};

//! Message class for FS_APPEND_PARTITION message.
class fsAppendPartitionMessage: public FAS_THREAD::SyncMessage
{
public:
    fsAppendPartitionMessage();
    virtual ~fsAppendPartitionMessage();

    std::string dev_path;               // [IN] The device we're interested in.
    LAYOUT_PARTITION_VEC layout;        // [IN] The vector of partition layout information.
};

//! Message class for FS_CLEAR_CACHE message.
class fsClearCacheMessage: public FAS_THREAD::SyncMessage
{
public:
    fsClearCacheMessage();
    virtual ~fsClearCacheMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
};

//! Message class for FS_DELETE_CAPTURE message.
class fsDeleteCaptureMessage: public FAS_THREAD::SyncMessage
{
public:
    fsDeleteCaptureMessage();
    virtual ~fsDeleteCaptureMessage();

    std::string dev_path;               // [IN] The device we're interested in.
    uint64_t part_lba;                  // [IN] The LBA of the partition we're interested in.
    uint64_t capture_lba;               // [IN] The address of the capture header we want to remove.
};

//! Message class for FS_DELETE_PARTITION message.
class fsDeletePartitionMessage: public FAS_THREAD::SyncMessage
{
public:
    fsDeletePartitionMessage();
    virtual ~fsDeletePartitionMessage();

    std::string dev_path;               // [IN] The device we're interested in.
    uint64_t part_lba;                  // [IN] The LBA of the partition we're interested in.
};

//! Message class for FS_GET_BOOKMARK message.
class fsGetBookmarkMessage : public FAS_THREAD::SyncMessage
{
public:
    fsGetBookmarkMessage();
    virtual ~fsGetBookmarkMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    uint64_t capture_lba;               // [IN] The address of the capture header we want the bookmark information from.
    std::vector<uint8_t> bk_buffer;     // [OUT] The bookmark(s) from the drive.
};

//! Message class for FS_GET_CORRECTION_TABLE message.
class fsGetCorrectionTableMessage : public FAS_THREAD::SyncMessage
{
public:
    fsGetCorrectionTableMessage();
    virtual ~fsGetCorrectionTableMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    uint64_t capture_lba;               // [IN] The address of the capture header we want the correction table from.
    std::vector<uint8_t> ct_buffer;     // [OUT] The correction table from the drive.
};

//! Message class for FS_GET_GAIN_TABLE message.
class fsGetGainTableMessage : public FAS_THREAD::SyncMessage
{
public:
    fsGetGainTableMessage();
    virtual ~fsGetGainTableMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    uint64_t capture_lba;               // [IN] The address of the capture header we want the gain table from.
    std::vector<uint8_t> gt_buffer;     // [OUT] The gain table from the drive.
};

//! Message class for FS_GET_INFO message.
class fsGetInfoMessage : public FAS_THREAD::SyncMessage
{
public:
    fsGetInfoMessage();
    virtual ~fsGetInfoMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    std::string name;                   // [OUT] The name of the device.
    PARTITION_INFO_VEC part_info;       // [OUT] The partition information we have for the device.
    uint64_t assigned_space;            // [OUT] The number of sectors assigned for the device.
};

//! Message class for FS_GET_PM_METADATA message.
class fsGetPmMetadataMessage : public FAS_THREAD::SyncMessage
{
public:
    fsGetPmMetadataMessage();
    virtual ~fsGetPmMetadataMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    uint64_t capture_lba;               // [IN] The address of the capture header we want the pm_metadata from.
    std::vector<uint8_t> metadata;      // [OUT] The metadata from the drive.
};

//! Message class for FS_LAYOUT message.
class fsLayoutMessage : public FAS_THREAD::SyncMessage
{
public:
    fsLayoutMessage();
    virtual ~fsLayoutMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    std::string name;                   // [IN] The name for the device.
    uint64_t num_sectors;               // [IN] Total number of sectors to use.
    LAYOUT_PARTITION_VEC layout;        // [IN] The layout information to apply to the drive.
};

//! Message class for FS_RENAME message.
class fsRenameMessage : public FAS_THREAD::SyncMessage
{
public:
    fsRenameMessage();
    virtual ~fsRenameMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    std::string name;                   // [IN] The name for the device.
};

//! Message class for FS_RESIZE_CAPTURE message.
class fsResizeCaptureMessage : public FAS_THREAD::SyncMessage
{
public:
    fsResizeCaptureMessage();
    virtual ~fsResizeCaptureMessage();

    std::string dev_path;               // [IN] The device we're interested in.
    uint64_t cap_lba;                   // [IN] The LBA of the capture we're interested in.
    uint64_t resize_amount;             // [IN] The amount to resize.
};

//! Message class for FS_RESIZE_PARTITION message.
class fsResizePartitionMessage : public FAS_THREAD::SyncMessage
{
public:
    fsResizePartitionMessage();
    virtual ~fsResizePartitionMessage();

    std::string dev_path;               // [IN] The device we're interested in.
    uint64_t part_lba;                  // [IN] The LBA of the partition we're interested in.
    uint64_t resize_amount;             // [IN] The amount to resize.
};

//! Message class for FS_SET_BOOKMARK message.
class fsSetBookmarkMessage : public FAS_THREAD::SyncMessage
{
public:
    fsSetBookmarkMessage();
    virtual ~fsSetBookmarkMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    uint64_t capture_lba;               // [IN] The address of the capture header we want the bookmark information to go to.
    std::vector<uint8_t> bk_buffer;     // [IN] The bookmark(s) from the drive.
};

//! Message class for FS_SET_CORRECTION_INFO message.
class fsSetCorrectionInfoMessage : public FAS_THREAD::SyncMessage
{
public:
    fsSetCorrectionInfoMessage();
    virtual ~fsSetCorrectionInfoMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    uint64_t capture_lba;               // [IN] The address of the capture header we want the correction table to go to.

    std::vector<uint8_t> ct_buffer;     // [IN] The correction table from the drive.
    std::vector<uint8_t> gt_buffer;     // [IN] The gain table from the drive.
};

//! Message class for FS_SET_PM_METADATA message.
class fsSetPmMetadataMessage : public FAS_THREAD::SyncMessage
{
public:
    fsSetPmMetadataMessage();
    virtual ~fsSetPmMetadataMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    uint64_t capture_lba;               // [IN] The address of the capture header we want the pm_metadata from.
    std::vector<uint8_t> metadata;      // [IN] The metadata from the drive.
};

//! Message class for FS_TRY_LAYOUT message.
class fsTryLayoutMessage : public FAS_THREAD::SyncMessage
{
public:
    fsTryLayoutMessage();
    virtual ~fsTryLayoutMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    uint64_t num_sectors;               // [IN] Total number of sectors to use.
    LAYOUT_PARTITION_VEC layout;        // [IN] The layout information to apply to the drive.
    PARTITION_INFO_VEC part_info;       // [OUT] The partition information we calculate from the layout information.
};

//! Message class for FS_UPDATE_CAPTURE message.
class fsUpdateCaptureMessage : public FAS_THREAD::SyncMessage
{
public:
    fsUpdateCaptureMessage();
    virtual ~fsUpdateCaptureMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    CAPTURE_DATA capture_info;          // [IN] The capture information we want to update.
};

//! Message class for FS_WRITE_PARTITION message.
class fsWritePartitionMessage : public FAS_THREAD::SyncMessage
{
public:
    fsWritePartitionMessage();
    virtual ~fsWritePartitionMessage();

    std::string dev_path;               // [IN] The device we're interested in. Empty to clear all.
    uint64_t num_sectors;               // [IN] Total number of sectors to use.
    PARTITION_INFO_VEC part_info;       // [IN] The partition information we calculate from the layout information.
};

//--------------------------------------------------------------------
//! @brief The Fastec Filesystem working class.
//--------------------------------------------------------------------
class fsWorker : public FAS_THREAD::MsgThread
{
public:
    virtual ~fsWorker();
    void messageHandle(FAS_THREAD::Message *message) override;

    // Singletons should not be cloneable.
    fsWorker(fsWorker& other) = delete;

    // Singeltons should not be assignable.
    void operator=(const fsWorker&) = delete;

    static fsWorker *factory();

    constexpr std::string& getDrivePath()           { return dev_path; }
    bool setDrivePath(const std::string& path);
    constexpr uint64_t getNumSectors()              { return num_sectors; }
    void setNumSectors(uint64_t count)              { num_sectors = count; }

private:
    fsWorker();

    bool analyzeDrive(DRIVE_HD *hdr_ptr, PARTITION_INFO_VEC& drive_info);
    bool analyzePartition(const uint8_t *hdr_ptr, PARTITION_INFO& part_info);
    bool analyzeAvailPartition(const uint8_t *hdr_ptr, PARTITION_INFO& part_info);
    bool analyzeLrCapture(const uint8_t *hdr_ptr, CAPTURE_DATA& cap_info, uint64_t& next_lba);
    bool analyzeLrPartition(const uint8_t *hdr_ptr, PARTITION_INFO& part_info);
    bool analyzeLrPartialCapture(LR_CAPTURE_HDR *cap);
    bool analyzeStorCapture(const uint8_t *hdr_ptr, CAPTURE_DATA& cap_info, uint64_t& next_lba);
    bool analyzeStorPartition(const uint8_t *hdr_ptr, PARTITION_INFO& part_info);

    bool appendLrCapture(const LR_PARTITION_HDR *part_ptr, uint64_t part_lba, LAYOUT_CAPTURE_VEC& layout);
    bool appendStorCapture(const STOR_PARTITION_HDR *part_ptr, uint64_t part_lba, LAYOUT_CAPTURE_VEC& layout);

    bool buildLayout(const std::string& dev_path, uint64_t num_sectors, const LAYOUT_PARTITION_VEC& part_info, DRIVE_HD& drv, PARTITION_INFO_VEC& layout_info);
    bool buildAvailLayout(const LAYOUT_PARTITION *part_info, uint64_t part_lba, uint64_t part_sectors, PARTITION_INFO& cache_info);
    bool buildLrLayout(const LAYOUT_PARTITION *part_info, uint64_t part_lba, uint64_t part_sectors, PARTITION_INFO& cache_info);
    bool buildStorLayout(const LAYOUT_PARTITION *part_info, uint64_t part_lba, uint64_t part_sectors, PARTITION_INFO& cache_info);
    bool writeLayout(const DRIVE_HD *drv_ptr, const PARTITION_INFO_VEC& part_info);

    CAPTURE_DATA *cacheFindCapture(const std::string& dev_path, uint64_t cap_lba);
    CAPTURE_DATA *cacheGetCapture(const std::string& dev_path, uint64_t part_lba, uint64_t cap_lba);
    CAPTURE_DATA *cacheGetCapture(PARTITION_INFO *part_info, uint64_t cap_lba);
    PARTITION_INFO *cacheGetPartition(const std::string& dev_path, uint64_t part_lba, int& part_index);
    void cacheInvalidate(const std::string& dev_path);

    void createAvailPartHdr(AVAIL_PARTITION_HDR *dpart, uint64_t start_lba, uint64_t num_sectors);
    void createLrCaptureHdr(LR_CAPTURE_HDR *cap, uint64_t dpart_lba, uint64_t previous_cap_lba, uint64_t cap_lba, uint64_t image_sectors, uint64_t fmeta_lba, uint64_t fmeta_sectors, uint32_t max_ct_size);
    void createLrPartHdr(LR_PARTITION_HDR *dpart, uint64_t start_lba, uint64_t num_sectors, uint32_t min_frame_size);
    void createStorCaptureHdr(STOR_CAPTURE_HDR *cap, uint64_t dpart_lba, uint64_t previous_cap_lba, uint64_t cap_lba, uint64_t data_sectors);
    void createStorPartHdr(STOR_PARTITION_HDR *dpart, uint64_t start_lba, uint64_t num_sectors);

    bool findLrCaptureEnd(LR_CAPTURE_HDR *cap);
    bool findLrCaptureWrap(LR_CAPTURE_HDR *cap);
    int  findPartition(const DRIVE_HD& drive_hdr, uint64_t part_lba);
    uint32_t *getFrameMetadata(const LR_CAPTURE_HDR *cap, uint32_t *sector_buf, uint32_t abs_frame);

    bool resizeAvailPartition(PARTITION_INFO *part_info, uint64_t resize_amount);
    bool resizeLrCapture(LR_CAPTURE_HDR& cap_hdr, uint64_t resize_amount, const PARTITION_INFO *part_info);
    bool resizeLrPartition(PARTITION_INFO *part_info, uint64_t resize_amount);
    bool resizeStorCapture(STOR_CAPTURE_HDR& cap_hdr, uint64_t resize_amount, const PARTITION_INFO *part_info);
    bool resizeStorPartition(PARTITION_INFO *part_info, uint64_t resize_amount);

    void simulateEndFrameNoWrap(LR_CAPTURE_HDR *cap, const uint32_t *fmeta_buf, uint32_t frame_num);
    void simulateEndFrameWrap(LR_CAPTURE_HDR *cap, const uint32_t *fmeta_buf, uint32_t frame_num);

    void handleAnalyzeRequest(fsAnalyzeMessage *msg);
    void handleAppendCaptureRequest(fsAppendCaptureMessage *msg);
    void handleAppendPartitionRequest(fsAppendPartitionMessage *msg);
    void handleClearCacheRequest(fsClearCacheMessage *msg);
    void handleDeleteCaptureRequest(fsDeleteCaptureMessage *msg);
    void handleDeletePartitionRequest(fsDeletePartitionMessage *msg);
    void handleGetBookmarkRequest(fsGetBookmarkMessage *msg);
    void handleGetCorrectionTableRequest(fsGetCorrectionTableMessage *msg);
    void handleGetGainTableRequest(fsGetGainTableMessage *msg);
    void handleGetInfoRequest(fsGetInfoMessage *msg);
    void handleGetPmMetadataRequest(fsGetPmMetadataMessage *msg);
    void handleLayoutRequest(fsLayoutMessage *msg);
    void handleRenameRequest(fsRenameMessage *msg);
    void handleResizeCaptureRequest(fsResizeCaptureMessage *msg);
    void handleResizePartitionRequest(fsResizePartitionMessage *msg);
    void handleSetBookmarkRequest(fsSetBookmarkMessage *msg);
    void handleSetCorrectionInfoRequest(fsSetCorrectionInfoMessage *msg);
    void handleSetPmMetadataRequest(fsSetPmMetadataMessage *msg);
    void handleTryLayout(fsTryLayoutMessage *msg);
    void handleUpdateCapture(fsUpdateCaptureMessage *msg);
    void handleWritePartition(fsWritePartitionMessage *msg);

    void setBlockChecksum(FI_SIG *sig);
    bool verifyCaptureHeader(FI_SIG *sig, uint32_t part_sig);
    bool verifyDriveHeader(FI_SIG *sig);
    bool verifyPartitionHeader(FI_SIG *sig);

    bool readSectors(uint64_t start_sector, uint8_t *buf, uint32_t num_sectors);
    bool writeSectors(uint64_t start_sector, const uint8_t *buf, uint32_t num_sectors);

    uint32_t writeCorrectionInfo(const CAPTURE_DATA *cap_data, fsSetCorrectionInfoMessage *msg);

    constexpr uint64_t fmetaTime(const uint32_t *metadata_buf);
    std::string reqToString(int req_code) override;
    const char *signatureToString(uint32_t signature);

    uint8_t *md5Calculate(const unsigned char *buf, size_t buf_size, unsigned char *md);
    int md5Length();

private:
    static std::unique_ptr<fsWorker> fs_proc;       //!< Our filesystem worker thread.
    static std::mutex fs_mutex;                     //!< Mutex to protect our singleton object.

    uint64_t    num_sectors;                        //!< The number of sectors we can use.
    int         dev_fd;                             //!< When the device path is set, the device is opened and made ready for reading.
    std::string dev_path;                           //!< Where to access the drive (e.g., "/dev/nvme0n1").
    std::unique_ptr<uint8_t[]> sector_buf;          //!< Page aligned buffer to hold sector data.

    //! The cache of information for each device we've analyzed. The map is keyed
    //! by the dev_path. The cache holds all the layout information for the device
    //! and allows us to traverse the filesystem without having to perform disk
    //! accesses (slow).
    std::map<std::string, FS_CACHE_ENTRY> cache;
};


} /* namespace FAS_FILESYSTEM */

#endif // FS_WORKER_H_

