/*!
 * @file fas_filesystem.cpp
 * @brief The main implementation file for the Fastec Filesystem library.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <iterator>
#include <map>
#include <memory>
#include <string>
#include <vector>

// Windows or Linux?
#if _WIN32
//
// Windows
#include "fas_trace.h"
#include "gige_msg.h"
#include "fas_log.h"
#include "format_macros.h"
#else
//
// Linux & MacOS
#include <fastec/gige_msg.h>
#include <fas_trace.h>
#include <fas_log.h>
#endif

#include "fas_filesystem.h"
#include "fs_worker.h"

namespace FAS_FILESYSTEM
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

enum FS_TRACE
{
    FS_FUNC             = 0x00000010,
    FS_FUNC_NOISY       = 0x00000020
};

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Variable Declaration
//--------------------------------------------------------------------

} /* namespace FAS_FILESYSTEM */

using namespace FAS_FILESYSTEM;

// Our instance and the mutex required to be thread-safe.
std::unique_ptr<FiFilesystem> FiFilesystem::instance;
std::mutex FiFilesystem::instance_mutex;

// For tracing, we need this local control defined outside the namespace.
static const uint32_t traceControl = 0 |
                                     FS_FUNC |
//                                     FS_FUNC_NOISY |
//                                     TRACE_0;        // Set our trace level to TRACE_0 to trace.
                                     TRACE_3;        // Set our trace level to TRACE_3 to not trace.


//--------------------------------------------------------------------
// Implementation: FiFilesystem
//--------------------------------------------------------------------

/*!
 * @brief Base class for Fastec Filesystem.
 *
 * The Fastec Filesystem base class.
 *
 */
FiFilesystem::FiFilesystem()
{
    TP(FS_FUNC, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Destructor for the Fastec Filesystem class object.
 */
FiFilesystem::~FiFilesystem()
{
    TP(FS_FUNC, "%s", __PRETTY_FUNCTION__);
}

/*!
 * @brief Retrieve our singleton instance.
 * @return The pointer to the instance object.
 */
FiFilesystem *FiFilesystem::getInstance()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    TP(FS_FUNC_NOISY, "%s", __PRETTY_FUNCTION__);

    // Do we already have our instance?
    if (instance.get() == nullptr)
    {
        // Create our instance.
        instance = std::unique_ptr<FiFilesystem>(new FiFilesystem);
    }

    // Return our instance.
    return instance.get();
}

/*!
 * @brief Analyze the contents of the device located at "dev_path" to see
 *        if it is a valid Fastec Filesystem.
 *
 * In addition to doing a check for a valid Fastec Filesystem, this also
 * builds up our filesystem "cache".
 *
 * @param dev_path  [IN] The device path to analyze.
 * @return True if we have a valid Fastec Filesystem, false otherwise.
 */
bool FiFilesystem::analyze(const std::string& dev_path)
{
    bool rc;
    fsAnalyzeMessage msg;

    TP(FS_FUNC, "%s: dev_path=[%s]", __PRETTY_FUNCTION__, dev_path.c_str());

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Build up our message.
    msg.dev_path = dev_path;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: analyzed successful, is_fastec=%c", __PRETTY_FUNCTION__, (msg.is_fastec ? 'T' : 'F'));
        rc = msg.is_fastec;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Append a capture to an existing partition.
 *
 * This assumes that the partition specified by "part_lba" will have sufficient
 * space at the end to accomodate the new capture.
 *
 * @note We currently only support appending a SINGLE capture to the partition.
 *       In the future we may add support for appending multiple captures in
 *       one call.
 *
 * @param dev_path  [IN] The device we are working with.
 * @param part_lba  [IN] The partition's LBA, uniquely identifies the partition.
 * @param layout    [IN] The description of the new capture we want to create.
 * @return True on success, false otherwise.
 */
bool FiFilesystem::appendCapture(const std::string& dev_path, uint64_t part_lba, const LAYOUT_CAPTURE_VEC& layout)
{
    bool rc;
    fsAppendCaptureMessage msg;

    TP(FS_FUNC, "%s: dev_path=[%s], part_lba=%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), part_lba);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }
    if (layout.size() != 1)
    {
        fi_log(LOG_DEBUG, "%s: We only support appending a single capture.", __PRETTY_FUNCTION__);
        return false;
    }

    // Build up our message.
    msg.dev_path = dev_path;
    msg.part_lba = part_lba;
    msg.layout = layout;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: capture appended successfully", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Append a partition to an existing filesystem on the device specified by "dev_path".
 *
 * This assumes that the drive will have sufficient space at the end to accomodate the
 * new partition.
 *
 * @note We currently only support appending a SINGLE partition to the drive.
 *       In the future we may add support for appending multiple partitions
 *       in a single call.
 *
 * @param dev_path  [IN] The device we are working with.
 * @param layout    [IN] The description of the new partition we want to create.
 * @return True if successful, flase otherwise.
 */
bool FiFilesystem::appendPartition(const std::string& dev_path, const LAYOUT_PARTITION_VEC& layout)
{
    bool rc;
    fsAppendPartitionMessage msg;

    TP(FS_FUNC, "%s: dev_path=[%s]", __PRETTY_FUNCTION__, dev_path.c_str());

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }
    if (layout.size() != 1)
    {
        fi_log(LOG_DEBUG, "%s: We only support appending a single partition.", __PRETTY_FUNCTION__);
        return false;
    }

    // Build up our message.
    msg.dev_path = dev_path;
    msg.layout = layout;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: partition appended successfully", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Remove an entry from the cache, or remove all entries from the cache.
 *
 * @param dev_path  [IN] The device to remove from the cache. If empty (default), remove all entries.
 * @return True on success, false on failure.
 */
bool FiFilesystem::clearCache(const std::string& dev_path)
{
    bool rc;
    fsClearCacheMessage msg;

    TP(FS_FUNC, "%s: dev_path=[%s]", __PRETTY_FUNCTION__, dev_path.c_str());

    // Build up our message.
    msg.dev_path = dev_path;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: cache cleared.", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Delete the capture specified by "capture_lba" from the drive specified by "dev_path" in the
 *        partition specified by "part_lba".
 *
 * @param dev_path      [IN] The device we are working with.
 * @param part_lba      [IN] The partition's LBA, uniquely identifies the partition.
 * @param capture_lba   [IN] The LBA of the capture we want to delete.
 * @return True if we were successful, false otherwise.
 */
bool FiFilesystem::deleteCapture(const std::string& dev_path, uint64_t part_lba, uint64_t capture_lba)
{
    bool rc;
    fsDeleteCaptureMessage msg;

    TP(FS_FUNC, "%s: dev_path:%s, part_lba:%" PRIu64 ", capture_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), part_lba, capture_lba);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Build up our message.
    msg.dev_path = dev_path;
    msg.part_lba = part_lba;
    msg.capture_lba = capture_lba;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: capture deleted successfully", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Delete the partition specified by "part_lba" from the drive specified by "dev_path".
 *
 * @param dev_path  [IN] The device we are working with.
 * @param part_lba  [IN] The LBA of the partition we want to delete.
 * @return True if we were successful, false otherwise.
 */
bool FiFilesystem::deletePartition(const std::string& dev_path, uint64_t part_lba)
{
    bool rc;
    fsDeletePartitionMessage msg;

    TP(FS_FUNC, "%s: dev_path:%s, part_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), part_lba);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Build up our message.
    msg.dev_path = dev_path;
    msg.part_lba = part_lba;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: partition deleted successfully", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}


/*!
 * @brief Format the drive using the specified paramters.
 *
 * @param dev_path          [IN] The path to the device (or file) to be formatted, e.g. "/dev/nvme0n1".
 * @param volume_label      [IN] The name for the device.
 * @param num_sectors       [IN] Set up to use this many sectors for the drive (may not be the entire drive).
 * @param layout            [IN] The parameters to use for the layout of each partition.
 * @return True if everything went well, false otherwise.
 */
bool FiFilesystem::format(const std::string& dev_path, const std::string& volume_label, uint64_t num_sectors, const LAYOUT_PARTITION_VEC& layout)
{
    bool rc;
    fsLayoutMessage msg;

    // Build up our message.
    msg.num_sectors = num_sectors;
    msg.dev_path = dev_path;
    msg.name = volume_label;
    msg.layout = layout;
    TP(FS_FUNC, "%s: num_sectors=%u, dev_path=[%s], name=[%s], layout=%d elements", __PRETTY_FUNCTION__, msg.num_sectors, msg.dev_path.c_str(), msg.name.c_str(), msg.layout.size());

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: drive formatted.", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Retrieve the bookmark table information located in the capture found at "capture_lba".
 *
 * @param dev_path      [IN] The drive we are interested in.
 * @param capture_lba   [IN] The location of the capture we are interested.
 * @param bk_buffer     [OUT] The vector to copy the bookmark information to.
 * @return True if we were able to succesfully get the bookmark information, false otherwise.
 */
bool FiFilesystem::getCaptureBK(const std::string& dev_path, uint64_t capture_lba, std::vector<uint8_t>& bk_buffer)
{
    bool rc;
    fsGetBookmarkMessage msg;

    TP(FS_FUNC, "%s: dev_path:%s, capture_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), capture_lba);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure the buffer is cleared of any data.
    bk_buffer.clear();

    // Build up our message.
    msg.dev_path = dev_path;
    msg.capture_lba = capture_lba;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: bookmark(s) retrieved.", __PRETTY_FUNCTION__);
        bk_buffer = msg.bk_buffer;
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Retrieve the correction table information located in the capture found at "capture_lba".
 *
 * @param dev_path      [IN] The drive we are interested in.
 * @param capture_lba   [IN] The location of the capture we are interested.
 * @param ct_buffer     [OUT] The vector to copy the correction table information to.
 * @return True if we were able to succesfully get the correction table information, false otherwise.
 */
bool FiFilesystem::getCaptureCT(const std::string& dev_path, uint64_t capture_lba, std::vector<uint8_t>& ct_buffer)
{
    bool rc;
    fsGetCorrectionTableMessage msg;

    TP(FS_FUNC, "%s: dev_path:%s, capture_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), capture_lba);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure the buffer is cleared of any data.
    ct_buffer.clear();

    // Build up our message.
    msg.dev_path = dev_path;
    msg.capture_lba = capture_lba;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: correction table retrieved.", __PRETTY_FUNCTION__);
        ct_buffer = msg.ct_buffer;
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Retrieve the gain table information located in the capture found at "capture_lba".
 *
 * @param dev_path      [IN] The drive we are interested in.
 * @param capture_lba   [IN] The location of the capture we are interested.
 * @param gt_buffer     [OUT] The vector to copy the gain table information to.
 * @return True if we were able to succesfully get the gain table information, false otherwise.
 */
bool FiFilesystem::getCaptureGT(const std::string& dev_path, uint64_t capture_lba, std::vector<uint8_t>& gt_buffer)
{
    bool rc;
    fsGetGainTableMessage msg;

    TP(FS_FUNC, "%s: dev_path:%s, capture_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), capture_lba);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure the buffer is cleared of any data.
    gt_buffer.clear();

    // Build up our message.
    msg.dev_path = dev_path;
    msg.capture_lba = capture_lba;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: gain table retrieved.", __PRETTY_FUNCTION__);
        gt_buffer = msg.gt_buffer;
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Retrieve the pm_metadata (i.e., the capture's "partition metadata") information located in the
 *        capture found at "capture_lba".
 *
 * @param dev_path      [IN] The drive we are interested in.
 * @param capture_lba   [IN] The location of the capture we are interested.
 * @param pm_buffer     [OUT] The vector to copy the pm_metadata information to.
 * @return True if we were able to succesfully get the pm_metadata information, false otherwise.
 */
bool FiFilesystem::getCapturePM(const std::string& dev_path, uint64_t capture_lba, std::vector<uint8_t>& pm_buffer)
{
    bool rc;
    fsGetPmMetadataMessage msg;

    TP(FS_FUNC, "%s: dev_path:%s, capture_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), capture_lba);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Make sure the buffer is cleared of any data.
    pm_buffer.clear();

    // Build up our message.
    msg.dev_path = dev_path;
    msg.capture_lba = capture_lba;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: pm_metadata retrieved.", __PRETTY_FUNCTION__);
        pm_buffer = msg.metadata;
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Query the worker thread for the specified drive's partition information.
 *
 * NOTE: This assumes the system has already done an "FiFilesystem::analyze()" on
 *        the drive to build up the partition information.
 *
 * @param dev_path      [IN] The drive we are interested in.
 * @param part_info     [OUT] The FAS_FILESYSTEM::PARTITION_INFO vector containing the information.
 * @param drive_space   [OUT] The number of sectors available on the drive.
 * @return True if we were able to acquire the drive's partitioning information, false otherwise.
 */
bool FiFilesystem::getDriveInfo(const std::string& dev_path, PARTITION_INFO_VEC& part_info, uint64_t& drive_space, std::string& volume_label)
{
    bool rc;
    fsGetInfoMessage msg;

    TP(FS_FUNC, "%s: dev_path=[%s]", __PRETTY_FUNCTION__, dev_path.c_str());

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Build up our message.
    msg.dev_path = dev_path;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // Copy the data over to the user.
        part_info = msg.part_info;
        drive_space = msg.assigned_space;
        volume_label = msg.name;

        // It's all good.
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Merge two captures (specified by the merge_lbas vector) located in the partition
 *        specified by "part_lba" on the drive specified by "dev_path".
 *
 * @param dev_path      [IN] The device we are working with.
 * @param part_lba      [IN] The partition's LBA, uniquely identifies the partition.
 * @param merge_lbas    [IN] The LBAs of the captures to merge.
 * @return True if we were successful, false otherwise.
 */
bool FiFilesystem::mergeCaptures(const std::string& dev_path, uint64_t part_lba, const std::vector<uint64_t>& merge_lbas)
{
    (void)dev_path;
    (void)part_lba;
    (void)merge_lbas;

    TP(FS_FUNC, "%s: dev_path:%s, part_lba:%" PRIu64 ", merge_lbas:%" PRIu64 ", %" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), part_lba,
       merge_lbas[0], merge_lbas[1]);

    return false;
}

/*!
 * @brief Rename the drive specified by "dev_path" to "name".
 *
 * @param dev_path      [IN] The device we are working with.
 * @param name          [IN] The new name to use.
 * @return True if we were successful, false otherwise.
 */
bool FiFilesystem::rename(const std::string& dev_path, const std::string& volume_label)
{
    bool rc;
    fsRenameMessage msg;

    // Build up our message.
    msg.dev_path = dev_path;
    msg.name = volume_label;
    TP(FS_FUNC, "%s: dev_path=[%s], new name=[%s]", __PRETTY_FUNCTION__, msg.dev_path.c_str(), msg.name.c_str());

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: drive renamed.", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Resize the capture located at "cap_lba". Resizing can be to a specified size (in sectors)
 *        or to fit the size of already existing data.
 *
 * If the intent is to grow the size of the capture, the new size must fit within the partition. In
 * other words, the partition size will not be modified (see FiFilesystem::resizePartition).
 *
 * If the intent is to shrink the size of the capture, existing data MAY be destroyed. Also, the new
 * size must be large enough to hold the minimum number of frames (FAS_FILESYSTEM::MIN_FRAMES_REQUIRED).
 *
 * @note Currently we can only resize the LAST capture in a partition!
 *
 * @param dev_path          [IN] The device we are working with.
 * @param capture_lba       [IN] The location of the capture header.
 * @param resize_amount     [IN] The amount to resize. 0: fit to existing data, otherwise the new size (in sectors.)
 * @return True on success, false otherwise.
 */
bool FiFilesystem::resizeCapture(const std::string& dev_path, uint64_t capture_lba, uint64_t resize_amount)
{
    bool rc;
    fsResizeCaptureMessage msg;

    TP(FS_FUNC, "%s: dev_path=[%s], lba=%" PRIu64 ", resize:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), capture_lba, resize_amount);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Build up our message.
    msg.dev_path = dev_path;
    msg.cap_lba = capture_lba;
    msg.resize_amount = resize_amount;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: capture resized successfully", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Resize the partition located at "part_lba". Resizing can be to a specified size (in sectors)
 *        or to fit the size of already existing data.
 *
 * If the intent is to grow the size of the partition, the new size must fit within the drive.
 *
 * If the intent is to shrink the size of the partition, existing data MAY be destroyed. Also, the new
 * size must be large enough to hold at least one capture.
 *
 * @note Currently we can only resize the LAST partition in a drive!
 *
 * @param dev_path          [IN] The device we are working with.
 * @param part_lba          [IN] The location of the partition header.
 * @param resize_amount     [IN] The amount to resize. 0: fit to existing data, otherwise the new size (in sectors.)
 * @return True on success, false otherwise.
 */
bool FiFilesystem::resizePartition(const std::string& dev_path, uint64_t part_lba, uint64_t resize_amount)
{
    bool rc;
    fsResizePartitionMessage msg;

    TP(FS_FUNC, "%s: dev_path=[%s], lba=%" PRIu64 ", resize:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), part_lba, resize_amount);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Build up our message.
    msg.dev_path = dev_path;
    msg.part_lba = part_lba;
    msg.resize_amount = resize_amount;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: partition resized successfully", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Set the bookmark table information located in the capture found at "capture_lba".
 *
 * @param dev_path      [IN] The drive we are interested in.
 * @param capture_lba   [IN] The location of the capture we are interested.
 * @param bk_buffer     [IN] The vector containing the bookmark information.
 * @return True if we were able to succesfully write the bookmark information, false otherwise.
 */
bool FiFilesystem::setCaptureBK(const std::string& dev_path, uint64_t capture_lba, const std::vector<uint8_t>& bk_buffer)
{
    bool rc;
    fsSetBookmarkMessage msg;

    TP(FS_FUNC, "%s: dev_path:%s, capture_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), capture_lba);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Build up our message.
    msg.dev_path = dev_path;
    msg.capture_lba = capture_lba;
    msg.bk_buffer = bk_buffer;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: bookmark(s) saved.", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Set the correction table information located in the capture found at "capture_lba".
 *
 * @param dev_path      [IN] The drive we are interested in.
 * @param capture_lba   [IN] The location of the capture we are interested.
 * @param ct_buffer     [IN] The correction table information.
 * @param gt_buffer     [IN] The gain table information.
 * @return True if we were able to succesfully write the correction table information, false otherwise.
 */
bool FiFilesystem::setCaptureInfo(const std::string& dev_path, uint64_t capture_lba, const std::vector<uint8_t>& ct_buffer, const std::vector<uint8_t>& gt_buffer)
{
    bool rc;
    fsSetCorrectionInfoMessage msg;

    TP(FS_FUNC, "%s: dev_path:%s, capture_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), capture_lba);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Build up our message.
    msg.dev_path = dev_path;
    msg.capture_lba = capture_lba;
    msg.ct_buffer = ct_buffer;
    msg.gt_buffer = gt_buffer;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: correction table saved.", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Set the pm_metadata (i.e., the capture's "partition metadata") information located in the
 *        capture found at "capture_lba".
 *
 * @param dev_path      [IN] The drive we are interested in.
 * @param capture_lba   [IN] The location of the capture we are interested.
 * @param pm_buffer     [IN] The vector to copy the pm_metadata information to.
 * @return True if we were able to succesfully write the pm_metadata information, false otherwise.
 */
bool FiFilesystem::setCapturePM(const std::string& dev_path, uint64_t capture_lba, const std::vector<uint8_t>& pm_buffer)
{
    bool rc;
    fsSetPmMetadataMessage msg;

    TP(FS_FUNC, "%s: dev_path:%s, capture_lba:%" PRIu64, __PRETTY_FUNCTION__, dev_path.c_str(), capture_lba);

    // Sanity check.
    if (dev_path.empty())
    {
        fi_log(LOG_DEBUG, "%s: dev_path is empty.", __PRETTY_FUNCTION__);
        return false;
    }

    // Build up our message.
    msg.dev_path = dev_path;
    msg.capture_lba = capture_lba;
    msg.metadata = pm_buffer;

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: pm_metadata retrieved.", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Prototype formatting the drive using the specified paramters, without actually writing to the drive.
 *
 * @param dev_path          [IN] The path to the device (or file) to be formatted, e.g. "/dev/nvme0n1".
 * @param num_sectors       [IN] Set up to use this many sectors for the drive (may not be the entire drive).
 * @param layout            [IN] The parameters to use for the layout of each partition.
 * @param part_info         [OUT] The FAS_FILESYSTEM::PARTITION_INFO vector containing the information.
 * @return True if everything went well, false otherwise.
 */
bool FiFilesystem::tryLayout(const std::string& dev_path, uint64_t num_sectors, const LAYOUT_PARTITION_VEC& layout, PARTITION_INFO_VEC& part_info)
{
    bool rc;
    fsTryLayoutMessage msg;

    // Build up our message.
    msg.num_sectors = num_sectors;
    msg.dev_path = dev_path;
    msg.layout = layout;
    TP(FS_FUNC, "%s: num_sectors=%u, dev_path=[%s], layout=%d elements", __PRETTY_FUNCTION__, msg.num_sectors, msg.dev_path.c_str(), msg.layout.size());

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: drive format prototype complete.", __PRETTY_FUNCTION__);
        part_info = msg.part_info;
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Write the partition information, typically generated by a call to FiFilesystem::tryLayout().
 *
 * @param dev_path          [IN] The path to the device (or file) to be formatted, e.g. "/dev/nvme0n1".
 * @param capture_info      [IN] The FAS_FILESYSTEM::CAPTURE_DATA containing the information to be written.
 * @return True if everything went well, false otherwise.
 */
bool FiFilesystem::updateCapture(const std::string& dev_path, const CAPTURE_DATA& capture_info)
{
    bool rc;
    fsUpdateCaptureMessage msg;

    // Build up our message.
    msg.dev_path = dev_path;
    msg.capture_info = capture_info;
    TP(FS_FUNC, "%s: dev_path=[%s], capture@%" PRIu64, __PRETTY_FUNCTION__, msg.dev_path.c_str(), msg.capture_info.lba);

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: capture updated.", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Write the partition information, typically generated by a call to FiFilesystem::tryLayout().
 *
 * @param dev_path          [IN] The path to the device (or file) to be formatted, e.g. "/dev/nvme0n1".
 * @param num_sectors       [IN] Set up to use this many sectors for the drive (may not be the entire drive).
 * @param part_info         [IN] The FAS_FILESYSTEM::PARTITION_INFO vector containing the information to be written.
 * @return True if everything went well, false otherwise.
 */
bool FiFilesystem::writePartition(const std::string& dev_path, uint64_t num_sectors, const PARTITION_INFO_VEC& part_info)
{
    bool rc;
    fsWritePartitionMessage msg;

    // Build up our message.
    msg.num_sectors = num_sectors;
    msg.dev_path = dev_path;
    msg.part_info = part_info;
    TP(FS_FUNC, "%s: num_sectors=%u, dev_path=[%s], part_info=%d elements", __PRETTY_FUNCTION__, msg.num_sectors, msg.dev_path.c_str(), msg.part_info.size());

    // Send it on.
    msg.doSync(fsWorker::factory());

    // Success?
    if (msg.resp == FAS_THREAD::OK)
    {
        // It's all good.
        TP(FS_FUNC_NOISY, "%s: drive formatted.", __PRETTY_FUNCTION__);
        rc = true;
    }
    else
    {
        // There was a problem.
        TP(FS_FUNC_NOISY, "%s: request failed.", __PRETTY_FUNCTION__);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Get the number of reserved sectors for the specified header.
 *
 * @param type  The type of header we are interested in.
 * @return The number of sectors reserved for the header.
 */
int FiFilesystem::getReserved(uint32_t type)
{
    int reserved_size = 0;

    // What type of header are we interested in?
    switch (type)
    {
    case LR_CAPTURE_SIG:
        // The number of sectors used for the capture header is the one sector for
        // the actual header, the sectors for the capture metadata, and the sectors
        // for the bookmark table. The correction table information MAY be added later.
        reserved_size = (BYTES_TO_SECTORS(CAPTURE_METADATA_SIZE) +
                         BYTES_TO_SECTORS(CAPTURE_BOOKMARK_SIZE) + 1);
        break;
    case DRIVE_SIG:
    case AVAIL_PART_SIG:
    case LR_PART_SIG:
    case STOR_CAPTURE_SIG:
    case STOR_PART_SIG:
        // These headers only requires the 1 sector.
        reserved_size = 1;
        break;
    default:
        break;
    }

    // Return our result.
    return reserved_size;
}
