/*!
 * @file fi_rdi_motionampl.h
 * @brief Definitions and declarations for Fastec Imaging RDI motion
 *        amplification logic.
 *
 * Copyright 2023 (c) RDI Technologies as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * RDI Technologies. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with RDI Technologies.
 */

#ifndef FI_RDI_MOTIONAMPL_
#define FI_RDI_MOTIONAMPL_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>

#include <memory>
#include <mutex>
#include <queue>
#include <vector>

namespace FAS_RDI
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

class FiRDIMotionAmpl
{
public:
    FiRDIMotionAmpl();
    ~FiRDIMotionAmpl();

    /*!
     * @brief Add the 'frame' to the reference frame.
     *
     * @note The FiRDIMotionAmpl class object takes over ownership of the IMG_POINTER.
     *       After the method returns, the 'frame' value is reset and points to null.
     *
     * @param frame     [IN] The frame we are going to add to the system.
     */
    void addFrame(uint8_t *frame);

    /*!
     * @brief Check to see if frames can be amplified yet.
     * @return True if the amplifier is ready, false otherwise.
     */
    bool amplifyAvailable();

    /*!
     * @brief Amplify the frame last submitted via addFrame().
     *
     * @param ampFrame  [OUT] The amplified image.
     * @return True if successful, false otherwise.
     */
    bool amplifyFrame(uint8_t *ampFrame);

    /*!
     * @brief Calculate and return the current reference frame based on the frames in
     *        the current reference window.
     *
     * @param refFrame  [OUT] The reference "image".
     * @return True if successful, false otherwise.
     */
    bool getReference(uint8_t *refFrame);

    /*!
     * @brief Return the reference window size. This is the number of frames that need
     *        to go into the summation logic to build the reference frame. Amplification
     *        is not available until after that condition is met.
     *
     * @return The number of frames that go into the building of the reference frame.
     */
    constexpr int getRefWindowSize()    { return ref_window; }

    /*!
     * @brief Retrieve the width of the images.
     */
    constexpr int getWidth()            { return width; }

    /*!
     * @brief Retrieve the height of the images.
     */
    constexpr int getHeight()           { return height; }

    /*!
     * @brief Call this method to setup the live MA feed.
     *
     * @param framerate             Framerate of associated camera.
     * @param minFreqOfInterest     Minimum frequency of interest (provided by user).
     * @param width                 Width of frame.
     * @param height                Height of frame.
     * @param factor                Amplification factor.
     */
    void setParameters(double framerate, double minFreqOfInterest, int width, int height, double factor);

    /*!
     * @brief Quick setter if you only want to change the amplification factor.
     * @param factor    The new amplification factor to be used.
     */
    constexpr void setAmplificationFactor(double factor)        { ampl_factor = factor; }

private:
    void reset();

    /*!
     * @brief An utility operator to subtract the right operand (rhs) from the left (lhs)
     *        but not to go below 0.
     *
     * @param lhs   The left hand value (i.e., the "minuend").
     * @param rhs   The right hand value (i.e., the "subtrahend").
     * @return If lhs > rhs, return (lhs - rhs). Otherwise, return 0.
     */
    static uint8_t minus_floor(const uint8_t& lhs, const uint8_t& rhs);

    /*!
     * @brief An utility operator to add the right operand (rhs) to the left (lhs)
     *        but not to exceed 255 (0xFF).
     *
     * @param lhs   The left hand value (i.e., the first "addend").
     * @param rhs   The right hand value (i.e., the second "addend").
     * @return If (lhs + rhs) > 255, return 255. Otherwise, return (lhs + rhs).
     */
    static uint8_t plus_clip(const uint8_t& lhs, const uint8_t& rhs);


private:
    static const int NUM_CYCLES_REQD;

    double ampl_factor;         //!< The amplification factor to use.
    int ref_window;             //!< The number of frames that go into the reference frame.
    int width;                  //!< Width of the image(s).
    int height;                 //!< Height of the image(s).

    std::mutex instance_mutex;  //!< Mutex to prevent re-entrancy of the instance.
    std::queue<uint8_t *> ref_frame_queue;              //!< The queue of frames that went into the sum_frame.
    std::unique_ptr<std::vector<uint32_t>> sum_frame;   //!< The summation of all frames in the ref_window.
};

} // End of FAS_RDI namespace.

#endif  /* FI_RDI_MOTIONAMPL_ */

