/*!
 * @file fas_internal.cpp
 * @brief Public facing entry points for the Fastec Internal Proprietary library.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <chrono>
#include <memory>
#include <vector>

#include "fi_log.h"
#include "fi_rdi_packed.h"
#include "fi_stream_rdi_packed.h"
#include "fi_rdi_motionampl.h"

namespace FAS_INTERNAL
{
//--------------------------------------------------------------------
// Type Definitions
//--------------------------------------------------------------------

typedef std::unique_ptr<FAS_RDI::FiRDIPackedStream> PACKED_STREAM_PTR;
typedef std::unique_ptr<FAS_RDI::FiRDIMotionAmpl> MA_PTR;

//--------------------------------------------------------------------
// Local Variable Declarations
//--------------------------------------------------------------------
static std::vector<PACKED_STREAM_PTR> streamVector;
static std::vector<MA_PTR> maVector;

} // End of FAS_INTERNAL

using namespace FAS_INTERNAL;

// ===================================================================

uint32_t fiGetPackedSize(uint32_t width, uint32_t height, int bitsPerPixel)
{
    uint32_t size_per_frame;

    // For anything other than 8bpp, we have 2 pixels/3 bytes.
    if (bitsPerPixel == 8)
        size_per_frame = (width * height) + FI_RDI_PACKED::FiRDIPacked::getHeaderSize();
    else
        size_per_frame = ((width * height * 3) / 2) + FI_RDI_PACKED::FiRDIPacked::getHeaderSize();

    return size_per_frame;
}

FIP_HANDLE fiCreatePackStreamer(const std::string& streamName, const FIP_STREAM_CONFIGURATION& configInfo)
{
    int i;
    FIP_HANDLE streamHandle(0);
    PACKED_STREAM_PTR streamPtr;
    FIP_STATUS status;

    // Allocate the streamer class object.
    streamPtr = PACKED_STREAM_PTR(new FAS_RDI::FiRDIPackedStream());
    if (streamPtr.get() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_ERR, "%s: Unable to allocate the FiRDIPackedStream object!", FI_FUNCTION);
        goto create_bail;
    }

    // Setup the streamer.
    status = streamPtr->prepareForWriting(streamName, configInfo);
    if (status != FIP_STATUS::STREAM_READY)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: prepareForWriting() failed [status=%d]", FI_FUNCTION, (int)status);

        // Clean up and bail.
        streamPtr.reset();
        goto create_bail;
    }

    // Everything is looking good. Find an available slot in our vector.
    for (i = 0; i < (int)streamVector.size(); ++i)
    {
        // Is this entry available?
        if (streamVector.at(i).get() == nullptr)
            break;
    }
    if (i == (int)streamVector.size())
    {
        // Add a new entry on the vector.
        streamVector.push_back(std::move(streamPtr));
    }
    else
    {
        // Overwrite the available entry.
        streamVector.at(i).reset(std::move(streamPtr.release()));
    }

    // Our handle will be the index of the entry plus 1 to keep
    // it 1-based as a 0 is nullptr.
    streamHandle = (FIP_HANDLE)(i + 1);

create_bail:

    // Return the handle.
    return streamHandle;
}

FIP_STATUS fiClosePackStreamer(FIP_HANDLE streamHandle)
{
    int index;
    FIP_STATUS status;
    FAS_RDI::FiRDIPackedStream *streamPtr;

    // Sanity check.
    if (streamHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle is 0!", FI_FUNCTION);
        return INVALID_HANDLE;
    }

    // Get the index into our vector.
    index = ((int)streamHandle) - 1;
    if ((index < 0) || (index > (int)streamVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Retrieve the pointer.
    streamPtr = streamVector.at(index).get();
    if (streamPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Close the streamer.
    status = streamPtr->closeStream();

    // Delete the stream object and empty the vector entry.
    streamVector.at(index).reset();

    // Return our status.
    return status;
}

FIP_STATUS fiIsPackStreamerFinished(FIP_HANDLE streamHandle)
{
    int index;
    FAS_RDI::FiRDIPackedStream *streamPtr;

    // Sanity check.
    if (streamHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle is 0!", FI_FUNCTION);
        return INVALID_HANDLE;
    }

    // Get the index into our vector.
    index = ((int)streamHandle) - 1;
    if ((index < 0) || (index > (int)streamVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Retrieve the pointer.
    streamPtr = streamVector.at(index).get();
    if (streamPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Return the status.
    return streamPtr->isStreamWriterComplete();
}

FIP_STATUS fiAddPackedData(FIP_HANDLE streamHandle, const uint8_t *imagePtr)
{
    int index;
    uint32_t dataLength;
    std::unique_ptr<FI_RDI_PACKED::FiRDIPacked> packedImage;
    FAS_RDI::FiRDIPackedStream *streamPtr;

    // Sanity check.
    if (streamHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle is 0!", FI_FUNCTION);
        return INVALID_HANDLE;
    }

    // Get the index into our vector.
    index = ((int)streamHandle) - 1;
    if ((index < 0) || (index > (int)streamVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Retrieve the pointer.
    streamPtr = streamVector.at(index).get();
    if (streamPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Allocate our image object.
    packedImage = std::unique_ptr<FI_RDI_PACKED::FiRDIPacked>(new FI_RDI_PACKED::FiRDIPacked());
    if (packedImage.get() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: Unable to allocate FI_RDI_PACKED::FiRDIPacked object.", FI_FUNCTION);
        return ALLOCATION_FAILED;
    }

    // Create our packed image.
    dataLength = fiGetPackedSize(streamPtr->getWidth(), streamPtr->getHeight(), streamPtr->getBitsPerPixel());
    packedImage->setPackedData(imagePtr, dataLength, streamPtr->getWidth(), streamPtr->getHeight(), streamPtr->getBitsPerPixel());

    // Add the packed image to our streamer.
    return streamPtr->addFrame(std::move(packedImage));
}


FIP_STATUS fiAddTiffImageData(FIP_HANDLE streamHandle, const uint8_t *imagePtr)
{
    int index;
    std::unique_ptr<FI_RDI_PACKED::FiRDIPacked> packedImage;
    FAS_RDI::FiRDIPackedStream *streamPtr;

    // Sanity check.
    if (streamHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle is 0!", FI_FUNCTION);
        return INVALID_HANDLE;
    }

    // Get the index into our vector.
    index = ((int)streamHandle) - 1;
    if ((index < 0) || (index > (int)streamVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Retrieve the pointer.
    streamPtr = streamVector.at(index).get();
    if (streamPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Allocate our image object.
    packedImage = std::unique_ptr<FI_RDI_PACKED::FiRDIPacked>(new FI_RDI_PACKED::FiRDIPacked());
    if (packedImage.get() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: Unable to allocate FI_RDI_PACKED::FiRDIPacked object.", FI_FUNCTION);
        return ALLOCATION_FAILED;
    }

    // Create our packed image.
    packedImage->setTiffImageData(imagePtr, streamPtr->getWidth(), streamPtr->getHeight(), streamPtr->getBitsPerPixel());
    if (packedImage->getPackedData() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: setTiffImageData() failed.", FI_FUNCTION);
        return CODEC_ERROR;
    }

    // Add the packed image to our streamer.
    return streamPtr->addFrame(std::move(packedImage));
}

FIP_STATUS fiAddPlaceholder(FIP_HANDLE streamHandle)
{
    int index;
    FAS_RDI::FiRDIPackedStream *streamPtr;

    // Sanity check.
    if (streamHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle is 0!", FI_FUNCTION);
        return INVALID_HANDLE;
    }

    // Get the index into our vector.
    index = ((int)streamHandle) - 1;
    if ((index < 0) || (index > (int)streamVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Retrieve the pointer.
    streamPtr = streamVector.at(index).get();
    if (streamPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Add the placeholder.
    return streamPtr->fiAddPlaceholderFrame();
}

FIP_STATUS fiReplaceFrame(FIP_HANDLE streamHandle, uint32_t seqNumber, const uint8_t *imagePtr)
{
    int index;
    std::unique_ptr<FI_RDI_PACKED::FiRDIPacked> packedImage;
    FAS_RDI::FiRDIPackedStream *streamPtr;

    // Sanity check.
    if (streamHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle is 0!", FI_FUNCTION);
        return INVALID_HANDLE;
    }

    // Get the index into our vector.
    index = ((int)streamHandle) - 1;
    if ((index < 0) || (index > (int)streamVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Retrieve the pointer.
    streamPtr = streamVector.at(index).get();
    if (streamPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Allocate our image object.
    packedImage = std::unique_ptr<FI_RDI_PACKED::FiRDIPacked>(new FI_RDI_PACKED::FiRDIPacked());
    if (packedImage.get() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: Unable to allocate FI_RDI_PACKED::FiRDIPacked object.", FI_FUNCTION);
        return ALLOCATION_FAILED;
    }

    // Create our packed image.
    packedImage->setTiffImageData(imagePtr, streamPtr->getWidth(), streamPtr->getHeight(), streamPtr->getBitsPerPixel());
    if (packedImage->getPackedData() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: setTiffImageData() failed.", FI_FUNCTION);
        return CODEC_ERROR;
    }

    // Add the packed image to our streamer.
    return streamPtr->editVideoFrame(seqNumber, std::move(packedImage));
}

FIP_STATUS fiGetPackStatistics(FAS_INTERNAL::FIP_HANDLE streamHandle, uint32_t& frameCount, uint32_t& lastFrame, uint32_t& encodedFrames)
{
    int index;
    FAS_RDI::FiRDIPackedStream *streamPtr;

    // Sanity check.
    if (streamHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle is 0!", FI_FUNCTION);
        return INVALID_HANDLE;
    }

    // Get the index into our vector.
    index = ((int)streamHandle) - 1;
    if ((index < 0) || (index > (int)streamVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Retrieve the pointer.
    streamPtr = streamVector.at(index).get();
    if (streamPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: streamHandle (%d) is invalid!", FI_FUNCTION, streamHandle);
        return INVALID_HANDLE;
    }

    // Get the statistics.
    frameCount = streamPtr->getFrameCount();
    lastFrame = streamPtr->getLastFrameAdded();
    encodedFrames = streamPtr->getFramesEncodedCount();

    return FIP_OK;
}

FIP_STATUS fiConvertTiffImageData(uint8_t *imageInPtr, uint8_t *imageOutPtr, int frameSeqNum, uint32_t width, uint32_t height, int bitsPerPixel)
{
    std::unique_ptr<FI_RDI_PACKED::FiRDIPacked> packedImage;

    // Sanity check.
    if (imageInPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: imageInPtr is nullptr!", FI_FUNCTION);
        return NO_STREAM_IMAGE;
    }
    if (imageOutPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: imageOutPtr is nullptr!", FI_FUNCTION);
        return NO_STREAM_IMAGE;
    }

    // Allocate our image object.
    packedImage = std::unique_ptr<FI_RDI_PACKED::FiRDIPacked>(new FI_RDI_PACKED::FiRDIPacked());
    if (packedImage.get() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: Unable to allocate FI_RDI_PACKED::FiRDIPacked object.", FI_FUNCTION);
        return ALLOCATION_FAILED;
    }

    // Create our packed image.
    packedImage->setTiffImageData(imageInPtr, width, height, bitsPerPixel);
    if (packedImage->getPackedData() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: setTiffImageData() failed.", FI_FUNCTION);
        return CODEC_ERROR;
    }

    // Update the image's header information.
    if (!packedImage->isHeaderInitialized())
    {
        // Get the current time. We will make that the frame time.
        auto tp_now = std::chrono::system_clock::now();
        auto tp_secs = std::chrono::duration_cast<std::chrono::seconds>(tp_now.time_since_epoch());
        auto tp_usecs = std::chrono::duration_cast<std::chrono::microseconds>(tp_now.time_since_epoch()) - std::chrono::duration_cast<std::chrono::microseconds>(tp_secs);

        // Setup the header information.
        packedImage->setHeader(frameSeqNum, tp_secs.count(), tp_usecs.count());
    }

    // Copy the packed data into the buffer.
    memcpy(imageOutPtr, packedImage->getHeader(), packedImage->getHeaderSize());
    memcpy(imageOutPtr + packedImage->getHeaderSize(), packedImage->getPackedData(), packedImage->getPackedDataSize());

    // All done.
    return FIP_OK;
}

FIP_HANDLE fiCreateMotionAmplifier(const FIP_MA_CONFIGURATION& configInfo)
{
    int i;
    FIP_HANDLE maHandle(0);
    MA_PTR maPtr;

    // Allocate the motion amplification class object.
    maPtr = MA_PTR(new FAS_RDI::FiRDIMotionAmpl());
    if (maPtr.get() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_ERR, "%s: Unable to allocate the FiRDIMotionAmpl object!", FI_FUNCTION);
        goto create_bail;
    }

    // Setup the object.
    maPtr->setParameters(configInfo.framerate, configInfo.minFreqOfInterest, configInfo.width, configInfo.height, configInfo.factor);

    // Everything is looking good. Find an available slot in our vector.
    for (i = 0; i < (int)maVector.size(); ++i)
    {
        // Is this entry available?
        if (maVector.at(i).get() == nullptr)
            break;
    }
    if (i == (int)maVector.size())
    {
        // Add a new entry on the vector.
        maVector.push_back(std::move(maPtr));
    }
    else
    {
        // Overwrite the available entry.
        maVector.at(i).reset(std::move(maPtr.release()));
    }

    // Our handle will be the index of the entry plus 1 to keep
    // it 1-based as a 0 is nullptr.
    maHandle = (FIP_HANDLE)(i + 1);

create_bail:

    // Return the handle.
    return maHandle;
}

FIP_STATUS fiCloseMotionAmplifier(FIP_HANDLE maHandle)
{
    int index;
    FAS_RDI::FiRDIMotionAmpl *maPtr;

    // Sanity check.
    if (maHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle is nullptr!", FI_FUNCTION);
        return INVALID_HANDLE;
    }

    // Get the index into our vector.
    index = ((int)maHandle) - 1;
    if ((index < 0) || (index > (int)maVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return INVALID_HANDLE;
    }

    // Retrieve the pointer.
    maPtr = maVector.at(index).get();
    if (maPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return INVALID_HANDLE;
    }

    // Delete the object and empty the vector entry.
    maVector.at(index).reset();

    // Return our status.
    return MA_CLOSED;
}

FAS_INTERNAL::FIP_STATUS fiMAAddFrame(FAS_INTERNAL::FIP_HANDLE maHandle, uint8_t *framePtr)
{
    int index;
    FAS_RDI::FiRDIMotionAmpl *maPtr;

    // Sanity check.
    if (maHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle is 0!", FI_FUNCTION);
        return INVALID_HANDLE;
    }
    if (framePtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: framePtr is nullptr!", FI_FUNCTION);
        return BAD_PARAMETER;
    }

    // Get the index into our vector.
    index = ((int)maHandle) - 1;
    if ((index < 0) || (index > (int)maVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%d) is invalid!", FI_FUNCTION, maHandle);
        return INVALID_HANDLE;
    }

    // Retrieve the pointer.
    maPtr = maVector.at(index).get();
    if (maPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%d) is invalid!", FI_FUNCTION, maHandle);
        return INVALID_HANDLE;
    }

    // Pass it on.
    maPtr->addFrame(framePtr);

    // All is well.
    return FIP_OK;
}

bool fiMAAvailable(FAS_INTERNAL::FIP_HANDLE maHandle)
{
    int index;
    FAS_RDI::FiRDIMotionAmpl *maPtr;

    // Sanity check.
    if (maHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle is nullptr!", FI_FUNCTION);
        return false;
    }

    // Get the index into our vector.
    index = ((int)maHandle) - 1;
    if ((index < 0) || (index > (int)maVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return false;
    }

    // Retrieve the pointer.
    maPtr = maVector.at(index).get();
    if (maPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return false;
    }

    // Return if the MA engine can amplify frames yet.
    return maPtr->amplifyAvailable();
}

FAS_INTERNAL::FIP_STATUS fiMAGetFrame(FAS_INTERNAL::FIP_HANDLE maHandle, uint8_t *amplifiedFrame)
{
    int index;
    FAS_RDI::FiRDIMotionAmpl *maPtr;

    // Sanity check.
    if (maHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle is nullptr!", FI_FUNCTION);
        return INVALID_HANDLE;
    }

    // Get the index into our vector.
    index = ((int)maHandle) - 1;
    if ((index < 0) || (index > (int)maVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return INVALID_HANDLE;
    }

    // Retrieve the pointer.
    maPtr = maVector.at(index).get();
    if (maPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return INVALID_HANDLE;
    }

    // Get the amplified frame.
    if (!maPtr->amplifyFrame(amplifiedFrame))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: amplifyFrame() failed!", FI_FUNCTION);
        return MA_FAULT;
    }

    // Copy the data over.
    return FIP_OK;
}

FAS_INTERNAL::FIP_STATUS fiMAGetRefFrame(FAS_INTERNAL::FIP_HANDLE maHandle, uint8_t *referenceFrame)
{
    int index;
    FAS_RDI::FiRDIMotionAmpl *maPtr;

    // Sanity check.
    if (maHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle is nullptr!", FI_FUNCTION);
        return INVALID_HANDLE;
    }

    // Get the index into our vector.
    index = ((int)maHandle) - 1;
    if ((index < 0) || (index > (int)maVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return INVALID_HANDLE;
    }

    // Retrieve the pointer.
    maPtr = maVector.at(index).get();
    if (maPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return INVALID_HANDLE;
    }

    // Get the reference frame.
    if (!maPtr->getReference(referenceFrame))
        return MA_FAULT;

    // All is well.
    return FIP_OK;
}

int fiMAGetRefWindowSize(FAS_INTERNAL::FIP_HANDLE maHandle)
{
    int index;
    FAS_RDI::FiRDIMotionAmpl *maPtr;

    // Sanity check.
    if (maHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle is nullptr!", FI_FUNCTION);
        return 0;
    }

    // Get the index into our vector.
    index = ((int)maHandle) - 1;
    if ((index < 0) || (index > (int)maVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return 0;
    }

    // Retrieve the pointer.
    maPtr = maVector.at(index).get();
    if (maPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return INVALID_HANDLE;
    }

    // Return the reference window size.
    return maPtr->getRefWindowSize();
}

void fiMAUpdateAmpFactor(FAS_INTERNAL::FIP_HANDLE maHandle, int newAmpFactor)
{
    int index;
    FAS_RDI::FiRDIMotionAmpl *maPtr;

    // Sanity check.
    if (maHandle == 0)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle is nullptr!", FI_FUNCTION);
        return;
    }

    // Get the index into our vector.
    index = ((int)maHandle) - 1;
    if ((index < 0) || (index > (int)maVector.size()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return;
    }

    // Retrieve the pointer.
    maPtr = maVector.at(index).get();
    if (maPtr == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: maHandle (%p) is invalid!", FI_FUNCTION, maHandle);
        return;
    }

    // Return the reference window size.
    maPtr->setAmplificationFactor((double)newAmpFactor);
}
