/*!
 * @file fi_log.cpp
 * @brief Generic logging functions for debug or production code.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef _WIN32
#include <io.h>
#else
#include <unistd.h>
#include <syslog.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <array>
#include <string>

#include "fi_log.h"

#if defined(_WIN32) && !defined(__MINGW32__)
// Turn off the "This function may be unsafe" warning.
#  pragma warning( disable: 4996 )
#endif

//--------------------------------------------------------------------
// Global Variable declarations
//--------------------------------------------------------------------

namespace FI_LOG
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

#if defined(_WIN32) && !defined(__MINGW32__)
#  define open      _open
#  define snprintf  _snprintf
#  define write     _write

#  define S_IRUSR   _S_IREAD
#  define S_IWUSR   _S_IWRITE
#  define S_IRGRP   _S_IREAD
#  define S_IWGRP   _S_IWRITE
#endif

//--------------------------------------------------------------------
// Variable declarations
//--------------------------------------------------------------------


// ==============================================================
// Static methods should be defined outside the class.
// ==============================================================

std::unique_ptr<FiLogger> FiLogger::instance;
std::mutex FiLogger::instance_mutex;

//============================================================================
// Implementation
//============================================================================


FiLogger::FiLogger()
{
    // Flag to indicate messages from debug_log() are be displayed.
    debug_to_log = false;

    // Flag to indicate if the syslog facility should be used.
#ifdef DEBUG
    use_syslog = false;
#else
    use_syslog = true;
#endif

    // Flag to indicate if log messages are to be sent to STDOUT.
    use_stdout = true;

    // File descriptor for the log file.
    logFileFd = -1;
}

FiLogger::~FiLogger()
{
    // Do we still have the log file open?
    if (logFileFd >= 0)
        loggerFile(nullptr);
}

FiLogger *FiLogger::factory()
{
    std::lock_guard<std::mutex> lock(instance_mutex);

    // Check to see if we have our singleton.
    if (instance.get() == nullptr)
    {
        // Create the one and only instance.
        instance = std::unique_ptr<FiLogger>(new FiLogger());
    }

    return instance.get();
}

// Logger control API
void FiLogger::loggerUseSyslog(bool enable)
{
    use_syslog = enable;
}

void FiLogger::loggerUseStdout(bool enable)
{
    use_stdout = enable;
}

int FiLogger::loggerFile(const char *filename)
{
    int fd;

    if (filename != nullptr)
    {
        // Enabling output to a private log file.
        if (!logFileName.empty())
        {
            // Already open; not allowed to change the name
            return -1;
        }
        // Open (or create) file for writing
        if ((fd = open(filename, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP)) < 0)
        {
            return errno;
        }

        // Successfully opened file; hang on to a copy of the file name
        logFileName.assign(filename);
        logFileFd = fd;
    }
    else
    {
        // Turn off output to private log file
        fd = logFileFd;
        logFileFd = -1;
        logFileName.clear();

#if defined(_WIN32)
        FILE *fp = fdopen(fd, "a");
        fflush(fp);
        fclose(fp);
#else
        fsync(fd);
        close(fd);
#endif
    }

    // All is well.
    return 0;
}

void FiLogger::loggerReopen()
{
    int fd;

    // Close the old (presumably rename by logrotate) file descriptor
    fd = logFileFd;
    logFileFd = -1;

#if defined(_WIN32)
    FILE *fp = fdopen(fd, "a");
    fflush(fp);
    fclose(fp);
#else
    fsync(fd);
    close(fd);
#endif

    // Re-open (create) new log file using the previously configured name
    if ((fd = open(logFileName.c_str(), O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP)) < 0)
    {
        //! @todo Invent some way to deal with errors here
        return;
    }

    logFileFd = fd;
}

void FiLogger::fi_log(unsigned severity, const char *fmt, ...)
{
    std::array<char, 1024> formatted;
    std::string msg;

    va_list args;
    va_start(args, fmt);
    vsnprintf(formatted.data(), formatted.size(), fmt, args);
    va_end(args);

    // Move to a string for ease of use.
    msg.assign(formatted.data());

#ifdef _WIN32
    (void)severity; // Unused.
#else
    if (use_syslog != 0)
    {
        syslog(severity, "%s", msg.c_str());
    }
#endif /* !_WIN32 */

    msg.append("\n");
    if (use_stdout != 0)
    {
        fwrite(msg.c_str(), msg.length(), 1, stdout);
        fflush(stdout);
    }

    //! @todo Deal with possible race here: loggerFile() and loggerReopen()
    if (logFileFd > 0)
    {
        write(logFileFd, msg.c_str(), (unsigned int)msg.length());
    }
}

} /* namespace FI_LOG */


