/*!
 * @file fi_rdi_motionampl.cpp
 * @brief Contains the implementation of the FiRDIMotionAmpl class.
 *
 * Copyright 2023 (c) RDI Technologies as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * RDI Technologies. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with RDI Technologies.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <string.h>
#include <stddef.h>
#include <stdint.h>
#include <errno.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>

#include <algorithm>
#include <execution>
#include <functional>
#include <vector>

#include "fi_rdi_motionampl.h"
#include "fi_log.h"

// Make things easier for ourselves.
using namespace FAS_RDI;

//--------------------------------------------------------------------
// Static declarations
//--------------------------------------------------------------------
const int FiRDIMotionAmpl::NUM_CYCLES_REQD = 3;

uint8_t FiRDIMotionAmpl::minus_floor(const uint8_t& lhs, const uint8_t& rhs)
{
    return ((lhs < rhs) ? 0 : (lhs - rhs));
}

uint8_t FiRDIMotionAmpl::plus_clip(const uint8_t& lhs, const uint8_t& rhs)
{
    uint32_t result = lhs + rhs;
    return ((result > 0xFF) ? 0xFF : result);
}

/*!
 * @brief Constructor for the FiRDIMotionAmpl image object.
 */
FiRDIMotionAmpl::FiRDIMotionAmpl() :
    ampl_factor(1.0),
    ref_window(1),
    width(0),
    height(0)
{
}

FiRDIMotionAmpl::~FiRDIMotionAmpl()
{
    // Clear out our buffers.
    reset();
}

void FiRDIMotionAmpl::addFrame(uint8_t *frame)
{
    std::unique_lock<std::mutex> locker(instance_mutex);

    // Sanity check.
    if (sum_frame.get() == nullptr)
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: unexpectedly null summer!", FI_FUNCTION);
        return;
    }

    // Do we need to remove the oldest frame?
    if (amplifyAvailable())
    {
        // Pop our oldest frame off the queue.
        std::unique_ptr<uint8_t> oldest_frame;
        oldest_frame.reset(ref_frame_queue.front());
        ref_frame_queue.pop();

        // Subtract it out of the summation frame.
        std::transform(std::execution::par, sum_frame->cbegin(), sum_frame->cend(), oldest_frame.get(), sum_frame->begin(), std::minus<uint32_t>());
    }

    // Add in the new frame.
    std::transform(std::execution::par, sum_frame->cbegin(), sum_frame->cend(), frame, sum_frame->begin(), std::plus<uint32_t>());

    // Push it onto our queue.
    ref_frame_queue.push(frame);
}

bool FiRDIMotionAmpl::amplifyAvailable()
{
    // True if we have enough frames for a reference frame.
//     FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: ref_window=%d, queue=%d", FI_FUNCTION, ref_window, ref_frame_queue.size());
    return ((ref_window > 0) && (ref_frame_queue.size() >= (size_t)(ref_window - 1)));
}

bool FiRDIMotionAmpl::amplifyFrame(uint8_t *ampFrame)
{
    // Defer the lock until after we get the reference frame.
    std::unique_lock<std::mutex> locker(instance_mutex, std::defer_lock);

    // Get our reference frame. Note that the getReference() method grabs the
    // lock, so we need to hold off grabbing the mutex until after the call.
    std::unique_ptr<uint8_t[]> ref_frame = std::unique_ptr<uint8_t[]>(new uint8_t[width * height]);;
    if (!getReference(ref_frame.get()))
    {
        FI_LOG::FiLogger::factory()->fi_log(LOG_DEBUG, "%s: Unable to get the reference frame!", FI_FUNCTION);
        return false;
    }

    // OK, time to lock things up.
    locker.lock();

    // We want to amplify the last frame (i.e., the tail or back).
    int index;
    uint8_t *original_frame = ref_frame_queue.back();

    // *INDENT-OFF*
#pragma omp parallel for
    // *INDENT-ON*
    for (index = 0; index < (width * height); ++index)
    {
        // Grab the original pixel value, so it's easier to work with.
        uint8_t orig_pixel = original_frame[index];

        // Amplify the pixel and store it.
        ampFrame[index] = plus_clip(orig_pixel, (minus_floor(orig_pixel, ref_frame[index]) * ampl_factor));
    }

    // All is well.
    return true;
}

bool FiRDIMotionAmpl::getReference(uint8_t *refFrame)
{
    std::unique_lock<std::mutex> locker(instance_mutex);

    // Sanity check.
    if (!amplifyAvailable())
        return false;

    // The reference frame is the average of the frames in the queue.
    std::transform(std::execution::par, sum_frame->cbegin(), sum_frame->cend(), refFrame, [this](auto & pixel) { return (uint8_t)(pixel / this->ref_window); });

    // All is well.
    return true;
}

void FiRDIMotionAmpl::reset()
{
    // Standard way to clear a standard container.
    while (!ref_frame_queue.empty())
    {
        uint8_t *old_frame = ref_frame_queue.front();
        ref_frame_queue.pop();
        delete old_frame;
    }

    // Reset our sum values.
    if (sum_frame.get() != nullptr)
        sum_frame->assign(sum_frame->size(), 0);
}

void FiRDIMotionAmpl::setParameters(double framerate, double minFreqOfInterest, int width, int height, double factor)
{
    // Save off the settings for later use.
    this->width = width;
    this->height = height;
    this->ampl_factor = factor;

    // Calculate our reference window, how many frames we want to average over.
    int samplesPerCycle = (int)(framerate / minFreqOfInterest);
    this->ref_window = ((samplesPerCycle == 0) ? 1 : (samplesPerCycle * NUM_CYCLES_REQD));

    // Allocate our summing vector.
    sum_frame = std::make_unique<std::vector<uint32_t>>(width * height);

    // Clear out the buffers.
    reset();
}
