/*!
 * @file fi_rdi_packed.cpp
 * @brief Contains the implementation of the FiRDIPacked class.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <string.h>
#include <stddef.h>
#include <stdint.h>
#include <errno.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef _WIN32
#else
#include <unistd.h>
#endif

#include "fi_rdi_packed.h"
#include "fi_log.h"

namespace FI_RDI_PACKED
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

#pragma pack(1)
//! The contents of the small header in front of each image.
typedef struct IMAGE_HDR
{
    uint32_t    imageNumber;        //!< The image number for this frame.
    uint64_t    timeSinceEpoch;     //!< The number of seconds since epoch for image creation.
    uint32_t    microSeconds;       //!< The number of microseconds for image creation.
} IMAGE_HDR;
#ifdef _WIN32
#pragma pack()
#else
#pragma pack(0)
#endif

} // End of FI_RDI_PACKED

// Make things easier for ourselves.
using namespace FI_RDI_PACKED;

/*!
 * @brief Constructor for the FiRDIPacked image object.
 */
FiRDIPacked::FiRDIPacked() :
    bitsPerPixel(0),
    packedDataSize(0),
    height(0),
    width(0)
{
}

/*!
 * @brief A public accessor to get the size of the image header.
 * @return The number of bytes in the image header.
 */
uint32_t FiRDIPacked::getHeaderSize()
{
    return sizeof(IMAGE_HDR);
}

/*!
 * @brief Retrieve the image header.
 * @return Pointer to the header bytes, or nullptr.
 */
const uint8_t *FiRDIPacked::getHeader()
{
    // Make sure we return a nullptr if the size of the header is 0.
    if (header.size() == 0)
        return nullptr;

    // Return the pointer to the header bytes.
    return header.data();
}

/*!
 * @brief Retrieve the byte stream of the image.
 * @return Pointer to the image bytes, or nullptr.
 */
const uint8_t *FiRDIPacked::getPackedData()
{
    return packedData.get();
}

/*!
 * @brief Create the image header.
 *
 * @param frameNumber       The frame number for this image.
 * @param timeSeconds       The time (in seconds) to associate with this image.
 * @param timeMicroseconds  The microsecond part of time to associate with this image.
 */
void FiRDIPacked::setHeader(uint32_t frameNumber, time_t timeSeconds, uint32_t timeMicroseconds)
{
    IMAGE_HDR hdr;

    // Initialize our structre.
    hdr.imageNumber = frameNumber;
    hdr.timeSinceEpoch = timeSeconds;
    hdr.microSeconds = timeMicroseconds;

    // Allocate our byte array.
    header.resize(sizeof(IMAGE_HDR));

    // Copy the data over.
    memcpy(header.data(), &hdr, sizeof(IMAGE_HDR));
}

/*!
 * @brief Store the image data (pixels) into the FiRDIPacked object.
 *
 * This loads the data from a TIFF image into an RDI packed formatted data stream.
 *
 * @note The imageBpp parameter is NOT the size of the TIFF data (which may be 8
 *       or 16) but is the actual bits-per-pixel of the image data (8, 10, or 12).
 *
 * @param imagePtr      Pointer to the image data pixels.
 * @param imageWidth    Number of pixels per row (width of the image).
 * @param imageHeight   Number of rows (height of the image).
 * @param imageBpp      The pixel size.
 */
void FiRDIPacked::setTiffImageData(const uint8_t *imagePtr, uint32_t imageWidth, uint32_t imageHeight, int imageBpp)
{
    int image_align_bytes;
    uint32_t column;
    uint32_t row;

    // Save the image information.
    bitsPerPixel = imageBpp;
    height = imageHeight;
    width = imageWidth;

    // Are we dealing with an 8 or 16 bit TIFF image?
    if (imageBpp == 8)
    {
        uint8_t *packed_ptr;

        // Allocate our image stream.
        packedDataSize = (imageWidth * imageHeight);
        packedData = std::unique_ptr<uint8_t[]>(new uint8_t[packedDataSize]());
        if (packedData.get() == nullptr)
        {
            // ALM_TODO: SQUAWK! Error!
            return;
        }

        // Get a pointer to our destination.
        packed_ptr = packedData.get();

        // Figure out how much padding we need, if any.
        if ((width & 3) != 0)
            image_align_bytes = 4 - (width & 3);
        else
            image_align_bytes = 0;

        // *INDENT-OFF*
#pragma omp parallel for collapse(2)
        // *INDENT-ON*
        // Loop through each row of the data.
        for (row = 0; row < height; ++row)
        {
            // Loop through each column of the data.
            for (column = 0; column < width; ++column)
            {
                // The packed data has no padding per row, where as the source may.
                const int dest_offset = (int)(row * width) + (int)(column);
                const int src_offset = (int)(row * (width + image_align_bytes)) + (int)(column);

                // Copy the data.
                packed_ptr[dest_offset] = imagePtr[src_offset];
            }
        }
    }
    else
    {
        uint8_t *packed_ptr;

        // Allocate our image stream. We are packing 2 pixels per 3 bytes.
        packedDataSize = (imageWidth * imageHeight * 3) / 2;
        packedData = std::unique_ptr<uint8_t[]>(new uint8_t[packedDataSize]());
        if (packedData.get() == nullptr)
        {
            // ALM_TODO: SQUAWK! Error!
            return;
        }

        // Get a pointer to our destination.
        packed_ptr = packedData.get();

        // In 10- or 12-bit mode, since we expand to 16-bit pixels, an alignment can only
        // be two bytes.
        if (((width * 2) & 3) != 0)
            image_align_bytes = 2;
        else
            image_align_bytes = 0;

        // *INDENT-OFF*
#pragma omp parallel for collapse(2)
        // *INDENT-ON*
        // Loop through each row of the data.
        for (row = 0; row < height; ++row)
        {
            // Loop through each column of the data. Due to the use of openmp, we really want
            // a single increment per loop, so that means the width needs to be half because
            // each loop consumes 2 pixels.
            for (column = 0; column < (width / 2); ++column)
            {
                uint16_t pixel_value;
                uint8_t packed[3];

                // The packed data has no padding per row, where as the source may, specified by image_align_bytes.
                // In calculating the source offset, we are calculating the byte offset to the source pixel. Each
                // source pixel is 2 bytes, but the "width" value is in pixels, so we need to adjust. Also, the
                // column is incrementing by one, but that is one per every 2 pixels, so we need to adjust for that
                // as well.
                const int dest_offset = (((int)(row * width) + (int)(column * 2)) * 3) / 2;
                const int src_offset = (int)(row * ((width * 2) + image_align_bytes)) + (int)((column * 2) * 2);

                // Get the first pixel value.
                memcpy(&pixel_value, &imagePtr[src_offset], 2);

                // Shift down as appropriate.
#ifdef ALM_TODO_PACKED
                // While this is ACCURATE, the RDI software doesn't know what to do
                // with 10bpp images. When displayed, they are too dark. So the expedient
                // fix is to gain the 10bpp pixels to 12bpp.
                pixel_value >>= (16 - imageBpp);
#else
                pixel_value >>= 4;
#endif /* ALM_TODO_PACKED */

                // Pack the data. First pack data is the MS byte of pixel data.
                packed[0] = (uint8_t)((pixel_value >> 4) & 0xFF);

                // Second pack data contains the LS nibble.
                packed[1] = (uint8_t)(pixel_value & 0xF);

                // Get the second pixel value.
                memcpy(&pixel_value, &imagePtr[src_offset + 2], 2);

                // Shift down as appropriate.
#ifdef ALM_TODO_PACKED
                // While this is ACCURATE, the RDI software doesn't know what to do
                // with 10bpp images. When displayed, they are too dark. So the expedient
                // fix is to gain the 10bpp pixels to 12bpp.
                pixel_value >>= (16 - imageBpp);
#else
                pixel_value >>= 4;
#endif /* ALM_TODO_PACKED */

                // Pack the data. 3rd pack data is the MS byte of pixel data.
                packed[2] = (uint8_t)((pixel_value >> 4) & 0xFF);

                // Second pack data contains the LS nibble combined with the previous value.
                packed[1] |= (uint8_t)((pixel_value & 0xF) << 4);

                // Save the data.
                memcpy(&packed_ptr[dest_offset], packed, 3);
            }
        }
    }
}

/*!
 * @brief Store the packed image data into the FiRDIPacked object.
 *
 * This loads the data from a pre-packed image into our object. The assumption
 * is that this is a fully formed (with header) packed image.
 *
 * @param dataPtr           Pointer to the packed data stream.
 * @param dataLength        The number of bytes in the packed data stream.
 * @param imageWidth        Number of pixels per row (width) of the original image.
 * @param imageHeight       Number of rows (height) of the original image.
 * @param imageBpp          The original pixel size.
 */
void FiRDIPacked::setPackedData(const uint8_t *dataPtr, uint32_t dataLength, uint32_t imageWidth, uint32_t imageHeight, int imageBpp)
{
    // Save the image information.
    bitsPerPixel = imageBpp;
    height = imageHeight;
    width = imageWidth;
    packedDataSize = dataLength - getHeaderSize();

    // Allocate space for our header.
    header.resize(getHeaderSize());

    // Allocate our local memory. As stated, the assumption is that the
    // data (and length) include the header and packed image data.
    packedData = std::unique_ptr<uint8_t[]>(new uint8_t[packedDataSize]());
    if (packedData.get() == nullptr)
    {
        // ALM_TODO: SQUAWK! Error!
        return;
    }

    // If we got a data pointer, copy it over. Otherwise, if it was a nullptr, just zero out the contents.
    if (dataPtr == nullptr)
    {
        memset(header.data(), 0, getHeaderSize());
        memset(packedData.get(), 0, packedDataSize);
    }
    else
    {
        memcpy(header.data(), dataPtr, getHeaderSize());
        memcpy(packedData.get(), &dataPtr[getHeaderSize()], packedDataSize);
    }
}
