/*!
 * @file log.cpp
 * @brief Generic logging functions for debug or production code.
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef _WIN32
#include <io.h>
#else
#include <unistd.h>
#include <syslog.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <fmt/format.h>
#include <fmt/printf.h>

#include <chrono>
#include <ctime>
#include <iomanip>
#include <iostream>
#include <memory>
#include <sstream>
#include <string>

#include "fas_log.h"

#if defined(_WIN32) && !defined(__MINGW32__)
// Turn off the "This function may be unsafe" warning.
#  pragma warning( disable: 4996 )
#endif

//--------------------------------------------------------------------
// Global Variable declarations
//--------------------------------------------------------------------

// PDEBUG verbosity flags.
unsigned verbosity = 0;

namespace FI_LOG
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

#if defined(_WIN32) && !defined(__MINGW32__)
#  define open      _open
#  define snprintf  _snprintf
#  define write     _write

#  define S_IRUSR   _S_IREAD
#  define S_IWUSR   _S_IWRITE
#  define S_IRGRP   _S_IREAD
#  define S_IWGRP   _S_IWRITE
#endif

//--------------------------------------------------------------------
// Variable declarations
//--------------------------------------------------------------------

// Flag to indicate messages from debug_log() are be displayed.
int debug_to_log = 0;

// Flag to indicate if the syslog facility should be used.
#ifdef DEBUG
int use_syslog = 0;
#else
int use_syslog = 1;
#endif

// Flag to indicate if log messages are to be sent to STDOUT.
int use_stdout = 1;

// Private log file name.
std::string logFileName;

// Flag to indicate if log messages are to be saved to a file.
int logFileFd = -1;

} /* namespace FI_LOG */

using namespace FI_LOG;

//============================================================================
// Implementation
//============================================================================

/**
 * @brief Output a message logged via PDEBUG.
 */
void logMsg(unsigned severity, std::string msg)
{
#ifdef _WIN32
    (void)severity; // Unused.
#else
    if (use_syslog != 0)
    {
        syslog(severity, "%s", msg.c_str());
    }
#endif /* !_WIN32 */

    msg.append("\n");
    if (use_stdout != 0)
    {
        fwrite(msg.c_str(), msg.length(), 1, stdout);
        fflush(stdout);
    }

    //! @todo Deal with possible race here: loggerFile() and loggerReopen()
    if (logFileFd > 0)
    {
        write(logFileFd, msg.c_str(), (unsigned int)msg.length());
    }
}

/*
 * @brief Output a message for a panic-y situation.
 */
void fi_error_log_unlikely(const char *func, int line, int errno_value)
{
    fiLog("ERROR#%d at %s#%d: %s\n", errno_value, func, line, strerror(errno_value));
}

void logTagMsg(std::string file, int line, std::string msg)
{
    std::string output_msg;
    std::stringstream ss;

    // Get a timepoint for now.
    auto now = std::chrono::system_clock::now();

    // Get the time in time_t format for use in the localtime call.
    auto now_time_t = std::chrono::system_clock::to_time_t(now);

    // Get the seconds to subtract from the total nanoseconds. This will
    // give us the number of nanoseconds since the second mark.
    auto secs = std::chrono::time_point_cast<std::chrono::seconds>(now);
    auto ns = std::chrono::time_point_cast<std::chrono::nanoseconds>(now) -
              std::chrono::time_point_cast<std::chrono::nanoseconds>(secs);

    // Build up our string.
    ss << std::put_time(std::localtime(&now_time_t), "%Y%m%d %H%M%S.");
    ss << std::setfill('0') << std::setw(6) << (int)(ns.count() / 1000) << ": ";

    output_msg = ss.str();
    output_msg.append(file);
    output_msg.append(fmt::format(", {0}: ", line));
    output_msg.append(msg);

    // Successful case: send fully formatted string to logger
    logMsg(LOG_DEBUG, output_msg);
}

// Logger control API
void loggerUseSyslog(int enable)
{
    use_syslog = enable;
}

void loggerUseStdout(int enable)
{
    use_stdout = enable;
}

int loggerFile(const char *filename)
{
    int fd;

    if (filename != nullptr)
    {
        // Enabling output to a private log file.
        if (!logFileName.empty())
        {
            // Already open; not allowed to change the name
            return -1;
        }
        // Open (or create) file for writing
        if ((fd = open(filename, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP)) < 0)
        {
            return errno;
        }

        // Successfully opened file; hang on to a copy of the file name
        logFileName.assign(filename);
        logFileFd = fd;
    }
    else
    {
        // Turn off output to private log file
        fd = logFileFd;
        logFileFd = -1;
        logFileName.clear();

#if defined(_WIN32)
        FILE *fp = fdopen(fd, "a");
        fflush(fp);
        fclose(fp);
#else
        fsync(fd);
        close(fd);
#endif
    }

    // All is well.
    return 0;
}

void loggerReopen()
{
    int fd;

    // Close the old (presumably rename by logrotate) file descriptor
    fd = logFileFd;
    logFileFd = -1;

#if defined(_WIN32)
    FILE *fp = fdopen(fd, "a");
    fflush(fp);
    fclose(fp);
#else
    fsync(fd);
    close(fd);
#endif

    // Re-open (create) new log file using the previously configured name
    if ((fd = open(logFileName.c_str(), O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP)) < 0)
    {
        //! @todo Invent some way to deal with errors here
        return;
    }

    logFileFd = fd;
}

