/*!
 * @file  fas_sharedmem.h
 * @brief Local Streaming Shared Memory interface
 *
 * Copyright 2023 (c) RDI Technologies as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of RDI
 * Technologies. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement with
 * RDI Technologies.
 */

#ifndef FAS_SHAREDMEM_H_
#define FAS_SHAREDMEM_H_
#pragma once

#include <stdint.h>
#include <time.h>

#ifdef __cplusplus
extern "C" {
#endif

// --------------------------------------------------------------------
// Constant and Type Definitions
//--------------------------------------------------------------------

//! Possible fault values.
typedef enum
{
    FSM_NO_FAULT = 0,                   //!< Fault is cleared, or no fault is present.

    // Server faults.
    FSM_SERVER_BASE_CHANGE,             //!< The base memory information has changed (size and/or address).
    FSM_SERVER_COUT_CHANGE,             //!< The descriptor count has changed.

    // Client faults.
    FSM_CLIENT_NOT_CONNECTED            //!< The client fault field is initialized to this. Client must change to FSM_NO_FAULT.

} FSM_FAULTS;

//! An opaque data type which holds the reference to the control structure.
typedef void *FSM_HANDLE;

/*!
 * @brief Add DLL spcs for Windows builds.
 *
 * This marcro adds DLL export/import specifications needed for Windows builds.
 * It evaluates to an empty string on other patforms.
 */
#if defined(WIN32) && defined(FI_DLL)
#  ifdef FI_EXPORT
#    define FI_API      __declspec(dllexport)
#  else
#    define FI_API      __declspec(dllimport)
#  endif // FI_EXPORT
#else
#  define FI_API
#endif // WIN32 && FI_DLL

//! @name Functions for accessing buffer information.
//! @{

//! @name Functions for writer, which is the server.
//! @{
extern FI_API bool fsmIsFull(FSM_HANDLE hSharedMem);
extern FI_API bool fsmIsIndexAvailable(FSM_HANDLE hSharedMem, uint32_t check_index);
extern FI_API void fsmSetFrameAddress(FSM_HANDLE hSharedMem, uint64_t im_addr);
extern FI_API void fsmSetFrameNum(FSM_HANDLE hSharedMem, int32_t frame_num);
extern FI_API void fsmSetFrameSize(FSM_HANDLE hSharedMem, uint32_t frame_size);
extern FI_API void fsmSetInfo(FSM_HANDLE hSharedMem, uint64_t im_addr, uint32_t frame_size, int32_t frame_num);
extern FI_API uint32_t fsmWriteDone(FSM_HANDLE hSharedMem);
//
//! @}

//! @name Functions for reader, which is the client.
//! @{
extern FI_API bool fsmIsEmpty(FSM_HANDLE hSharedMem);
extern FI_API uint64_t fsmGetBaseAddr(FSM_HANDLE hSharedMem);
extern FI_API uint32_t fsmGetDescCount(FSM_HANDLE hSharedMem);
extern FI_API int32_t  fsmGetFrameNum(FSM_HANDLE hSharedMem);
extern FI_API uint32_t fsmGetFrameSize(FSM_HANDLE hSharedMem);
extern FI_API uint8_t *fsmGetImage(FSM_HANDLE hSharedMem);
extern FI_API void fsmReadDone(FSM_HANDLE hSharedMem);
//
//! @}

//! Check server fault status.
extern FI_API int32_t fsmGetServerFault(FSM_HANDLE hSharedMem);

//! Check client fault status.
extern FI_API int32_t fsmGetClientFault(FSM_HANDLE hSharedMem);
//
//! @}

//! @name Functions for server only.
//! @{
extern FI_API FSM_HANDLE fsmServerInit(uint64_t base_addr, uint64_t base_len, uint32_t count);
extern FI_API int  fsmServerReset(FSM_HANDLE hSharedMem);
extern FI_API void fsmServerCleanup(FSM_HANDLE hSharedMem);
extern FI_API int  fsmServerSignal(FSM_HANDLE hSharedMem);
extern FI_API int  fsmServerWait(FSM_HANDLE hSharedMem, uint64_t timeoutPeriod);
extern FI_API int  fsmServerWaitUntil(FSM_HANDLE hSharedMem, const timespec *absTimeout);
extern FI_API int  fsmServerClearFault(FSM_HANDLE hSharedMem);
extern FI_API int  fsmServerSetFault(FSM_HANDLE hSharedMem, int32_t fault);
extern FI_API void fsmServerUpdateMemInfo(FSM_HANDLE hSharedMem, uint64_t base_addr, uint64_t base_len);
//
//! @}

//! @name Functions for client only.
//! @{
extern FI_API FSM_HANDLE fsmClientInit();
extern FI_API void fsmClientCleanup(FSM_HANDLE hSharedMem);
extern FI_API int  fsmClientSignal(FSM_HANDLE hSharedMem);
extern FI_API int  fsmClientWait(FSM_HANDLE hSharedMem, uint64_t timeoutPeriod);
extern FI_API int  fsmClientClearFault(FSM_HANDLE hSharedMem);
extern FI_API int  fsmClientSetFault(FSM_HANDLE hSharedMem, int32_t fault);
//
//! @}
//! @}

#ifdef __cplusplus
}
#endif
#endif /* FAS_SHAREDMEM_H_ */

