/*!
 * @file fas_sharedmem.c
 *
 * @brief Local Stream Shared Memory interfaces
 *
 * Copyright 2023 (c) RDI Technologies as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * RDI Technologies. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with RDI Technologies.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <inttypes.h>

#include <fcntl.h>
#include <pthread.h>
#include <signal.h>

#include <sys/mman.h>
#include <sys/stat.h>

#include <atomic>

// Windows or Linux?
#if _WIN32
//
// Windows
#include "fas_log.h"

#else
//
// Linux and MacOS
#include <fas_log.h>
#endif

#include "fas_sharedmem.h"

namespace FAS_SHAREDMEM
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Name of shared memory region.
#define LOCALSTREAM_SHM_NAME    "/fsm_localstream"

// --------------------------------------------------------------------
// Constant and Type Definitions
//--------------------------------------------------------------------

//! @brief Shared buffer descriptor.
typedef struct FSM_BUF_DESC
{
    uint64_t im_addr;       //!< Address in system memory.
    uint32_t frame_size;    //!< Size of processed image in bytes.
    int32_t  frame_num;     //!< Frame number (0 to framecount-1), negative illegal.
} FSM_BUF_DESC;

//! @brief The shared memory control structure.
//! An opaque type is used to access through the functional interfaces.
typedef struct FSM_CONTROL
{
    pthread_mutex_t stream_lock;    //!< Process-shared mutex.
    pthread_cond_t server_wakeup;   //!< Condition variable for signaling server.
    pthread_cond_t client_wakeup;   //!< Condition variable for signaling client.

    uint32_t server_fault;          //!< Fault indicator from server; client clears (non-zero => fault)
    uint32_t client_fault;          //!< Fault indicator from client; server clears (non-zero => fault)

    std::atomic_uint32_t rd_index;  //!< Index for the next buffer descriptor to be read.
    std::atomic_uint32_t wr_index;  //!< Index for the next buffer descriptor to be written.

    uint64_t im_base_addr;          //!< Physical address of the image memory buffer(s).
    uint64_t im_base_size;          //!< Number of bytes in the image memory space.

    uint32_t desc_count;            //!< The number of descriptors in the 'buf_desc' array.
    FSM_BUF_DESC buf_desc[1];       //!< The ring buffer of buffer descriptors.
} FSM_CONTROL;

// --------------------------------------------------------------------
// Local variable declarations
//--------------------------------------------------------------------

//! @name Client specific variables
//! @{
static int im_mmap_fd;              //!< File descriptor for the image memory mmap.
static uint8_t *im_base_ptr;        //!< The base memory pointer for the image memory space.
//
//! @}

// --------------------------------------------------------------------
// Local utility functions
//--------------------------------------------------------------------

/*!
 * @brief Take the shared memory mutex. This is a blocking call.
 *
 * @param hSharedMem    [IN] Either client's or server's shared memory handle.
 * @returns non-zero for error, 0 for success.
 */
static int lockControl(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    int rc;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // Try to lock the control block.
    rc = pthread_mutex_trylock(&control_ptr->stream_lock);
    if (rc != 0)
    {
        // If the error is because the mutex is busy, go ahead and block.
        if (rc == EBUSY)
        {
            // Block on the semaphore.
            rc = pthread_mutex_lock(&control_ptr->stream_lock);
        }
    }

    // See if we got the lock.
    if (rc != 0)
    {
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Same as lockControl(), but with a timeout.
 *
 * @note If absTimeout is a nullptr, then the behavior defaults to lockControl().
 *
 * @param hSharedMem    [IN] Either client's or server's shared memory handle.
 * @param absTimeout    [IN] The timeout shall expire when the absolute time specified by absTimeout passes.
 *
 * @returns Non-zero for error or timeout (ETIMEDOUT), 0 for success.
 */
static int lockControlTimed(FSM_HANDLE hSharedMem, const struct timespec *absTimeout)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    int rc;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // Try locking the mutex to see if we get lucky.
    rc = pthread_mutex_trylock(&control_ptr->stream_lock);
    if (rc != 0)
    {
        // Is the mutex busy (i.e., already locked)?
        if (rc == EBUSY)
        {
            // Use the appropriate mechanism to wait.
            if (absTimeout == nullptr)
                rc = pthread_mutex_lock(&control_ptr->stream_lock);
            else
                rc = pthread_mutex_timedlock(&control_ptr->stream_lock, absTimeout);
        }
    }

    // See if we got the lock.
    if (rc != 0)
    {
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));
    }

    return rc;
}

/*!
 * @brief Release the shared memory mutex.
 *
 * @param hSharedMem    [IN] Either client's or server's shared memory handle.
 *
 * @returns non-zero for error, 0 for success.
 */
static int unlockControl(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    int rc;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    rc = pthread_mutex_unlock(&control_ptr->stream_lock);
    if (rc < 0)
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));

    return rc;
}


/*!
 * @brief Initialize a process-sharable mutex.
 *
 * @param pm    [IN] Pointer to mutex to initialize.
 * @return Zero for success, non-zero for failure.
 */
static int setupSharedMutex(pthread_mutex_t *pm)
{
    pthread_mutexattr_t mutex_attr;
    int rc;

    // Sanity check.
    if (pm == nullptr)
    {
        fi_log(LOG_ERR, "%s: mutex pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    /* Initialize mutex attribute variable. */
    rc = pthread_mutexattr_init(&mutex_attr);
    if (rc != 0)
    {
        fi_log(LOG_ERR, "%s: pthread_mutexattr_init() failed, rc=%d!", FI_FUNCTION, rc);
        return rc;
    }

    // We do not want recursive for mutex used with condition variable.
    rc = pthread_mutexattr_settype(&mutex_attr, PTHREAD_MUTEX_NORMAL);
    if (rc != 0)
    {
        fi_log(LOG_ERR, "%s: pthread_mutexattr_settype() failed, rc=%d!", FI_FUNCTION, rc);
        goto bail;
    }

// ROBUST option is not available the way we are compiled.
//  rc = pthread_mutexattr_setrobust( &mutex_attr, PTHREAD_MUTEX_ROBUST );
//  if ( rc < 0 )
//      goto bail;

    rc = pthread_mutexattr_setpshared(&mutex_attr, PTHREAD_PROCESS_SHARED);
    if (rc != 0)
    {
        fi_log(LOG_ERR, "%s: pthread_mutexattr_setpshared() failed, rc=%d!", FI_FUNCTION, rc);
        goto bail;
    }

    // Create the shared memory mutex in our shared memory region.
    // We will use this same mutex for the server wakeup condition.
    rc = pthread_mutex_init(pm, &mutex_attr);
    if (rc != 0)
    {
        fi_log(LOG_ERR, "%s: pthread_mutex_init() failed, rc=%d!", FI_FUNCTION, rc);
        goto bail;
    }

    rc = 0;

bail:
    pthread_mutexattr_destroy(&mutex_attr);
    return rc;
}


/*!
 * @brief Initialize a process-sharable condition variable.
 *
 * @param pc    [IN] Pointer to condition variable to initialize.
 *
 * @return Zero for success, non-zero for failure.
 */
static int setupSharedCond(pthread_cond_t *pc)
{
    pthread_condattr_t cond_attr;
    int rc;

    // Sanity check.
    if (pc == nullptr)
    {
        fi_log(LOG_ERR, "%s: condition pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    rc = pthread_condattr_init(&cond_attr);
    if (rc != 0)
    {
        fi_log(LOG_ERR, "%s: pthread_condattr_init() failed, rc=%d!", FI_FUNCTION, rc);
        return rc;
    }

    // Set up attributes for condition variable.
    rc = pthread_condattr_setpshared(&cond_attr, PTHREAD_PROCESS_SHARED);
    if (rc != 0)
    {
        fi_log(LOG_ERR, "%s: pthread_condattr_setpshared() failed, rc=%d!", FI_FUNCTION, rc);
        goto bail;
    }

    // Create server's condition variable in our shared memory region.
    rc = pthread_cond_init(pc, &cond_attr);
    if (rc != 0)
    {
        fi_log(LOG_ERR, "%s: pthread_cond_init() failed, rc=%d!", FI_FUNCTION, rc);
        goto bail;
    }

    rc = 0;

bail:
    pthread_condattr_destroy(&cond_attr);
    return rc;
}

/*!
 * @brief Initialize the Local Stream Shared Memory variables.
 *
 * @param control_ptr   [IN] Pointer to state structure in shared memory
 * @param desc_count    [IN] The number of descriptors to allocate.
 *
 * @returns 0 for success, non-zero for error.
 */
static int sharedMemInit(FSM_CONTROL *control_ptr, uint64_t base_addr, uint64_t base_len, int desc_count)
{
    int rc;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        rc = EINVAL;
        goto bail;
    }

    // Initialize our mutex.
    rc = setupSharedMutex(&control_ptr->stream_lock);
    if (rc != 0)
        goto bail;

    // Initialize the server's condition variable.
    rc = setupSharedCond(&control_ptr->server_wakeup);
    if (rc != 0)
    {
        pthread_mutex_destroy(&control_ptr->stream_lock);
        goto bail;
    }

    // Initialize the client's condition variable.
    rc = setupSharedCond(&control_ptr->client_wakeup);
    if (rc != 0)
    {
        pthread_cond_destroy(&control_ptr->server_wakeup);
        pthread_mutex_destroy(&control_ptr->stream_lock);
        goto bail;
    }

    // Reset the fault indicators.
    control_ptr->server_fault = FSM_NO_FAULT;
    control_ptr->client_fault = FSM_CLIENT_NOT_CONNECTED;

    // Reset our indicies.
    control_ptr->rd_index.store(0);
    control_ptr->wr_index.store(0);

    // Initialize our base information.
    control_ptr->im_base_addr = base_addr;
    control_ptr->im_base_size = base_len;

    // Zero out the descriptors.
    control_ptr->desc_count = desc_count;
    memset(&control_ptr->buf_desc, 0, sizeof(FSM_BUF_DESC) * desc_count);

    // All is well.
    rc = 0;

bail:
    // Log any errors.
    if (rc != 0)
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));

    // Return our result.
    return rc;
}

/*!
 * @brief Calculate the size of the control structure.
 *
 * @param count     [IN] The number of descriptors to allocate.
 * @return The number of bytes for the control structure.
 */
static int getControlSize(uint32_t count)
{
    int control_size;
    int page_size = sysconf(_SC_PAGESIZE);

    // Sanity check.
    if (count < 1)
        count = 1;

    // How big is the control structure we want to allocate. The size of the FSM_CONTROL
    // structure already includes one descriptor, so decrement the count by one.
    control_size = sizeof(FSM_CONTROL) + (sizeof(FSM_BUF_DESC) * (count - 1));

    // Make sure the size is a multiple of the system page size.
    control_size = ((control_size + page_size - 1) / page_size) * page_size;

    // Return our size.
    return control_size;
}

} /* End of namespace FAS_SHAREDMEM */

using namespace FAS_SHAREDMEM;

// --------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

/*!
 * @brief Wake up the server using the condition variable in shared memory.
 *
 * This needs to be a separate function from the lock/unlock to support proper
 * mutual exclusion when signalling:
 *
 *   lock
 *   set up memory content, e.g. add message to queue or return a buffer
 *   signal condition
 *   unlock
 *
 * @param hSharedMem    [IN] Server's handle to the control structure.
 * @returns Non-zero for error, 0 for success.
 */
int fsmServerSignal(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    int rc;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // Send the signal.
    rc = pthread_cond_signal(&control_ptr->server_wakeup);
    if (rc < 0)
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));

    return rc;
}

/*!
 * @brief This is used by the server to clear the fault code set by the client.
 *
 * After clearing the client's fault code, the client is signaled to wake up.
 *
 * @param hSharedMem    [IN] Server's handle to the control structure.
 * @returns Non-zero for error, 0 for success.
 */
int fsmServerClearFault(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    int rc;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // Clear the fault setting.
    control_ptr->client_fault = FSM_NO_FAULT;

    // Wake up the client. We do NOT need to own the mutex in order to call the
    // signaling method.
    rc = pthread_cond_signal(&control_ptr->client_wakeup);
    if (rc != 0)
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));

    return rc;
}

/*!
 * @brief This is used by the server to set the server's fault code.
 *
 * After setting the fault code, the client is signaled to wake up.
 *
 * @param hSharedMem    [IN] Server's handle to the control structure.
 * @param fault         [IN] The fault code to set into the control structure.
 *
 * @returns Non-zero for error, 0 for success.
 */
int fsmServerSetFault(FSM_HANDLE hSharedMem, int32_t fault)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    int rc;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // fi_log(LOG_ERR, "%s: Send server fault (%d) indication to client", FI_FUNCTION, fault);
    control_ptr->server_fault = fault;

    // Wake up the client. We do NOT need to own the mutex in order to call the
    // signaling method.
    rc = pthread_cond_signal(&control_ptr->client_wakeup);
    if (rc != 0)
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));

    return rc;
}

/*!
 * @brief Server call to wait for event or timeout.
 *
 * @param hSharedMem    [IN] Server's handle to the control structure.
 * @param absTimeout    [IN] The timeout shall expire when the absolute time specified by absTimeout passes.
 *
 * @returns Non-zero for error or timeout (ETIMEDOUT), 0 for success.
 */
int fsmServerWaitUntil(FSM_HANDLE hSharedMem, const struct timespec *absTimeout)
{
    int rc;
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;

    // Sanity check.
    if (absTimeout == nullptr)
    {
        fi_log(LOG_ERR, "%s: wakeup time pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // Get ownership of the mutex.
    rc = lockControlTimed(hSharedMem, absTimeout);
    if (rc == 0)
    {
        // Wait for awhile.
        rc = pthread_cond_timedwait(&control_ptr->server_wakeup, &control_ptr->stream_lock, absTimeout);

        // If we were successfull or not, we need to release the mutex.
        unlockControl(hSharedMem);
    }
    if ((rc != 0) && (rc != ETIMEDOUT))
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(rc));

    return rc;
}

/*!
 * @brief Server call to wait for event or timeout
 *
 * @param hSharedMem    [IN] Server's handle to the control structure.
 * @param timeoutPeriod [IN] The timeout in microseconds. A '0' value means to wait indefinitely.
 *
 * @returns Non-zero for error or timeout (ETIMEDOUT), 0 for success.
 */
int fsmServerWait(FSM_HANDLE hSharedMem, uint64_t timeoutPeriod)
{
    int rc;
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // Are we waiting for a maximum period of time, or forever?
    if (timeoutPeriod == 0)
    {
        // Get ownership of the mutex.
        rc = lockControl(hSharedMem);
        if (rc == 0)
        {
            // Wait for a response.
            rc = pthread_cond_wait(&control_ptr->server_wakeup, &control_ptr->stream_lock);

            // If we were successfull or not, we need to release the mutex.
            unlockControl(hSharedMem);
        }
    }
    else
    {
        struct timespec ts;

        // Build up our time structure.
        clock_gettime(CLOCK_REALTIME, &ts);
        ts.tv_nsec += ((timeoutPeriod % 1000000) * 1000);
        while (ts.tv_nsec >= 1000 * 1000 * 1000)
        {
            ts.tv_nsec -= 1000 * 1000 * 1000;
            ts.tv_sec += 1;
        }
        ts.tv_sec += (timeoutPeriod / 1000000);

        // Wait for a maximum period of time.
        rc = fsmServerWaitUntil(hSharedMem, &ts);
    }

    if ((rc != 0) && (rc != ETIMEDOUT))
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));

    return rc;
}

int32_t fsmGetServerFault(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return 0;
    }

    // This should rarely change, so it should be safe to simply read it.
    return control_ptr->server_fault;
}

int32_t fsmGetClientFault(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return 0;
    }

    // This should rarely change, so it should be safe to simply read it.
    return control_ptr->client_fault;
}

/*!
 * @brief Check to see if the ring buffer is empty.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @returns True if the ring buffer is empty, false otherwise.
 */
bool fsmIsEmpty(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    int rd_index;
    int wr_index;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return true;
    }

    // Get our indicies.
    rd_index = control_ptr->rd_index.load();
    wr_index = control_ptr->wr_index.load();

    // Check for an empty condition.
    return (wr_index == rd_index);
}

/*!
 * @brief Check to see if the ring buffer is full.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @returns True if the ring buffer is full, false otherwise.
 */
bool fsmIsFull(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    int rc;
    int rd_index;
    int wr_index;
    int num_bufs;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return true;
    }

    // Get our indicies.
    rd_index = control_ptr->rd_index.load();
    wr_index = control_ptr->wr_index.load();

    // This rarely changes, so it should be safe to simply read it.
    num_bufs = control_ptr->desc_count;

    // Calculate the next write index.
    rc = (wr_index + 1) % num_bufs;

    // Check for a full condition.
    return (rc == rd_index);
}

/*!
 * @brief Check to see if the provided 'check_index' is available to
 *        be written into.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @param check_index   [IN] Index to check.
 * @return True if the index can be written to, false otherwise.
 */
bool fsmIsIndexAvailable(FSM_HANDLE hSharedMem, uint32_t check_index)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    uint32_t rd_index;
    uint32_t wr_index;
    bool rc;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return false;
    }
    if (check_index >= control_ptr->desc_count)
    {
        fi_log(LOG_ERR, "%s: bad check index=%d!", FI_FUNCTION, check_index);
        return false;
    }

    // Get the indicies from the control block.
    rd_index = control_ptr->rd_index.load();
    wr_index = control_ptr->wr_index.load();

    // Sanity check.
    if (rd_index >= control_ptr->desc_count)
    {
        fi_log(LOG_ERR, "%s: bad read index=%d!", FI_FUNCTION, rd_index);
        return false;
    }
    if (wr_index >= control_ptr->desc_count)
    {
        fi_log(LOG_ERR, "%s: bad write index=%d!", FI_FUNCTION, rd_index);
        return false;
    }

    // Is the index in the "available-to-be-written" section of the ring?
    // Drawing a picture helps, if you are trying to follow along...
    if (wr_index > rd_index)
    {
        rc = ((check_index >= wr_index) || (check_index < rd_index));
    }
    else if (wr_index < rd_index)
    {
        rc = ((check_index >= wr_index) && (check_index < rd_index));
    }
    else
    {
        // The ring is empty.
        rc = true;
    }
    return rc;
}

/*!
 * @brief Get the physical base memory address corresponding to the buffer
 *        available to the reader.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @returns Base memory address for the buffer.
 */
uint64_t fsmGetBaseAddr(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    uint64_t im_addr(0);
    uint32_t rd_index;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        goto fsm_ba_bail;
    }

    // Make sure we aren't empty.
    if (fsmIsEmpty(hSharedMem))
    {
        fi_log(LOG_DEBUG, "%s: shared memory ring is empty!", FI_FUNCTION);
        goto fsm_ba_bail;
    }

    // Get our read index.
    rd_index = control_ptr->rd_index.load();

    // Sanity check.
    if (rd_index >= control_ptr->desc_count)
    {
        fi_log(LOG_ERR, "%s: bad read index=%d!", FI_FUNCTION, rd_index);
        goto fsm_ba_bail;
    }

    // Compute the address for the specified buffer.
    im_addr = control_ptr->buf_desc[rd_index].im_addr;

fsm_ba_bail:
    // Return our address.
    return im_addr;
}

/*!
 * @brief Get the image memory address corresponding to the buffer
 *        available to the reader.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @returns Pointer to image memory, or nullptr on error.
 */
uint8_t *fsmGetImage(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    uint64_t im_addr;
    uint64_t base_addr;
    uint32_t rd_index;
    uint8_t *image_ptr = nullptr;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return nullptr;
    }

    // Make sure we aren't empty.
    if (fsmIsEmpty(hSharedMem))
    {
        fi_log(LOG_DEBUG, "%s: shared memory ring is empty!", FI_FUNCTION);
        return nullptr;
    }

    // Get our read index.
    rd_index = control_ptr->rd_index.load();

    // Sanity check.
    if (rd_index >= control_ptr->desc_count)
    {
        fi_log(LOG_ERR, "%s: bad read index=%d!", FI_FUNCTION, rd_index);
        return nullptr;
    }

    // Compute the address for the specified buffer.
    base_addr = control_ptr->im_base_addr;
    im_addr = control_ptr->buf_desc[rd_index].im_addr;

    // Sanity check.
    if ((im_addr - base_addr) > control_ptr->im_base_size)
    {
        fi_log(LOG_ERR, "%s: Invalid image address: 0x%" PRIx64, FI_FUNCTION, im_addr);
        return nullptr;
    }

    // Check to see if we need to map the image memory.
    if ((im_mmap_fd < 0) && (((FSM_CONTROL *)control_ptr)->im_base_size > 0))
    {
        // Memory map the image memory address.
        im_mmap_fd = open("/dev/mem", O_RDWR);
        if (im_mmap_fd < 0)
        {
            LOG_UNLIKELY_ERRNO(FI_FUNCTION, __LINE__, errno);
            return nullptr;
        }

        // Get a pointer to our buffer.
        im_base_ptr = (uint8_t *)mmap(NULL, ((FSM_CONTROL *)control_ptr)->im_base_size, PROT_READ | PROT_WRITE, MAP_SHARED, im_mmap_fd, ((FSM_CONTROL *)control_ptr)->im_base_addr);
        if (im_base_ptr == (uint8_t *)MAP_FAILED)
        {
            fi_log(LOG_ERR, "%s: Unable to map the image memory space.", FI_FUNCTION);

            // Close the file descriptor.
            close(im_mmap_fd);
            im_mmap_fd = -1;
            return nullptr;
        }
    }

    // Create a pointer.
    image_ptr = im_base_ptr + (im_addr - base_addr);

    // Return our address.
    return image_ptr;
}

/*!
 * @brief Set the physical image memory address corresponding to the image
 *        to be transferred.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @param im_addr       [IN] The physical image memory address of the frame being transferred.
 */
void fsmSetFrameAddress(FSM_HANDLE hSharedMem, uint64_t im_addr)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    uint32_t wr_index;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return;
    }

    // Make sure we aren't full.
    if (fsmIsFull(hSharedMem))
    {
        fi_log(LOG_DEBUG, "%s: shared memory ring is full!", FI_FUNCTION);
        return;
    }

    // Get our write index.
    wr_index = control_ptr->wr_index.load();

    // Sanity check.
    if (wr_index >= control_ptr->desc_count)
    {
        fi_log(LOG_ERR, "%s: bad write index=%d!", FI_FUNCTION, wr_index);
    }
    else
    {
        // Set the address for the specified buffer.
        control_ptr->buf_desc[wr_index].im_addr = im_addr;
    }
}

/*!
 * @brief Set the information corresponding to the image to be transferred.
 *
 * This combines the effect of fsmSetFrameAddress(), fsmSetFrameSize(), and fsmSetFrameNum().
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @param im_addr       [IN] The physical image memory address of the frame being transferred.
 * @param frame_size    [IN] The size to associate with the buffer.
 * @param frame_num     [IN] The frame number to associate with the buffer.
 */
void fsmSetInfo(FSM_HANDLE hSharedMem, uint64_t im_addr, uint32_t frame_size, int32_t frame_num)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    uint32_t wr_index;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return;
    }

    // Make sure we aren't full.
    if (fsmIsFull(hSharedMem))
    {
        fi_log(LOG_DEBUG, "%s: shared memory ring is full!", FI_FUNCTION);
        return;
    }

    // Get our write index.
    wr_index = control_ptr->wr_index.load();

    // Sanity check.
    if (wr_index >= control_ptr->desc_count)
    {
        fi_log(LOG_ERR, "%s: bad write index=%d!", FI_FUNCTION, wr_index);
        return;
    }

    // Set the information into our descriptor.
    control_ptr->buf_desc[wr_index].im_addr = im_addr;
    control_ptr->buf_desc[wr_index].frame_size = frame_size;
    control_ptr->buf_desc[wr_index].frame_num = frame_num;
}

/*!
 * @brief Get the number of descriptors the server has allocated.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @returns Number of descriptrs allocated.
 */
uint32_t fsmGetDescCount(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return -1;
    }

    // Return the count.
    return control_ptr->desc_count;
}

/*!
 * @brief Get the frame number corresponding to the buffer indexed by the control
 *        structure's rd_index.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @returns Frame number or negative for error.
 */
int32_t fsmGetFrameNum(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    uint32_t rd_index;
    uint32_t wr_index;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return -1;
    }

    // Get our read index.
    rd_index = control_ptr->rd_index.load();
    wr_index = control_ptr->wr_index.load();
    if (rd_index >= control_ptr->desc_count)
    {
        fi_log(LOG_ERR, "%s: bad rd_index=%d", FI_FUNCTION, rd_index);
        return -1;
    }
    if (rd_index == wr_index)
    {
        // The ring is empty.
        return -1;
    }

    // Return the frame number.
    return control_ptr->buf_desc[rd_index].frame_num;
}

/*!
 * @brief Set the frame number corresponding to the buffer indexed by the control
 *        structure's wr_index.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @param frame_num     [IN] The frame number to associate with the buffer.
 */
void fsmSetFrameNum(FSM_HANDLE hSharedMem, int32_t frame_num)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    uint32_t wr_index;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return;
    }

    // Make sure we aren't full.
    if (fsmIsFull(hSharedMem))
    {
        fi_log(LOG_DEBUG, "%s: shared memory ring is full!", FI_FUNCTION);
        return;
    }

    // Get our write index.
    wr_index = control_ptr->wr_index.load();
    if (wr_index >= control_ptr->desc_count)
    {
        fi_log(LOG_ERR, "%s: bad wr_index=%d.", FI_FUNCTION, wr_index);
    }
    else
    {
        control_ptr->buf_desc[wr_index].frame_num = frame_num;
    }
}

/*!
 * @brief Get the image size corresponding to the buffer indexed by the control
 *        structure's rd_index.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @returns Size in bytes or zero if error.
 */
uint32_t fsmGetFrameSize(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    uint32_t size = 0;
    uint32_t rd_index;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return 0;
    }

    // Get our read index.
    rd_index = control_ptr->rd_index.load();
    if (rd_index >= control_ptr->desc_count)
    {
        fi_log(LOG_ERR, "%s: bad rd_index=%d", FI_FUNCTION, rd_index);
    }
    else
    {
        size = control_ptr->buf_desc[rd_index].frame_size;
    }

    // Return our size.
    return size;
}

/*!
 * @brief Set the frame size corresponding to the buffer indexed by the control
 *        structure's wr_index.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @param frame_size    [IN] The size to associate with the buffer.
 */
void fsmSetFrameSize(FSM_HANDLE hSharedMem, uint32_t frame_size)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    uint32_t wr_index;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return;
    }

    // Make sure we aren't full.
    if (fsmIsFull(hSharedMem))
    {
        fi_log(LOG_DEBUG, "%s: shared memory ring is full!", FI_FUNCTION);
        return;
    }

    // Get our write index.
    wr_index = control_ptr->wr_index.load();
    if (wr_index >= control_ptr->desc_count)
    {
        fi_log(LOG_ERR, "%s: bad wr_index=%d.", FI_FUNCTION, wr_index);
    }
    else
    {
        control_ptr->buf_desc[wr_index].frame_size = frame_size;
    }
}

/*!
 * @brief The server may need to update the base address and length of the image memory
 *        being used.
 *
 * @param hSharedMem    [IN] The shared memory handle.
 * @param base_addr     [IN] The physical address of the image memory space.
 * @param base_len      [IN] The total number of bytes for the image memory space.
 */
void fsmServerUpdateMemInfo(FSM_HANDLE hSharedMem, uint64_t base_addr, uint64_t base_len)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return;
    }

    // Reset our base information.
    control_ptr->im_base_addr = base_addr;
    control_ptr->im_base_size = base_len;
}

/*!
 * @brief The writer is finished with the write buffer.
 *
 * When the server has finished writing to the buffer, the control structure's
 * wr_index is updated.
 *
 * @param hSharedMem  [IN] The shared memory handle.
 * @returns The index for the buffer just written, or (uint32_t)-1 for an error.
 */
uint32_t fsmWriteDone(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    uint32_t ret_index = (uint32_t) -1;
    uint32_t wr_index;
    bool wakeup_needed = false;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        goto fwd_bail;
    }

    // There's no need to go any further if there is no client connected.
    if (fsmGetClientFault(hSharedMem) == FSM_CLIENT_NOT_CONNECTED)
        goto fwd_bail;

    // Are we full?
    if (fsmIsFull(hSharedMem))
        goto fwd_bail;

    // Are we empty?
    wakeup_needed = fsmIsEmpty(hSharedMem);

    // Update our write index. The access of 'desc_count' should be safe, since
    // it doesn't update. We increment the index around the ring.
    ret_index = control_ptr->wr_index.load();
    wr_index = (ret_index + 1) % control_ptr->desc_count;
    control_ptr->wr_index.store(wr_index);

    // Wake up the client if we were empty.
    if (wakeup_needed)
        fsmClientSignal(hSharedMem);

fwd_bail:
    // Return the index.
    return ret_index;
}

/*!
 * @brief The reader has finished with a buffer and can now release
 *        it to the writer.
 *
 * @param hSharedMem  [IN] Server's handle to the control structure.
 */
void fsmReadDone(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    uint32_t rd_index;
    bool wakeup_needed = false;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return;
    }

    // Are we empty?
    if (fsmIsEmpty(hSharedMem))
        return;

    // Are we full?
    wakeup_needed = fsmIsFull(hSharedMem);

    // Update our read index. The access of 'desc_count' should be safe, since
    // it doesn't update. We increment the index around the ring.
    rd_index = (control_ptr->rd_index.load() + 1) % control_ptr->desc_count;
    control_ptr->rd_index.store(rd_index);

    // Wake up the server if we were full.
    if (wakeup_needed)
        fsmServerSignal(hSharedMem);
}

/*!
 * @brief Cleanup before server exit.
 *
 * @param hSharedMem    [IN] Shared memory pointer as seen from server.
 */
void fsmServerCleanup(FSM_HANDLE hSharedMem)
{
    uint32_t desc_count;
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return;
    }

    // Cleanup our condition variables and mutex.
    pthread_cond_destroy(&control_ptr->server_wakeup);
    pthread_cond_destroy(&control_ptr->client_wakeup);
    pthread_mutex_destroy(&control_ptr->stream_lock);

    // Unmap our shared memory and unlink it.
    desc_count = control_ptr->desc_count;
    munmap(hSharedMem, getControlSize(desc_count));
    shm_unlink(LOCALSTREAM_SHM_NAME);
}

/*!
 * @brief Reset the ring buffer, this is a server-side only operation.
 *
 * @param hSharedMem    [IN] Shared memory pointer as seen from server.
 * @return Zero for success, non-zero for failure
 */
int fsmServerReset(FSM_HANDLE hSharedMem)
{
    int rc;
    struct timespec ts;
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // Wait long enough...but not too long.
    clock_gettime(CLOCK_REALTIME, &ts);
    ts.tv_sec += 2;

    // Lock access. This is more of a courtesy on the behalf of the server.
    // We are going to reset whether the client agrees or not.
    rc = lockControlTimed(hSharedMem, &ts);
    if (rc != 0)
        fi_log(LOG_ERR, "%s: failed to lock (%d) %s", FI_FUNCTION, rc, strerror(rc));

    // Force the fault flags clear.
    control_ptr->server_fault = FSM_NO_FAULT;
    control_ptr->client_fault = FSM_NO_FAULT;

    // Reset our indecies to reflect an empty buffer.
    control_ptr->rd_index.store(0);
    control_ptr->wr_index.store(0);

    // Go ahead and unlock in case the error is that we already own the lock.
    if ((rc == 0) || (rc == EDEADLK))
        rc = unlockControl(hSharedMem);

    return rc;
}

/*!
 * @brief Initialize the Local Stream Shared Memory for the server.
 *
 * The server is expected to run before the client starts and does this server initialization
 * when it first starts up so that the shared memory is ready for the client to connect to it.
 *
 * While the server doesn't actually use the base_addr/base_len information, it needs to pass
 * this information to the client. The user can enter place holders (0) at this time, but must
 * make certain that actual values are in place before the first fsmGetImage() is called.
 * The client side has to memory map that space in order to be able to access the image data.
 *
 * @sa fsmServerUpdateMemInfo
 *
 * @param base_addr [IN] The physical address of the image memory space.
 * @param base_len  [IN] The total number of bytes for the image memory space.
 * @param count     [IN] The number of descriptors to initialize.
 * @returns The handle for the control structure, or nullpltr on error.
 */
FSM_HANDLE fsmServerInit(uint64_t base_addr, uint64_t base_len, uint32_t count)
{
    int fd;
    int rc;
    int control_size;
    FSM_HANDLE control_ptr;

    // Open up our shared memory anchor.
    fd = shm_open(LOCALSTREAM_SHM_NAME, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
    if (fd < 0)
    {
        fi_log(LOG_ERR, "%s: shm_open failed (%d) %s", FI_FUNCTION, errno, strerror(errno));
        return nullptr;
    }

    // Get our control size.
    control_size = getControlSize(count);

    // Make sure we set the size of the shared memory area.
    if (ftruncate(fd, control_size) < 0)
    {
        fi_log(LOG_ERR, "%s: ftruncate to size (%d) failed: (%d) %s", FI_FUNCTION, control_size, errno, strerror(errno));
        close(fd);
        return nullptr;
    }

    // Map our shared memory anchor into our memory space.
    control_ptr = mmap(0, control_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (control_ptr == MAP_FAILED)
    {
        fi_log(LOG_ERR, "%s: mmap failed (%d) %s", FI_FUNCTION, errno, strerror(errno));
        close(fd);
        return nullptr;
    }

    // Don't need this any more once the mmap is done.
    close(fd);

    // Initialize the shared memory control structures.
    rc = sharedMemInit((FSM_CONTROL *)control_ptr, base_addr, base_len, count);
    if (rc != 0)
    {
        munmap(control_ptr, control_size);
        shm_unlink(LOCALSTREAM_SHM_NAME);
        control_ptr = nullptr;
    }

    // Return the pointer (aka "FSM_HANDLE") to the control structure.
    return control_ptr;
}


/*!
 * @brief Wake up the client using the condition variable in shared memory.
 *
 * This needs to be a separate function from the lock/unlock to support proper
 * mutual exclusion when signaling:
 *
 *   lock
 *   set up memory content, e.g. get buffer, set frame number or address, etc.
 *   signal condition
 *   unlock
 *
 * @param hSharedMem    [IN] Client's handle to the control structure.
 * @returns Non-zero for error, 0 for success.
 */
int fsmClientSignal(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    int rc;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // Wake up!
    rc = pthread_cond_signal(&control_ptr->client_wakeup);
    if (rc != 0)
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));

    // Return our result.
    return rc;
}

/*!
 * @brief Client call to wait for event or timeout.
 *
 * @param hSharedMem    [IN] Client's handle to the control structure.
 * @param timeoutPeriod [IN] The timeout in microseconds. A '0' value means to wait indefinitely.
 *
 * @returns Non-zero for error or timeout (ETIMEDOUT), 0 for success.
 */
int fsmClientWait(FSM_HANDLE hSharedMem, uint64_t timeoutPeriod)
{
    int rc;
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // Are we waiting for a maximum period of time, or forever?
    if (timeoutPeriod == 0)
    {
        // Get ownership of the mutex.
        rc = lockControl(hSharedMem);
        if (rc == 0)
        {
            // Wait for a response.
            rc = pthread_cond_wait(&control_ptr->client_wakeup, &control_ptr->stream_lock);

            // If we were successfull or not, we need to release the mutex.
            unlockControl(hSharedMem);
        }
    }
    else
    {
        struct timespec ts;

        // Build up our time structure.
        clock_gettime(CLOCK_REALTIME, &ts);
        ts.tv_nsec += ((timeoutPeriod % 1000000) * 1000);
        while (ts.tv_nsec >= 1000 * 1000 * 1000)
        {
            ts.tv_nsec -= 1000 * 1000 * 1000;
            ts.tv_sec += 1;
        }
        ts.tv_sec += (timeoutPeriod / 1000000);

        // Get ownership of the mutex.
        rc = lockControlTimed(hSharedMem, &ts);
        if (rc == 0)
        {
            // Wait for awhile.
            rc = pthread_cond_timedwait(&control_ptr->client_wakeup, &control_ptr->stream_lock, &ts);

            // If we were successfull or not, we need to release the mutex.
            unlockControl(hSharedMem);
        }
    }

    if ((rc != 0) && (rc != ETIMEDOUT))
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));

    return rc;
}

/*!
 * @brief This is used by the client to clear the fault code set by the server.
 *
 * After clearing the server's fault code, the server is signaled to wake up.
 *
 * @param hSharedMem    [IN] Clients's handle to the control structure.
 * @returns Non-zero for error, 0 for success.
 */
int fsmClientClearFault(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    int rc;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // Clear the fault indication.
    control_ptr->server_fault = FSM_NO_FAULT;

    // Wake up the server. We do NOT need to own the mutex in order to call the
    // signaling method.
    rc = pthread_cond_signal(&control_ptr->server_wakeup);
    if (rc != 0)
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));

    return rc;
}

/*!
 * @brief This is used by the client to set the client's fault code.
 *
 * After setting the fault code, the server is signaled to wake up.
 *
 * @param hSharedMem    [IN] Client's handle to the control structure.
 * @param fault         [IN] The fault code to set into the control structure.
 *
 * @returns Non-zero for error, 0 for success.
 */
int fsmClientSetFault(FSM_HANDLE hSharedMem, int32_t fault)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;
    int rc;

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
        return EINVAL;
    }

    // fi_log(LOG_ERR, "%s: Send client fault (%d) indication to server", FI_FUNCTION, fault);
    control_ptr->client_fault = fault;

    // Wake up the server. We do NOT need to own the mutex in order to call the
    // signaling method.
    rc = pthread_cond_signal(&control_ptr->server_wakeup);
    if (rc != 0)
        fi_log(LOG_ERR, "%s: failed (%d) %s", FI_FUNCTION, errno, strerror(errno));

    return rc;
}

/*!
 * @brief Initialize the shared memory for the client.
 *
 * The server runs before the client starts and does server initialization
 * when it first starts up.
 *
 * It should always be ready for the client to connect, the client should find
 * the share existing already. If not, the client should wait for short period
 * and then try again.
 *
 * @returns non-zero for error, 0 for success.
 */
FSM_HANDLE fsmClientInit()
{
    int fd;
    int desc_count;
    int minimum_size;
    int actual_size;
    FSM_HANDLE control_ptr;

    // Try opening the shared memory.
    fd = shm_open(LOCALSTREAM_SHM_NAME, O_RDWR, S_IRUSR | S_IWUSR);
    if (fd < 0)
    {
        fi_log(LOG_ERR, "%s: shm_open failed (%d) %s", FI_FUNCTION, errno, strerror(errno));
        return nullptr;
    }

    // Calculate the minimum size of the control structure.
    minimum_size = getControlSize(1);

    // Map in only the control structure so we can see how much we REALLY need to map in.
    control_ptr = mmap(0, minimum_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (control_ptr == MAP_FAILED)
    {
        fi_log(LOG_ERR, "%s: mmap failed (%d) %s", FI_FUNCTION, errno, strerror(errno));
        close(fd);
        return nullptr;
    }

    // Get the number of descriptors the server has set up.
    desc_count = ((FSM_CONTROL *)control_ptr)->desc_count;

    // Now that we know the descriptor count, calculate the actual size required.
    actual_size = getControlSize(desc_count);

    // Do we need to remap the memory?
    if (minimum_size < actual_size)
    {
        // Unmap the temporary mapping.
        munmap(control_ptr, minimum_size);

        // Now map in the full structure.
        control_ptr = mmap(0, actual_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
        if (control_ptr == MAP_FAILED)
        {
            fi_log(LOG_ERR, "%s: mmap failed (%d) %s", FI_FUNCTION, errno, strerror(errno));
            close(fd);
            return nullptr;
        }
    }

    // Don't need this any more once the mmap is done.
    close(fd);

    // Make sure we actually have image memory to map in.
    im_mmap_fd = -1;
    im_base_ptr = nullptr;

    // Indicate that we have connected.
    fsmClientSetFault(control_ptr, FSM_NO_FAULT);

    // Return our mapped pointer, aka "FSM_HANDLE".
    return control_ptr;
}

/*!
 * @brief Cleanup when client is shutting down a session; release all buffers.
 *
 * @param hSharedMem    Shared memory pointer as seen from server.
 */
void fsmClientCleanup(FSM_HANDLE hSharedMem)
{
    FSM_CONTROL *control_ptr = (FSM_CONTROL *)hSharedMem;

    // Are we mapped?
    if (im_mmap_fd >= 0)
    {
        // Unmap our image memory space.
        munmap(im_base_ptr, control_ptr->im_base_size);
        close(im_mmap_fd);
        im_mmap_fd = -1;
    }

    // Sanity check.
    if (control_ptr == nullptr)
    {
        fi_log(LOG_ERR, "%s: shared memory pointer is null!", FI_FUNCTION);
    }
    else
    {
        // Return any buffers that we own. We do this by setting the
        // rd_index equal to the wr_index. These are atomic, so we
        // don't need to lock the structure.
        control_ptr->rd_index.store(control_ptr->wr_index.load());
    }
}
