/*!
 * @file fas_thread.cpp
 * @brief The main implementation file for the Fastec Thread library.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#ifndef _WIN32
// For the abi::__forced_unwind exception.
#include <cxxabi.h>
#endif

#include <chrono>

#include "fas_thread.h"
#include "fi_exception.h"
#include "log_if.h"

namespace FAS_THREAD
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

} /* namespace FAS_THREAD */

using namespace FAS_THREAD;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

/*! @brief Base class for Fastec Threads
 *
 * The Fastec thread base class.
 *
 * Classes derived from FAS_THREAD::Thread must define a Thread::run()
 * method.
 */
Thread::Thread(bool is_detached) :
    stop_request(false),
    detached(is_detached)
{
}

/*!
 * @brief Standard destructor.
 */
Thread::~Thread()
{
}

/*!
 * @brief Thread startup method.
 */
void Thread::start_thread(void *t)
{
    // Sanity check.
    if (t == nullptr)
    {
        fi_log(LOG_ERR, "%s: NULL thread pointer passed!", __FUNCTION__);
        return;
    }

    // Run the thread. This will not return until the thread finishes.
    ((Thread *)t)->run();
}

/*!
 * @brief The thread is not actually started until this method is
 *    called. This method does all the setup and creation of
 *    the thread.
 */
void Thread::start(THREAD_STARTUP threadFn)
{
    // Initialize our stop request flag.
    stop_request = false;

    // Is there a supplied thread entry, or are we using the default?
    if (threadFn == nullptr)
        thread_ptr = std::unique_ptr<std::thread>(new std::thread(start_thread, this));
    else
        thread_ptr = std::unique_ptr<std::thread>(new std::thread(threadFn, this));

    // Handle the differernt processing required if we are creating
    // a detached thread or one that can be joined.
    if (detached)
    {
        LOG("%s: detaching and cleaningup the thread", __FUNCTION__);

        // Detach the thread.
        thread_ptr->detach();

        // Since the thread is detached, we no longer have any
        // control over it. Go ahead and get rid of the pointer.
        thread_ptr.reset(nullptr);
    }
}

/*!
 * @brief Calling this method will wait until the thread completes.
 */
void Thread::join()
{
    if ((thread_ptr.get() != nullptr) && thread_ptr->joinable())
        thread_ptr->join();
}

/*!
 * @brief Request the thread to stop.
 * @note This will work with detached threads too.
 */
void Thread::stop()
{
    // Initialize our stop request flag.
    stop_request = true;

    // Wake up our thread.
    sleep_cond.notify_one();
}

/*!
 * @brief Have the thread sleep for the period specified.
 *
 * This function simply converts the period speicifed into
 * milliseconds and calls Thread::sleep_ms().
 *
 * @param delay_ptr A timespec structure.
 * @return 0 on success, non-zero if there's a problem.
 */
int Thread::sleep(const struct timespec *delay_ptr)
{
    uint32_t delay_ms;

    // How many milliseconds are we talking about?
    delay_ms = static_cast<uint32_t>((delay_ptr->tv_sec * 1000) + (delay_ptr->tv_nsec / 1000000));

    // Now that we have it in milliseconds, pass on to the "real" sleep method.
    return sleep_ms(delay_ms);
}

/*!
 * @brief Have the thread sleep for the period specified.
 *
 * This function simply converts the period speicifed into
 * milliseconds and calls Thread::sleep_ms().
 *
 * @param sleep_sec The number of seconds to sleep.
 * @return 0 on success, non-zero if there's a problem.
 */
int Thread::sleep(int sleep_sec)
{
    // Convert to milliseconds and pass it on to the "real" sleep method.
    return sleep_ms(static_cast<uint32_t>(sleep_sec * 1000));
}

/*!
 * @brief Have the thread sleep for the period specified.
 *
 * @note The thread can be awakened by "notifying" the
 *   sleep_cond variable. @sa Thread::stop().
 *
 * @param delay_ms  The number of milliseconds to sleep.
 * @return 0 on success, non-zero if there's a problem.
 */
int Thread::sleep_ms(uint32_t delay_ms)
{
    std::unique_lock<std::mutex> locker(sleep_mutex);

    // No sleeping if we're supposed to be stopping.
    if (stop_request)
        return 0;

    // Time to take a nap.
    std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
    using ms = std::chrono::milliseconds;
    if (sleep_cond.wait_until(locker, now + ms(delay_ms)) == std::cv_status::timeout)
        return ETIME;

    // All done.
    return 0;
}

/*!
 * @brief Wait for the "done" condition variable to be notified.
 *
 * @param delay_ms  Number of milliseconds to wait. 0 to wait forever.
 * @return 0 if OK, non-zero otherwise.
 *
 * @sa Thread::notify_done()
 */
int Thread::wait_for_done(uint32_t delay_ms)
{
    std::unique_lock<std::mutex> locker(wait_for_done_mutex);

    // Make sure we're not trying to stop.
    if (stop_request)
        return 0;

    // Time to take a nap.
    if (delay_ms == 0)
    {
        // Wait forever for the condition variable to be notified.
        wait_for_done_cond.wait(locker);
    }
    else
    {
        // Wait for either the condition variable to be notified or the
        // specified number of milliseconds.
        std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
        using ms = std::chrono::milliseconds;
        if (wait_for_done_cond.wait_until(locker, now + ms(delay_ms)) == std::cv_status::timeout)
            return ETIME;
    }

    // All done.
    return 0;
}

/*!
 * @brief Wait for the "ready" condition variable to be notified.
 *
 * @param delay_ms  Number of milliseconds to wait. 0 to wait forever.
 * @return 0 if OK, non-zero otherwise.
 *
 * @sa Thread::notify_ready()
 */
int Thread::wait_for_ready(uint32_t delay_ms)
{
    std::unique_lock<std::mutex> locker(wait_for_ready_mutex);

    // Make sure we're not trying to stop.
    if (stop_request)
        return 0;

    // Time to take a nap.
    if (delay_ms == 0)
    {
        // Wait forever for the condition variable to be notified.
        wait_for_ready_cond.wait(locker);
    }
    else
    {
        // Wait for either the condition variable to be notified or the
        // specified number of milliseconds.
        std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
        using ms = std::chrono::milliseconds;
        if (wait_for_ready_cond.wait_until(locker, now + ms(delay_ms)) == std::cv_status::timeout)
            return ETIME;
    }

    // All done.
    return 0;
}

