/*!
 * @file fi_exception.cpp
 * @brief The implementation file for the Exception class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <cstdarg>
#include "fi_exception.h"

namespace FAS_THREAD
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

} /* namespace FAS_THREAD */

using namespace FAS_THREAD;

/*!
 * @brief Fastec Imaging exception base class default constructor.
 */
Exception::Exception() :
    Note()
{
}

/*!
 * @brief Explicit copy constructor, since std::stringstream does not copy.
 *
 * @param e LogicError instance to copy
 */
Exception::Exception(const Exception& e) :
    Note(e)
{
}

/*!
 * @brief Exception which creates a Note.
 *
 * Creates a new exception, with a one-line Note that includes the creation
 * timestamp, file and line number tags, and caller's printf()-formatted
 * one-line message.
 *
 * @param file   File name tag (for example from __FILE__ macro)
 * @param line   Line number (for example from __LINE__ macro)
 * @param fmt    printf()-style format string, followed by arguments as required
 */
Exception::Exception(const char *file, int line, const char *fmt, ...) :
    Note()
{
    va_list ap;

    // Build the formatted string. The "vappend()" is implemented
    // in the Note class.
    va_start(ap, fmt);
    vappend(file, line, fmt, ap);
    va_end(ap);
}

void Exception::append(const char *fmt, ...)
{
    va_list ap;

    // Build the formatted string. The "vappend()" is implemented
    // in the Note class.
    va_start(ap, fmt);
    vappend(fmt, ap);
    va_end(ap);
}

/*!
 * @brief Add formatted message string to existing exception
 *
 * This method is intended for use with the FI_EXC_APPEND(...) macro.
 * It adds a line to the exception's message, embedding the file name and
 * line number at which the macro was invoked in front of the formatted
 * message string.
 *
 * @param file  File name (normally supplied by __FILE__ macro)
 * @param line  Line number (normally supplied by the __LINE__ macro)
 * @param fmt   printf()-style format string, followed by zero or more arguments
 */
void Exception::append(const char *file, int line, const char *fmt, ...)
{
    va_list ap;

    // Build the formatted string. The "vappend()" is implemented
    // in the Note class.
    va_start(ap, fmt);
    vappend(file, line, fmt, ap);
    va_end(ap);
}

// ----------------------------------------------------------------------------

/*!
 * @brief Fastec Imaging "logic error" exception class
 *
 * This is the sub-class of Fastec::Exception intended to denote logic flaws
 * in the software detected at runtime ("can't happen" conditions that
 * happen anyway.
 *
 * This class does not add any functionality to the underlying base
 * class. It is declared only to make a distinct type that can be
 * distinguished in catch() clauses.
 */
LogicError::LogicError() :
    Exception()
{
}

/*!
 * @brief Explicit copy constructor, since std::stringstream does not copy
 *
 * @param e LogicError instance to copy.
 */
LogicError::LogicError(const LogicError& e) :
    Exception(e)
{
}

/*!
 * @brief Note constructor
 *
 * Creates a new exception, with a one-line Note that includes the creation
 * timestamp, file and line number tags, and caller's printf()-formatted
 * one-line message.
 *
 * @param file   File name tag (for example from __FILE__ macro)
 * @param line   Line number (for example from __LINE__ macro)
 * @param fmt    printf()-style format string, followed by arguments as required
 */
LogicError::LogicError(const char *file, int line, const char *fmt, ...) :
    Exception()
{
    va_list ap;

    // Build the formatted string. The "vappend()" is implemented
    // in the Note class.
    va_start(ap, fmt);
    vappend(file, line, fmt, ap);
    va_end(ap);
}

