/*!
 * @file fi_note.cpp
 * @brief The implementation file for the Note class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdlib.h>
#include <stdarg.h>

#include <chrono>
#include <ctime>
#include <iomanip>
#include <memory>

#include "fi_note.h"

#if defined(_WIN32) && !defined(__MINGW32__)
// Turn off the "This function may be unsafe" warning.
#  pragma warning( disable: 4996 )

#  define snprintf  _snprintf
#endif

namespace FAS_THREAD
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

} /* namespace FAS_THREAD */

using namespace FAS_THREAD;

/*!
 * @brief Fastec Imaging "Note" base class
 *
 * This utility class holds an "extendible," human-readable, text note. It is intended
 * to support trace/log messages, and messages carried in exceptions (that often end up
 * as log messages).
 *
 * The base class supports the following features:
 *
 *     - printf()-style formatting for messages
 *     - ostream-style string handling ("<<" operator)
 *     - Multi-line messages, formatted nicely for addition to log files
 *     - Automatic time-stamping of messages
 *     - Support (via associated macros) for File/Line tagging
 *
 * The first line of text added to a Note will be automatically pre-pended with a
 * timestamp, of the form "YYYYMMDD HHMMSS.ssssss," recording the microsecond resolution
 * time at which the first line was added to the Note.
 */
Note::Note()
{
    msg = std::unique_ptr<std::stringstream>(new std::stringstream());
}

/*!
 * @brief Explicit copy constructor, since std::stringstream does not copy.
 *
 * @param n Note instance to copy
 */
Note::Note(const Note& n) noexcept
{
    msg = std::unique_ptr<std::stringstream>(new std::stringstream());
    *msg << n.msg->rdbuf();
}

/*!
 * @brief Note constructor
 *
 * Creates a new one-line Note that includes the creation
 * timestamp, file and line number tags, and caller's printf()-formatted
 * one-line message.
 *
 * @param file   File name tag (for example from __FILE__ macro)
 * @param line   Line number (for example from __LINE__ macro)
 * @param fmt    printf()-style format string, followed by arguments as required
 */
Note::Note(const char *file, int line, const char *fmt, ...)
{
    va_list ap;

    msg = std::unique_ptr<std::stringstream>(new std::stringstream());

    // Compute length of formatted string
    va_start(ap, fmt);
    vappend(file, line, fmt, ap);
    va_end(ap);
}

Note& Note::operator<<(const char *s)
{
    *msg << s;
    return *this;
}

/*!
 * @brief Add formatted message string to existing Note
 *    (without file/line # tag; no timestamp).
 *
 * @param format  printf()-style format string, followed by zero or more arguments.
 */
void Note::append(const char *fmt, ...)
{
    va_list ap;

    // Compute length of formatted string
    va_start(ap, fmt);
    vappend(fmt, ap);
    va_end(ap);
}

/*!
 * @brief Add formatted message string to existing Note.
 *
 * This method is intended for use with a macro (*_APPEND(...)).
 * It adds a line to the exception's message, embedding the file name and
 * line number at which the macro was invoked in front of the formatted
 * message string.
 *
 * @param file  File name (normally supplied by __FILE__ macro).
 * @param line  Line number (normally supplied by the __LINE__ macro).
 * @param fmt   printf()-style format string, followed by zero or more arguments.
 */
void Note::append(const char *file, int line, const char *fmt, ...)
{
    va_list ap;

    // Compute length of formatted string
    va_start(ap, fmt);
    vappend(file, line, fmt, ap);
    va_end(ap);
}

void Note::vappend(const char *fmt, va_list ap)
{
    int n;
    va_list ap2;

    // Make a copy of the arg-list. It is assumed that 'ap' has already
    // had a va_start called with it.
    va_copy(ap2, ap);

    // Is this first line of Note?
    if (msg->str().length() > 0)
    {
        // Insert newline and two spaces to indent new stuff
        *msg << std::endl << "  ";
    }

    // Compute length of formatted string
    if ((n = vsnprintf(nullptr, 0, fmt, ap) + 1) == 1)
    {
        *msg << "empty message.";
    }
    else
    {
        try
        {
            int m;

            // Get a buffer of the requisite size
            std::unique_ptr<char[]> buffer(new char[static_cast<uint64_t>(n)]);

            m = vsnprintf(buffer.get(), static_cast<size_t>(n), fmt, ap2);
            if (m != (n - 1))
            {
                *msg << "vsnprintf() fault in processing exception message.";
            }
            else
            {
                // Successful case: add caller's formatted message to exception
                *msg << buffer.get();
            }
        }
        catch (...)
        {
            *msg << "could not allocate " << n << " bytes of storage for exception message.";
        }
    }

    // Finish the arg-list processing on ap2. We assume our caller will take
    // care of 'ap'.
    va_end(ap2);
}

void Note::vappend(const char *file, int line, const char *fmt, va_list ap)
{
    int n;
    va_list ap2;

    // Make a copy of the arg-list. It is assumed that 'ap' has already
    // had a va_start called with it.
    va_copy(ap2, ap);

    // Is this first line of Note?
    if (msg->str().length() > 0)
    {
        // Insert two spaces to indent new stuff
        *msg << std::endl << "  ";
    }
    else
    {
        // Insert timestamp
        *msg << get_datetime() << ": ";
    }

    // Compute length of formatted string
    if ((n = vsnprintf(nullptr, 0, fmt, ap) + 1) == 1)
    {
        *msg << file << ", " << line << ": empty exception message.";
    }
    else
    {
        try
        {
            int m;

            // Get a buffer of the requisite size
            std::unique_ptr<char[]> buffer(new char[static_cast<uint64_t>(n)]);

            m = vsnprintf(buffer.get(), static_cast<size_t>(n), fmt, ap2);
            if (m != (n - 1))
            {
                *msg << file << ", " << line << ": vsnprintf() fault in processing exception message.";
            }
            else
            {
                // Successful case: add caller's formatted message to exception
                *msg << file << ", " << line << ": " << buffer.get();
            }
        }
        catch (...)
        {
            *msg << file << ", " << line << ": could not allocate " << n << " bytes of storage for exception message.";
        }
    }

    // Finish the arg-list processing on ap2. We assume our caller will take
    // care of 'ap'.
    va_end(ap2);
}

/*!
 * @brief A helper function to return the date and time in a format
 *    for use in the log message.
 *
 * @return A std::string in the format: "YYYYMMDD HHMMSS.ssssss".
 */
std::string Note::get_datetime()
{
    // Get a timepoint for now.
    auto now = std::chrono::system_clock::now();

    // Get the time in time_t format for use in the localtime call.
    auto now_time_t = std::chrono::system_clock::to_time_t(now);

    // Get the seconds to subtract from the total nanoseconds. This will
    // give us the number of nanoseconds since the second mark.
    auto secs = std::chrono::time_point_cast<std::chrono::seconds>(now);
    auto ns = std::chrono::time_point_cast<std::chrono::nanoseconds>(now) -
              std::chrono::time_point_cast<std::chrono::nanoseconds>(secs);

    // Build up our string.
    std::stringstream ss;
    ss << std::put_time(std::localtime(&now_time_t), "%Y%m%d %H%M%S.");
    ss << std::setfill('0') << std::setw(6) << (int)(ns.count() / 1000);

    // Return the string.
    return ss.str();
}

/*!
 * @brief Extract accumulated message(s).
 *
 * This is suitable for use with iostreams, as in "cerr << e.what() << endl," or with
 * C-style output as in "fprintf(stderr, "%s\n", e.what().c_str())."
 *
 * @return Complete formatted message string.
 */
const std::string Note::what()
{
    return msg->str();
}
