/*!
 * @file fi_sync_message.cpp
 * @brief The main implementation file for the SyncMessage class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <system_error>

#include "fi_exception.h"
#include "fi_sync_message.h"

namespace FAS_THREAD
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

} /* namespace FAS_THREAD */

using namespace FAS_THREAD;

/*!
 * @brief SyncMessage constructor.
 */
SyncMessage::SyncMessage() :
    Message(),
    msgBusy(false)
{
    sync_mutex = std::unique_ptr<std::mutex>(new std::mutex());
    sync_status = std::unique_ptr<std::condition_variable>(new std::condition_variable());
}

/*!
 * @brief Launch a synchronous message
 *
 * The calling thread blocks until the message is forwarded with
 * an empty route table.
 *
 * @param t (Initial) destination thread
 */
void SyncMessage::doSync(FAS_THREAD::MsgThread *t)
{
    std::unique_lock<std::mutex> locker(*sync_mutex);

    // Sanity check.
    if (msgBusy)
    {
        // Throw an error.
        FI_LOGIC_ERROR("Sync fault");
    }

    // Put the message on the thread's queue.
    msgBusy = true;
    t->messageAdd(this);

    // Block until we're not busy.
    while (msgBusy)
    {
        // Wait for the condition variable to be released.
        sync_status->wait(locker);
    }
}

/*!
 * @brief "Forward" the message.
 *
 * The forward() method, like its base class counterpart, will
 * route the message to the next thread in the route vector.
 *
 * Handling differs however when the route vector is empty. Rather
 * than posting the message to the originating thread's input queue,
 * we awaken the blocked thread that issued the original doSync()
 * call to launch the message.
 */
void SyncMessage::forward()
{
    // Acquire message lock; released when 'locker' goes out of scope.
    std::unique_lock<std::mutex> locker(*sync_mutex);

    // Sanity check.
    if (!msgBusy)
    {
        // Throw an error.
        FI_LOGIC_ERROR("Sync fault");
    }

    // Check to see if there are anymore threads in the route.
    if (route->empty())
    {
        // Un-block waiting thread
        msgBusy = false;
        sync_status->notify_one();
    }
    else
    {
        // Send to next thread on route
        FAS_THREAD::MsgThread *t = route->top();
        route->pop();
        t->messageAdd(this);
    }
}

