/*!
 * @file fas_video.cpp
 * @brief The main implementation file for the Fastec Video library.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
//#include <sys/stat.h> // Was added for linux but need an access mode let's leave that to caller.
#include <dirent.h>
#include <string>
#include <memory>
#include <chrono>
#include <fmt/printf.h>
#include <list>

#include "fi_image.h"
#include "fi_bmp.h"
#include "fi_dng.h"
#include "fi_tiff.h"
#include "fas_video.h"
#include "fi_stream.h"
#include "fi_stream_avi.h"
#include "fi_stream_mp4.h"
#include "fi_frame.h"
#include "ffmpeg_encoder.h"
#include "log_if.h"

#ifdef INCLUDE_RDI_PACKED
#include "fi_stream_rdi_packed.h"
#endif /* INCLUDE_RDI_PACKED */

namespace FAS_VIDEO
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

// Defines the bit mode strings for use in the metadata text.
// These are deprecated in favor of the start-/end-bit description.
static const char *TS3_BIT_MODE_LOWER_8 = "low 8";
static const char *TS3_BIT_MODE_MIDDLE_8 = "middle 8";
static const char *TS3_BIT_MODE_UPPER_8 = "upper 8";
static const char *TS3_BIT_MODE_FULL_10 = "full 10";
static const char *TS3_BIT_MODE_UNKNOWN = "unknown";

// Defines the scale factor for converting microseconds to nanoseconds.
static const uint32_t MICRO_TO_NANO_SCALE = 1000;

#pragma pack(1)
//! Xcitex binary ".cfg" file format.
typedef struct
{
    uint8_t string_length;  // Set to 0, we put no string following structure

    uint32_t record_rate;
    uint32_t trigger_setting;
    uint32_t zero_position;
    uint32_t shutter_speed;
    uint32_t burst_frames;
    uint32_t depth_mode;
    uint32_t burst_offset;
    uint32_t roc_mode;
    uint32_t saved_with_skip_rate;
    uint32_t saved_with_h_flip;
    uint32_t saved_with_v_flip;
    uint32_t saved_with_rotation;
} XCFG_METADATA_REC;
#pragma pack()

// Define a boolean to hold state of initialization for log file used by fi_log().
static bool loggerInitialized = false;

// Define the name and extension of the log file we will use for our library.
const std::string logFileName   = "fas_video"s;
const std::string logFileExt    = ".txt"s;

} // End of FAS_VIDEO

using namespace FAS_VIDEO;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------



/*!
 * @brief Convert straight-binary-seconds to a struct tm, with same code as CS uses.
 *
 * NOTE_JLC - revisit this incorporation with FIXME in mind.
 * It was put here as part of the metadata string generation.
 *
 * FIXME This is not an accurate conversion of Unix time. We almost certainly have problems
 * if we ever record over a year turnover, but it is otherwise internally consistent.
 */
static void sbs_to_tm(uint32_t sbs, struct tm *tm_ptr)
{
    // Clear out the structure.
    memset(tm_ptr, 0, sizeof(struct tm));

    // Get years.
    tm_ptr->tm_year = sbs / (365 * 24 * 60 * 60);
    sbs -= (uint32_t)(tm_ptr->tm_year * (365 * 24 * 60 * 60));
    tm_ptr->tm_yday = sbs / (24 * 60 * 60);
    sbs -= (uint32_t)(tm_ptr->tm_yday * (24 * 60 * 60));
    tm_ptr->tm_hour = sbs / (60 * 60);
    sbs -= (uint32_t)(tm_ptr->tm_hour * (60 * 60));
    tm_ptr->tm_min = sbs / 60;
    tm_ptr->tm_sec = (int)sbs - (tm_ptr->tm_min * 60);
}

/*!
 * @brief Generate the proper header data for a type of image.
 *
 * @param videoType     One of the FAS_VIDEO::VIDEO_CODEC values for images.
 * @param fourcc        One of the supported FAS_VIDEO @ref fv_fourcc "FourCC" codes.
 * @param meta          Pointer to the images metadata.
 *
 * @returns A std::unique_ptr<char> containing the image file header.
 */
std::unique_ptr<FAS_VIDEO::HEADER_DATA> FiVideo::CreateImageHeader(FAS_VIDEO::VIDEO_CODEC videoType, uint32_t fourcc, pm_metadata *meta)
{
    std::unique_ptr<FAS_VIDEO::HEADER_DATA> header_ptr;

    // The FI Image Pointer object created based on the FAS_VIDEO::VIDEO_CODEC type.
    std::unique_ptr<FAS_VIDEO::FiImage> fiImagePtr;

    // Sanity check.
    if (meta == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: meta pointer is NULL!?", __PRETTY_FUNCTION__);
        header_ptr = std::unique_ptr<FAS_VIDEO::HEADER_DATA>();
        goto gh_exit;
    }

#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "%s: Creating Header", __PRETTY_FUNCTION__);
#endif

    // Which header type are we trying to get?
    switch (videoType)
    {
    case FI_BMP_STACK_CODEC:
        fiImagePtr = FAS_VIDEO::FiImage::CreateImageObject(FAS_VIDEO::FI_BMP_IMAGE);
        break;
    case FI_DNG_STACK_CODEC:
        fiImagePtr = FAS_VIDEO::FiImage::CreateImageObject(FAS_VIDEO::FI_DNG_IMAGE);
        break;
    case FI_TIFF_STACK_CODEC:
        fiImagePtr = FAS_VIDEO::FiImage::CreateImageObject(FAS_VIDEO::FI_TIF_IMAGE);
        break;
    case FI_JPG_STACK_CODEC:
        // JPEGs are a little special. They don't have a header, but rather than
        // truly having NO header, let's return a 0-length header. This way we
        // can differentiate between an empty header (JPEG) and an error.
        header_ptr = std::unique_ptr<FAS_VIDEO::HEADER_DATA>(new FAS_VIDEO::HEADER_DATA());
        break;
    case FI_PLANAR_STACK_CODEC:
        fiImagePtr = FAS_VIDEO::FiImage::CreateImageObject(FAS_VIDEO::FI_BMP_PLANAR_IMAGE);
        break;
    default:
        header_ptr = std::unique_ptr<FAS_VIDEO::HEADER_DATA>();
        break;
    }

    //  If FiImage object was created then generate and retrieve header data.
    if (fiImagePtr != nullptr)
    {
        if (fiImagePtr->CreateHeader(fourcc, meta) > 0)
        {
            header_ptr = fiImagePtr->GetHeader();
        }
    }

gh_exit:
    return header_ptr;
}

/*!
* @brief Factory to create an AVI video stream writer for Fastec Imaging Videos.
*
* @param videoNameToCreate  Name of the video supporting the stream.
* @param frameRate          How many frames to render per second of video.
* @param frameWidth         Width of the frame image.
* @param frameHeight        Number of rows in frame image.
* @param frameCount         The number of frames for this video.
*
* @returns A std::unique_ptr<FAS_VIDEO::FiVideoStream> containing the video streaming object.
*/
std::unique_ptr<FAS_VIDEO::FiVideoStream> FiVideo::CreateAVIVideoStreamWriter(const std::string& videoName,
                                                                              uint32_t frameRate,
                                                                              uint32_t frameWidth,
                                                                              uint32_t frameHeight,
                                                                              uint32_t frameCount)
{
    FAS_VIDEO::STREAM_STATUS streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_EMPTY;
    std::unique_ptr<FAS_VIDEO::FiVideoStream> videoStreamWriter;

#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "%s: Creating Video Stream Writer for AVI Video", __PRETTY_FUNCTION__);
#endif

    try
    {
        // Create the AVI Video stream writer.
        videoStreamWriter = std::unique_ptr<FAS_VIDEO::FiVideoStream>(new FAS_AVI::FiVideoAVIStream());

        // Make sure it was created properly.
        if (videoStreamWriter != nullptr)
        {
            // Prepare the video stream for writing.
            streamStatus = videoStreamWriter->PrepareForWriting(videoName, frameRate, frameWidth, frameHeight, frameCount);

            // If the stream is not ready report the problem!
            if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_READY)
            {
                fi_log(LOG_ERR, "%s: Unable to Create AVI Video Stream Writer.", __PRETTY_FUNCTION__);
                videoStreamWriter->LogStreamStatus();
            }
        }
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while preparing AVI video stream writer.", __PRETTY_FUNCTION__, e.what());
        if (videoStreamWriter != nullptr)
            videoStreamWriter.reset();
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught  while preparing AVI video stream writer.", __PRETTY_FUNCTION__);
        if (videoStreamWriter != nullptr)
            videoStreamWriter.reset();
    }

    return videoStreamWriter;
}


/*!
* @brief Factory to create a MP4 video stream writer for Fastec Imaging Videos.
*
* @param videoNameToCreate  Name of the video supporting the stream.
* @param frameRate          How many frames to render per second of video.
* @param frameWidth         Width of the frame image.
* @param frameHeight        Number of rows in frame image.
* @param frameCount         The number of frames for this video.
*
* @returns A std::unique_ptr<FAS_VIDEO::FiVideoStream> containing the video streaming object.
*/
std::unique_ptr<FAS_VIDEO::FiVideoStream> FiVideo::CreateMP4VideoStreamWriter(const std::string& videoName,
                                                                              uint32_t frameRate,
                                                                              uint32_t frameWidth,
                                                                              uint32_t frameHeight,
                                                                              uint32_t frameCount,
                                                                              uint32_t h264QualityPct,
                                                                              FAS_VIDEO::H264_SPEEDS h264Speed,
                                                                              uint32_t keyFrameIntervalToUse)
{
    FAS_VIDEO::STREAM_STATUS streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_EMPTY;
    std::unique_ptr<FAS_VIDEO::FiVideoStream> videoStreamWriter;

#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "%s: Creating Video Stream Writer for MP4 Video", __PRETTY_FUNCTION__);
#endif

    try
    {
        // Create the MP4 Video stream writer.
        videoStreamWriter = std::unique_ptr<FAS_VIDEO::FiVideoStream>(new FAS_MP4::FiVideoMP4Stream(h264QualityPct, h264Speed, keyFrameIntervalToUse));

        // Make sure it was created properly.
        if (videoStreamWriter != nullptr)
        {
            // Prepare the video stream for writing.
            streamStatus = videoStreamWriter->PrepareForWriting(videoName, frameRate, frameWidth, frameHeight, frameCount);

            // If the stream is not ready report the problem!
            if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_READY)
            {
                fi_log(LOG_ERR, "%s: Unable to Create MP4 Video Stream Writer.", __PRETTY_FUNCTION__);
                videoStreamWriter->LogStreamStatus();
            }
        }
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while preparing MP4 video stream writer.", __PRETTY_FUNCTION__, e.what());
        if (videoStreamWriter != nullptr)
            videoStreamWriter.reset();
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught  while preparing MP4 video stream writer.", __PRETTY_FUNCTION__);
        if (videoStreamWriter != nullptr)
            videoStreamWriter.reset();
    }

    return videoStreamWriter;
}

#ifdef INCLUDE_RDI_PACKED
/*!
* @brief Factory to create an RDI Packed Image video stream writer.
*
* @param videoNameToCreate  Name of the video supporting the stream.
* @param frameRate          How many frames to render per second of video.
* @param frameWidth         Width of the frame image.
* @param frameHeight        Number of rows in frame image.
* @param frameCount         The number of frames for this video.
* @param rawBPP             Specifies the bit depth resolution inside of 16 bits per pixel image data.
*
* @returns A std::unique_ptr<FAS_VIDEO::FiVideoStream> containing the video streaming object.
*/
std::unique_ptr<FAS_VIDEO::FiVideoStream> FiVideo::CreateRDIPackedVideoStreamWriter(const std::string& videoName,
        uint32_t frameRate,
        uint32_t frameWidth,
        uint32_t frameHeight,
        uint32_t frameCount,
        const pm_metadata_type *pm_ptr)
{
    FAS_VIDEO::STREAM_STATUS streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_EMPTY;
    std::unique_ptr<FAS_VIDEO::FiVideoStream> videoStreamWriter;

    // Sanity check.
    if (pm_ptr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: Missing pm_metadata!", __PRETTY_FUNCTION__);
        return videoStreamWriter;
    }

    try
    {
        // Create the RDI Packed Video stream writer and install the rawBPP through the constructor.
        videoStreamWriter = std::unique_ptr<FAS_VIDEO::FiVideoStream>(new FAS_RDI::FiVideoRDIPackedStream(pm_ptr));

        // Make sure it was created properly.
        if (videoStreamWriter != nullptr)
        {
            streamStatus = videoStreamWriter->PrepareForWriting(videoName, frameRate, frameWidth, frameHeight, frameCount);

            // If the stream is not ready report the problem!
            if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_READY)
            {
                fi_log(LOG_ERR, "%s: Unable to Create RDI Packed Video Stream Writer.", __PRETTY_FUNCTION__);
                videoStreamWriter->LogStreamStatus();
            }
        }
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while preparing AVI video stream writer.", __PRETTY_FUNCTION__, e.what());
        if (videoStreamWriter != nullptr)
            videoStreamWriter.reset();
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught  while preparing AVI video stream writer.", __PRETTY_FUNCTION__);
        if (videoStreamWriter != nullptr)
            videoStreamWriter.reset();
    }

    return videoStreamWriter;
}
#endif /* INCLUDE_RDI_PACKED */

/*!
* @brief Factory method to create a video stream reader for Fastec Imaging AVI Videos.
*
* @param videoNameToOpen      Name of the video supporting the AVI stream.
*
* @returns A std::unique_ptr<FAS_VIDEO::FiVideoStream> containing the video streaming object.
*/
std::unique_ptr<FAS_VIDEO::FiVideoStream> FiVideo::CreateAVIVideoStreamReader(const std::string& videoNameToOpen)
{
    std::unique_ptr<FAS_VIDEO::FiVideoStream> videoStreamReader;

#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "%s: Creating AVI Video Stream Reader in Factory.", __PRETTY_FUNCTION__);
#endif

    try
    {
        // Create the AVI Video Stream object.
        videoStreamReader = std::unique_ptr<FAS_VIDEO::FiVideoStream>(new FAS_AVI::FiVideoAVIStream());

        // Prepare video stream for writing.
        videoStreamReader->PrepareForReading(videoNameToOpen);

        // If the stream is not ready report the problem!
        if (!videoStreamReader->IsStreamReady())
        {
            fi_log(LOG_ERR, "%s: PrepareForReading for[%s] FAILED and Stream Status on next line.", __PRETTY_FUNCTION__, videoNameToOpen.c_str());
            videoStreamReader->LogStreamStatus();
        }
    }
    // Look for a specific resource allocation error to set state of the reader object.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while preparing AVI video stream reader.", __PRETTY_FUNCTION__, e.what());
        return nullptr;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught while preparing AVI video stream reader.", __PRETTY_FUNCTION__);
        return nullptr;
    }

    return videoStreamReader;
}

/*!
 * @brief Initialize a log file for fas_video library use.
 *
 * @param logDirectory where to locate our log file and it
 * expects all directory separation characters to use '/' and
 * it also expects the directory to exist on the system.
 *
 * @param maxLogsB4Rotate maximum number of log files in directory
 * before we start reusing the oldest one's spot.
 *
 * @returns True if successfully set up the log file, false if not.
 *
 */
bool FiVideo::InitializeLogger(std::string& logDirectory, int maxLogsBeforeRotate)
{
    bool initializedLogger  = false;

    // Grab our time of initialization.
    std::chrono::system_clock::time_point startTime;

    std::string theDirSlash = "/";
    std::string theLogFullPath = logDirectory;  // Full path to log file folder.
    std::string theLogFileNameToCreate;         // Filename_ + "yyyy_MMdd_hhmmss" + .txt

    // These are used for directory information enquiries.
    DIR *dp;
    struct dirent *ep;

    int nCount          = 0;    // Number of log files in directory.
    int nOver           = 0;    // Number of log files over maxLogsBeforeRotate.
#if 0
    int mkdirResult     = 0;    // Result code from a mkdir event.
#endif

    // If have to delete log files the information is stored here a and reused
    // for all older logfiles over maxLogsBeforeRotate one.
    std::string fileToDelete;
    std::string fullPathToFileToDelete;

    // If have to delete existing log files because over maxLogsBeforeRotate
    // then this list will contain the matching log information for each file found.
    std::list<std::string> existingLogFiles;

    // Is there something in the directory?
    if (logDirectory.size() == 0)
        return initializedLogger;   // There is no directory information available.

    // If the logger is already initialized for this library cannot
    // initialize it a second time.
    if (loggerInitialized)
        return initializedLogger;

    // Does the directory have an ending slash?
    if (logDirectory.back() != '/')
    {
        theLogFullPath = theLogFullPath + theDirSlash;
    }

    //QString dt_format = QString::fromUtf8("yyyy_MMdd_hhmmss");

    startTime = std::chrono::system_clock::now();
    std::time_t tt = std::chrono::system_clock::to_time_t(startTime);
    //tm utc_tm = *gmtime(&tt);
    tm localTime = *localtime(&tt);

    theLogFileNameToCreate = fmt::sprintf(FMT_STRING("%s_%4d_%02d%02d_%02d%02d%02d%s"),
                                          logFileName.c_str(),
                                          (localTime.tm_year + 1900), (localTime.tm_mon + 1), localTime.tm_mday,
                                          localTime.tm_hour, localTime.tm_min, localTime.tm_sec,
                                          logFileExt.c_str());

    // Now query the full path for information.
    dp = opendir(theLogFullPath.c_str());

    // Will leave the requirement up to caller to make the directory
    // because of the differences between linux and windows on permission usage.
#if 0
    // If unable to open directory then try to make it.
    if (dp == nullptr)
    {
        mkdirResult = mkdir(theLogFullPath.c_str());    // Windows Version.

        // Try to open the directory one more time now that it's been created.
        // Want to catch the failure again as terminal too so check for nullptr
        // one more time afterwards.
        dp = opendir(theLogFullPath.c_str());
    }
#endif

    // If have directory opened then loop through files looking for matches to
    // our own log files in the folder.
    if (dp != nullptr)
    {
        do
        {
            ep = readdir(dp);
            if (ep != nullptr)
            {
                std::string theName = ep->d_name;
                if (theName.size() == theLogFileNameToCreate.size())
                {
                    // first part of this should be the file name.
                    size_t thePos = theName.find_first_of(logFileName.c_str());
                    if ((thePos == 0) && (thePos != std::string::npos))
                    {
                        // Make sure there is an extension of txt.
                        size_t theExtPos = theName.find_first_of(logFileExt.c_str());
                        if (theExtPos != std::string::npos)
                        {
                            // Add the name to a existing log files.
                            existingLogFiles.push_back(theName);
                            nCount++;
                            // If the count is over or equal to the maxLogsBeforeRotate
                            // then we bump up the count of the number over.  Remember
                            // that it must have one slot for new log file.
                            if (nCount >= maxLogsBeforeRotate)
                                nOver++;
                        }
                    }
                }
            }
        } while (ep != nullptr);

        // Close the open directory.
        (void) closedir(dp);

        // If have to remove a log file then sort the list and do so...
        if (nOver > 0)
        {
            int nDeleted        = 0;    // Number of log files deleted.

            existingLogFiles.sort();
            while (nDeleted < nOver)
            {
                fileToDelete = existingLogFiles.front();
                existingLogFiles.pop_front();
                fullPathToFileToDelete = theLogFullPath + fileToDelete;
                int removeResult = std::remove(fullPathToFileToDelete.c_str());
                if (removeResult != 0)
                {
                    fi_log(LOG_ERR, "%s: Unable to remove log file: %s.",
                           __PRETTY_FUNCTION__, fullPathToFileToDelete.c_str());
                }

                // Always increment so the loop terminates even if it failed.
                nDeleted++;
            }
        }
    }
    else
    {
        fi_log(LOG_ERR, "%s: Unable to query information for directory %s.",
               __PRETTY_FUNCTION__, theLogFullPath.c_str());
    }

    // Set up the full path with the log filename to create as one string.
    theLogFullPath = theLogFullPath + theLogFileNameToCreate;

    // Set up the fi_log object to write to this file.
    if (loggerFile(theLogFullPath.c_str()) == 0)
    {
        initializedLogger = loggerInitialized = true;
    }
    else
    {
        fi_log(LOG_ERR, "%s: Unable to create loggerFile using log file: %s.",
               __PRETTY_FUNCTION__, theLogFullPath.c_str());
    }

    return initializedLogger;
}

/*!
 * @brief Translates the H264 setting for compression speed to a user readable string.
 *
 * @param speedText     A place to put the string.
 * @param speedSetting  The speed setting to translate.
 *
 */
void FiVideo::H264SpeedSettingToText(std::string& speedText, FAS_VIDEO::H264_SPEEDS speedSetting)
{
    switch (speedSetting)
    {
    case FAS_VIDEO::H264_SPEEDS::H264_SLOWER:
        speedText = "Slow";
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_MEDIUM:
        speedText = "Medium";
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_FAST:
        speedText = "Fast";
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_VERY_FAST:
    default:
        speedText = "Very Fast";
        break;
    }
}


/*!
 * @brief Fills a buffer with information from the metadata provided.
 *
 * The frame count and trigger frame values are adjusted based on startFrame and
 * endFrame values.
 *
 * @param metadataStringRef     Caller provides std::string that will be formatted internally here.
 * @param pmMetadataPtr         A read-only pointer to a pm_metadata data structure.
 * @param pmMetadataSize        How many uint8_t bytes pointed to by the pmMetadataPtr.
 * @param startFrame            The starting frame number for the capture.
 * @param endFrame              The ending frame number for the capture.
 * @param gainsPtr              The read-only video gain values used for the capture.
 *
 * @returns The number of bytes written or a negative number on failure.
 */
int32_t FiVideo::CreateMetadataString(std::string& metadataStringRef,
                                      const void *pmMetadataPtr, uint32_t pmMetaDataSize,
                                      uint32_t startFrame, uint32_t endFrame,
                                      const FAS_VIDEO::VIDEO_GAINS *gainsPtr)
{
    int32_t resultCode = -1;
    int32_t first = 0;
    int32_t last = 0;

    uint32_t red_gain;
    uint32_t green_gain;
    uint32_t blue_gain;

    uint32_t frame_count;
    uint32_t trig_frame_num;

    struct tm sbs_tm;

    std::string timeStr;                // Builds up the time string.
    std::string triggerSetStr;          // Builds up the trigger set string.
    std::string bitModeTypeStr;         // Holds user readable bitmode string.
    std::string optionsStr;             // Builds up the sensor options string.

    // Grab a type-defined pointer to the passed in read-only void* pm_metadata.
    const pm_metadata *meta = (const pm_metadata *)pmMetadataPtr;

    // Make sure the passed metadataStringRef is empty.
    metadataStringRef.clear();

    // Sanity checks.
    if (meta == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: pm_metadata is required!", __PRETTY_FUNCTION__);
        goto exitCreateMetadataString;
    }
    if (pmMetaDataSize != sizeof(pm_metadata))
    {
        fi_log(LOG_DEBUG, "%s: pm_metadata is the wrong size!", __PRETTY_FUNCTION__);
        goto exitCreateMetadataString;
    }

    // Wrap this logic in a try/catch block.
    try
    {
        // Setup a bitmode user readable string based on metadata bit_sel setting.
        switch (meta->bit_sel)
        {
        case GIGE_BITMODE_LOWER_8_BITS:
            bitModeTypeStr = TS3_BIT_MODE_LOWER_8;
            break;
        case GIGE_BITMODE_MIDDLE_8_BITS:
            bitModeTypeStr = TS3_BIT_MODE_MIDDLE_8;
            break;
        case GIGE_BITMODE_UPPER_8_BITS:
            bitModeTypeStr = TS3_BIT_MODE_UPPER_8;
            break;
        case GIGE_BITMODE_10_BITS:
            bitModeTypeStr = TS3_BIT_MODE_FULL_10;
            break;
        case GIGE_BITMODE_7_0:  first = 0;  last = 7;   break;
        case GIGE_BITMODE_8_1:  first = 1;  last = 8;   break;
        case GIGE_BITMODE_9_2:  first = 2;  last = 9;   break;
        case GIGE_BITMODE_10_3: first = 3;  last = 10;  break;
        case GIGE_BITMODE_11_4: first = 4;  last = 11;  break;
        case GIGE_BITMODE_9_0:  first = 0;  last = 9;   break;
        case GIGE_BITMODE_10_1: first = 1;  last = 10;  break;
        case GIGE_BITMODE_11_2: first = 2;  last = 11;  break;
        case GIGE_BITMODE_11_0: first = 0;  last = 11;  break;
        default:
            bitModeTypeStr.assign(TS3_BIT_MODE_UNKNOWN);
            break;
        }

        // Skip if we already have a pre-defined string in place. Otherwise,
        // we'll want to build a string from the information.
        if (bitModeTypeStr.size() == 0)
        {
            bitModeTypeStr = fmt::sprintf(FMT_STRING("%d:%d"), last, first);
        }

        // Build up our string describing the binning and subsampling options.
        optionsStr = fmt::sprintf(FMT_STRING("bin%ux:subs%ux"),
                                  (meta->sensor_opts & GIGE_SENSOPT_BIN_MASK) >> GIGE_SENSOPT_BIN_SHIFT,
                                  (meta->sensor_opts & GIGE_SENSOPT_SUBS_MASK) >> GIGE_SENSOPT_SUBS_SHIFT);

        // Construct a string describing the trigger position, either percent or just frames.
        if (meta->trig_set <= 100)
            triggerSetStr = fmt::sprintf(FMT_STRING("%u%%"), meta->trig_set);
        else
            triggerSetStr = fmt::sprintf(FMT_STRING("%u"), meta->trig_set_frames);

        // Calculate time/date of trigger
        {
            sbs_to_tm(meta->trig_time_s, &sbs_tm);

            timeStr = fmt::sprintf("%02u:%03u:%02u:%02u:%02u.%06u",
                                   sbs_tm.tm_year, sbs_tm.tm_yday + 1, sbs_tm.tm_hour, sbs_tm.tm_min, sbs_tm.tm_sec,
                                   ((meta->trig_time_ns + 500) / 1000));
        }

        // Adjust frame count and trigger frame for selected clip
        frame_count = endFrame - startFrame + 1;

        // Now calculate the "relative" trigger offset. If our skip_count is non-zero,
        // we must be loading a partial CAP file. In this case, the trigger frame is
        // set properly in the metadata.
        if (meta->skip_count != 0)
            trig_frame_num = meta->trig_frame;

        // Otherwise, if the trigger frame is in front of the first frame, then we have
        // a "wrap" condition and we need to "straighten" that out.
        else if (meta->trig_frame < meta->first_frame)
            trig_frame_num = meta->frame_count - (meta->first_frame - meta->trig_frame);

        // Finally, for a straight forward capture, we just need to normalize the frame number.
        else
            trig_frame_num = meta->trig_frame - meta->first_frame;

        // Generate the Image meta data string.
        metadataStringRef = fmt::sprintf("[Image]\n"
                                         "\troi_x=%u\n"
                                         "\troi_y=%u\n"
                                         "\twidth=%u\n"
                                         "\theight=%u\n"
                                         "\tbit_mode=%s\n"
                                         "\tsensor_options=%s\n"
                                         "\tframe_count=%u\n"
                                         "\ttrigger_frame=%d\n"
                                         "\tstart_frame=%d\n"
                                         "\tend_frame=%d\n"
                                         "\ttime_stamp=%s\n"
                                         "\tcomment=%s\n"
                                         "\n",
                                         meta->roi_x, meta->roi_y, meta->roi_width, meta->roi_height, bitModeTypeStr.c_str(), optionsStr.c_str(),
                                         frame_count, trig_frame_num, startFrame, endFrame, timeStr.c_str(), (const char *)meta->comment);

        // Generate the Camera meta data string.
        metadataStringRef += fmt::sprintf("[Camera]\n"
                                          "\tmake=%s\n"
                                          "\tmodel=%s\n"
                                          "\tfpga_rev=0x%08x\n"
                                          "\tsoftware_version=%s\n"
                                          "\tmac_address=%02x:%02x:%02x:%02x:%02x:%02x\n"
                                          "\tcamera_name=%s\n"
                                          "\tsensor_type=%4s\n"
                                          "\n",
                                          meta->man_name, meta->model_name, meta->fpga_rev,
                                          meta->app_ver, meta->mac_addr[0], meta->mac_addr[1], meta->mac_addr[2],
                                          meta->mac_addr[3], meta->mac_addr[4], meta->mac_addr[5],
                                          meta->camera_name, (const char *)(&meta->sensor_type));

        // Generate the Record meta data string.
        metadataStringRef += fmt::sprintf("[Record]\n"
                                          "\tfps=%u\n"
                                          "\tshutter_speed=%u\n"
                                          "\tmulti_slope=%u:%u\n"
                                          "\ttrigger_setting=%s\n"
                                          "\tsync_in=0x%x\n"
                                          "\tsync_out=0x%x\n"
                                          "\n",
                                          meta->fps, meta->shutter_speed, meta->multslope[0], meta->multslope[1],
                                          triggerSetStr.c_str(), meta->sync_in_set, meta->sync_out_set);


        // Set up the individual red, green and blue gain values.
        if (gainsPtr != nullptr)
        {
            red_gain = gainsPtr->red;
            green_gain = gainsPtr->green;
            blue_gain = gainsPtr->blue;
        }
        else
        {
            red_gain = meta->red;
            green_gain = meta->green;
            blue_gain = meta->blue;
        }

        // Generate the Normalization meta data string.
        metadataStringRef += fmt::sprintf("[Normalization]\n"
                                          "\tred_balance=%d\n"
                                          "\tblue_balance=%d\n"
                                          "\tgreen_balance=%d\n"
                                          "\tbrightness=%d\n"
                                          "\tcontrast=%d\n"
                                          "\tgamma=%d\n"
                                          "\tsensor_gain=%u\n"
                                          "\tred_gain=%u.%u\n"
                                          "\tblue_gain=%u.%u\n"
                                          "\tgreen_gain=%u.%u\n"
                                          "\tred_matrix=[%d,%d,%d]\n"
                                          "\tblue_matrix=[%d,%d,%d]\n"
                                          "\tgreen_matrix=[%d,%d,%d]\n"
                                          "\tred_offset=%d\n"
                                          "\tgreen_offset=%d\n"
                                          "\tblue_offset=%d\n",
                                          meta->wb_red, meta->wb_blu, meta->wb_grn, meta->brightness,
                                          meta->contrast, meta->gamma, meta->sensor_gain,
                                          (red_gain >> 8), (((((red_gain & 0xFF) * 10000) / 256) + 5) / 10),
                                          (blue_gain >> 8), (((((blue_gain & 0xFF) * 10000) / 256) + 5) / 10),
                                          (green_gain >> 8), (((((green_gain & 0xFF) * 10000) / 256) + 5) / 10),
                                          (int16_t)meta->matrix_coeff1_red, (int16_t)meta->matrix_coeff2_red,/**/
                                          (int16_t)meta->matrix_coeff3_red, (int16_t)meta->matrix_coeff1_blu,/**/
                                          (int16_t)meta->matrix_coeff2_blu, (int16_t)meta->matrix_coeff3_blu,/**/
                                          (int16_t)meta->matrix_coeff1_grn, (int16_t)meta->matrix_coeff2_grn,/**/
                                          (int16_t)meta->matrix_coeff3_grn,
                                          (int16_t)meta->dc_offset_red, (int16_t)meta->dc_offset_grn, (int16_t)meta->dc_offset_blu);
    }

    // Look for a specific resource allocation error to set state of the reader object.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while preparing metadata string.", __PRETTY_FUNCTION__, e.what());
        goto exitCreateMetadataString;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught while metadata string.", __PRETTY_FUNCTION__);
        goto exitCreateMetadataString;
    }

    // If make it here the metadataStringRef should have been created correctly.
    resultCode = (int32_t) metadataStringRef.size();

exitCreateMetadataString:
    return resultCode;
}

/*!
 * @brief Fills a buffer with information from the metadata in a format specific
 * to the xcitex binary cfg file format.
 *
 * meta is of type pm_metadata, void * is used to prevent the interface from
 * being dependent upon gige_msg.h which is inside the api and because of
 * copyright issues can't be used outside the api.
 *
 * @param xcfgMetaDataInBytes   Caller provides reference to metadata vector that will be formatted internally here.
 * @param pmMetadataPtr         A read-only pointer to a pm_metadata data structure.  See comment above.
 * @param pmMetadataSize        How many uint8_t bytes pointed to by the metaPtr.
 * @param startFrame            The starting frame number for the capture.
 * @param endFrame              The ending frame number for the capture.
 *
 * @returns the number of bytes written or a negative number on failure.
 */
int32_t FiVideo::CreateX_CFG_MetadataString(std::unique_ptr<FAS_VIDEO::X_CFG_METADATA_VECTOR>& xcfgMetaDataInBytes,
                                            const void *pmMetadataPtr, uint32_t pmMetadataSize,
                                            uint32_t startFrame, uint32_t endFrame)
{
    int32_t resultSize = -1;

    // Pull the read-only metadata from the camera system.
    const pm_metadata *meta = (const pm_metadata *)pmMetadataPtr;

    // Create the specific xcite metadata record.
    XCFG_METADATA_REC xcfgMetadataRec;

    // Create a local variable for the record size.
    uint32_t xcfgMetadataRecSize = sizeof(XCFG_METADATA_REC);

    // Unused...
    (void)startFrame;
    (void)endFrame;

    // Sanity checks.
    if (pmMetadataPtr == nullptr)
    {
        fi_log(LOG_ERR, "%s: pmMetadataPtr passed was nullptr.", __PRETTY_FUNCTION__);
        goto exitCreateX_CFG_MetadataString;
    }
    if (pmMetadataSize != sizeof(pm_metadata))
    {
        fi_log(LOG_ERR, "%s: %u=pmMetadataSize != sizeof(pm_metadata[%lu]).",
               __PRETTY_FUNCTION__, pmMetadataSize, (unsigned long)sizeof(pm_metadata));
        goto exitCreateX_CFG_MetadataString;
    }


    // Wrap this logic in a try/catch block.
    try
    {
        // Set the record to zeroes.
        memset(&xcfgMetadataRec, 0, xcfgMetadataRecSize);

        // The record rate is our frames-per-second.
        xcfgMetadataRec.record_rate = meta->fps;

        // Assume "zero_position" means trigger frame.
        //
        // If we have a skip count, then we're working with a parital CAP file.
        // In this case, the trigger frame is set properly in the metadata.
        if (meta->skip_count != 0)
        {
            xcfgMetadataRec.zero_position = meta->trig_frame;
        }

        // Otherwise, calculate the trigger frame as follows:
        //  We can have 1 of 2 situations illustrated below:
        //   A)   0   first_frame     trigger_frame       frame_count
        //   B)   0   trigger_frame   first_frame         frame_count
        else if (meta->first_frame <= meta->trig_frame)
        {
            xcfgMetadataRec.zero_position = meta->trig_frame - meta->first_frame;
        }
        else
        {
            xcfgMetadataRec.zero_position = (meta->frame_count - meta->first_frame) + meta->trig_frame;
        }

        // Zero-position is now calculated as an INDEX. However, Midas
        // wants a count, so bump our number by 1.
        xcfgMetadataRec.zero_position++;

        // Convert the shutter speed.
        xcfgMetadataRec.shutter_speed = meta->shutter_speed * MICRO_TO_NANO_SCALE;

        // Make sure we have a frame count.
        if (meta->frame_count > 0)
            xcfgMetadataRec.burst_frames = meta->frame_count;
        else
        {
            xcfgMetadataRec.burst_frames = 1;
            fi_log(LOG_WARNING, "%s: Metadata frame count = 0?!.", __PRETTY_FUNCTION__);
        }

        // Copy the bit-depth field over.
        xcfgMetadataRec.depth_mode = meta->bit_sel;

        // Calculate trigger percentage rounding to nearest percent
        xcfgMetadataRec.trigger_setting = ((xcfgMetadataRec.zero_position * 100) + (xcfgMetadataRec.burst_frames / 2)) / xcfgMetadataRec.burst_frames;

        // Create the vector to hold the contents of the data structure.
        xcfgMetaDataInBytes = std::unique_ptr<FAS_VIDEO::X_CFG_METADATA_VECTOR>(new FAS_VIDEO::X_CFG_METADATA_VECTOR(xcfgMetadataRecSize));
        if (xcfgMetaDataInBytes.get() == nullptr)
        {
            fi_log(LOG_DEBUG, "%s: Unable to allocate xcfgMetaDataInBytes!", __PRETTY_FUNCTION__);
            goto exitCreateX_CFG_MetadataString;
        }

        // Make sure the size of the vector matches the size expected.
        if (xcfgMetaDataInBytes->size() != xcfgMetadataRecSize)
        {
            fi_log(LOG_ERR, "%s: xcfgMetaDataInBytes[%u] != xcfgMetadataRecSize[%u].",
                   __PRETTY_FUNCTION__, xcfgMetaDataInBytes->size(), xcfgMetadataRecSize);
            goto exitCreateX_CFG_MetadataString;
        }

        // Copy the data in the data structure into the vector returned.
        memcpy((uint8_t *) xcfgMetaDataInBytes->data(), (uint8_t *) &xcfgMetadataRec, xcfgMetadataRecSize);

        // If make it here the data was transferred.
        resultSize = (int32_t) xcfgMetadataRecSize;
    }
    // Look for a specific resource allocation error to set state of the reader object.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while preparing xcite metadata string.", __PRETTY_FUNCTION__, e.what());
        goto exitCreateX_CFG_MetadataString;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught while xcite metadata string.", __PRETTY_FUNCTION__);
        goto exitCreateX_CFG_MetadataString;
    }

exitCreateX_CFG_MetadataString:
    return resultSize;
}

/*!
 * @brief Utility function to return a human readable string for a FourCC code.
 *
 * @param fourccToTranslate   The FourCC code to translate into a string.
 * @returns The human readable string.
 */
const std::string FiVideo::GetFourCCReadableString(uint32_t fourccToTranslate)
{
    std::string readableFourCCStr;
    uint8_t readableFourCC[5];
    uint8_t *pFourCC = (uint8_t *) &fourccToTranslate;
    readableFourCC[0] = *(pFourCC + 0);
    readableFourCC[1] = *(pFourCC + 1);
    readableFourCC[2] = *(pFourCC + 2);
    readableFourCC[3] = *(pFourCC + 3);
    readableFourCC[4] = 0;

    // Plop it into the string.
    readableFourCCStr = (const char *)readableFourCC;

    return readableFourCCStr;
}

/*!
 * @brief Utility function to return the string we append to video files
 * when the stream writer wasn't able to complete the video due to an error
 * or when the user cancels the save operation.
 *
 * @returns The string our stream writer uses.
 */
const std::string& FiVideo::GetIncompleteVideoFileString()
{
    return FiVideoStream::GetIncompleteVideoFileString();
}

