/*!
 * @file ffmpeg_encoder.cpp
 * @brief Interface for FFmpeg encoder object.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <sys/stat.h>
#include <string>
#include <memory>
#include <mutex>
#include <deque>
#include <chrono>
#include <fmt/printf.h>

#ifdef  __cplusplus
extern "C" {
#endif

#include "libavformat/avformat.h"
#include "libavformat/avio.h"
#include "libavutil/pixfmt.h"
#include "libavutil/opt.h"
#include "libavcodec/avcodec.h"

#ifdef  __cplusplus
}
#endif

#include "fi_durations.h"
#include "ffmpeg_container.h"
#include "ffmpeg_encoder.h"
#include "ffmpeg_filter.h"
#include "ffmpeg_status.h"

#include "log_if.h"
#include "fas_thread.h"

// Make things easier for usage.
using namespace std;
using namespace FAS_VIDEO;
using namespace FI_FFMPEG;
using namespace FAS_THREAD;

// Turn these debugging loggers on by uncommenting the define - to remove it comment the line.
//#define LOG_ENCODER_FRAME_INFO 1

namespace FI_FFMPEG
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! If you want to turn on verbose logging for this object set this to 1
//static const bool VERBOSE_SAVE_LOG = true;
const bool VERBOSE_SAVE_LOG = false;

//! If we want to check for hardware encoder support this can be set to true.
//! This will allow for the moment a compile-specific option for managing the
//! support at the build-level.
const bool CHECK_FOR_HW_ACCELERATION = true;
//const bool CHECK_FOR_HW_ACCELERATION = false;

const static std::string HW_CODEC_EXISTS           = "FFmpeg HW Codec Available"s;
const static std::string HW_CODEC_NOT_AVAILABLE    = "FFmpeg HW Codec Not Available"s;
const static std::string HW_CODEC_NO_CHECKING      = "Not set to check for HW Codecs"s;

//! Create a constant to use for logging representing the Frame Writer Thread.
const std::string FRAME_WRITER_THREAD_PROCESS_NAME = "[FrameWriterThread]"s;

//! When adding frames to the framesToWrite queue the stream writer
//! may send them faster than the frameWriterThread can process them.
//! When that happens the process that adds the frame to the framesToWrite
//! Queue will relieve the backpressure and wait a little bit to allow for
//! the frameWriterThread to catch up processing the framesToWrite until
//! it is reduced to a threshold.  But what if the frameWriterThread never
//! gets its queue down to a FRAME_QUEUE_RESUME_SIZE?
const uint32_t MS_THRESHOLD_FOR_FRAME_QUEUE_RESUME = (1000 * 4);

//! Define some ceiling for the number of frames added to the FIFO
//! frame queue for encoding.  Once that max is reached the FIFO will
//! not accept any more frames until enough have been processed by the
//! FrameWriter thread and the size of the FIFO queue is at resume size.
const uint32_t FRAME_QUEUE_MAX_SIZE = 80;
const uint32_t FRAME_QUEUE_RESUME_SIZE = 60;
const uint32_t FRAME_QUEUE_HOLD_MS = 100;       //!< How long to sleep for if holding.

//! These types identify a frame being added to the FFmpegFrameQueue
//! object that maintains the "stream" between the mp4 stream writer and
//! the FrameWriterThread.
typedef enum ENCODER_FRAME_TYPES
{
    NO_TYPE = 0,
    DATA_FRAME_TO_ENCODE,
    EMPTY_FRAME_TO_FLUSH
} ENCODER_FRAME_TYPES;

//! The FFmpegEncoderFrame Class is a specialized class containing information about
//! the AVFrame of data to write to the FFmpeg's codec stream for encoding.  The proper
//! factory creation method is invoked by the FFmpegEncoder class itself whenever it
//! receives an external FiImage to write out to the streamwriter.
class FFmpegEncoderFrame
{
public:
    FFmpegEncoderFrame();
    virtual ~FFmpegEncoderFrame();

    //! This object is non-copyable and non-copy assignable.
    FFmpegEncoderFrame(const FFmpegEncoderFrame& i) = delete;
    FFmpegEncoderFrame& operator=(const FFmpegEncoderFrame& i) = delete;

    //! Factory creation methods.
    static std::shared_ptr<FFmpegEncoderFrame> CreateFFmpegEncoderFrame(AVFrame *theAVFrameToWrite);
    static std::shared_ptr<FFmpegEncoderFrame> CreateFFmpegFlushFrame();

    bool isADataFrame()     { return (myType == ENCODER_FRAME_TYPES::DATA_FRAME_TO_ENCODE); }
    bool isAFlushFrame()    { return (myType == ENCODER_FRAME_TYPES::EMPTY_FRAME_TO_FLUSH); }

    uint32_t frameId;                   //!< One-based frame id for order added to queue.

    AVFrame *GetFrameToWrite()  { return frameToWrite; }
    uint32_t GetFrameStats(bool markNow);

private:

    AVFrame *frameToWrite;                  //!< The filtered frame for writing to the encoder.
    ENCODER_FRAME_TYPES myType;             //!< This frame can be in multiple forms.  This identifies those forms

    // Keep track of this frame's lifetime.
    FiDurations frameStats;
};

//! The FFmpegFrameQueue Class is a specialized class containing information
//! about a queue of FFmpegEncoderFrame objects.  This queue will be used by
//! both the FFmpegEncoder and the FFmpegFrameWriterThread objects and the
//! frames on it are managed in a FIFO manner.
class FFmpegFrameQueue
{
private:
    FFmpegFrameQueue()
    {
        dataFramesAdded         = 0;
        dataFramesExpected      = 0;
        doneStreamingFrames     = false;
        flushFramesAdded        = 0;
        queueHighMark           = 0;
        queueLongestWait        = 0;
    }

public:

    ~FFmpegFrameQueue();

    //! This object is non-copyable and non-copy assignable.
    FFmpegFrameQueue(const FFmpegFrameQueue& i) = delete;
    FFmpegFrameQueue& operator=(const FFmpegFrameQueue& i) = delete;

    //! Factory creation method.
    static std::unique_ptr<FFmpegFrameQueue> CreateFFmpegFrameQueue(uint32_t frameCountExpected);

    //! @name Public Property Accessor Functions.
    //!
    //! These are the public functions used to access the properties of this object.
    //! @{
    //!
    bool        AddFrameToQueue(std::shared_ptr<FFmpegEncoderFrame> ffmpegFrameToAdd);
    bool        FrameStreamHasStopped();
    uint32_t    LargestQueueSize();
    uint32_t    LongestFrameWait();
    void        LogStatistics();
    uint32_t    NumberOfFramesInQueue();
    std::shared_ptr<FFmpegEncoderFrame> WaitAndPopFirstFrame();

    //! @}

private:
    //! Protect access to frameQueueHasData condition variable.
    mutable std::mutex frameQueueMutex;
    std::condition_variable frameQueueHasData;

    //! Protect access to ffmpegFrames FIFO queue.
    mutable std::mutex ffmpegFramesMutex;
    std::deque<std::shared_ptr<FFmpegEncoderFrame>> ffmpegFrames;

    //! How many frames have been added to this queue.  It is used to set
    //! the frameId (zero-based) of frames added in the queue.
    uint32_t dataFramesAdded;

    //! How many data frames are expected?
    uint32_t dataFramesExpected;

    //! When there is no longer a need to look for queued up frames this
    //! flag will be set.
    bool doneStreamingFrames;

    //! How many flush frames have been added to the queue.  There should be 1.
    uint32_t flushFramesAdded;

    uint32_t queueHighMark;     //!< Highest number of frames in this queue at one time.
    uint32_t queueLongestWait;  //!< Longest wait time for life of an FFmpegEncoderFrame.
};

//! The Frame Writer Thread will be a background process spawned when the
//! the ffmpeg encoder class is prepared.
class FrameWriterThread : public FAS_THREAD::Thread
{
public:
    virtual ~FrameWriterThread();

    static std::unique_ptr<FI_FFMPEG::FrameWriterThread> CreateFrameWriter();

    void    cleanup();
    bool    is_stop_requested()     { return stop_request; }
    bool    is_running()            { return thread_running; }

    virtual void start(THREAD_STARTUP threadFn = nullptr);
    virtual void stop();

    bool EncodeAndWriteAVFrame(std::shared_ptr<FI_FFMPEG::FFmpegEncoderFrame> theFrameToEncodeAndWrite);

    // Manages the frame writer status record structure.
    static void InitStatusRec(FRAME_WRITER_STATUS_REC& status);
    void GetStatusRec(FRAME_WRITER_STATUS_REC& status);
    void LogStatistics();

private:
    FrameWriterThread();

    // This object is non-copyable and non-copy assignable.
    FrameWriterThread(const FrameWriterThread& i) = delete;
    FrameWriterThread& operator=(const FrameWriterThread& i) = delete;

    virtual void run();

private:
    volatile bool thread_running;           //!< Flag to indicate if we are running or not.
    volatile bool stopFrameWriterThread;    //!< Flag to indicate this thread is shutting down.
    volatile bool processedFlushFrame;      //!< All streams need a flush frame has it been processed?

    static std::mutex frameWriterStatusRecMutex;

    uint32_t            frameCount;         //!< How many frames are we expecting?
    uint32_t            encodedPacketCount; //!< How many packets encoded to mp4 stream.
    AVFormatContext    *avFormatContext;    //!< The Encoder's FFmpeg AVFormatContext record.
    AVCodecContext     *avCodecContext;     //!< The Encoder's FFmpeg AVCodecContext record.

    FRAME_WRITER_STATUS_REC myStatus;       //!< Keep tracks of the status within this frame writer thread.

    std::shared_ptr<FiDurations> threadStats;       //!< Keep statistics of run time for thread.
    std::shared_ptr<FiDurations> dataFrameStats;    //!< Keep statistics of data frame encode time.
    std::shared_ptr<FiDurations> flushFrameStats;   //!< Keep statistics of flush frame encode time.
    std::shared_ptr<FAS_VIDEO::FiDurations> totalEncoderTime;

    void PushStatusUpdate();
};

} // End of the namespace FI_FFMPEG

//--------------------------------------------------------------------
// Static Instances privately managed in this module.
//--------------------------------------------------------------------
extern std::unique_ptr<FFmpegFrameQueue> framesToWrite;
extern std::unique_ptr<FI_FFMPEG::FrameWriterThread> frameWriterThreadInstance;
extern std::mutex FFmpegEncoderInstanceMutex;

//! The framesToWrite is a FIFO FFmpegFrameQueue designed to collect and manage
//! FFmpegEncoderFrame data created by a FFmpegEncoder class and consumed by the
//! FI_FFMPEG::FrameWriterThread class.
//!
//! The FFmpegEncoder will create and add FFMpegEncoderFrames as fast as
//! it can and the frameWriterThreadInstance will consume and process
//! these frames as fast as it can in the background.
std::unique_ptr<FFmpegFrameQueue> framesToWrite = nullptr;

//! The frameWriterThreadInstance is the FI_FFMPEG::FrameWriterThread that
//! manages the actual encoding process for the FFmpegEncoder class.  It runs
//! in the background and it operates on frames added to the framesToWrite
//! FIFO queue.
// The one the only me!  The FrameWriterThread.
std::unique_ptr<FI_FFMPEG::FrameWriterThread> frameWriterThreadInstance = nullptr;

//!< Broker access to frameWriterStatusRec.
std::mutex FrameWriterThread::frameWriterStatusRecMutex;

//! theFrameWriterStatus record is shared between the singleton FFmpegEncoder
//! "FFmpegEncoderInstance" and the frameWriterThreadInstance.
static FRAME_WRITER_STATUS_REC frameWriterStatusRec;

//! The FFmpegEncoder is a singleton and a static instance declared here.  There will always
//! only ever be one associated with a stream writer and only one stream writer is ever active
//! at a single time.  If this changes this will not accomodate that architecture.
std::shared_ptr<FFmpegEncoder> FFmpegEncoder::FFmpegEncoderInstance = nullptr;
std::mutex FFmpegEncoderInstanceMutex;   //!< Broker access to FFmpegEncoderInstance.

//! If using hardware acceleration this is FFmpeg hardware device ctx returned
//! by the FFmpeg library call av_hwdevice_ctx_create() when the encoder
//! object is being created in its factory.
AVBufferRef *FFmpegEncoder::HWDeviceCTX = nullptr;

//--------------------------------------------------------------------
// Class implementation of FFmpegEncoderFrame.
//--------------------------------------------------------------------
/*!
 * @brief Construction of FFmpegEncoderFrame
 */
FFmpegEncoderFrame::FFmpegEncoderFrame()
{
    frameToWrite            = nullptr;
    myType                  = ENCODER_FRAME_TYPES::NO_TYPE;
    frameId                 = 0;        // Initialize to zero but it will be set by framesToWrite queue object.
}

FFmpegEncoderFrame::~FFmpegEncoderFrame()
{

    // Log the life of this frame.
    if (VERBOSE_SAVE_LOG)
    {
        std::string infoStr = "[~FFmpegEncoderFrame()]";
        fi_log(LOG_DEBUG, "%s: Deleting myType[%s] frameId[%d] and AVFrameToWrite[%p]",
               infoStr.c_str(), myType == ENCODER_FRAME_TYPES::DATA_FRAME_TO_ENCODE ? "D" : "F",
               frameId, (void *)frameToWrite);
        frameStats.LogDurationInfo(infoStr.c_str());
    }

    // This is where the filtered ffmpeg AVFrame object is freed.
    if (frameToWrite != nullptr)
    {
        av_frame_free(&frameToWrite);
    }
    frameToWrite = nullptr;
}

/*!
 * @brief Construction of FFmpegEncoderFrame object.
 *
 * @param theAVFrameToWrite: The pointer to the "filtered" AVFrame created by the FFmpegEncoder
 * object and ownership will transfer to this object and is freed in the destructor.
 *
 * @returns a shared pointer of allocated FFmpegEncoderFrame object.
 */
std::shared_ptr<FFmpegEncoderFrame> FFmpegEncoderFrame::CreateFFmpegEncoderFrame(
    AVFrame *theAVFrameToWrite)
{
    std::shared_ptr<FFmpegEncoderFrame> ffmpegEncoderFrame = nullptr;

    if (theAVFrameToWrite == nullptr)
    {
        fi_log(LOG_ERR, "%s: theAVFrameToWrite is NULL!", __PRETTY_FUNCTION__);
    }
    else
    {
        // Create the shared pointer for the encoder frame wrapping the AVFrame pointer.
        ffmpegEncoderFrame = std::shared_ptr<FFmpegEncoderFrame>(new FI_FFMPEG::FFmpegEncoderFrame());
        ffmpegEncoderFrame->frameToWrite = theAVFrameToWrite;
        // The AVFrame is not null and the user is wanting to encode it in the stream.
        ffmpegEncoderFrame->myType = ENCODER_FRAME_TYPES::DATA_FRAME_TO_ENCODE;
    }
    return ffmpegEncoderFrame;
}

/*!
 * @brief Construction of FFmpegEncoderFrame object for a flusable frame.
 *
 * @returns a shared pointer of allocated FFmpegEncoderFrame object.
 */
std::shared_ptr<FFmpegEncoderFrame> FFmpegEncoderFrame::CreateFFmpegFlushFrame()
{
    std::shared_ptr<FFmpegEncoderFrame> ffmpegEncoderFrame = nullptr;

    // Create the shared pointer for the "empty" encoder frame to flush.
    ffmpegEncoderFrame = std::shared_ptr<FFmpegEncoderFrame>(new FI_FFMPEG::FFmpegEncoderFrame());
    ffmpegEncoderFrame->frameToWrite = nullptr;
    ffmpegEncoderFrame->myType = ENCODER_FRAME_TYPES::EMPTY_FRAME_TO_FLUSH;

    return ffmpegEncoderFrame;
}

/*!
 * @brief Called to get a time difference from now() to construction.
 *
 * @returns The number of milleseconds to process through encoder.
 */
uint32_t FFmpegEncoderFrame::GetFrameStats(bool markNow)
{
    if (markNow)
        frameStats.MarkInterval();

    uint32_t timeInMilliSec = frameStats.GetTotalIntervalTimeInMS();
    return timeInMilliSec;
}

//--------------------------------------------------------------------
// Class implementation of FFmpegFrameQueue.
//--------------------------------------------------------------------

/*!
 * @brief FFmpegFrameQueue destructor.
 */
FFmpegFrameQueue::~FFmpegFrameQueue()
{
    if (VERBOSE_SAVE_LOG)
    {
        fi_log(LOG_DEBUG, "FFmpegFrameQueue: Destruction.");
    }


    // Will be set to any residual frames in this queue.
    std::shared_ptr<FFmpegEncoderFrame> frameInQueue = nullptr;

    // Cleanup the encoder frames in this queue.
    // This *SHOULD* not have any frames in it.  If it does then we
    // should cleanup the frames ourselves.
    while (!ffmpegFrames.empty())
    {
        frameInQueue = ffmpegFrames.front();

        if (VERBOSE_SAVE_LOG)
        {
            int frameIdSet;
            if (frameInQueue)
                frameIdSet = (int)frameInQueue->frameId;
            else
                frameIdSet = -1;
            fi_log(LOG_DEBUG, "FFMPEG_FRAME_QUEUE: MANUALLY DELETING FRAME[%d] IN QUEUE.", frameIdSet);
        }

        // Remove the frame entry from the queue.
        ffmpegFrames.pop_front();
        if (frameInQueue)
            frameInQueue.reset();
    }
}

/*!
 * @brief Construction of FFmpegFrameQueue object.
 *
 * @returns a unique pointer of allocated FFmpegFrameQueue object.
 */
std::unique_ptr<FFmpegFrameQueue> FFmpegFrameQueue::CreateFFmpegFrameQueue(uint32_t frameCountExpected)
{
    std::unique_ptr<FFmpegFrameQueue> ffmpegFrameQueue = nullptr;
    // Create the FFmpeg FIFO Frame Queue.
    ffmpegFrameQueue = std::unique_ptr<FFmpegFrameQueue>(new FI_FFMPEG::FFmpegFrameQueue());

    // Save off the expected number of data frames for safeguard.
    ffmpegFrameQueue->dataFramesExpected = frameCountExpected;

    return ffmpegFrameQueue;
}

/*!
 * @brief Returns the number of frames in this queue.
 * @returns Number of frames or zero if empty.
 */
uint32_t FFmpegFrameQueue::NumberOfFramesInQueue()
{
    uint32_t queueSize = (uint32_t)ffmpegFrames.size();
    return queueSize;
}

/*!
 * @brief Add a frame to the queue.
 * @param ffmpegFrameToAdd:  The frame to add to the queue at the end (the queue
 * is a FIFO so the first in must remain there and we add to end).
 * @returns True if added, false if error.
 */
bool FFmpegFrameQueue::AddFrameToQueue(std::shared_ptr<FFmpegEncoderFrame> ffmpegFrameToAdd)
{
    bool theFrameWasAdded = false;

    // There should be a frame writer thread instance running at this
    // point.  If it is not we cannot add anymore to this queue.
    if (frameWriterThreadInstance)
    {
        if (frameWriterThreadInstance->is_running())
        {
            // Push the frame to the end of the queue.
            if (ffmpegFrameToAdd)
            {
                // Lock access to our frame queue.
                {
                    std::lock_guard<std::mutex> guard(ffmpegFramesMutex);
                    ffmpegFrames.push_back(ffmpegFrameToAdd);
                }

                theFrameWasAdded = true;

                // Only increment the data frames added to the queue.
                // A flushing frame is operational not data specific.
                if (ffmpegFrameToAdd->isADataFrame())
                {
                    dataFramesAdded++;
                    // The frame ID represents a one-based frame id.
                    ffmpegFrameToAdd->frameId = dataFramesAdded;
                }
                else if (ffmpegFrameToAdd->isAFlushFrame())
                {
                    flushFramesAdded++;
                    ffmpegFrameToAdd->frameId = dataFramesAdded + flushFramesAdded;
                }

                // Set the condition variable.
                {
                    std::lock_guard<std::mutex> guard(frameQueueMutex);
                    frameQueueHasData.notify_one();
                }
            }
        }
        else
        {
            fi_log(LOG_WARNING, "%s: The frameWriterThreadInstance is not running! dataFramesAdded[%d]",
                   __PRETTY_FUNCTION__, dataFramesAdded);
        }
    }
    else
    {
        fi_log(LOG_WARNING, "%s: There is no frameWriterThreadInstance! dataFramesAdded[%d]",
               __PRETTY_FUNCTION__, dataFramesAdded);
    }
    return theFrameWasAdded;
}

/*!
 * @brief Pop off the first frame in the queue for processing.
 * @returns frame in queue for processing or nullptr if nothing there.
 */
std::shared_ptr<FFmpegEncoderFrame> FFmpegFrameQueue::WaitAndPopFirstFrame()
{
    // Create a flag to let us know when we are done waiting here.
    bool keepWaiting = true;

    std::shared_ptr<FFmpegEncoderFrame> theFrameToReturn = nullptr;

    // Wait for either a frame to be put on the FIFO OR doneStreamingFrames
    // flag to be set.
    while (keepWaiting)
    {
        // Create a scope to mutex lock the wait_for call on frameQueueHasData.
        {
            // Lock access to our frame queue.
            std::unique_lock<std::mutex> guard(frameQueueMutex);
            frameQueueHasData.wait_for(guard, 100ms,  [this] {return (doneStreamingFrames || !ffmpegFrames.empty());});
        }

        // NOTE:  I don't want to process any frames in the queue if the flag
        // doneStreamingFrames is set because the stream is no longer being
        // processed and we don't want to provide any data.  The frames will
        // be released in the normal way during destruction.

        // We don't keep waiting if the doneStreamingFrames flag is set.
        if (doneStreamingFrames)
            keepWaiting = false;
        else
        {
            // Check for a frame to process from the stream.
            if (!ffmpegFrames.empty())
            {
                std::unique_lock<std::mutex> guard(ffmpegFramesMutex);
                uint32_t queueSize = (uint32_t)ffmpegFrames.size();

                // Mark a high-mark if we have hit one.
                if (queueSize > queueHighMark)
                    queueHighMark = queueSize;

                // A frame should exist pull the first one in queue.
                theFrameToReturn = ffmpegFrames.front();

                // Mark the longest time in queue if one is hit.
                if (theFrameToReturn)
                {
                    uint32_t timeInQueue = theFrameToReturn->GetFrameStats(true);
                    if (timeInQueue > queueLongestWait)
                        queueLongestWait = timeInQueue;
                }

                // Removes the entry from the queue and the ownership
                // of the frame moves to the consumer.
                ffmpegFrames.pop_front();

                // Stop waiting now that a frame is ready for processing.
                keepWaiting = false;
            }
        }
    }
    return theFrameToReturn;
}

/*!
 * @brief Returns the longest wait for an FFmpegEncoderFrame in the queue.
 */
uint32_t FFmpegFrameQueue::LongestFrameWait()
{
    return queueLongestWait;
}

/*!
 * @brief The frame writer thread is no longer processing the frame
 * stream so this frame queue will not be receiving any more frames.
 */
bool FFmpegFrameQueue::FrameStreamHasStopped()
{
    bool processedOk = false;

    //fi_log(LOG_DEBUG, "%s: We are stopping wake up our condition variable.", __PRETTY_FUNCTION__);

    doneStreamingFrames = true;

    // Wake up the frame queue condition variable.
    frameQueueHasData.notify_one();

    //fi_log(LOG_DEBUG, "%s: Notification sent we are done here.", __PRETTY_FUNCTION__);

    processedOk = true;

    return processedOk;
}

/*!
 * @brief Returns the largest queue size (the high water mark if you will)
 * while processing the FFmpegEncoderFrames.
 */
uint32_t FFmpegFrameQueue::LargestQueueSize()
{
    return queueHighMark;
}

/*!
 * @brief Logs out the statistics for the frame queue.
 */
void FFmpegFrameQueue::LogStatistics()
{
    if (VERBOSE_SAVE_LOG)
    {
        uint32_t queueSize = 0;
        std::string typeStr = "[FFMPEG FRAME FIFO]";

        // Do we still have frames in the FIFO queue?
        if (!ffmpegFrames.empty())
        {
            queueSize = (uint32_t)ffmpegFrames.size();
        }

        // Write the log information.
        fi_log(LOG_DEBUG, "%s dataFrames[%d] flushFrames[%d] framesLeft[%d]",
               typeStr.c_str(), dataFramesAdded, flushFramesAdded, queueSize);
        fi_log(LOG_DEBUG, "%s queueHighMark[%d] queueEntryLongestWait[%dms]",
               typeStr.c_str(), queueHighMark, queueLongestWait);
    }
}

//--------------------------------------------------------------------
// Class implementation: FrameWriterThread
//--------------------------------------------------------------------

FrameWriterThread::FrameWriterThread() :
    FAS_THREAD::Thread(false)   // Don't want a detached thread it CRASHES FasMotion!
{
    // Following lead from example code in camera server.
    thread_running = false;
    InitStatusRec(myStatus);
}

FrameWriterThread::~FrameWriterThread()
{
    if (VERBOSE_SAVE_LOG)
    {
        fi_log(LOG_DEBUG, "FrameWriterThread: Destruction.");
    }
    if (thread_running)
    {
        fi_log(LOG_CRIT, "FFMPEG FRAME WRITER THREAD WAS STILL RUNNING IN DESTRUCTOR!!!");
        FI_FFMPEG::FrameWriterThread::stop();
    }

    // Cleanup framesToWrite queue as thread is going away.
    if (framesToWrite)
        framesToWrite.reset();
}

/*!
 * @brief Clean up the thread on our way out.
 *
 * @param arg   Pointer to the worker class object.
 */
void FrameWriterThread::cleanup()
{
    // We're not running anymore.
    thread_running = false;
    stop_request = false;
    processedFlushFrame = false;
    stopFrameWriterThread = false;

    notify_done();
}

/*!
 * @brief The FrameWriterThread run method.
 *
 */
void FrameWriterThread::run()
{
    // When all frames are encoded we going into a flush frame waiting
    // condition - we  may force it or not tbd.
    bool waitingForFlushFrame = false;

    // The maximum frames expected in this video stream.
    frameCount          = 0;
    encodedPacketCount  = 0;
    avFormatContext     = nullptr;
    avCodecContext      = nullptr;

    try
    {
        // Make sure we're not running yet.
        if (thread_running)
        {
            fi_log(LOG_WARNING, "%s: Cannot allow multiple instances to run!", __PRETTY_FUNCTION__);
            return;
        }
        thread_running = true;  // We are "running" now.
        fi_log(LOG_DEBUG, "%s: Running Frame Writer Thread.", __PRETTY_FUNCTION__);

        if (threadStats == nullptr)
            threadStats = std::shared_ptr<FiDurations>(new FiDurations());
        else
            threadStats->MarkStart();   // Want to catch time through the loop.

        InitStatusRec(myStatus);        // Reset all status data.
        myStatus.frameWriterThreadStatus   = FRAME_WRITER_STATUS::ALL_OK;

        PushStatusUpdate(); // Set the shared status record at start.

        std::shared_ptr<FFmpegEncoder> ffmpegEncoder = FFmpegEncoder::GetEncoderInstance();
        if (!ffmpegEncoder)
        {
            fi_log(LOG_WARNING, "%s: Missing FFmpegEncoder Instance!", __PRETTY_FUNCTION__);
            myStatus.ffmpegError = AVERROR(EINVAL);
        }
        else
        {
            // Set the information from the encoder instance.
            frameCount          = ffmpegEncoder->GetFrameCount();
            avFormatContext     = ffmpegEncoder->GetAVFormatContext();
            avCodecContext      = ffmpegEncoder->GetAVCodecContext();

            // Make sure the encoder was initialized and ready to go.
            if (avFormatContext == nullptr)
            {
                fi_log(LOG_WARNING, "%s: Missing FFmpegEncoder AVFormatContext Pointer!", __PRETTY_FUNCTION__);
                myStatus.ffmpegError = AVERROR(EINVAL);
            }
            if (avCodecContext == nullptr)
            {
                fi_log(LOG_WARNING, "%s: Missing FFmpegEncoder AVCodecContext Pointer!", __PRETTY_FUNCTION__);
                myStatus.ffmpegError = AVERROR(EINVAL);
            }
            if (frameCount == 0)
            {
                fi_log(LOG_WARNING, "%s: frameCount was ZERO!", __PRETTY_FUNCTION__);
                myStatus.ffmpegError = AVERROR(EINVAL);
            }
        }

        // If unable to initialize the FrameWriterThread can't continue.
        if (myStatus.ffmpegError != 0)
        {
            goto exitFrameWriterThreadRun;
        }

#ifdef LOG_ENCODER_FRAME_INFO
        fi_log(LOG_DEBUG, "%s: B4 WHILE ffmpegEOF[%d] framesEncoded[%d] frameCount[%d] ffmpegError[%d]!",
               FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), myStatus.ffmpegEOF, myStatus.framesEncoded, frameCount, myStatus.ffmpegError);
#endif

        // This thread needs to continually monitor the framesToWrite queue
        // and processing frames as they become available until the encoder
        // returns the end of file condition or the thread is terminated.
        while (!stopFrameWriterThread && !processedFlushFrame && !myStatus.ffmpegEOF && (myStatus.ffmpegError == 0))
        {
            // Because the FIFO frame queue is set up to be a thread-safe
            // data queue we can simply call the wait-and-pop-first logic
            // here in this thread.
            std::shared_ptr<FFmpegEncoderFrame> encoderFramePtr = framesToWrite->WaitAndPopFirstFrame();

            // Check for stop mode in case it came through while we were waiting.
            if (stopFrameWriterThread)
            {
                fi_log(LOG_DEBUG, "%s: stopFrameWriterThread is true will now exit run().", __PRETTY_FUNCTION__);
                goto exitFrameWriterThreadRun;
            }

            // Process the encoder frame from the framesToWrite FIFO.
            if (encoderFramePtr)
            {
                if (totalEncoderTime == nullptr)
                    totalEncoderTime = std::shared_ptr<FiDurations>(new FiDurations());
                else
                    totalEncoderTime->MarkStart();    // Building stats on the Encode and Write Frame.

                // Is this frame a good data frame to send to the encoder?
                if (encoderFramePtr->isADataFrame())
                {
                    // Start a duration statistic gathering of data frame encoding process.
                    if (dataFrameStats == nullptr)
                        dataFrameStats = std::shared_ptr<FiDurations>(new FiDurations());
                    else
                        dataFrameStats->MarkStart();    // Building stats on the Encode and Write Frame.

                    // Send the frame to the encoder.
                    if (!EncodeAndWriteAVFrame(encoderFramePtr))
                    {
                        fi_log(LOG_ERR, "%s: EncodeAndWriteAVFrame on Data Frame FAILED!", __PRETTY_FUNCTION__);
                    }

                    if (dataFrameStats)
                        dataFrameStats->MarkInterval();     // Mark a duration interval.
                }
                else if (encoderFramePtr->isAFlushFrame())
                {
                    // Start a duration statistic gathering of flush frame process.
                    // There should only be one in the queue.
                    flushFrameStats = std::shared_ptr<FiDurations>(new FiDurations());

                    // A null AVFrame pointer tells the FFMpeg encoder to flush all
                    // data in it's queue as no more frames will be sent.

                    // Send the empty flush frame to the encoder.
                    if (!EncodeAndWriteAVFrame(encoderFramePtr))
                    {
                        fi_log(LOG_ERR, "%s: EncodeAndWriteAVFrame on flush frame FAILED!",
                               FRAME_WRITER_THREAD_PROCESS_NAME.c_str());
                    }

                    processedFlushFrame = true;
                    waitingForFlushFrame = false;
                }
                else
                {
                    // This is just a programmatic or corruption type error.
                    fi_log(LOG_ERR, "%s: Unknown FFmpegEncoderFrame Type! Setting Invalid Argument Error Condition!",
                           FRAME_WRITER_THREAD_PROCESS_NAME.c_str());
                    myStatus.frameWriterThreadStatus = FRAME_WRITER_STATUS::INVALID_ENCODER_FRAME_TYPE;
                    myStatus.ffmpegError = AVERROR(EINVAL);
                }

                // Can free the encoder frame at this point.
                encoderFramePtr.reset();

                // Push status update at the end of every frame processed to share status
                // information with the FFmpegEncoder class.
                PushStatusUpdate();

                if (totalEncoderTime)
                    totalEncoderTime->MarkInterval();     // Mark a duration interval.

            }  // End of the process frame condition.

        } // End of the run condition - the while loop...

        fi_log(LOG_DEBUG, "%s Run() has completed! stopFrameWriterThread[%s] ffmpegEOF[%s] ffmpegError[%d] waitingForFlushFrame[%s]",
               FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), stopFrameWriterThread ? "Y" : "N", myStatus.ffmpegEOF ? "T" : "F", myStatus.ffmpegError,
               waitingForFlushFrame ? "T" : "F");
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s Exception %s caught in frame writer thread.", FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), e.what());
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s (...) Exception caught in frame writer thread.", FRAME_WRITER_THREAD_PROCESS_NAME.c_str());
    }

exitFrameWriterThreadRun:
    PushStatusUpdate();     // Set the exiting status information.

    if (threadStats)
        threadStats->MarkInterval();    // Mark the end of the run thread.

    // Clean up on the way out.
    cleanup();
}

/*!
 * @brief Get the thread running.
 */
void FrameWriterThread::start(THREAD_STARTUP /*threadFn*/)
{
    fi_log(LOG_DEBUG, "%s Starting FrameWriterThread.", FRAME_WRITER_THREAD_PROCESS_NAME.c_str());

    // Make sure we're not running yet.
    if (thread_running)
    {
        fi_log(LOG_WARNING, "%s Cannot allow multiple FrameWriterThread instances to run!", FRAME_WRITER_THREAD_PROCESS_NAME.c_str());
        return;
    }

    // Clear our own "in_stop_mode" flag.
    stopFrameWriterThread = false;

    // Clear out the processed flushing frame flag.
    processedFlushFrame = false;

    // Start the thread.
    // This will end up calling start_thread() which ends up calling FrameWriterThread::run().
    FAS_THREAD::Thread::start();
}

/*!
 * @brief Stop the frame writer thread.
 */
void FrameWriterThread::stop()
{
    fi_log(LOG_DEBUG, "%s: Stopping FrameWriterThread.", __PRETTY_FUNCTION__);

    // User wants to stop this thread from running.
    stopFrameWriterThread = true;

    // Let the framesToWrite queue object know we are stopping.
    framesToWrite->FrameStreamHasStopped();

    // Issue the standard stop request.
    FAS_THREAD::Thread::stop();
}

/*!
 * @brief Factory to create a frame writer thread object.
 */
std::unique_ptr<FI_FFMPEG::FrameWriterThread> FrameWriterThread::CreateFrameWriter()
{
    std::unique_ptr<FI_FFMPEG::FrameWriterThread> theFrameWriterThread;
    theFrameWriterThread = std::unique_ptr<FI_FFMPEG::FrameWriterThread>(new FI_FFMPEG::FrameWriterThread());

    return theFrameWriterThread;
}

/*!
 * @brief Sets up a clean frame writer status record.
 */
void FrameWriterThread::InitStatusRec(FRAME_WRITER_STATUS_REC& status)
{
    status.framesEncoded            = 0;
    status.ffmpegEOF                = false;
    status.ffmpegError              = 0;
    status.frameWriterThreadStatus  = FI_FFMPEG::FRAME_WRITER_STATUS::ALL_OK;
    status.wroteAVTrailer           = false;
}

/*!
 * @brief Interface method to poll for status of the frame writer thread.
 */
void FrameWriterThread::GetStatusRec(FRAME_WRITER_STATUS_REC& status)
{
    std::lock_guard<std::mutex> guard(frameWriterStatusRecMutex);
    status.ffmpegEOF                = frameWriterStatusRec.ffmpegEOF;
    status.ffmpegError              = frameWriterStatusRec.ffmpegError;
    status.framesEncoded            = frameWriterStatusRec.framesEncoded;
    status.frameWriterThreadStatus  = frameWriterStatusRec.frameWriterThreadStatus;
    status.wroteAVTrailer           = frameWriterStatusRec.wroteAVTrailer;
}

/*!
 * @brief Pushes the current frame writer status to the shared status record.
 */
void FrameWriterThread::PushStatusUpdate()
{
    // Lock access to our data to read the status.
    std::lock_guard<std::mutex> guard(frameWriterStatusRecMutex);
    frameWriterStatusRec.ffmpegEOF                  = myStatus.ffmpegEOF;
    frameWriterStatusRec.ffmpegError                = myStatus.ffmpegError;
    frameWriterStatusRec.framesEncoded              = myStatus.framesEncoded;
    frameWriterStatusRec.frameWriterThreadStatus    = myStatus.frameWriterThreadStatus;
    frameWriterStatusRec.wroteAVTrailer             = myStatus.wroteAVTrailer;
}

/*!
 * @brief Logs important statistics from the frame writer.
 */
void FrameWriterThread::LogStatistics()
{
    std::lock_guard<std::mutex> guard(frameWriterStatusRecMutex);

    if (VERBOSE_SAVE_LOG)
    {
        std::string typeStr = FRAME_WRITER_THREAD_PROCESS_NAME.c_str();
        fi_log(LOG_DEBUG, "%s framesEncoded[%d] wroteAVTrailer[%s]", typeStr.c_str(),
               myStatus.framesEncoded, myStatus.wroteAVTrailer ? "Y" : "N");

        if (threadStats)
        {
            typeStr = "[FrameWriterThread::Run Lifetime of Thread]";
            threadStats->LogDurationInfo(typeStr.c_str());
        }

        typeStr = "[EncodeAndWriteAVFrame FFmpeg Encoder]";
        if (dataFrameStats)
            dataFrameStats->LogDurationInfo(typeStr.c_str());

        typeStr = "[Flushframe Encoding]";
        if (flushFrameStats)
            flushFrameStats->LogDurationInfo(typeStr.c_str());

        typeStr = "[Total FFmpeg Encode Time]";
        if (totalEncoderTime)
            totalEncoderTime->LogDurationInfo(typeStr.c_str());

        // Log statistics from out FIFO queue.
        if (framesToWrite)
            framesToWrite->LogStatistics();
    }
}

/*!
 * @brief Encodes and Writes frame packets out of FFmpeg Encoder.
 * Common logic to push the frame to encode to the FFmpeg encoder and pull and write
 * any encoded frame to the video stream using the FFmpeg write interleave function.
 *
 * @param theFrameToEncodeAndWrite is the FFmpegEncoderFrame object wrapping the AVFrame.
 * This can be a nullptr and in that case it's a flushing all frames to stream operation
 * when the stream no longer has input frames to process.
 *
 * @returns a true state frame was processed, false if error condition and ffmpegError was set.
 */
bool FrameWriterThread::EncodeAndWriteAVFrame(std::shared_ptr<FI_FFMPEG::FFmpegEncoderFrame> theFrameToEncodeAndWrite)
{
    bool keepReading            = true;         // True if we are going to keep reading encoded AVFrames.
    bool flushingFrame          = false;        // True if the flushing frame is being processed.
    bool frameWasProcessed      = false;        // True if frame is processed through FFmpeg.
    AVPacket *packetReceivedPtr = nullptr;      // The encoded AVPacket object.
    int receivePacketResult     = 0;            // Keeps track of whether or not the AVPacket was received.
    int avSendFrameResult       = 0;            // Keeps track of the av send frame result.

    try
    {
#ifdef LOG_ENCODER_FRAME_INFO
        fi_log(LOG_DEBUG, "%s EncodeAndWriteFrames ENTRY. framesEncoded[%d] encodedEOF[%d] ffmpegError[%d].",
               FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), myStatus.framesEncoded, myStatus.ffmpegEOF, myStatus.ffmpegError);
#endif

        // Allocate the packet for use with this frame.
        packetReceivedPtr = av_packet_alloc();
        if (packetReceivedPtr == nullptr)
        {
            // If cannot allocate the frame packet buffer we can't continue.
            myStatus.ffmpegError = AVERROR(ENOMEM);
            fi_log(LOG_ERR, "%s Unable to allocate AVPacket object.", FRAME_WRITER_THREAD_PROCESS_NAME.c_str());
            goto exitEncodeAndWriteAVFrame;
        }

        // The ffmpeg encoder takes in the AVFrame data but it may not provide an
        // encoded packet right away.  What happens is that the encoder needs to
        // consume (via avcodec_send_frame) enough frames to build up what internal
        // data it needs for its compression algorithms.  When requesting encoded
        // frames from ffmpeg encoder if the ffmpeg_result of call is EAGAIN setting
        // that indicates the encoder doesn't have enough data to provide an encoded
        // frame for the output stream yet.

        // First the frame is sent to the encoder's codec.
        avSendFrameResult = avcodec_send_frame(avCodecContext, theFrameToEncodeAndWrite->GetFrameToWrite());
        if (avSendFrameResult < 0)
        {
            log_ffmpeg_error_str(avSendFrameResult);
            myStatus.ffmpegError = avSendFrameResult;
            fi_log(LOG_ERR, "%s avcodec_send_frame() Failed.", FRAME_WRITER_THREAD_PROCESS_NAME.c_str());
            goto exitEncodeAndWriteAVFrame;
        }

        // Is this a flushing frame?
        if (theFrameToEncodeAndWrite->isAFlushFrame())
        {
            flushingFrame = true;
            if (flushFrameStats)
                flushFrameStats->MarkStart();   // Mark the duration starting point.
        }

        // Loop to receive encoded packets and write them to the output
        // stream for as long as data is available or there is an unexpected error.
        // There is no way of knowing how many frames the encoder will provide
        // to this loop but they all must be consumed and processed until out
        // of data for this run.
        // Note:  This loop is exited by a goto to exit.
        while (keepReading)
        {
            // Note:  Each time through make sure to wipe the AVPacket structure
            // by calling av_packet_unref(packetReceivedPtr).
            receivePacketResult = avcodec_receive_packet(avCodecContext, packetReceivedPtr);

            // If ffmpeg_result is zero - there is a writable encoded frame packet.
            if (receivePacketResult == 0)
            {
#ifdef LOG_ENCODER_FRAME_INFO
                fi_log(LOG_DEBUG, "%s **MUX** frame[%lld] frameCount[%d] framesEncoded[%d] encodedEOF[%d] Flushing[%s]",
                       FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), packetReceivedPtr->pts, frameCount, myStatus.framesEncoded, myStatus.ffmpegEOF, flushingFrame ? "T" : "F");
#endif
                // Set up the time base information.  It is a void function and returns nothing.
                av_packet_rescale_ts(packetReceivedPtr, avCodecContext->time_base, avFormatContext->streams[0]->time_base);

                // Write out the encoded frame to the output stream.
                int interleavedWriteFrameResult = av_interleaved_write_frame(avFormatContext, packetReceivedPtr);

                // Record any error from the interleaved frame write as this should not fail and leave this
                // while loop to exit the function.
                if (interleavedWriteFrameResult != 0)
                {
                    log_ffmpeg_error_str(interleavedWriteFrameResult);
                    myStatus.ffmpegError = interleavedWriteFrameResult;
                    fi_log(LOG_ERR, "%s av_interleaved_write_frame() failed.", FRAME_WRITER_THREAD_PROCESS_NAME.c_str());
                    goto exitEncodeAndWriteAVFrame;
                }
                // Increase the number of encoded frames counter by one for GUI updates.
                myStatus.framesEncoded++;

                // Increase the real number of encoded packets received.
                // Note that these packet counts may not equal exactly 1
                // frame and may encompass two or more.  It jumps around
                // when it compresses I,P,B frames.
                encodedPacketCount++;

#ifdef LOG_ENCODER_FRAME_INFO
                fi_log(LOG_DEBUG, "%s Encoded packet count[%d] Packet->ptsframe[%lld].", FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), encodedPacketCount, packetReceivedPtr->pts);
#endif
                // Reset the AVPacket object to prepare for next receive.
                av_packet_unref(packetReceivedPtr);

                // The frame was definitely consumed by the encoder.
                frameWasProcessed = true;

                // Push a status update after the frame was processed if this is flushing frame.
                // If you don't do this the gui thread won't get updates.
                if (flushingFrame)
                {
                    flushFrameStats->MarkInterval();     // Set a duration interval.
                    PushStatusUpdate();
                    flushFrameStats->MarkStart();
                }

                // Terminate the while loop here and continue back up for another pass to see if there are more
                // encoded frames waiting to read from the encoder stream.
                continue;
            } // End of the receive packet (frame) block.

            // Note:  Falling through to this logic means that the call to avcodec_receive_packet
            // did not report an ffmpeg error condition and that the received packet (frame) is ready
            // to output to the stream.

            // If receivePacketResult is EAGAIN it means that the encoder needs more
            // input frames before it can return any encoded AV Packets for writing.
            if (receivePacketResult == AVERROR(EAGAIN))
            {
                frameWasProcessed = true;   // Frame was consumed there just isn't one to write yet.
#ifdef LOG_ENCODER_FRAME_INFO
                fi_log(LOG_DEBUG, "%s EncodeAndWriteFrames EAGAIN Need More Frames. framesEncoded[%d] encodedEOF[%d] ffmpegError[%d].", FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), myStatus.framesEncoded, myStatus.ffmpegEOF, myStatus.ffmpegError);
#endif
                // Done processing AV Packets now it is time to exit to consume more frames from input stream.
                goto exitEncodeAndWriteAVFrame;
            }

            // If the received packet result is AVERROR_EOF then the FFMpeg encoder has indicated end of
            // encoded data and video is complete.
            else if (receivePacketResult == AVERROR_EOF)
            {
                frameWasProcessed = true;   // Frame was consumed without error.
                myStatus.ffmpegEOF = true;  // Encoded video is completed time to mark the EOF condition.

#ifdef LOG_ENCODER_FRAME_INFO
                fi_log(LOG_DEBUG, "%s EncodeAndWriteFrames AVERROR_EOF. framesEncoded[%d] encodedEOF[%d] ffmpegError[%d].", FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), myStatus.framesEncoded, myStatus.ffmpegEOF, myStatus.ffmpegError);
#endif
                // If this was the flushing frame the trailer needs to be written to the video file.
                if (flushingFrame)
                {
                    // Only write out the FFMpeg AVTrailer if it hasn't been written already.
                    if (!myStatus.wroteAVTrailer)
                    {
                        int avWriteTrailerResult = av_write_trailer(avFormatContext);

                        // This may not really be a severe error just an EOF but TBD.
                        if (avWriteTrailerResult < 0)
                        {
                            log_ffmpeg_error_str(avWriteTrailerResult);
                            myStatus.ffmpegError = avWriteTrailerResult;
                            fi_log(LOG_ERR, "%s Received negative ffmpeg_result from av_write_trailer().", FRAME_WRITER_THREAD_PROCESS_NAME.c_str());
                        }

                        // Note: Once av_write_trailer is called don't call it again.
                        // Also it could have failed but we want to keep this flag set as it
                        // only indicates the call was made.  The error is preserved in
                        // the myStatus.ffmpegError flag state.
                        myStatus.wroteAVTrailer = true;

                        // If the encoder indicates the EOF flag and the number of frames
                        // encoded is less than frameCount it is not an error condition
                        // because of the nature of the h.264 compression.  If the encoder
                        // has returned EOF then the file is complete.  The reason these
                        // could be off is because we are incrementing based on the packet
                        // received - it is possible the packet contains more data to write
                        // out than the one frame's worth of data.
                        if (myStatus.framesEncoded != frameCount)
                        {
                            // Log the difference between packets written and frames read.
                            fi_log(LOG_DEBUG, "%s: h.264 encoder reports EOF. FFmpeg encoded [%d] packets and read [%d] frames.",
                                   FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), myStatus.framesEncoded, frameCount);

                            // Note: There is a reason the frames encoded is not a match one to one to the frame count
                            // and that has to do with the nature of the h.264 encoder.  This will be considered
                            // ok and in order for the GUI to match the two we will force equalivancy.
                            // The video is complete.
                            myStatus.framesEncoded = frameCount;
                        }
                    }
                }
//#ifdef LOG_ENCODER_FRAME_INFO
                fi_log(LOG_DEBUG, "%s EncodeAndWriteFrames EOF CONDITION. packetsEncoded[%d] framesEncoded[%d] encodedEOF[%d] ffmpegError[%d] flushingFrame[%s] wroteAVTrailer[%s].",
                       FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), encodedPacketCount, myStatus.framesEncoded, myStatus.ffmpegEOF, myStatus.ffmpegError, flushingFrame ? "Y" : "N", myStatus.wroteAVTrailer ? "Y" : "N");
//#endif
                goto exitEncodeAndWriteAVFrame;
            } // This brace marks the end of EOF processing.

            // The error was neither EAGAIN nor AVERROR_EOF so it is unexpected.
            else
            {
                log_ffmpeg_error_str(receivePacketResult);
                myStatus.ffmpegError = receivePacketResult;
                fi_log(LOG_DEBUG, "%s EncodeAndWriteFrames UNEXPECTED FFMPEG ERROR. framesEncoded[%d] encodedEOF[%d] ffmpegError[%d].",
                       FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), myStatus.framesEncoded, myStatus.ffmpegEOF, myStatus.ffmpegError);
                keepReading = false;    // This is unexpected.
            }
        } // End of the Packet Receive Loop
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught in frame writer thread.", __PRETTY_FUNCTION__, e.what());
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught in frame writer thread.", __PRETTY_FUNCTION__);
    }

exitEncodeAndWriteAVFrame:

    // Cleanup and reset the AVPacket if it exists.
    if (packetReceivedPtr != nullptr)
        av_packet_unref(packetReceivedPtr);

#ifdef LOG_ENCODER_FRAME_INFO
    fi_log(LOG_DEBUG, "%s EncodeAndWriteFrames EXIT ffmpegResult[%d] packetsEncoded[%d] framesEncoded[%d] encodedEOF[%d] ffmpegError[%d].",
           FRAME_WRITER_THREAD_PROCESS_NAME.c_str(), ffmpeg_result,
           encodedPacketCount, myStatus.framesEncoded, myStatus.ffmpegEOF, myStatus.ffmpegError);
#endif

    return frameWasProcessed;
}

//--------------------------------------------------------------------
// Class implementation: FFmpegEncoder
//--------------------------------------------------------------------

FFmpegEncoder::~FFmpegEncoder()
{
    if (VERBOSE_SAVE_LOG)
    {
        fi_log(LOG_DEBUG, "FFmpegEncoder: Destruction.");
    }
}

/*!
 * @brief Prepares a hardware acceleration codec context for operation.
 *
 * @param ctx is the AVCodecContext created for hardware acceleration
 * @param hw_device_ctx is the AVBufferRef created for hardware acceleration.
 * @param swAVPixFmtOfFrame tells the AVPixelFormat for input frames.
 *
 * @returns 0 if the operation was successful otherwise -1 if error condition.
 */
int FFmpegEncoder::PrepareHWDeviceContext(AVCodecContext *ctx, AVBufferRef *hw_device_ctx, AVPixelFormat swAVPixFmtOfFrame)
{
    bool allocatedHW                = false;
    AVBufferRef *hw_frames_ref      = nullptr;
    AVHWFramesContext *frames_ctx   = nullptr;
    bool swAVPixFmtOk               = false;
    int ffmpegError                 = 0;

    fi_log(LOG_DEBUG, "%s: HW Input Frames using Software AV Pixel Format[%s].", __PRETTY_FUNCTION__, FFmpegContainer::GetStringForAVPixelFormat(swAVPixFmtOfFrame));

    // Validate the Software AV Pixel Format provided.
    switch (swAVPixFmtOfFrame)
    {
    case AV_PIX_FMT_NV12:
        swAVPixFmtOk = true;
        break;
    // Most HW encoders do not take in frames in this format it's for testing only.
    case AV_PIX_FMT_YUV420P:
        swAVPixFmtOk = true;
        break;
    default:
        swAVPixFmtOk = false;
        break;
    }

    // If the Software AV Pixel Format is not supported cannot use HW Device.
    if (!swAVPixFmtOk)
    {
        fi_log(LOG_ERR, "%s: Unsupported AVPixelFormat[%s] for software input frame.",
               __PRETTY_FUNCTION__, FFmpegContainer::GetStringForAVPixelFormat(swAVPixFmtOfFrame));
        ffmpegError = -1;
        goto exit_set_hwframe_ctx;
    }

    // This is supposed to be sending an enumeration not the hw_device_ctx.
    // Something is wrong with the example.
    hw_frames_ref = av_hwframe_ctx_alloc(hw_device_ctx);
    if (hw_frames_ref == nullptr)
    {
        fi_log(LOG_ERR, "%s: av_hwframe_ctx_alloc() failed to create FFmpeg HW frame context.", __PRETTY_FUNCTION__);
        ffmpegError = -1;
        goto exit_set_hwframe_ctx;
    }
    else
    {
        allocatedHW = true;
    }
    frames_ctx = (AVHWFramesContext *)(hw_frames_ref->data);

    //frames_ctx->format    = AV_PIX_FMT_VAAPI;
    // This is the device AVPixelFormat Value.
    frames_ctx->format    = ctx->pix_fmt;

    // The input frames should be either AV_PIX_FMT_NV12 or AV_PIX_FMT_GRAY8.
    frames_ctx->sw_format = swAVPixFmtOfFrame;
    frames_ctx->width     = ctx->width;
    frames_ctx->height    = ctx->height;

    // If ROI is less than or equal to really small frames create a lot in the pool.
    if ((ctx->width * ctx->height) <= (256 * 256))
        frames_ctx->initial_pool_size = 1000;
    // If ROI is less than or equal to standard definition set high pool size as frames come fast.
    else if ((ctx->width * ctx->height) <= (854 * 480))
        frames_ctx->initial_pool_size = 500;
    else
        frames_ctx->initial_pool_size = 250;

    // NOTE:  For the AMF Encoder using the D3D11 device it is important
    // to set the initial_pool_size to zero so that the device manages its
    // surface allocations.  When this is set to anything other than zero the
    // device returns a texture creation error.
    if (frames_ctx->format == AV_PIX_FMT_D3D11)
    {
        frames_ctx->initial_pool_size = 0;
    }

    //3840x2160 2160p   4K
    //2560x1440 1440p   2K
    //1920x1080 1080p   Maximum resolution for HD
    //1280x720  720p    Minimum resolution for HD
    //854x480   480p    Standard definition
    //640x360   360p    Traditional website resolution
    //426x240   240p    Minimum YouTube video size

#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "%s: calling av_hwframe_ctx_init()  initial_pool_size[%d].",
           __PRETTY_FUNCTION__, frames_ctx->initial_pool_size);
#endif

    // See this implementation in Hwcontext.c of ffmpeg source.
    if ((ffmpegError = av_hwframe_ctx_init(hw_frames_ref)) < 0)
    {
        fi_log(LOG_ERR, "%s: av_hwframe_ctx_init() failed to initialize FFmpeg HW frame context.", __PRETTY_FUNCTION__);
        goto exit_set_hwframe_ctx;
    }

#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "%s: calling av_buffer_ref().", __PRETTY_FUNCTION__);
#endif

    ctx->hw_frames_ctx = av_buffer_ref(hw_frames_ref);
    if (!ctx->hw_frames_ctx)
    {
        fi_log(LOG_ERR, "%s: hw_frames_ctx is nullptr after av_buffer_ref().", __PRETTY_FUNCTION__);
        ffmpegError = AVERROR(ENOMEM);
        goto exit_set_hwframe_ctx;
    }

    fi_log(LOG_DEBUG, "%s: Successful Initialization HWContext().", __PRETTY_FUNCTION__);
exit_set_hwframe_ctx:

    if (allocatedHW)
    {
        // Looks like this gets unreferenced without regard to error condition.
        av_buffer_unref(&hw_frames_ref);
    }
    return ffmpegError;
}

/*!
 * @brief Prepares the encoder object for FFmpeg encoding operation.
 *
 * @param videoType The type of video being created.
 *
 * @returns The shared pointer for the encoder.
 */
std::shared_ptr<FFmpegEncoder> FFmpegEncoder::CreateEncoder(CONTAINER_TYPE videoType,
                                                            const std::string& videoName,
                                                            enum AVPixelFormat avHWDeviceToUseIfAvailable,
                                                            enum AVPixelFormat avInPixelFormatToUse,
                                                            enum AVPixelFormat avOutPixelFormatToUse,
                                                            uint32_t frameRateToUse,
                                                            uint32_t frameCountToUse,
                                                            uint32_t frameWidthToUse,
                                                            uint32_t frameHeightToUse,
                                                            uint32_t crfValueToUse,
                                                            const std::string& presetValueToUse,
                                                            uint32_t keyFrameIntervalToUse)
{
    bool isPrepared = false;
    bool isHWAccelerated = false;

    // If we have HW acceleration available but the input isn't NV12 we can use the
    // FFmpeg software filtering process to generate an NV12 to feed the HW encoder.
    enum AVPixelFormat hwAcceleratedFilter =  AV_PIX_FMT_NV12;

    // Create a string out of the crfValueToUse variable.
    std::string crfValueString = std::to_string(crfValueToUse);

    // Set the codec ID default for the h.264 software encoder.
    enum AVCodecID swCodecID = AV_CODEC_ID_H264;

    // Create a user-readable string for default encoder_id being processed.
    std::string encoderNameStr = FFmpegContainer::GetStringForAVCodecID(swCodecID);
    std::string myNameStr = "[FFmpegEncoder::CreateEncoder]";

    try
    {
        bool createdFilter  = false;
        int ffmpegError     = 0;

        // Protect the encoder instance.
        std::lock_guard<std::mutex> guard(FFmpegEncoderInstanceMutex);

resetEncoderForSoftware:
        isPrepared          = false;
        isHWAccelerated     = false;
        encoderNameStr  = FFmpegContainer::GetStringForAVCodecID(swCodecID);

        // Validate the video type is supported.  Right now only MP4 encoding is implemented.
        if (videoType != CONTAINER_TYPE::MP4_CONTAINER)
        {
            fi_log(LOG_ERR, "%s: The videoType provided: %d is not supported.",
                   myNameStr.c_str(), (uint32_t)videoType);
            goto exitPrepareEncoder;
        }

        // Create this singleton encoder object.
        FFmpegEncoderInstance = std::shared_ptr<FFmpegEncoder>(new FI_FFMPEG::FFmpegEncoder());

        // If this encoder cannot be created that's a critical error condition.
        if (!FFmpegEncoderInstance)
        {
            fi_log(LOG_ERR, "%s: The FFmpeg Encoder could not be allocated.", myNameStr.c_str());
            goto exitPrepareEncoder;
        }

        // Set the key interval used for KeyFrames in the encoded video.
        FFmpegEncoderInstance->keyFrameInterval = keyFrameIntervalToUse;

        // Make sure the HWDeviceCtx is nullptr.
        HWDeviceCTX = nullptr;

        // Does the caller want this encoder prepared to utilize HW encoding?
        if ((avHWDeviceToUseIfAvailable == AV_PIX_FMT_VAAPI)
            || (avHWDeviceToUseIfAvailable == AV_PIX_FMT_QSV)
            || (avHWDeviceToUseIfAvailable == AV_PIX_FMT_CUDA)
            || (avHWDeviceToUseIfAvailable == H264_AMD_AV_PIX_FMT)
           )
        {
            // Right now the only input pixel format that works with
            // the h264_vaapi hw encoder is AV_PIX_FMT_NV12 - but we
            // don't always stream that format as input because our
            // BMP planar frames could also be AV_PIX_FMT_GRAY (Y800)
            // The actual interface in FFmpeg does not have a profile
            // supporting the Y800 single plane and fails to open
            // the h264_vaapi codec later in this process.  Therefore what
            // was decided was that we would only send over the BMP Planar
            // NV12 format - OR - the standard BMP RBG format that we always
            // support.  When it comes time to actually create a "filtered"
            // frame later in the encoder process if the frame coming in is
            // not BMP Planar NV12 we will convert the BMP RGB format to that
            // layout then create the filtered frame.  So we do not need to
            // rely on input frames at this point in the process to be BMP Planar
            // NV12.  The condition is removed but left in place to document.
            //#HWTEST
            //if (avInPixelFormatToUse == AV_PIX_FMT_NV12)
            {
                int ffmpegHWCtxCreateResult = 0;    // Make sure set to zero before testing HW encoder creation.

                switch (avHWDeviceToUseIfAvailable)
                {
                case AV_PIX_FMT_VAAPI:
                    // Create the hardware accelerated context data.
                    ffmpegHWCtxCreateResult = av_hwdevice_ctx_create(&HWDeviceCTX,
                                                                     AV_HWDEVICE_TYPE_VAAPI,
                                                                     nullptr, nullptr, 0);
                    encoderNameStr = H264_VAAPI_CODEC_NAME;
                    break;
                case AV_PIX_FMT_QSV:
                    // Create the hardware accelerated context data.
                    ffmpegHWCtxCreateResult = av_hwdevice_ctx_create(&HWDeviceCTX,
                                                                     AV_HWDEVICE_TYPE_QSV,
                                                                     H264_INTEL_QSV_DEVICE.c_str(), nullptr, 0);
                    encoderNameStr = H264_INTEL_QSV_NAME;
                    break;
                case AV_PIX_FMT_CUDA:
                    //AV_PIX_FMT_D3D11VA_VLD or  AV_PIX_FMT_CUDA    in the avctx->pix_fmt  otherwise EINVAL
                    // Create the hardware accelerated context data.
                    ffmpegHWCtxCreateResult = av_hwdevice_ctx_create(&HWDeviceCTX,
                                                                     AV_HWDEVICE_TYPE_CUDA,
                                                                     nullptr, nullptr, 0);
                    encoderNameStr = H264_NVIDIA_NVENC_NAME;
                    break;
                case AV_PIX_FMT_D3D11VA_VLD:
                case AV_PIX_FMT_D3D11:
                    //AV_PIX_FMT_D3D11VA_VLD,AV_HWDEVICE_TYPE_D3D11VA
                    // Create the hardware accelerated context data.
                    ffmpegHWCtxCreateResult = av_hwdevice_ctx_create(&HWDeviceCTX,
                                                                     AV_HWDEVICE_TYPE_D3D11VA,
                                                                     H264_AMD_AMF_DEVICE.c_str(), nullptr, 0);
                    encoderNameStr = H264_AMD_AMF_NAME;
                    break;
                case AV_PIX_FMT_DXVA2_VLD:
                    //AV_PIX_FMT_DXVA2_VLD,AV_HWDEVICE_TYPE_DXVA2
                    // Create the hardware accelerated context data.
                    ffmpegHWCtxCreateResult = av_hwdevice_ctx_create(&HWDeviceCTX,
                                                                     AV_HWDEVICE_TYPE_DXVA2,
                                                                     H264_AMD_AMF_DEVICE.c_str(), nullptr, 0);
                    encoderNameStr = H264_AMD_AMF_NAME;
                    break;
                default:
                    avHWDeviceToUseIfAvailable = AV_PIX_FMT_NONE;
                    isHWAccelerated = false;
                    break;
                }

                // If the hw device attempted creation what is its status?
                if (avHWDeviceToUseIfAvailable != AV_PIX_FMT_NONE)
                {
                    // If ffmpegHWCtxCreateResult is set to zero then the HW encoder device was
                    // created without an error.
                    if (ffmpegHWCtxCreateResult == 0)
                    {
                        isHWAccelerated = true;
                    }
                    else
                    {
                        // Log the FFmpeg error condition.
                        log_ffmpeg_error_str(ffmpegHWCtxCreateResult);
                        avHWDeviceToUseIfAvailable = AV_PIX_FMT_NONE;
                    }
                }

                fi_log(LOG_DEBUG, "%s: av_hwdevice_ctx_create(%s) ffmpegError[%d] isHWAccelerated[%s].",
                       myNameStr.c_str(), encoderNameStr.c_str(),  ffmpegHWCtxCreateResult, isHWAccelerated ? "T" : "F");
            }

            // If not hardware accelerated restore to normal libx264 software encoding.
            if (!isHWAccelerated)
            {
                // There is no HW encoder available, set for software encoded out pixfmt.
                avOutPixelFormatToUse = AV_PIX_FMT_YUV420P;
                encoderNameStr = FFmpegContainer::GetStringForAVCodecID(swCodecID);
            }
        }

        // At this point the in/out/hw pixel format settings are validated for encoder.
        FFmpegEncoderInstance->avInPixelFormat = avInPixelFormatToUse;
        FFmpegEncoderInstance->avOutPixelFormat = avOutPixelFormatToUse;
        FFmpegEncoderInstance->avHWDevicePixelFormat = avHWDeviceToUseIfAvailable;

        // Set up the avr encoder properties.
        FFmpegEncoderInstance->avrAspectRatio = {1, 1};
        FFmpegEncoderInstance->avrTimeBase.num = FFmpegEncoderInstance->avrFrameRate.den = 1;
        FFmpegEncoderInstance->avrTimeBase.den = FFmpegEncoderInstance->avrFrameRate.num = (int) frameRateToUse;

        // Set up the frame dimension.
        FFmpegEncoderInstance->frameWidth = frameWidthToUse;
        FFmpegEncoderInstance->frameHeight = frameHeightToUse;

        // Set up the frame counters.
        FFmpegEncoderInstance->frameCount = frameCountToUse;
        FFmpegEncoderInstance->framesRead = 0;

        FrameWriterThread::InitStatusRec(FFmpegEncoderInstance->theFrameWriterStatus);

        // Set the status to a non-error condition.
        FFmpegEncoderInstance->ffmpegStatusFromCreation = 0;

        // NOTE:  All video encoding must have an AVFormatContext and AVStream object
        // whether codec is hw accelerated or not.

        // Create the output format context.  Use avformat_free_context() to free it.
        // NOTE:  Difference here was passing AVOutputFormat* to call - transcode.c uses null and a filename.
        FFmpegEncoderInstance->ffmpegStatusFromCreation = avformat_alloc_output_context2(&FFmpegEncoderInstance->fmt_ctx, nullptr, nullptr, videoName.c_str());
        if (FFmpegEncoderInstance->ffmpegStatusFromCreation < 0)
        {
            // Unable to create the format context for the encoded output.
            log_ffmpeg_error_str(FFmpegEncoderInstance->ffmpegStatusFromCreation);
            fi_log(LOG_ERR, "%s: Unable to create AVFormatContext for AVCodec[%s].",
                   myNameStr.c_str(), encoderNameStr.c_str());
            goto exitPrepareEncoder;
        }

        // Create the ffmpeg output stream context.
        FFmpegEncoderInstance->stream = avformat_new_stream(FFmpegEncoderInstance->fmt_ctx, nullptr);
        if (!FFmpegEncoderInstance->stream)
        {
            fi_log(LOG_ERR, "%s: Unable to create AVStream* for %s using AVCodec[%s].",
                   myNameStr.c_str(), videoName.c_str(), encoderNameStr.c_str());
            goto exitPrepareEncoder;
        }

        // If this is a hardware acceleration the codec setup is slightly different.
        if (isHWAccelerated)
        {
            // In order to use hardware acceleration must look for codec by name.
            FFmpegEncoderInstance->codec = avcodec_find_encoder_by_name(encoderNameStr.c_str());
            if (!FFmpegEncoderInstance->codec)
            {
                fi_log(LOG_ERR, "%s: Unable to find FFmpeg encoder codec[%s].",
                       myNameStr.c_str(), encoderNameStr.c_str());
                goto exitPrepareEncoder;
            }

        }
        // Perform the regular software encoder setup.
        else
        {
            // The codec id returned set up from "mp4" above is AV_CODEC_ID_H264.
            FFmpegEncoderInstance->codec = avcodec_find_encoder(swCodecID);
            if (!FFmpegEncoderInstance->codec)
            {
                fi_log(LOG_ERR, "%s: Unable to find FFmpeg encoder codec[%s].",
                       myNameStr.c_str(), encoderNameStr.c_str());
                goto exitPrepareEncoder;
            }
        }

        // Create the encoder codec context structure.
        FFmpegEncoderInstance->codec_ctx = avcodec_alloc_context3(FFmpegEncoderInstance->codec);
        if (!FFmpegEncoderInstance->codec_ctx)
        {
            fi_log(LOG_ERR, "%s: Unable to find FFmpeg encoder codec[%s] context data.",
                   myNameStr.c_str(), encoderNameStr.c_str());
            goto exitPrepareEncoder;
        }

        // Install the codec context information for frames expected.
#if 0
        // Test this theory out.
        // Yes - odd number height causes ffmpeg libraries to return a generic error.
        // This has been fixed in outside code but I'm leaving this comment in place as
        // well as the code used to test it.
        if ((frameHeight % 2) != 0)
            frameHeight = frameHeight - 1;
        // or to generate an error on purpose use this:
        //if ((frameHeight % 2) == 0)
        //    frameHeight = frameHeight - 1;
#endif
        FFmpegEncoderInstance->codec_ctx->width = (int) FFmpegEncoderInstance->frameWidth;
        FFmpegEncoderInstance->codec_ctx->height = (int) FFmpegEncoderInstance->frameHeight;
        FFmpegEncoderInstance->codec_ctx->sample_aspect_ratio = FFmpegEncoderInstance->avrAspectRatio;

        // Set the codec context objects pix_fmt.  If hw encoding this will be set to the
        // hardware device AVPixelFormat, if sw encoding it's set to the output AVPixelFormat.
        if (isHWAccelerated)
        {
            FFmpegEncoderInstance->codec_ctx->pix_fmt = FFmpegEncoderInstance->avHWDevicePixelFormat;
        }
        else
        {
            FFmpegEncoderInstance->codec_ctx->pix_fmt = FFmpegEncoderInstance->avOutPixelFormat;
        }

        // Note:  This is not being set in the transcode.c example.
        //codec_ctx->bit_rate = 40000;
        FFmpegEncoderInstance->codec_ctx->time_base = FFmpegEncoderInstance->avrTimeBase;
        //codec_ctx->framerate = avrFrameRate;

        // The gop_size is equal to the key frame interval setting.
        FFmpegEncoderInstance->codec_ctx->gop_size = (int)keyFrameIntervalToUse;

        // Set codec options for hardware encoders.  This will need a bunch of cleanup
        // but for now leave it in place with all the comments.
        if (isHWAccelerated)
        {
            // The compression level is being looked at in the vaapi code.  When I set it to 8 the error message below pops out
            // and I can see this code in the function vaapi_encode_init_quality() of vaapi_encode.c (not the cpp sample the c interface)
            // However when I set it to something smaller I am not seeing that default message go away but perhaps it is earlier I haven't
            // been able to find the source of that "No quality level set; using default (x)." anywhere in the source yet.
            //[h264_vaapi @ 0x7fa50c015f00] No quality level set; using default (20).
            //[h264_vaapi @ 0x7fa50c015f00] Invalid quality level: valid range is 0-7, using 7.

            // Some pre-codec creation setups.
            switch (FFmpegEncoderInstance->codec_ctx->pix_fmt)
            {
            case AV_PIX_FMT_QSV:
                // When this is not set to the CRF value - the quality is bad
                // and the default setup is to use an average variable bitrate.
                FFmpegEncoderInstance->codec_ctx->gop_size = 15;
                FFmpegEncoderInstance->codec_ctx->global_quality = (int)crfValueToUse;
                break;
            case H264_AMD_AV_PIX_FMT:
                FFmpegEncoderInstance->codec_ctx->gop_size = 15;
                FFmpegEncoderInstance->codec_ctx->global_quality = (int)crfValueToUse;
                break;
            default:
                break;
            }

            // Prepare the hardware device context.
            // NOTE:  The codec_ctx->pix_fmt must equal the device AVPixelFormat and
            // the last parameter is the AVPixelFormat of the expected output frames
            // that feed the hardware encoder.  Typically this is NV12.
            if ((ffmpegError = FFmpegEncoder::PrepareHWDeviceContext(FFmpegEncoderInstance->codec_ctx, HWDeviceCTX, hwAcceleratedFilter)) < 0)
            {
                fi_log(LOG_ERR, "%s: AVCodec[%s] PrepareHWDeviceContext() FAILED! ffmpegError[%d]",
                       myNameStr.c_str(), encoderNameStr.c_str(), ffmpegError);

                // Unable to use the HW Encoder device.  Let's try to recover here
                // by flipping back to the SW Encoder.
                avOutPixelFormatToUse = AV_PIX_FMT_YUV420P;
                avHWDeviceToUseIfAvailable = AV_PIX_FMT_NONE;

                // Attempt to close down the encoder instance in progress.
                // Reset the instance!
                FFmpegEncoderInstance.reset();
                FFmpegEncoderInstance = nullptr;    // Should already be nullptr from reset so this is overkill.

                fi_log(LOG_DEBUG, "%s: Attempting to reset SW Encoder.",
                       myNameStr.c_str(), encoderNameStr.c_str());

                // Try to recreate with the software encoder.
                goto resetEncoderForSoftware;
            }
            else
            {
                fi_log(LOG_INFO, "%s: AVCodec[%s] PrepareHWDeviceContext() SUCCESS.",
                       myNameStr.c_str(), encoderNameStr.c_str());
            }

            // This compression_level setting does nothing in the hw acceleration pathway.
            //FFmpegEncoderInstance->codec_ctx->compression_level = quality;

            // This does make a difference and I've tested it both with 13 and 40 -
            // Setting it to 13 made the file large and the bitrate much higher than default.
            // Setting it to 40 made the file large but the bitrate much smaller and it looks
            // like the 40 created B frames that were really big hence the large file size.
            // Setting it to 20 provides the same result as the default.
            // For match the crf setting here - it should be within the range of 13 and 40
            switch (FFmpegEncoderInstance->codec_ctx->pix_fmt)
            {
            case AV_PIX_FMT_VAAPI:
            case AV_PIX_FMT_CUDA:
                av_opt_set(FFmpegEncoderInstance->codec_ctx->priv_data, "qp", crfValueString.c_str(), 0);
                break;
            case H264_AMD_AV_PIX_FMT:
                av_opt_set(FFmpegEncoderInstance->codec_ctx->priv_data, "qp", crfValueString.c_str(), 0);
                av_opt_set(FFmpegEncoderInstance->codec_ctx->priv_data, "log_to_debug", "1", 0);
                av_opt_set(FFmpegEncoderInstance->codec_ctx->priv_data, "quality", "quality", 0);
                av_opt_set(FFmpegEncoderInstance->codec_ctx->priv_data, "rc", "cqp", 0);
                av_opt_set(FFmpegEncoderInstance->codec_ctx->priv_data, "coder", "cabac", 0);
                av_opt_set(FFmpegEncoderInstance->codec_ctx->priv_data, "profile", "high", 0);
                break;
            default:
                break;
            }
        }
        if (!isHWAccelerated)
        {
            // Setup the libx264 options for software encoding.
            // Right now only applying two H.264 option settings to the encoder.
            // These are the crf which is the Constant Rate Factor and the preset
            // which tells the encoder how efficiently compressed (which has a
            // direct impact on filesize) to make the video.  Note that the "slower"
            // the preset the more efficiently compressed and smaller the video file
            // but the use of the string slow versus fast indicate how long it
            // takes to encode the video as well.  A "slower" presets take longer to
            // encode while a "faster" preset takes less time to encode.

            fi_log(LOG_DEBUG, "%s: AVCodec[%s] h264.preset[%s] h.264.crf[%s].",
                   myNameStr.c_str(), encoderNameStr.c_str(), presetValueToUse.c_str(), crfValueString.c_str());

            av_opt_set(FFmpegEncoderInstance->codec_ctx->priv_data, "preset", presetValueToUse.c_str(), 0);
            av_opt_set(FFmpegEncoderInstance->codec_ctx->priv_data, "crf", crfValueString.c_str(), 0);
        }

        // This was in Transcode.c and this appears to create a valid MP4 file header.  Without this
        // the Windows Media Player and other players on Windows report file is corrupted.  However
        // with the "corrupted" files VLC still reads them.  This puts in a "global" valid header
        // in the MP4 container and id critical to valid output.  Differences in header start at 0x6A
        if (FFmpegEncoderInstance->fmt_ctx->oformat->flags & AVFMT_GLOBALHEADER)
            FFmpegEncoderInstance->codec_ctx->flags |= AV_CODEC_FLAG_GLOBAL_HEADER;

        // Initialize the AVCodecContext to use the proper AVCodec.
        // One of the status errors is AVERROR_EXTERNAL which means the
        // library for the codec did not like the data in the codec_ctx.
        // This is a fatal error.
        // Note:  H.264 returns AVERROR_EXTERNAL libx264.c line 542 of X264_init(AVCodecContext *avctx)
        ffmpegError = avcodec_open2(FFmpegEncoderInstance->codec_ctx, FFmpegEncoderInstance->codec, nullptr);
        FFmpegEncoderInstance->ffmpegStatusFromCreation = ffmpegError;
        if (ffmpegError < 0)
        {
            log_ffmpeg_error_str(ffmpegError);
            fi_log(LOG_ERR, "%s: Unable to open AVCodec[%s] the call avcodec_open2() FAILED!",
                   myNameStr.c_str(), encoderNameStr.c_str());
            goto exitPrepareEncoder;
        }

        // Fill the parameters structure based on the values from the supplied codec context.
        // Any allocated fileds in par are freed and replaced with duplicates of the corresponding
        // fields in codec.
        FFmpegEncoderInstance->ffmpegStatusFromCreation = avcodec_parameters_from_context(FFmpegEncoderInstance->stream->codecpar, FFmpegEncoderInstance->codec_ctx);
        if (FFmpegEncoderInstance->ffmpegStatusFromCreation < 0)
        {
            log_ffmpeg_error_str(FFmpegEncoderInstance->ffmpegStatusFromCreation);
            fi_log(LOG_ERR, "%s: AVCodec[%s] avcodec_parameters_from_context() FAILED!",
                   myNameStr.c_str(), encoderNameStr.c_str());
            goto exitPrepareEncoder;
        }

        // From transcode.c.
        FFmpegEncoderInstance->stream->time_base = FFmpegEncoderInstance->codec_ctx->time_base;
        //FFmpegEncoderInstance->stream->cur_dts = AV_NOPTS_VALUE;

        // Create the output container file.
        if (!(FFmpegEncoderInstance->fmt_ctx->oformat->flags & AVFMT_NOFILE))
        {
            FFmpegEncoderInstance->ffmpegStatusFromCreation = avio_open(&FFmpegEncoderInstance->fmt_ctx->pb, videoName.c_str(), AVIO_FLAG_WRITE);
            if (FFmpegEncoderInstance->ffmpegStatusFromCreation < 0)
            {
                log_ffmpeg_error_str(FFmpegEncoderInstance->ffmpegStatusFromCreation);
                fi_log(LOG_ERR, "%s: AVCodec[%s] Unable to avio_open() the video %s.",
                       myNameStr.c_str(), encoderNameStr.c_str(), videoName.c_str());
                goto exitPrepareEncoder;
            }
        }

        // Write out the encoded header to file.
        FFmpegEncoderInstance->ffmpegStatusFromCreation = avformat_write_header(FFmpegEncoderInstance->fmt_ctx, nullptr);
        if (FFmpegEncoderInstance->ffmpegStatusFromCreation < 0)
        {
            log_ffmpeg_error_str(FFmpegEncoderInstance->ffmpegStatusFromCreation);
            fi_log(LOG_ERR, "%s: AVCodec[%s] Unable to avformat_write_header() the video %s.",
                   myNameStr.c_str(), encoderNameStr.c_str(), videoName.c_str());
            goto exitPrepareEncoder;
        }

        // Only use the FFmpeg filter object if performing software encoding.
        // TESTING a filter approach to generate NV12 if input is not planar.
        // Because NV12 is only way to get data through the hw h264 encoder.
        // Create the FFmpeg filter object as the encoder needs it to properly run
        // through the encode a frame pathway with or without filter operations.
        FFmpegEncoderInstance->filterPtr = std::shared_ptr <FFmpegFilters>(new FFmpegFilters());
        // If unable to create the filters object set status and exit.
        if (!FFmpegEncoderInstance->filterPtr)
        {
            fi_log(LOG_ERR, "%s: AVCodec[%s] Unable to create FFmpegFilters!",
                   myNameStr.c_str(), encoderNameStr.c_str());
            goto exitPrepareEncoder;
        }

        // If hardware accelerated the filter we want should require input as it comes and the output
        // in this case will be NV12.
        if (isHWAccelerated)
        {
            createdFilter = FFmpegEncoderInstance->filterPtr->InitializeFilters(FFmpegEncoderInstance->avrAspectRatio,
                                                                                FFmpegEncoderInstance->avrTimeBase,
                                                                                FFmpegEncoderInstance->frameWidth,
                                                                                FFmpegEncoderInstance->frameHeight,
                                                                                FFmpegEncoderInstance->avInPixelFormat,
                                                                                hwAcceleratedFilter);
        }
        else
        {
            createdFilter = FFmpegEncoderInstance->filterPtr->InitializeFilters(FFmpegEncoderInstance->avrAspectRatio,
                                                                                FFmpegEncoderInstance->avrTimeBase,
                                                                                FFmpegEncoderInstance->frameWidth,
                                                                                FFmpegEncoderInstance->frameHeight,
                                                                                FFmpegEncoderInstance->avInPixelFormat,
                                                                                FFmpegEncoderInstance->avOutPixelFormat);
        }

        // Was the FFmpeg filtering object created?
        if (!createdFilter)
        {
            std::string inPixelFormat = FFmpegContainer::GetStringForAVPixelFormat(FFmpegEncoderInstance->avInPixelFormat);
            std::string outPixelFormat = FFmpegContainer::GetStringForAVPixelFormat(FFmpegEncoderInstance->avOutPixelFormat);
            fi_log(LOG_ERR, "%s: AVCodec[%s] Unable to initialize FFmpegFilters for In[%s] and Out[%s].",
                   myNameStr.c_str(), encoderNameStr.c_str(), inPixelFormat.c_str(), outPixelFormat.c_str());
            goto exitPrepareEncoder;
        }

        // Create the frame queues used for processing the frames to encode
        // for the video being created.
        framesToWrite = FFmpegFrameQueue::CreateFFmpegFrameQueue(frameCountToUse);

        // If unable to create the frame queues report the error and encoder is unprepared.
        if (framesToWrite == nullptr)
        {
            fi_log(LOG_ERR, "%s: AVCodec[%s] Unable to create FFmpegFrameQueue framesToWrite!",
                   myNameStr.c_str(), encoderNameStr.c_str());
            goto exitPrepareEncoder;
        }

        // Create the frame writer thread.
        if (frameWriterThreadInstance == nullptr)
            frameWriterThreadInstance = FrameWriterThread::CreateFrameWriter();
        if (frameWriterThreadInstance == nullptr)
        {
            fi_log(LOG_ERR, "%s: AVCodec[%s] Unable to create frameWriterThreadInstance!",
                   myNameStr.c_str(), encoderNameStr.c_str());
            goto exitPrepareEncoder;
        }
        else
        {
            // If the frame writer thread is not running there is an issue!
            if (!frameWriterThreadInstance->is_running())
            {
                // Start running framewriter thread instance.
                frameWriterThreadInstance->start();
            }
            else
            {
                // This really should not be running - it would be very weird if it was.
                fi_log(LOG_CRIT, "%s: AVCodec[%s] How could a frameWriterThreadInstance be running already?!",
                       myNameStr.c_str(), encoderNameStr.c_str());
            }
        }

        // Finally create the encoder statistic object to monitor duration time to encode frames.
        FFmpegEncoderInstance->encoderStats = std::shared_ptr<FiDurations>(new FiDurations());

        // All is prepared theoretically so notify the caller by returning 0.  The caller will
        // then start to write frame data to the video stream.
        isPrepared = true;
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught while preparing encoder.", __PRETTY_FUNCTION__, e.what());
        goto exitPrepareEncoder;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught while preparing encoder.", __PRETTY_FUNCTION__);
        goto exitPrepareEncoder;
    }

exitPrepareEncoder:
    fi_log(LOG_DEBUG, "%s: exitPrepareEncoder: AVCodec[%s] isPrepared[%s].",
           myNameStr.c_str(), encoderNameStr.c_str(), isPrepared ? "T" : "F");

    FFmpegEncoderInstance->encoderPrepared = isPrepared;
    FFmpegEncoderInstance->encodeAVFrameTimeout = false;
    return FFmpegEncoderInstance;
}

/*!
 * @brief Provide access to the one and only FFmpegEncoder object.
 * @returns The shared pointer for the encoder.
 */
std::shared_ptr<FFmpegEncoder> FFmpegEncoder::GetEncoderInstance()
{
    std::lock_guard<std::mutex> guard(FFmpegEncoderInstanceMutex);
    return FFmpegEncoderInstance;
}

/*!
 * @brief Stats logging for this encoder.
 */
void FFmpegEncoder::LogStatistics()
{
    std::string typeStr;

    typeStr = "[FFmpegEncoder]";
    fi_log(LOG_DEBUG, "%s framesRead[%d] width[%d] height[%d]",
           typeStr.c_str(), frameCount, frameWidth, frameHeight);
    if (encoderStats)
    {
        typeStr += "[Lifetime of FFmpegEncoder]";
        encoderStats->LogDurationInfo(typeStr.c_str());
    }
    if (encodeAVFrameStats)
    {
        typeStr = "[EncodeAVFrame Durations]";
        encodeAVFrameStats->LogDurationInfo(typeStr.c_str());
    }
}

/*!
 * @brief Determines if the H264 VAAPI Encoder for HW acceleration exists.
 * AV_HWDEVICE_TYPE_VAAPI is the device type for FFmpeg.
 *
 * @returns true if the encoder is available, false if not.
 */
bool FFmpegEncoder::IsVAAPIAvailable()
{
    bool h264VAAPIExists                    = false;
    const AVCodec *h264HWCodec              = nullptr;

    // Make sure the container supports HW Acceleration.
    if (CHECK_FOR_HW_ACCELERATION)
    {
        // First check is to see if the encoder was built into library.
        h264HWCodec = avcodec_find_encoder_by_name(H264_VAAPI_CODEC_NAME.c_str());
        if (h264HWCodec != nullptr)
        {
            h264VAAPIExists = true;
        }
        fi_log(LOG_DEBUG, "%s: [%s] %s.",
               __PRETTY_FUNCTION__, H264_VAAPI_CODEC_NAME.c_str(), h264VAAPIExists ? HW_CODEC_EXISTS.c_str() : HW_CODEC_NOT_AVAILABLE.c_str());
    }
#ifdef USE_VERBOSE_INFO_DEBUG
    else
    {
        fi_log(LOG_DEBUG, "%s: [%s] %s.",
               __PRETTY_FUNCTION__, H264_VAAPI_CODEC_NAME.c_str(), HW_CODEC_NO_CHECKING.c_str());
    }
#endif

    return h264VAAPIExists;
}

/*!
 * @brief Determines if the H264 VAAPI Encoder for HW acceleration exists.
 * AV_HWDEVICE_TYPE_VAAPI is the device type for FFmpeg.
 *
 * @returns true if the encoder is available, false if not.
 */
bool FFmpegEncoder::IsNVIDIAAvailable()
{
    bool h264NVIDIAExists                   = false;
    const AVCodec *h264HWCodec              = nullptr;
    AVBufferRef *h264HWDeviceContext        = nullptr;
    //enum AVHWDeviceType h264HWDeviceType    = AV_HWDEVICE_TYPE_QSV; // doesn't seem to be a device type needed for NVIDIA.

    // Make sure the container supports HW Acceleration.
    if (CHECK_FOR_HW_ACCELERATION)
    {
        // First check is to see if the encoder was built into library.
        h264HWCodec = avcodec_find_encoder_by_name(H264_NVIDIA_NVENC_NAME.c_str());

        if (h264HWCodec != nullptr)
        {
            // The NVIDIA encoder may have the support built in to the library
            // but that doesn't mean the harware exists to drive it.  The only
            // way to determine this is to create a HW Device Context then free it.
            int ffmpegError = av_hwdevice_ctx_create(&h264HWDeviceContext,
                                                     AV_HWDEVICE_TYPE_CUDA,
                                                     nullptr, nullptr, 0);

            // Was the HW device created?
            if ((ffmpegError == 0) && h264HWDeviceContext)
            {
                h264NVIDIAExists = true;
            }
            // Close and cleanup the hardware device context.
            if (h264HWDeviceContext != nullptr)
            {
                av_buffer_unref(&h264HWDeviceContext);
                h264HWDeviceContext = nullptr;
            }
        }
        fi_log(LOG_DEBUG, "%s: [%s] %s.",
               __PRETTY_FUNCTION__, H264_NVIDIA_NVENC_NAME.c_str(), h264NVIDIAExists ? HW_CODEC_EXISTS.c_str() : HW_CODEC_NOT_AVAILABLE.c_str());
    }
#ifdef USE_VERBOSE_INFO_DEBUG
    else
    {
        fi_log(LOG_DEBUG, "%s: [%s] %s.",
               __PRETTY_FUNCTION__, H264_NVIDIA_NVENC_NAME.c_str(), HW_CODEC_NO_CHECKING.c_str());
    }
#endif

    return h264NVIDIAExists;
}

/*!
 * @brief Determines if the H264 AMF Encoder for HW acceleration exists.
 *
 * @returns true if the encoder is available, false if not.
 */
bool FFmpegEncoder::IsAMFAvailable()
{
    bool h264AMFExists                      = false;
    const AVCodec *h264HWCodec              = nullptr;
    AVBufferRef *h264HWDeviceContext        = nullptr;
    //AV_HWDEVICE_TYPE_DXVA2 or AV_HWDEVICE_TYPE_D3D11VA,
    enum AVHWDeviceType h264HWDeviceType    = AV_HWDEVICE_TYPE_DXVA2;

    // Make sure the container supports HW Acceleration.
    if (CHECK_FOR_HW_ACCELERATION)
    {
        // First check is to see if the encoder was built into library.
        h264HWCodec = avcodec_find_encoder_by_name(H264_AMD_AMF_NAME.c_str());

        if (h264HWCodec != nullptr)
        {
            // The AMF encoder may have the support built in to the library
            // but that doesn't mean the harware exists to drive it.  The only
            // way to determine this is to create a HW Device Context then free it.
            int ffmpegError = av_hwdevice_ctx_create(&h264HWDeviceContext,
                                                     h264HWDeviceType,
                                                     H264_AMD_AMF_DEVICE.c_str(), nullptr, 0);

            // Was the HW device created?
            if ((ffmpegError == 0) && h264HWDeviceContext)
            {
                h264AMFExists = true;
            }
            // Close and cleanup the hardware device context.
            if (h264HWDeviceContext != nullptr)
            {
                av_buffer_unref(&h264HWDeviceContext);
                h264HWDeviceContext = nullptr;
            }
        }
        fi_log(LOG_DEBUG, "%s: [%s] %s.",
               __PRETTY_FUNCTION__, H264_AMD_AMF_NAME.c_str(), h264AMFExists ? HW_CODEC_EXISTS.c_str() : HW_CODEC_NOT_AVAILABLE.c_str());
    }
#ifdef USE_VERBOSE_INFO_DEBUG
    else
    {
        fi_log(LOG_DEBUG, "%s: [%s] %s.",
               __PRETTY_FUNCTION__, H264_AMD_AMF_NAME.c_str(), HW_CODEC_NO_CHECKING.c_str());
    }
#endif

    return h264AMFExists;
}

/*!
 * @brief Determines if the H264 Encoder for HW acceleration exists.
 * AV_HWDEVICE_TYPE_QSV is the device type for FFmpeg.
 *
 * @returns true if the encoder is available, false if not.
 */
bool FFmpegEncoder::IsQSVAvailable()
{
    bool h264QSVExists                      = false;
    const AVCodec *h264HWCodec              = nullptr;
    AVBufferRef *h264HWDeviceContext        = nullptr;
    enum AVHWDeviceType h264HWDeviceType    = AV_HWDEVICE_TYPE_QSV;

    // If VAAPI is available we don't to setup QSV.
    if (CHECK_FOR_HW_ACCELERATION)
    {
        // Is the encoder was built into library?
        h264HWCodec = avcodec_find_encoder_by_name(H264_INTEL_QSV_NAME.c_str());
        if (h264HWCodec != nullptr)
        {
            // The QSV encoder may have the support built in to the library
            // but that doesn't mean the harware exists to drive it.  The only
            // way to determine this is to create a HW Device Context then free it.
            int ffmpegError = av_hwdevice_ctx_create(&h264HWDeviceContext,
                                                     h264HWDeviceType,
                                                     H264_INTEL_QSV_DEVICE.c_str(), nullptr, 0);

            // Was the HW device created?
            if ((ffmpegError == 0) && h264HWDeviceContext)
            {
                h264QSVExists = true;
            }
            // Close and cleanup the hardware device context.
            if (h264HWDeviceContext != nullptr)
            {
                av_buffer_unref(&h264HWDeviceContext);
                h264HWDeviceContext = nullptr;
            }
        }
        fi_log(LOG_DEBUG, "%s: [%s] %s.",
               __PRETTY_FUNCTION__, H264_INTEL_QSV_NAME.c_str(), h264QSVExists ? HW_CODEC_EXISTS.c_str() : HW_CODEC_NOT_AVAILABLE.c_str());
    }
#ifdef USE_VERBOSE_INFO_DEBUG
    else
    {
        fi_log(LOG_DEBUG, "%s: [%s] %s.",
               __PRETTY_FUNCTION__, H264_INTEL_QSV_NAME.c_str(), HW_CODEC_NO_CHECKING.c_str());
    }
#endif

    return h264QSVExists;
}

/*!
 * @brief Closedown the Encoder Instance and free resources.
 */
void FFmpegEncoder::CloseEncoderInstance()
{
    std::lock_guard<std::mutex> guard(FFmpegEncoderInstanceMutex);
    std::string closeInfoStr;
    std::string encoderInfoStr  = "[FFmpegEncoder]";
    std::string cleaningStart   = "[Cleanup Starts]";
    std::string cleaningEnd     = "[Cleanup Ends]";

    // Report encoder statistics out to the log file for this session.
    FiDurations::LogDurationSeparator();
    if (FFmpegEncoderInstance)
    {
        if (FFmpegEncoderInstance->encoderStats)
            FFmpegEncoderInstance->encoderStats->MarkInterval();    // Mark the time to closure.

        FFmpegEncoderInstance->LogStatistics();
    }
    if (frameWriterThreadInstance)
    {
        frameWriterThreadInstance->LogStatistics();
    }
    FiDurations::LogDurationSeparator();

    closeInfoStr = "[frameWriterThread]";

    // Cleanup the frame writer thread.
    if (frameWriterThreadInstance)
    {
        if (VERBOSE_SAVE_LOG)
            fi_log(LOG_DEBUG, "%s: %s.", closeInfoStr.c_str(), cleaningStart.c_str());

        if (frameWriterThreadInstance->is_running())
        {
            // NOTE:  The framesToWrite queue will be cleaned up in
            // the frameWriterThreadInstance::cleanup() function.
            frameWriterThreadInstance->stop();
        }
        // Join the thread so we can clean it up.
        if (VERBOSE_SAVE_LOG)
            fi_log(LOG_DEBUG, "%s: Waiting for frameWriterThreadInstance to Join.", closeInfoStr.c_str());

        frameWriterThreadInstance->join();

        if (VERBOSE_SAVE_LOG)
            fi_log(LOG_DEBUG, "%s: Joined frameWriterThreadInstance.", closeInfoStr.c_str());

        frameWriterThreadInstance.reset();

        if (VERBOSE_SAVE_LOG)
            fi_log(LOG_DEBUG, "%s: %s.", closeInfoStr.c_str(), cleaningEnd.c_str());
    }
    // Cleanup the FFmpegEncoder object.
    if (FFmpegEncoderInstance)
    {
        fi_log(LOG_DEBUG, "%s: %s.", encoderInfoStr.c_str(), cleaningStart.c_str());
        if (VERBOSE_SAVE_LOG)
            fi_log(LOG_DEBUG, "%s: Waiting for frameWriterThreadInstance to Join.", closeInfoStr.c_str());

        // Cleanup the Filter Object.
        if (FFmpegEncoderInstance->filterPtr)
        {
            closeInfoStr = "[FFmpeg Filter]";
            if (VERBOSE_SAVE_LOG)
                fi_log(LOG_DEBUG, "%s: %s.", closeInfoStr.c_str(), cleaningStart.c_str());

            FFmpegEncoderInstance->filterPtr.reset();
            FFmpegEncoderInstance->filterPtr = nullptr;
            if (VERBOSE_SAVE_LOG)
                fi_log(LOG_DEBUG, "%s: %s.", closeInfoStr.c_str(), cleaningEnd.c_str());
        }

        // Clean up the codec context.
        if (FFmpegEncoderInstance->codec_ctx != nullptr)
        {
            closeInfoStr = "[FFmpeg Codec Context]";
            if (VERBOSE_SAVE_LOG)
                fi_log(LOG_DEBUG, "%s: %s.", closeInfoStr.c_str(), cleaningStart.c_str());
            avcodec_free_context(&FFmpegEncoderInstance->codec_ctx);
            FFmpegEncoderInstance->codec_ctx = nullptr;
            if (VERBOSE_SAVE_LOG)
                fi_log(LOG_DEBUG, "%s: %s.", closeInfoStr.c_str(), cleaningEnd.c_str());
        }

        // Cleanup any HWDeviceCTX allocated.
        if (HWDeviceCTX != nullptr)
        {
            av_buffer_unref(&HWDeviceCTX);
            HWDeviceCTX = nullptr;
        }

        // Cleanup the ffmpeg format context.
        if (FFmpegEncoderInstance->fmt_ctx != nullptr)
        {
            closeInfoStr = "[FFmpeg Format Context]";
            if (VERBOSE_SAVE_LOG)
                fi_log(LOG_DEBUG, "%s: %s.", closeInfoStr.c_str(), cleaningStart.c_str());

            // Close the FFMpeg MP4 file.
            if (!(FFmpegEncoderInstance->fmt_ctx->oformat->flags & AVFMT_NOFILE))
            {
                int result = avio_closep(&FFmpegEncoderInstance->fmt_ctx->pb);
                if (result != 0)
                {
                    if (VERBOSE_SAVE_LOG)
                        fi_log(LOG_DEBUG, "%s: avio_closep() returned [%d].", __PRETTY_FUNCTION__, result);
                }
            }
            avformat_free_context(FFmpegEncoderInstance->fmt_ctx);
            FFmpegEncoderInstance->fmt_ctx = nullptr;
            if (VERBOSE_SAVE_LOG)
                fi_log(LOG_DEBUG, "%s: %s.", closeInfoStr.c_str(), cleaningEnd.c_str());
        }

        // Reset the instance!
        FFmpegEncoderInstance.reset();
        FFmpegEncoderInstance = nullptr;    // Should already be nullptr from reset so this is overkill.
        if (VERBOSE_SAVE_LOG)
            fi_log(LOG_DEBUG, "%s: %s.", encoderInfoStr.c_str(), cleaningEnd.c_str());
    }
}

/*!
 * @brief Runs the FFmpeg AVFrame data through any filters which pushes
 * the frame through the filter translation changing the bits passed in
 * into the bits needed for the encoder producing the output video.
 *
 * It is only usable when the encoder and the filter object are both
 * properly prepared for the container processing the FFmpeg video stream.
 *
 * @param sourceFrame is a pointer to an FFmpeg AVFrame object created with
 * a call to av_frame_alloc() it contains the planar data necessary for
 * the encoder to translate raw frame data and encode it into the stream.
 *
 * @return A point to the filtered frame created here or nullptr if error.
 *
 */
AVFrame *FFmpegEncoder::CreateFilteredAVFrame(AVFrame *sourceFrame, int& ffmpegResultRef)
{
    int ffmpeg_result = ffmpegResultRef = 0;
    AVFrame *filteredFrame = nullptr;
    std::string myNameStr = "[FFmpegEncoder::CreateFilteredAVFrame]";

    try
    {
        // If there is no sourceFrame cannot create a filtered frame.
        if (sourceFrame == nullptr)
        {
            ffmpeg_result = AVERROR(EINVAL);
            fi_log(LOG_ERR, "%s: sourceFrame was nullptr!", myNameStr.c_str());
            goto exitCreateFilteredAVFrame;
        }

        // Is this encoder prepared?
        if (!IsPrepared())
        {
            ffmpeg_result = AVERROR(EINVAL);
            fi_log(LOG_ERR, "%s: The encoder is not prepared properly.", myNameStr.c_str());
            goto exitCreateFilteredAVFrame;
        }

        // Is the FFmpeg filter object prepared?
        if (!filterPtr || (!filterPtr->is_filter_prepared()))
        {
            ffmpeg_result = AVERROR(EINVAL);
            fi_log(LOG_ERR, "%s: The encoder's ffmpeg filter is not prepared.", myNameStr.c_str());
            goto exitCreateFilteredAVFrame;
        }

        // For very first frame processed plop out an information frame.
        if (framesRead == 1)
        {
            std::string frameIdStr = myNameStr.c_str() + std::string("Pre-Filter Frame 1"s);
            FFmpegContainer::LogAVFrameInfo(sourceFrame, frameIdStr.c_str(), false);
        }

        // Add the source AVFrame to the ffmpeg filter object to start the transition.
        ffmpeg_result = av_buffersrc_add_frame_flags(filterPtr->get_buffer_src_ctx(), sourceFrame, 0);
        if (ffmpeg_result < 0)
        {
            fi_log(LOG_ERR, "%s: av_buffersrc_add_frame_flags FAILED.", myNameStr.c_str());
            goto exitCreateFilteredAVFrame;
        }

        // Allocate the memory for the resultant filtered frame object.
        filteredFrame = av_frame_alloc();
        if (filteredFrame == nullptr)
        {
            ffmpeg_result = AVERROR(ENOMEM);
            fi_log(LOG_ERR, "%s: av_frame_alloc() FAILED.", myNameStr.c_str());
            goto exitCreateFilteredAVFrame;
        }

        // Pull the filtered frame data from the buffer sink.
        ffmpeg_result = av_buffersink_get_frame(filterPtr->get_buffer_sink_ctx(), filteredFrame);

        // Expected return >= 0 if frame returned but check for end of frame data.
        if (ffmpeg_result < 0)
        {
            // If no more frames for output - returns AVERROR(EAGAIN).
            // If flushed and no more frames for output - returns AVERROR_EOF
            if (ffmpeg_result == AVERROR(EAGAIN) || ffmpeg_result == AVERROR_EOF)
            {
                if (ffmpeg_result == AVERROR(AVERROR_EOF))
                {
                    fi_log(LOG_ERR, "%s: av_buffer_sink_get_frame() returned AVERROR_EOF.", myNameStr.c_str());
                }
                ffmpeg_result = 0;
            }
            else
            {
                // Unexpected error condition.
                log_ffmpeg_error_str(ffmpeg_result);
                fi_log(LOG_ERR, "%s: av_buffer_sink_get_frame() returned unexpeced ffmpeg_result[%d].",
                       myNameStr.c_str(), ffmpeg_result);
                goto exitCreateFilteredAVFrame;
            }
        }

        // Write information from the filtered frame (this is the frame converted in
        // proper format for the stream) to the log.
        if (framesRead == 1)
        {
            std::string frameIdStr = myNameStr.c_str() + std::string("Post-Filter Frame 1"s);
            FFmpegContainer::LogAVFrameInfo(filteredFrame, frameIdStr.c_str(), false);
        }

        // NOTE_JLC - Followup after testing.  When this is set to AV_PICTURE_TYPE_I it causes the encoder
        // to force a keyframe for every frame scenario.  When the default setting of AV_PICTURE_TYPE_NONE
        // is set the H.264 encoder applies compression to the frames based on the keyFrameInterval setting.
        // When the keyFrameInterval size is larger the compression results make a difference to the bitrate
        // and the datasize of the file but there is a degration when "scrubbing" in a video player.
        // What we have decided is to give the user the ability to set keyframes for every frame or we will
        // force keyframes at a modulus of every FAS_VIDEO::DEFAULT_H264_KEYFRAME_INTERVAL (which is 60 right
        // now because that was a QuickTime usage and it provides reasonable "scrubbing" performance).
        if ((framesRead % keyFrameInterval) == 0)
        {
            //fi_log(LOG_DEBUG, "%s: framesRead[%] FORCING KEYFRAME HERE.", myNameStr.c_str(), framesRead);
            filteredFrame->pict_type = AV_PICTURE_TYPE_I;
        }
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught creating filtered AVFrame.", myNameStr.c_str(), e.what());
        goto exitCreateFilteredAVFrame;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught creating filtered AVFrame.", myNameStr.c_str());
        goto exitCreateFilteredAVFrame;
    }

exitCreateFilteredAVFrame:
    ffmpegResultRef = ffmpeg_result;
    return filteredFrame;
} // End of CreateFilteredAVFrame()


/*!
 * @brief Runs the FFmpeg AVFrame data through hardware acceleration to
 * transform the NV12 raw frame data into the H.264 frame data.
 *
 * It is only usable when the encoder and hardware acceleration is available.
 *
 * @param sourceFrame is a pointer to an FFmpeg AVFrame object created with
 * a call to av_frame_alloc() it contains the planar data necessary for
 * the encoder to translate raw frame data and encode it into the stream.
 *
 * @return A pointer to the hw encoded frame created here or nullptr if error.
 *
 */
AVFrame *FFmpegEncoder::CreateHWAVFrame(AVFrame *sourceFrame, int& ffmpegResultRef)
{
    int ffmpeg_result = ffmpegResultRef = 0;
    AVFrame *hwFrame = nullptr;
    try
    {
        // Increment the frames "read" counter at top when we receive a non-nullptr for the frame.
        if (sourceFrame != nullptr)
        {
            framesRead++;
        }
        else
        {
            fi_log(LOG_ERR, "%s: sourceFrame was nullptr!", __PRETTY_FUNCTION__);
            goto exitCreateHWAVFrame;
        }

        // Is this encoder prepared?
        if (!IsPrepared())
        {
            fi_log(LOG_ERR, "%s: The encoder is not prepared properly.", __PRETTY_FUNCTION__);
            goto exitCreateHWAVFrame;
        }

        if (!(hwFrame = av_frame_alloc()))
        {
            ffmpeg_result = AVERROR(ENOMEM);
            fi_log(LOG_ERR, "%s: av_frame_alloc() FAILED.", __PRETTY_FUNCTION__);
            goto exitCreateHWAVFrame;
        }
        if ((ffmpeg_result = av_hwframe_get_buffer(codec_ctx->hw_frames_ctx, hwFrame, 0)) < 0)
        {
            fi_log(LOG_ERR, "%s: av_hwframe_get_buffer() FAILED.", __PRETTY_FUNCTION__);
            log_ffmpeg_error_str(ffmpeg_result);
            goto exitCreateHWAVFrame;
        }
        if (!hwFrame->hw_frames_ctx)
        {
            ffmpeg_result = AVERROR(ENOMEM);
            fi_log(LOG_ERR, "%s: hwFrame->hw_frames_ctx is nullptr!", __PRETTY_FUNCTION__);
            goto exitCreateHWAVFrame;
        }
        if ((ffmpeg_result = av_hwframe_transfer_data(hwFrame, sourceFrame, 0)) < 0)
        {
            fi_log(LOG_ERR, "%s: Error while transferring sourceFrame to hw surface frame.", __PRETTY_FUNCTION__);
            log_ffmpeg_error_str(ffmpeg_result);
            goto exitCreateHWAVFrame;
        }

        // Write information from the hardware frame (this is the frame converted in
        // proper format for the stream) to the log.
        if (framesRead == 1)
        {
            FFmpegContainer::LogAVFrameInfo(hwFrame, "FFMpegEncoder::CreateHWAVFrame Post Filter First Frame", false);
        }

        // NOTE_JLC - Followup after testing.  When this is set to AV_PICTURE_TYPE_I it causes the encoder
        // to force a keyframe for every frame scenario.  When the default setting of AV_PICTURE_TYPE_NONE
        // is set the H.264 encoder applies compression to the frames based on the keyFrameInterval setting.
        // When the keyFrameInterval size is larger the compression results make a difference to the bitrate
        // and the datasize of the file but there is a degration when "scrubbing" in a video player.
        // What we have decided is to give the user the ability to set keyframes for every frame or we will
        // force keyframes at a modulus of every FAS_VIDEO::DEFAULT_H264_KEYFRAME_INTERVAL (which is 60 right
        // now because that was a QuickTime usage and it provides reasonable "scrubbing" performance).
        if ((framesRead % keyFrameInterval) == 0)
        {
            //fi_log(LOG_DEBUG, "%s: framesRead[%] FORCING KEYFRAME HERE.", __PRETTY_FUNCTION__, framesRead);
            hwFrame->pict_type = AV_PICTURE_TYPE_I;
        }

        // TRY to set this to see if it fixes one of the problems I see in Hw accelerated encoder
        // where the pts data is just garbage.

        // pts is the Presentation Time Stamp - time when frame should be shown to user.
        // THIS MUST be the frame number in the sequence.  It is how the time is interleaved
        // when the encoded frames start coming out of the ffmpeg encoder.  In this object the
        // inforamtion is maintained in the framesProcessed.
        hwFrame->pts = sourceFrame->pts;
        hwFrame->pkt_dts = sourceFrame->pkt_dts;
        hwFrame->sample_aspect_ratio.num = 1;
        hwFrame->sample_aspect_ratio.den = 1;
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught creating HW AVFrame.", __PRETTY_FUNCTION__, e.what());
        goto exitCreateHWAVFrame;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught creating HW AVFrame.", __PRETTY_FUNCTION__);
        goto exitCreateHWAVFrame;
    }

exitCreateHWAVFrame:
    ffmpegResultRef = ffmpeg_result;
    return hwFrame;
} // End of CreateHWAVFrame()

/*!
 * @brief Runs the FFmpeg AVFrame data through any filters which pushes
 * the frame through the filter translation changing the bits passed in
 * into the bits needed for the encoder producing the output video.
 *
 * It is only usable when the encoder and the filter object are both
 * properly prepared for the container processing the FFmpeg video stream.
 *
 * @param frame is a pointer to an FFmpeg AVFrame object created with
 * a call to av_frame_alloc() it contains the planar data necessary for
 * the encoder to translate raw frame data and encode it into the stream.
 *
 * @return The ffmpeg_result recorded here.
 *
 */
int FFmpegEncoder::EncodeAVFrame(AVFrame **frame_pp)
{
    int ffmpeg_result = 0;
    uint32_t numberOfFramesInQueue      = 0;

    // Keeps track of the size of the queue.  It should be reducing as the
    // backpressure is held off until the resume threshold is reached for
    // the frames to write.  Right now it is only being reported in log msg.
    uint32_t lastCheckQueueSize         = 0;

    AVFrame *frame = *frame_pp;

    AVFrame *filteredFrame = nullptr;
    std::shared_ptr<FFmpegEncoderFrame> filteredEncoderFrame;
    try
    {
        // If never created the addFrame Statistic Duration create it.
        if (encodeAVFrameStats == nullptr)
            encodeAVFrameStats = std::unique_ptr<FiDurations>(new FiDurations());
        else
            encodeAVFrameStats->MarkStart(); // Mark start of interval.

        // If the frame provided to encode is not null increment the framesRead counter.
        if (frame != nullptr)
        {
            framesRead++;
        }
        // Otherwise mark this as an invalid input value.
        else
        {
            ffmpeg_result = AVERROR(EINVAL);
            goto exitEncodeAVFrame;
        }

        // If operating with hardware encoding then create the proper hardware AVFrame* for the
        // hardware encoder.  If the source frame is not in the NV12 format it must be converted
        // prior to the call to CreateHWAVFrame as that function only operates on NV12 data.
        if (HWDeviceCTX)
        {
            AVFrame *tempNV12 = nullptr;
            //#HWTEST
            if (this->avInPixelFormat != AV_PIX_FMT_NV12)
            {
                // Can we filter the source frame into an NV12 AVFrame?
                tempNV12 = CreateFilteredAVFrame(frame, ffmpeg_result);
                if (tempNV12 != nullptr)
                {
                    filteredFrame = CreateHWAVFrame(tempNV12, ffmpeg_result);
                    av_frame_free(&tempNV12);
                    tempNV12 = nullptr;
                }
                else
                {
                    fi_log(LOG_ERR, "%s: Unable to generate an NV12 frame for CreateHWAVFrame() input AVPixFormat was [%d]!", __PRETTY_FUNCTION__, (int)this->avInPixelFormat);
                }
            }
            else    // We already have an NV12 Frame to use.
            {
                filteredFrame = CreateHWAVFrame(frame, ffmpeg_result);
            }
        }
        else    // Normal software encoder is being used so create the filtered AVFrame as specified.
        {
            // If in/out pixel formats match just use source frame and save
            // an allocate/filter/copy of an new AVFrame.
            if (this->avInPixelFormat == this->avOutPixelFormat)
            {
                // Grabbing ownership of source frame and resetting the
                // pointer to a nullptr as this object owns it now.
                filteredFrame = frame;
                *frame_pp = nullptr;
            }
            else
            {
                filteredFrame = CreateFilteredAVFrame(frame, ffmpeg_result);
            }
        }

        if (filteredFrame != nullptr)
        {
            // If frame writer thread instance isn't running there is an issue.
            if (!frameWriterThreadInstance)
            {
                ffmpeg_result = AVERROR(EINVAL);
                fi_log(LOG_ERR, "%s: No frameWriterThreadInstance!", __PRETTY_FUNCTION__);
                goto exitEncodeAVFrame;
            }

            // Create the data frame to add.
            filteredEncoderFrame = FFmpegEncoderFrame::CreateFFmpegEncoderFrame(filteredFrame);
            if (!filteredEncoderFrame)
            {
                ffmpeg_result = AVERROR(ENOMEM);
                fi_log(LOG_ERR, "%s: Unable to create filteredEncoderFrame.", __PRETTY_FUNCTION__);
                goto exitEncodeAVFrame;
            }

            // Is the frame writer's queue at a maximum size?
            if ((lastCheckQueueSize = framesToWrite->NumberOfFramesInQueue()) >= FRAME_QUEUE_MAX_SIZE)
            {
                // We will have a maximum time that we can hold on to this frame before
                // we start to impact any input frame streamer.  Currently that will be
                // about 6 seconds so we will assume we can be in here at least 5...
                bool timeIsUp = false;

                // Keep track of a backpressure scenario.
                bool backpressureHold = false;

                // Accumulator for the time we take in a backpressure loop.
                uint32_t waitTimeForBackpressure = 0;

                // Don't want to get into a situation where we never come out of this
                // sleep cycle if for instance the frame writer stops pulling frames
                // from the frame queue.  Let's keep track of the time in the loop.
                while (!timeIsUp && ((numberOfFramesInQueue = framesToWrite->NumberOfFramesInQueue()) > FRAME_QUEUE_RESUME_SIZE))
                {
                    backpressureHold = true;

                    // We need to be in a holding pattern which means that our thread is going to
                    // hold onto this current frame which has the intended effect of backing off the
                    // input stream handler until we release the buffer after this hold.
                    // We are trying to give our ffmpeg frame writer queue a chance to catch up with
                    // all of the input frames and we have to get the input stream to back-off.
                    // This is a balancing act at the moment because we do not have a way to notify
                    // the calling object about our internal holding pattern.  AT least at this point
                    // in time.  If we hold off here too long what happens is that the system's receive
                    // frame queue will start timing out as nothing is processing the input frames. The
                    // optimal solution would be to have the system know to pause the input stream but
                    // there is no mechanism for that at the moment. -- Right now the input stream handler
                    // expects a frame to be handled every 2 seconds and if not will time out.  This time
                    // out is allowed up to 3 times for a total of 6 seconds which is the window we have
                    // here to hold.  If we hold out here for 6 seconds we risk timing out the save video
                    // handler of the system which will close the input stream so we must facilitate a hold
                    // no longer than that time frame.
                    std::this_thread::sleep_for(std::chrono::milliseconds(FRAME_QUEUE_HOLD_MS));
                    waitTimeForBackpressure += FRAME_QUEUE_HOLD_MS;

                    // We can only hold for a certain amount of time.
                    if (waitTimeForBackpressure > MS_THRESHOLD_FOR_FRAME_QUEUE_RESUME)
                    {
                        timeIsUp = true;
                    }
                }

                // Did we have a backpressure hold?  Log it.
                if (backpressureHold)
                {
#if 0
                    fi_log(LOG_DEBUG, "[FFMPEG Encoder Releasing BACKPRESSURE HOLD]: WaitTime[%dms] MaxWait[%dms] framesInQueueNow[%d] resumeSizeWanted[%d] startingQueueSize[%d].",
                           waitTimeForBackpressure, (MS_THRESHOLD_FOR_FRAME_QUEUE_RESUME + 1000),
                           numberOfFramesInQueue, FRAME_QUEUE_RESUME_SIZE, lastCheckQueueSize);
#endif
                }

                // Is there a place to put this frame now that we tried to relieve backpressure?
                if (framesToWrite->NumberOfFramesInQueue() >= FRAME_QUEUE_MAX_SIZE)
                {
                    ffmpeg_result = AVERROR(ETIME);
                    encodeAVFrameTimeout = true;
                    fi_log(LOG_ERR, "%s: Unable to reduce the frame queue for ffmpeg encoder.  The framesToWrite[%d] buffer queue is full and cannot accomodate this frame!",
                           __PRETTY_FUNCTION__, framesToWrite->NumberOfFramesInQueue());
                    goto exitEncodeAVFrame;
                }
            }

            // Add it to the processing queue if make it here.
            framesToWrite->AddFrameToQueue(filteredEncoderFrame);

            if (encodeAVFrameStats)
                encodeAVFrameStats->MarkInterval();   // Mark the time spent here.
        }
    }
    // For exception handling just report the error.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception %s caught.", __PRETTY_FUNCTION__, e.what());
        goto exitEncodeAVFrame;
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught.", __PRETTY_FUNCTION__);
        goto exitEncodeAVFrame;
    }

exitEncodeAVFrame:

    return ffmpeg_result;
} // End of EncodeAVFrame()

bool FFmpegEncoder::IsPrepared()
{
    bool prepared = false;
    std::lock_guard<std::mutex> guard(FFmpegEncoderInstanceMutex);
    if (FFmpegEncoderInstance != nullptr)
        prepared = true;

    return prepared;
}

/*!
 * @brief Return the Frame Writer status information that will
 * inform the caller if the ffmpeg encoder (aka the Frame Writer Thread)
 * was successful.
 *
 * @param inError (output) - will be set to true if there is an ffmpegError status
 * that occurred during the frame writer thread's processing.  Otherwise it will
 * be set to false indicating that no ffmpeg errors occurred.
 *
 * @returns True if the frame writer thread wrote the AV Trailer indicating
 * that the MP4 video was created.
 */
bool FFmpegEncoder::IsComplete(bool& inError)
{
    bool isComplete = false;

    // Get the very latest frame writer thread's status information.  If it has already
    // terminated because the video is completed, it's ok because this returns the very
    // last status in that case.
    PollFrameWriterThreadStatus();

    // Set the error state of the ffmpeg encoder
    if (theFrameWriterStatus.ffmpegError != 0)
    {
        inError = true;
    }
    else
    {
        inError = false;
    }

    // And finally the true indicator that the ffmpeg encoder
    // completed the video will be whether or not the AV trailer
    // was written to the video file.
    if (theFrameWriterStatus.wroteAVTrailer)
    {
        isComplete = true;
    }

    return isComplete;
}

/*!
 * @brief Return the AVCodecContext for this encoder object.
 *
 * @returns The AVCodecContext pointer or nullptr if not set.
 */
AVCodecContext *FFmpegEncoder::GetAVCodecContext()
{
    AVCodecContext *theAVCodecContext = nullptr;
    theAVCodecContext = codec_ctx;
    return theAVCodecContext;
}

/*!
 * @brief Return the AVFormatContext for this encoder object.
 *
 * @returns The AVFormatContext pointer or nullptr if not set.
 */
AVFormatContext *FFmpegEncoder::GetAVFormatContext()
{
    AVFormatContext *theAVFormatContext = nullptr;
    theAVFormatContext = fmt_ctx;
    return theAVFormatContext;
}

/*!
 * @brief Return the FFmpeg Input AVPixelFormat record.
 *
 * @returns The FFmpeg AVPixelFormat value for incoming frames to encoder,
 * or AV_PIX_FMT_NONE if error.
 */
enum AVPixelFormat  FFmpegEncoder::GetPixelInputFormat()
{
    enum AVPixelFormat theAVPixelFormatToReturn = AV_PIX_FMT_NONE;

    if (!encoderPrepared)
        goto exitGetPixelFormat;

    theAVPixelFormatToReturn = avInPixelFormat;

exitGetPixelFormat:
    return theAVPixelFormatToReturn;
}

/*!
 * @brief Return the FFmpeg Output AVPixelFormat record.
 *
 * @returns The FFmpeg AVPixelFormat value for outgoing frames from encoder,
 * or AV_PIX_FMT_NONE if error.
 */
enum AVPixelFormat FFmpegEncoder::GetPixelOutputFormat()
{
    enum AVPixelFormat theAVPixelFormatToReturn = AV_PIX_FMT_NONE;

    if (!encoderPrepared)
        goto exitGetPixelFormat;

    if (codec_ctx == nullptr)
        goto exitGetPixelFormat;

    // Return the setting in the object.
    theAVPixelFormatToReturn = avOutPixelFormat;

exitGetPixelFormat:
    return theAVPixelFormatToReturn;
}

/*!
 * @brief Return the FFmpeg HW Device AVPixelFormat setting.
 *
 * @returns The FFmpeg AVPixelFormat value for hw encoder if set,
 * or AV_PIX_FMT_NONE if error.
 */
enum AVPixelFormat FFmpegEncoder::GetPixelHWFormat()
{
    enum AVPixelFormat theAVPixelFormatToReturn = AV_PIX_FMT_NONE;

    if (!encoderPrepared)
        goto exitGetPixelFormat;

    if (codec_ctx == nullptr)
        goto exitGetPixelFormat;

    // Return the setting in the object.
    theAVPixelFormatToReturn = avHWDevicePixelFormat;

exitGetPixelFormat:
    return theAVPixelFormatToReturn;
}

/*!
 * @brief Returns the AVRational setting for the aspect ratio.
 *
 * @param avrAspectRatioToSet   The resultant data record.
 *
 * @returns True if set properly, false if error.
 *
 */
bool FFmpegEncoder::GetAspectRatio(AVRational& avrAspectRatioToSet)
{
    avrAspectRatioToSet.num = avrAspectRatio.num;
    avrAspectRatioToSet.den = avrAspectRatio.den;

    // TODO_JLC - Add verification.
    return true;
}

/*!
 * @brief Returns the AVRational setting for the time base.
 *
 * @param avrTimeBaseToSet   The resultant data record.
 *
 * @returns True if set properly, false if error.
 *
 */
bool FFmpegEncoder::GetTimeBase(AVRational& avrTimeBaseToSet)
{
    avrTimeBaseToSet.num = avrTimeBase.num;
    avrTimeBaseToSet.den = avrTimeBase.den;
    // TODO_JLC - Add verification.
    return true;
}

/*!
 * @brief Returns the AVRational setting for the frame rate.
 *
 * @param avrFrameRateToSet   The resultant data record.
 *
 * @returns True if set properly, false if error.
 *
 */
bool FFmpegEncoder::GetFrameRate(AVRational& avrFrameRateToSet)
{
    avrFrameRateToSet.num = avrTimeBase.num;
    avrFrameRateToSet.den = avrTimeBase.den;
    // TODO_JLC - Add verification.
    return true;
}

/*!
 * @brief Returns number of frames expected to encode.
 *
 * @returns the stored value or 0 if object is not valid.
 */
uint32_t FFmpegEncoder::GetFrameCount()
{
    uint32_t frameCountToReturn = 0;

    if (IsPrepared())
        frameCountToReturn = frameCount;

    return frameCountToReturn;
}

/*!
 * @brief Returns frame width set when encoder prepared.
 *
 * @returns the stored value or 0 if object is not valid.
 *
 */
uint32_t FFmpegEncoder::GetFrameWidth()
{
    uint32_t frameWidthToReturn = 0;

    if (IsPrepared())
        frameWidthToReturn = frameWidth;

    return frameWidthToReturn;
}

/*!
 * @brief Returns frame height set when encoder prepared.
 *
 * @returns the stored value or 0 if object is not valid.
 *
 */
uint32_t FFmpegEncoder::GetFrameHeight()
{
    uint32_t frameHeightToReturn = 0;

    if (IsPrepared())
        frameHeightToReturn = frameHeight;

    return frameHeightToReturn;
}

/*!
 * @brief Returns the count of encoded frames processed and sets up the true
 * state of the encoder.  If the setting provided in isComplete is true then
 * the ffmpeg encoder wrote the AV Trailer and the video is truly completed.
 *
 * @param isPrepared(output) set to whether or not the ffmpeg encoder had been prepared.
 * @param isComplete(output) set to whether or not video is truly completed without error.
 * @param hasError(output) set true if there is an error condition in the ffmpeg encoder.
 *
 * @returns The count of encoded frames processed by encoder.
 */
uint32_t FFmpegEncoder::GetEncodedFrameCount(bool& isPrepared, bool& isComplete, bool& hasError)
{
    uint32_t theEncodedFrameCount = 0;
    isPrepared = IsPrepared();
    isComplete = false;
    hasError = false;

    if (isPrepared)
    {
        // The function IsComplete() polls for frame writer status and error condition.
        // It is also looking to see if the AV trailer was written which signals definitively
        // that the video is completed and all frames are encoded.
        isComplete = IsComplete(hasError);
        theEncodedFrameCount = theFrameWriterStatus.framesEncoded;
#if 0
        // FAKE OUT.
        if (theEncodedFrameCount >= 25)
        {
            hasError = true;    // Let's fake an error condition.
        }
#endif
    }
    return theEncodedFrameCount;
}

/*!
 * @brief Flush a frame into the encoder stream.
 *
 * @return True if the encoder was sent the flush encoder frame or
 * false if there is an error condition.
 *
 */
bool FFmpegEncoder::SendFlushEncoderFrame()
{
    bool sentFlushEncoderFrame = false;
    int ffmpeg_result = 0;
    std::string myNameStr = "[FFmpegEncoder::SendFlushEncoderFrame]";

    std::shared_ptr<FFmpegEncoderFrame> flushingEncoderFrame;

    // These FFMpeg calls can throw an exception and should gracefully
    // exit if this occurs.
    try
    {
        // Is this encoder prepared?
        if (!IsPrepared())
        {
            fi_log(LOG_DEBUG, "%s: Encoder was not prepared!!",
                   myNameStr.c_str());
            goto exitFlushEncoder;
        }

        // Need to create an empty AVFrame to send to through the system.
        flushingEncoderFrame = FFmpegEncoderFrame::CreateFFmpegFlushFrame();
        if (!flushingEncoderFrame)
        {
            ffmpeg_result = AVERROR(ENOMEM);
            fi_log(LOG_ERR, "%s: Unable to create flushingEncoderFrame!", myNameStr.c_str());
            goto exitFlushEncoder;
        }

        // Write it to the background frame writer thread.
        // This is another place where we need to sleep until EOF condition
        framesToWrite->AddFrameToQueue(flushingEncoderFrame);

        // At this point in time this particular thread can wait until
        // there is an error condition or the EOF condition is met.
        if (frameWriterThreadInstance)
        {
            fi_log(LOG_DEBUG, "%s: Sent the Flush-Frame and polling final status from Frame Writer Thread!", myNameStr.c_str());
            PollFrameWriterThreadStatus();
        }

        ffmpeg_result = theFrameWriterStatus.ffmpegError;

        if (ffmpeg_result == 0)
            sentFlushEncoderFrame = true;   // Frame was sent and there is no error.
    }
    // Look for a specific resource allocation error to set state of the reader object.
    catch (std::exception& e)
    {
        fi_log(LOG_ERR, "%s: Exception caught [%s].", myNameStr.c_str(), e.what());
    }
    catch (...)
    {
        fi_log(LOG_ERR, "%s: (...) Exception caught.", myNameStr.c_str());
    }

exitFlushEncoder:
    fi_log(LOG_DEBUG, "%s: {exit} encodedEOF[%d] framesRead[%d] framesEncoded[%d] frameCount[%d] wroteAVTrailer[%s] ffmpegResult[%d].",
           myNameStr.c_str(), theFrameWriterStatus.ffmpegEOF, framesRead, theFrameWriterStatus.framesEncoded,
           frameCount, theFrameWriterStatus.wroteAVTrailer ? "YES" : "FrameWriterThread still encoding", ffmpeg_result);

    return sentFlushEncoderFrame;

} // End of SendFlushEncoderFrame()

/*!
 * @brief Flush a frame into the encoder stream.
 *
 * @return A boolean true if was able to poll the writer thread for status.
 *
 */
bool  FFmpegEncoder::PollFrameWriterThreadStatus()
{
    bool polledOK = false;
    FRAME_WRITER_STATUS_REC localStatus;    // Avoid a local mutex lock.

    // Poll for status from the Frame Writer Thread.
    if (frameWriterThreadInstance)
    {
        frameWriterThreadInstance->GetStatusRec(localStatus);
        polledOK = true;
    }

    // If the local status was read set ours.
    if (polledOK)
    {
        theFrameWriterStatus = localStatus; // A shallow copy should work.
    }

    return polledOK;
}






///////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////CODE BELOW IS LEGACY IMPLEMENTATION AND SOME PULLED FFMPEG SOURCE MODULES FOR TESTING.
///////////////////////////////////////////////////////////////////////////////////////////////////////

// GRAVEYARD Of Ideas...

#if 0

/*!
 * @brief Encodes an FFmpeg AVFrame into the video stream.
 *
 * This could write more than one encoded frame to the video stream depending
 * on the internal state of the encoder.
 *
 * @param filteredFrame is a pointer to an FFmpeg AVFrame object created with
 * a call to av_frame_alloc() it contains the planar data necessary for
 * the encoder to translate raw frame data and encode it into the stream.
 * If it is set to a nullptr it means the caller is flushing the encoder and
 * is a valid invocation.
 *
 * @param writtenCount is a reference to a written count it will be set to zero
 * upon entry to this function and will have a count of the number of encoded
 * frames written to the output stream.
 *
 * @return A zero if successful, otherwise a negative value is an ffmpeg error code.
 *
 */
int FFmpegEncoder::EncodeAndWriteFrames(AVFrame *filteredFrame, uint32_t& writtenCount)
{
    int ffmpeg_result = 0;                      // Holds the FFmpeg libary results.

    AVPacket *packetReceivedPtr = nullptr;      // The encoded AVPacket object.
    AVCodecContext *enc_codec_ctx = nullptr;    // The FFmpeg encoder's codec context data.
    AVFormatContext *enc_fmt_ctx = nullptr;     // The FFmpeg encoder's format context data.

    // Set the writtenCount to 0.
    writtenCount = 0;

    // Note:  When flushing the encoder the filt_frame can be a nullptr.

    // If this object isn't prepared it's a setup failure or error condition.
    if (!IsPrepared())
    {
        // Consider this an invalid input setting.
        ffmpeg_result = AVERROR(EINVAL);
        goto exitEncodeAndWriteFrame;
    }

    // Get the encoder's codec and format data pointers.
    enc_codec_ctx = GetAVCodecContext();
    enc_fmt_ctx = GetAVFormatContext();

    // Make sure these ffmpeg data structure pointers are valid.
    if (enc_codec_ctx == nullptr || enc_fmt_ctx == nullptr)
    {
        ffmpeg_result = AVERROR(ENOMEM);
        goto exitEncodeAndWriteFrame;
    }

    // Allocate the AVPacket that will be used in encoding loop.
    packetReceivedPtr = av_packet_alloc();
    if (packetReceivedPtr == nullptr)
    {
        ffmpeg_result = AVERROR(ENOMEM);
        fi_log(LOG_ERR, "%s: Unable to allocate AVPacket object.", __PRETTY_FUNCTION__);
        goto exitEncodeAndWriteFrame;
    }

    // The ffmpeg encoder takes in the AVFrame data but it may not provide an
    // encoded packet right away.  What happens is that the encoder needs to
    // consume (via avcodec_send_frame) enough frames to build up what internal
    // data it needs for its compression algorithms.  When requesting encoded
    // frames from ffmpeg encoder if the ffmpeg_result of call is EAGAIN setting
    // that indicates the encoder doesn't have enough data to provide an encoded
    // frame for the output stream yet.

#ifdef LOG_ENCODER_FRAME_INFO
    FFmpegContainer::LogAVFrameInfo(filteredFrame, "FFmpegEncoder::EncodeAndWriteFrames Before SendFrame", true);
#endif

    // First the frame is sent to the encoder.
    ffmpeg_result = avcodec_send_frame(enc_codec_ctx, filteredFrame);
    if (ffmpeg_result < 0)
    {
        log_ffmpeg_error_str(ffmpeg_result);
        fi_log(LOG_ERR, "%s: avcodec_send_frame() Failed.", __PRETTY_FUNCTION__);
        goto exitEncodeAndWriteFrame;
    }

    // Loop to receive encoded packets and write them to the output
    // stream for as long as data is available or there is an error.
    // There is no way of knowing how many frames the encoder will provide
    // to this loop but they all must be consumed and processed until out
    // of data for this run.
    while (ffmpeg_result >= 0)
    {
        // Note:  Each time through make sure to wipe the AVPacket structure
        // by calling av_packet_unref(packetReceivedPtr).
        ffmpeg_result = avcodec_receive_packet(enc_codec_ctx, packetReceivedPtr);

        // If ffmpeg_result is zero - there is a writable encoded frame.
        if (ffmpeg_result == 0)
        {
            // TODO_JLC Will want to conditionalize this for verbosity.
#ifdef DEBUG_FRAME_LEVEL_MSGS
            fi_log(LOG_DEBUG, "%s: **MUX** pkt->pts[%lld 0x%llx] pkt->dts[0x%llx] st->cur_dts[0x%llx] st->first_dts[0x%llx] st->timebase[%d/%d] ctx->timebase[%d/%d]\n",
                   __PRETTY_FUNCTION__,
                   packetReceivedPtr->pts,
                   packetReceivedPtr->pts,
                   packetReceivedPtr->dts,
                   enc_fmt_ctx->streams[0]->cur_dts,
                   enc_fmt_ctx->streams[0]->first_dts,
                   enc_fmt_ctx->streams[0]->time_base.den,
                   enc_fmt_ctx->streams[0]->time_base.num,
                   enc_codec_ctx->time_base.num,
                   enc_codec_ctx->time_base.den);
#endif

#ifdef LOG_ENCODER_FRAME_INFO
            fi_log(LOG_DEBUG, "FFmpegEncoder::EncodeAndWriteFrames **MUX** frame[%lld] frameCount[%d] framesRead[%d] framesEncoded[%d] encodedEOF[%d]",
                   packetReceivedPtr->pts, frameCount, framesRead, framesEncoded, encodedEOF);
#endif

            // Set up the time base information.
            av_packet_rescale_ts(packetReceivedPtr, enc_codec_ctx->time_base, enc_fmt_ctx->streams[0]->time_base);

            // Write out the encoded frame.
            ffmpeg_result = av_interleaved_write_frame(enc_fmt_ctx, packetReceivedPtr);

            // Record any error from the write as this should not fail.
            if (ffmpeg_result != 0)
            {
                log_ffmpeg_error_str(ffmpeg_result);
                fi_log(LOG_ERR, "%s: av_interleaved_write_frame() failed.", __PRETTY_FUNCTION__);
                goto exitEncodeAndWriteFrame;
            }

            // Reset the AVPacket object to prepare for next receive.
            av_packet_unref(packetReceivedPtr);

            // Increase the number of encoded frames counter by one and the
            // number of frames written in this call returned to caller.
            framesEncoded++;
            writtenCount++;

            // Loop again to see if there are more encoded frames to write to the stream.
            continue;
        }

        // Falling down to this logic means that the call to avcodec_receive_packet
        // did not return an ffmpeg_result of zero.

        // If ffmpeg_result is EAGAIN it means that the encoder needs more
        // input frames before it can return any encoded frames for writing.
        if (ffmpeg_result == AVERROR(EAGAIN))
        {
            ffmpeg_result = 0;  // Don't consider this a terminal error.
            goto exitEncodeAndWriteFrame;
        }

        // FFMpeg encoder has indicated end of encoded data.
        if (ffmpeg_result == AVERROR_EOF)
        {
            encodedEOF = true;
            ffmpeg_result = 0;  // Don't consider this a terminal error.
            fi_log(LOG_DEBUG, "%s: avcodec_receive_packet() has reached encodedEOF state.", __PRETTY_FUNCTION__);
            goto exitEncodeAndWriteFrame;
        }

        // This error was not expected.  Report it to the log.
        if (ffmpeg_result < 0)
        {
            log_ffmpeg_error_str(ffmpeg_result);
            fi_log(LOG_DEBUG, "%s: Encoder reported unexpected error.", __PRETTY_FUNCTION__);
            goto exitEncodeAndWriteFrame;
        }

        // Make sure to unreference any packet processed.
        // No code path should end up here but keep it here anyway.
        av_packet_unref(packetReceivedPtr);
    }

exitEncodeAndWriteFrame:

    // Cleanup and reset the AVPacket if it exists.
    if (packetReceivedPtr != nullptr)
        av_packet_unref(packetReceivedPtr);

    return ffmpeg_result;
}








//--------------------------------------------------------------------
//! Flush encoder.
//--------------------------------------------------------------------
int FFmpegEncoder::flush_encoder()
{
    int ffmpeg_result = 0;
    uint32_t writtenCount = 0;
    uint32_t totalWrittenCount = 0;
    bool checkForMoreFrames = true;
    int pass = 1;

    // Flush the encoder in a loop until no more encoded frames.
    while (checkForMoreFrames && ffmpeg_result >= 0)
    {
        fi_log(LOG_DEBUG, "%s: Flushing encoder pass [%d].", __PRETTY_FUNCTION__, pass++);
        ffmpeg_result = EncodeAndWriteFrame(nullptr, writtenCount);

        // Keep a running total.
        totalWrittenCount += writtenCount;

        // If some frames were written keep the checkForMoreFrames true.
        checkForMoreFrames = (writtenCount > 0) ? true : false;

        // Reset the writtenCount to zero.
        writtenCount = 0;
    }

    // Record failure in log.
    if (ffmpeg_result < 0)
    {
        log_ffmpeg_error_str(ffmpeg_result);
        fi_log(LOG_ERR, "%s: Flushing failed.", __PRETTY_FUNCTION__);
    }

    fi_log(LOG_DEBUG, "%s: Writing trailer totalWrittenCount[%d].",
           __PRETTY_FUNCTION__, totalWrittenCount);

    // Wrote the trailer to the video stream.
    ffmpeg_result = av_write_trailer(get_fmt_ctx());

    // Report any error writing the video trailer.
    if (ffmpeg_result < 0)
    {
        log_ffmpeg_error_str(ffmpeg_result);

        fi_log(LOG_ERR, "%s: av_write_trailer failed.", __PRETTY_FUNCTION__);
    }

    return ffmpeg_result;
} // End of flush_encoder
#endif

#if 0
// FOR TESTING ONLY PULL IN SOME PRIV DATA FOR THE MJPEG
/**
 * MpegEncContext.
 */

/**
 * Scantable.
 */
typedef struct ScanTable
{
    const uint8_t *scantable;
    uint8_t permutated[64];
    uint8_t raster_end[64];
} ScanTable;

typedef struct ThreadFrame
{
    AVFrame *f;
    AVCodecContext *owner[2];
    // progress->data is an array of 2 ints holding progress for top/bottom
    // fields
    AVBufferRef *progress;
} ThreadFrame;


/**
 * Picture.
 */
typedef struct Picture
{
    struct AVFrame *f;
    ThreadFrame tf;

    AVBufferRef *qscale_table_buf;
    int8_t *qscale_table;

    AVBufferRef *motion_val_buf[2];
    int16_t(*motion_val[2])[2];

    AVBufferRef *mb_type_buf;
    uint32_t *mb_type;          ///< types and macros are defined in mpegutils.h

    AVBufferRef *mbskip_table_buf;
    uint8_t *mbskip_table;

    AVBufferRef *ref_index_buf[2];
    int8_t *ref_index[2];

    AVBufferRef *mb_var_buf;
    uint16_t *mb_var;           ///< Table for MB variances

    AVBufferRef *mc_mb_var_buf;
    uint16_t *mc_mb_var;        ///< Table for motion compensated MB variances

    int alloc_mb_width;         ///< mb_width used to allocate tables
    int alloc_mb_height;        ///< mb_height used to allocate tables

    AVBufferRef *mb_mean_buf;
    uint8_t *mb_mean;           ///< Table for MB luminance

    AVBufferRef *hwaccel_priv_buf;
    void *hwaccel_picture_private; ///< Hardware accelerator private data

    int field_picture;          ///< whether or not the picture was encoded in separate fields

    int64_t mb_var_sum;         ///< sum of MB variance for current frame
    int64_t mc_mb_var_sum;      ///< motion compensated MB variance for current frame

    int b_frame_score;
    int needs_realloc;          ///< Picture needs to be reallocated (eg due to a frame size change)

    int reference;
    int shared;

    uint64_t encoding_error[AV_NUM_DATA_POINTERS];
} Picture;

typedef struct PutBitContext
{
    uint32_t bit_buf;
    int bit_left;
    uint8_t *buf, *buf_ptr, *buf_end;
    int size_in_bits;
} PutBitContext;

enum OutputFormat
{
    FMT_MPEG1,
    FMT_H261,
    FMT_H263,
    FMT_MJPEG,
};

typedef struct ScratchpadContext
{
    uint8_t *edge_emu_buffer;     ///< temporary buffer for if MVs point to out-of-frame data
    uint8_t *rd_scratchpad;       ///< scratchpad for rate distortion mb decision
    uint8_t *obmc_scratchpad;
    uint8_t *b_scratchpad;        ///< scratchpad used for writing into write only buffers
} ScratchpadContext;
#define MAX_THREADS 32


typedef struct MpegEncContext
{
    AVClass *our_class;

    int y_dc_scale, c_dc_scale;
    int ac_pred;
    int block_last_index[12];  ///< last non zero coefficient in block
    int h263_aic;              ///< Advanced INTRA Coding (AIC)

    /* scantables */
    ScanTable inter_scantable; ///< if inter == intra then intra should be used to reduce the cache usage
    ScanTable intra_scantable;
    ScanTable intra_h_scantable;
    ScanTable intra_v_scantable;

    /* WARNING: changes above this line require updates to hardcoded
     *          offsets used in ASM. */

    struct AVCodecContext *avctx;
    /* the following parameters must be initialized before encoding */
    int width, height;///< picture size. must be a multiple of 16
    int gop_size;
    int intra_only;   ///< if true, only intra pictures are generated
    int64_t bit_rate; ///< wanted bit rate
    enum OutputFormat out_format; ///< output format
    int h263_pred;    ///< use MPEG-4/H.263 ac/dc predictions
    int pb_frame;     ///< PB-frame mode (0 = none, 1 = base, 2 = improved)

    /* the following codec id fields are deprecated in favor of codec_id */
    int h263_plus;    ///< H.263+ headers
    int h263_flv;     ///< use flv H.263 header

    enum AVCodecID codec_id;     /* see AV_CODEC_ID_xxx */
    int fixed_qscale; ///< fixed qscale if non zero
    int encoding;     ///< true if we are encoding (vs decoding)
    int max_b_frames; ///< max number of B-frames for encoding
    int luma_elim_threshold;
    int chroma_elim_threshold;
    int strict_std_compliance; ///< strictly follow the std (MPEG-4, ...)
    int workaround_bugs;       ///< workaround bugs in encoders which cannot be detected automatically
    int codec_tag;             ///< internal codec_tag upper case converted from avctx codec_tag
    /* the following fields are managed internally by the encoder */

    /* sequence parameters */
    int context_initialized;
    int input_picture_number;  ///< used to set pic->display_picture_number, should not be used for/by anything else
    int coded_picture_number;  ///< used to set pic->coded_picture_number, should not be used for/by anything else
    int picture_number;       //FIXME remove, unclear definition
    int picture_in_gop_number; ///< 0-> first pic in gop, ...
    int mb_width, mb_height;   ///< number of MBs horizontally & vertically
    int mb_stride;             ///< mb_width+1 used for some arrays to allow simple addressing of left & top MBs without sig11
    int b8_stride;             ///< 2*mb_width+1 used for some 8x8 block arrays to allow simple addressing
    int h_edge_pos, v_edge_pos;///< horizontal / vertical position of the right/bottom edge (pixel replication)
    int mb_num;                ///< number of MBs of a picture
    ptrdiff_t linesize;        ///< line size, in bytes, may be different from width
    ptrdiff_t uvlinesize;      ///< line size, for chroma in bytes, may be different from width
    Picture *picture;          ///< main picture buffer
    Picture **input_picture;   ///< next pictures on display order for encoding
    Picture **reordered_input_picture; ///< pointer to the next pictures in coded order for encoding

    int64_t user_specified_pts; ///< last non-zero pts from AVFrame which was passed into avcodec_encode_video2()
    /**
     * pts difference between the first and second input frame, used for
     * calculating dts of the first frame when there's a delay */
    int64_t dts_delta;
    /**
     * reordered pts to be used as dts for the next output frame when there's
     * a delay */
    int64_t reordered_pts;

    /** bit output */
    PutBitContext pb;

    int start_mb_y;            ///< start mb_y of this thread (so current thread should process start_mb_y <= row < end_mb_y)
    int end_mb_y;              ///< end   mb_y of this thread (so current thread should process start_mb_y <= row < end_mb_y)
    struct MpegEncContext *thread_context[MAX_THREADS];
    int slice_context_count;   ///< number of used thread_contexts

    /**
     * copy of the previous picture structure.
     * note, linesize & data, might not match the previous picture (for field pictures)
     */
    Picture last_picture;

    /**
     * copy of the next picture structure.
     * note, linesize & data, might not match the next picture (for field pictures)
     */
    Picture next_picture;

    /**
     * copy of the source picture structure for encoding.
     * note, linesize & data, might not match the source picture (for field pictures)
     */
    Picture new_picture;

    /**
     * copy of the current picture structure.
     * note, linesize & data, might not match the current picture (for field pictures)
     */
    Picture current_picture;    ///< buffer to store the decompressed current picture

    Picture *last_picture_ptr;     ///< pointer to the previous picture.
    Picture *next_picture_ptr;     ///< pointer to the next picture (for bidir pred)
    Picture *current_picture_ptr;  ///< pointer to the current picture
    int last_dc[3];                ///< last DC values for MPEG-1
    int16_t *dc_val_base;
    int16_t *dc_val[3];            ///< used for MPEG-4 DC prediction, all 3 arrays must be continuous
    const uint8_t *y_dc_scale_table;     ///< qscale -> y_dc_scale table
    const uint8_t *c_dc_scale_table;     ///< qscale -> c_dc_scale table
    const uint8_t *chroma_qscale_table;  ///< qscale -> chroma_qscale (H.263)
    uint8_t *coded_block_base;
    uint8_t *coded_block;          ///< used for coded block pattern prediction (msmpeg4v3, wmv1)
    int16_t(*ac_val_base)[16];
    int16_t(*ac_val[3])[16];      ///< used for MPEG-4 AC prediction, all 3 arrays must be continuous
    int mb_skipped;                ///< MUST BE SET only during DECODING
    uint8_t *mbskip_table;        /**< used to avoid copy if macroblock skipped (for black regions for example)
                                   and used for B-frame encoding & decoding (contains skip table of next P-frame) */
    uint8_t *mbintra_table;       ///< used to avoid setting {ac, dc, cbp}-pred stuff to zero on inter MB decoding
    uint8_t *cbp_table;           ///< used to store cbp, ac_pred for partitioned decoding
    uint8_t *pred_dir_table;      ///< used to store pred_dir for partitioned decoding

    ScratchpadContext sc;

    int qscale;                 ///< QP
    int chroma_qscale;          ///< chroma QP
    unsigned int lambda;        ///< Lagrange multiplier used in rate distortion
    unsigned int lambda2;       ///< (lambda*lambda) >> FF_LAMBDA_SHIFT
    int *lambda_table;
    int adaptive_quant;         ///< use adaptive quantization
    int dquant;                 ///< qscale difference to prev qscale
    int closed_gop;             ///< MPEG1/2 GOP is closed
    int pict_type;              ///< AV_PICTURE_TYPE_I, AV_PICTURE_TYPE_P, AV_PICTURE_TYPE_B, ...
    int vbv_delay;
    int last_pict_type; //FIXME removes
    int last_non_b_pict_type;   ///< used for MPEG-4 gmc B-frames & ratecontrol
    int droppable;
    int frame_rate_index;
    AVRational mpeg2_frame_rate_ext;
    int last_lambda_for[5];     ///< last lambda for a specific pict type
    int skipdct;                ///< skip dct and code zero residual

    /* motion compensation */
    int unrestricted_mv;        ///< mv can point outside of the coded picture
    int h263_long_vectors;      ///< use horrible H.263v1 long vector mode

} MpegEncContext;




uint32_t FFmpegEncoderFrame::LifetimeInMilliSecs(const char *infoStr)
{
#if 0
    typedef std::chrono::high_resolution_clock Time;
    typedef std::chrono::milliseconds ms;
    typedef std::chrono::duration<float> fsec;
    auto t0 = Time::now();
    auto t1 = Time::now();
    fsec fs = t1 - t0;
    ms d = std::chrono::duration_cast<ms>(fs);
    std::cout << fs.count() << "s\n";
    std::cout << d.count() << "ms\n";
#endif
    uint32_t timeInMilliSec = 0;

    auto diff = chrono::system_clock::now() - startTime;
    auto ms = chrono::duration_cast<chrono::milliseconds>(diff);
    timeInMilliSec = (uint32_t) ms.count();

    return timeInMilliSec;
}
#endif

#if 0
// This code was put into place to play around with QScale on Motion JPEG images
// being created by an FFMpeg encoder to try to produce JPEG images of different
// qualities - It never impacted the quality and therefore I think this code is broken
// in FFMpeg library - but we don't produce these anyway but I left this code in place
// for reference purposes...
#define SET_MJPEG_QSCALE 1
#ifdef SET_MJPEG_QSCALE
else if (codec->id == AV_CODEC_ID_MJPEG)
{
    //codec_ctx->flags |= AV_CODEC_FLAG_QSCALE;
    // The qscale value should be between 1 and 31 where
    // 1 is the highest quality and 31 is lower quality but
    // I am not seeing any changes happen - it just seems as
    // though the flag setting itself produces higher quality
    // output.
    //codec_ctx->global_quality = (int)( FF_QP2LAMBDA * 31.0 );
    //if (qscale >= 0) {
    //  ost->enc_ctx->flags |= AV_CODEC_FLAG_QSCALE;
    //  ost->enc_ctx->global_quality = FF_QP2LAMBDA * qscale;
    //}
    codec_ctx->compression_level = 100; // Does this do anything?
}
#endif
#endif


#if 0
if (codec->id == AV_CODEC_ID_MJPEG)
{
    // OK - so when the flag is set above this "fixed_qscale" becomes set to a 1 otherwise
    // it is set to zero.
    MpegEncContext *s = (MpegEncContext *)codec_ctx->priv_data;
    int our_qscale = s->fixed_qscale;
    //s->qscale = 30;   // play with this value - no effect.
}
#endif

#if 0
//THIS DATA REALLY SCREWS UP THE OUTPUT!!!!!
// Which makes sense because it's mucking with the gop_size and directing max b_frames.
// It originally came from an encoding example file.
AVRational avr;
avr.num = 1;
avr.den = 25;
codec_ctx->time_base = avr;
avr.num = 25;
avr.den = 1;
codec_ctx->framerate = avr;
codec_ctx->gop_size = 10;
codec_ctx->max_b_frames = 1;
#endif
