/*!
 * @file ffmpeg_filter.h
 * @brief Interface for FFmpeg filters object.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <sys/stat.h>
#include <string>
#include <memory>

#include "ffmpeg_filter.h"
#include "ffmpeg_status.h"
#include "ffmpeg_container.h"
#include "ffmpeg_encoder.h"

#include "log_if.h"

// Make things easier for usage.
using namespace std;
using namespace FAS_VIDEO;
using namespace FI_FFMPEG;

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! If you want to turn on verbose logging for this object set this to 1
//static const bool VERBOSE_SAVE_LOG = true;
const bool VERBOSE_SAVE_LOG = false;


FFmpegFilters::~FFmpegFilters()
{
    if (filter_graph != nullptr)
        avfilter_graph_free(&filter_graph);
}

bool FFmpegFilters::InitializeFilters(AVRational& avrAspectRatio,
                                      AVRational& avrTimeBase,
                                      uint32_t frameWidth,
                                      uint32_t frameHeight,
                                      AVPixelFormat inFormat,
                                      AVPixelFormat outFormat)
{
    FFmpegStatus = 0;
    std::string myNameStr = "[FFmpegFilters::InitializeFilters]";

    // From transcode.c example code -
    // this is an empty pass through of the filter spec -
    // assuming we can set scales etc here.
    char filter_spec[] = { "null" };
    //char filter_spec[] = { "edgedetect[out]" }; // let's try a default setting first.
    //char filter_spec[] = { "edgedetect=low=0:high=0.1[out]" }; // let's try a default setting first.
    //char filter_spec[] = { "edgedetect=low=0:high=0.1:mode=canny[out]" }; // let's try a default setting first.
    //char filter_spec[512];

    // The filter_spec[] is how you manipulate input to output frame data.
    // This filter_spec[] is used in the call to avfilter_graph_parse_ptr

    // These are the old filter strings that were manipulated for simple filtering.
    // Right now there are no filters in place at all.
    //const char *filter_edge = "edgedetect[out]";
    //const char *filter_eq = "eq=contrast=1.0:brightness=-0.0:saturation=1.0 ";
    //const char *filter_edge = "edgedetect=low=0:high=0.1:mode=colormix "; canny and wire other modes

    //! args is used to setup the actual image pixel format modification which we do
    //! utilize in this filter object.
    char pixel_fmt_filter[512];
    const AVFilter *buffersrc = nullptr;
    const AVFilter *buffersink = nullptr;
    AVFilterInOut *outputs = nullptr;
    AVFilterInOut *inputs = nullptr;

    enum AVPixelFormat inputAVPixelFormat = inFormat;
    enum AVPixelFormat outputAVPixelFormat = outFormat;

    std::string inPixelFormatStr;
    std::string outPixelFormatStr;

    // Are these filters already set up for use - cannot call this a second time.
    if (filtersPrepared)
    {
        FFmpegStatus = AVERROR(EINVAL);
        log_ffmpeg_error_str(FFmpegStatus);
        fi_log(LOG_CRIT, "%s: FFmpeg filter object was previously initialized!",
               __PRETTY_FUNCTION__);
        goto exitPrepareFilters;
    }

    // Allocate using the ffmpeg library supporting functions.
    outputs = avfilter_inout_alloc();
    inputs = avfilter_inout_alloc();
    filter_graph = avfilter_graph_alloc();

    // Make sure the filter objects were allocated properly.
    if (!outputs || !inputs || !filter_graph)
    {
        FFmpegStatus = AVERROR(ENOMEM);
        log_ffmpeg_error_str(FFmpegStatus);
        fi_log(LOG_ERR, "%s: The allocations for filter in, out and graph failed!", myNameStr.c_str());
        goto exitPrepareFilters;
    }

    buffersrc = avfilter_get_by_name("buffer");
    buffersink = avfilter_get_by_name("buffersink");
    if (!buffersrc || !buffersink)
    {
        FFmpegStatus = AVERROR(ENOMEM);
        log_ffmpeg_error_str(FFmpegStatus);
        fi_log(LOG_ERR, "%s: Filtering source or sink element not found in avfilter_get_by_name().", myNameStr.c_str());
        goto exitPrepareFilters;
    }

    // This is a filter specification
    snprintf(pixel_fmt_filter, sizeof(pixel_fmt_filter),
             "video_size=%dx%d:pix_fmt=%d:time_base=%d/%d:pixel_aspect=%d/%d",
             frameWidth, frameHeight, inputAVPixelFormat,
             avrTimeBase.num, avrTimeBase.den,
             avrAspectRatio.num,
             avrAspectRatio.den);

    //Create the buffer source.
    FFmpegStatus = avfilter_graph_create_filter(&buffersrc_ctx, buffersrc, "in",
                                                pixel_fmt_filter, nullptr, filter_graph);
    if (FFmpegStatus < 0)
    {
        log_ffmpeg_error_str(FFmpegStatus);
        fi_log(LOG_ERR, "%s: Cannot create buffer source.", myNameStr.c_str());
        goto exitPrepareFilters;
    }
    else if (VERBOSE_SAVE_LOG)
    {
        fi_log(LOG_DEBUG, "%s avfilter_graph_create[%s].", myNameStr.c_str(), pixel_fmt_filter);
    }

    // Create the buffer sink.
    FFmpegStatus = avfilter_graph_create_filter(&buffersink_ctx, buffersink, "out",
                                                nullptr, nullptr, filter_graph);
    if (FFmpegStatus < 0)
    {
        log_ffmpeg_error_str(FFmpegStatus);
        fi_log(LOG_ERR, "%s: Cannot create buffer sink.", myNameStr.c_str());
        goto exitPrepareFilters;
    }

    // Install the encoder's pix_fmt.

    // Use the output AVPixelFormat for filter to encode frame properly.
    FFmpegStatus = av_opt_set_bin(buffersink_ctx, "pix_fmts",
                                  (uint8_t *)&outputAVPixelFormat, sizeof(outputAVPixelFormat),
                                  AV_OPT_SEARCH_CHILDREN);
    if (FFmpegStatus < 0)
    {
        log_ffmpeg_error_str(FFmpegStatus);
        fi_log(LOG_ERR, "%s: Cannot set output AVPixelFormat[%s].",
               myNameStr.c_str(), FFmpegContainer::GetStringForAVPixelFormat(outputAVPixelFormat));
        goto exitPrepareFilters;
    }

    // Keep these here to provide some samples of how the filter_spec can be modified to support
    // things like -filter_complex type of filtering.
    //snprintf(filter_spec, sizeof(filter_spec), "[in]split[a][b];[a]edgedetect=low=0:high=0.1:mode=canny[A1];[A1]colorkey=green[A2];[A2][b]overlay[out]");
    //snprintf(filter_spec, sizeof(filter_spec), "[in]split[a][b];[a]edgedetect=low=0:high=0.1,vflip[A1];[A1]colorkey=black[A2];[A2][b]blend=all_mode=addition:all_opacity=0.5[out]");
    //snprintf(filter_spec, sizeof(filter_spec), "[in]edgedetect=low=0:high=0.1[A1];[A1]vflip[out]");
    //snprintf(filter_spec, sizeof(filter_spec), "[in]split[a][b];[a]edgedetect=low=0:high=0.1,vflip[A1];[A1][b]blend=all_mode=overlay[out]");

    /* Endpoints for the filter graph. */
    outputs->name = av_strdup("in");
    outputs->filter_ctx = buffersrc_ctx;
    outputs->pad_idx = 0;
    outputs->next = nullptr;

    inputs->name = av_strdup("out");
    inputs->filter_ctx = buffersink_ctx;
    inputs->pad_idx = 0;
    inputs->next = nullptr;

    if (!outputs->name || !inputs->name)
    {
        FFmpegStatus = AVERROR(ENOMEM);
        log_ffmpeg_error_str(FFmpegStatus);
        fi_log(LOG_ERR, "%s: Cannot set the endpoints for filter graph.", myNameStr.c_str());
        goto exitPrepareFilters;
    }

    //if( avfilter_graph_parse2
    // Note:  It is here that the filter_spec (any additional filtering) would be specified
    // currently it is set to null.
    if ((FFmpegStatus = avfilter_graph_parse_ptr(filter_graph, filter_spec,
                                                 &inputs, &outputs, nullptr)) < 0)
    {
        log_ffmpeg_error_str(FFmpegStatus);
        fi_log(LOG_ERR, "%s: avfilter_graph_parse_ptr() FAILED!", myNameStr.c_str());
        goto exitPrepareFilters;
    }

    if ((FFmpegStatus = avfilter_graph_config(filter_graph, nullptr)) < 0)
    {
        log_ffmpeg_error_str(FFmpegStatus);
        fi_log(LOG_ERR, "%s: avfilter_graph_config() FAILED!", myNameStr.c_str());
        goto exitPrepareFilters;
    }

    // Free up temporary data for input and output points.
    if (inputs)
        avfilter_inout_free(&inputs);

    if (outputs)
        avfilter_inout_free(&outputs);

    // If make it here all is good and the filters are ready to go.
    filtersPrepared = true;
    inPixelFormatStr = FFmpegContainer::GetStringForAVPixelFormat(inFormat);
    outPixelFormatStr = FFmpegContainer::GetStringForAVPixelFormat(outFormat);
    fi_log(LOG_INFO, "%s: Successfully created FFMpeg Filters In[%s] Out[%s].",
           myNameStr.c_str(), inPixelFormatStr.c_str(), outPixelFormatStr.c_str());

exitPrepareFilters:
    return filtersPrepared;
}


bool FFmpegFilters::is_filter_prepared()
{
    return filtersPrepared;
}

AVFilterContext *FFmpegFilters::get_buffer_src_ctx()
{
    return buffersrc_ctx;
}

AVFilterContext *FFmpegFilters::get_buffer_sink_ctx()
{
    return buffersink_ctx;
}

AVFilterGraph *FFmpegFilters::get_filter_graph()
{
    return filter_graph;
}
