/*!
 * @file fi_dng.cpp
 * @brief Contains the implementation of the FiDNG class.
 *
 * DNG specific terms
 * - DNG: Digital Negative
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <string.h>
#include <stdint.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>

#include <algorithm>

#ifdef _WIN32
#include "fi_sensor.h"

#define FI_SENSOR_LINCE5M_COLOR         (FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR)
#define FI_SENSOR_LINCE5M_MONO          (FI_SENSOR::SENSOR_TYPE_LINCE5M_MONO)
#define FI_SENSOR_LUX13HS_COLOR         (FI_SENSOR::SENSOR_TYPE_LUX13HS_COLOR)
#define FI_SENSOR_LUX19HS_COLOR         (FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR)
#define FI_SENSOR_LUX160_COLOR          (FI_SENSOR::SENSOR_TYPE_LUX160_COLOR)
#define FI_SENSOR_LUX2100_COLOR         (FI_SENSOR::SENSOR_TYPE_LUX2100_COLOR)

#else
#include <fastec/fi_sensor_types.h>
#endif

#include "fi_dng.h"
#include "log_if.h"

namespace FI_DNG
{
//--------------------------------------------------------------------
// Constants, defines, Typedefs and Macros
//--------------------------------------------------------------------

// We need this here to assist in macro expansion.
using namespace FI_TIFF;

// Tags used in DNG spec.
#define DNG_CFA_REPEAT_PATTERN_DIM  33421
#define DNG_EP_CFA_PATTERN          33422

#define DNG_VERSION                 50706
#define DNG_UNIQUE_CAMERA_MODEL     50708
#define DNG_CFA_PLANE_COLOR         50710
#define DNG_CFA_LAYOUT              50711
#define DNG_BLACK_LEVEL             50714
#define DNG_DEFAULT_CROP_ORIGIN     50719
#define DNG_DEFAULT_CROP_SIZE       50720
#define DNG_COLOR_MATRIX1           50721
#define DNG_COLOR_MATRIX2           50722
#define DNG_AS_SHOT_NEUTRAL         50728
#define DNG_AS_SHOT_WHITE           50729
#define DNG_CALIBRATION_ILLUMINANT1 50778
#define DNG_CALIBRATION_ILLUMINANT2 50779
#define DNG_HUESATMAP_DIMS          50937
#define DNG_HUESATMAP_DATA1         50938
#define DNG_HUESATMAP_DATA2         50939

// Other DNG defines
const uint16_t DNG_CFA_PHOTOMETRIC_INTERPRETATION = 32803;

// Defines the count of SRATIONAL values contained in the color tranformation
// matrix which is a 3x3 matrix of SRATIONAL values stored in row scan order.
const uint32_t COLOR_MATRIX_VALUE_COUNT = 9;

// Defines the size of the colormatrix transformation for color DNG images.
const uint32_t COLOR_MATRIX_SIZE    = (COLOR_MATRIX_VALUE_COUNT *TIFF_SRATIONAL_SIZE);

// In a rational number, our denominator is 1.0.
//const uint32_t COLOR_MATRIX_DENOMINATOR = 10000;  // This is not currently used but left in place for posterity.

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

// The color parameters.
#include "lince_dng_constants.h"
#include "lux19hs_dng_constants.h"

//--------------------------------------------------------------------
// Function Definitions
//--------------------------------------------------------------------

TIFF_TAG_SIZE_FUNCTION(DNG_CFA_REPEAT_PATTERN_DIM)
{
    uint32_t ret_size = sizeof(uint32_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

TIFF_TAG_READ_FUNCTION(DNG_CFA_REPEAT_PATTERN_DIM)
{
    uint16_t val = 2;

    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    ifd_ptr->count = 2;
    memcpy(buffer, &val, sizeof(uint16_t));
    memcpy(buffer + sizeof(uint16_t), &val, sizeof(uint16_t));
}


TIFF_TAG_SIZE_FUNCTION(DNG_EP_CFA_PATTERN)
{
    uint32_t ret_size = sizeof(uint32_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

TIFF_TAG_READ_FUNCTION(DNG_EP_CFA_PATTERN)
{
    uint32_t cfa_pattern;

    (void)tiff;
    (void)fourcc;

    if (meta == nullptr)
        return;
    if (buffer == nullptr)
        return;

    // 0=Red, 1=Green, 2=Blue
    // Set in 32-bit word as low byte to high byte corresponding to (0,0) to (1,1) pixels.
    switch (meta->sensor_type)
    {
    case FI_SENSOR_LINCE5M_COLOR:
    case FI_SENSOR_LUX13HS_COLOR:
    case FI_SENSOR_LUX19HS_COLOR:
    case FI_SENSOR_LUX160_COLOR:
    case FI_SENSOR_LUX2100_COLOR:
        // GRBG
        cfa_pattern = (1 << 0) | (0 << 8) | (2 << 16) | (1 << 24);
        break;
    default:
        fi_log(LOG_DEBUG, "%s: Unsupported sensor type [0x%08x]", __PRETTY_FUNCTION__, meta->sensor_type);
        return;
    }

    // Doing it this way gives us the correct byte ordering
    ifd_ptr->count = 4;
    memcpy(buffer, &cfa_pattern, sizeof(uint32_t));
}


TIFF_TAG_SIZE_FUNCTION(DNG_VERSION)
{
    uint32_t ret_size = sizeof(uint32_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

TIFF_TAG_READ_FUNCTION(DNG_VERSION)
{
    uint32_t val = 0x00000101;

    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    ifd_ptr->count = sizeof(uint32_t);
    memcpy(buffer, &val, sizeof(uint32_t));
}

TIFF_TAG_SIZE_FUNCTION(DNG_UNIQUE_CAMERA_MODEL)
{
    uint32_t ret_size;

    (void)tiff;
    (void)fourcc;

    if (meta == nullptr)
        return 0;

    ret_size = static_cast<uint32_t>(strlen(meta->model_name) + 1);
    if (ret_size > GIGE_DSC_ACK_MODEL_B_LEN)
        ret_size = GIGE_DSC_ACK_MODEL_B_LEN;
    return TIFF_ALIGNED_SIZE(ret_size);
}

TIFF_TAG_READ_FUNCTION(DNG_UNIQUE_CAMERA_MODEL)
{
    uint32_t model_str_len;

    (void)tiff;
    (void)fourcc;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    model_str_len = static_cast<uint32_t>(strlen(meta->model_name) + 1);
    if (model_str_len > GIGE_DSC_ACK_MODEL_B_LEN)
        model_str_len = GIGE_DSC_ACK_MODEL_B_LEN;

    ifd_ptr->count = model_str_len;
    strncpy(buffer, meta->model_name, GIGE_DSC_ACK_MODEL_B_LEN);
}


TIFF_TAG_SIZE_FUNCTION(DNG_CFA_PLANE_COLOR)
{
    uint32_t ret_size = 3;

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

TIFF_TAG_READ_FUNCTION(DNG_CFA_PLANE_COLOR)
{
    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    buffer[0] = 0;
    buffer[1] = 1;
    buffer[2] = 2;

    ifd_ptr->count = 3;
}

TIFF_TAG_SIZE_FUNCTION(DNG_CFA_LAYOUT)
{
    uint32_t ret_size = sizeof(uint16_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

TIFF_TAG_READ_FUNCTION(DNG_CFA_LAYOUT)
{
    uint16_t val = 1;

    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint16_t));
}

TIFF_TAG_SIZE_FUNCTION(DNG_COLOR_MATRIX1)
{
    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(COLOR_MATRIX_SIZE);
}

TIFF_TAG_READ_FUNCTION(DNG_COLOR_MATRIX1)
{
    (void)tiff;
    (void)fourcc;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    switch (meta->sensor_type)
    {
    case FI_SENSOR_LINCE5M_COLOR:
        ifd_ptr->count = COLOR_MATRIX_VALUE_COUNT;
        memcpy(buffer, lince_dng_color_matrix1, COLOR_MATRIX_SIZE);
        break;
    case FI_SENSOR_LUX19HS_COLOR:
        ifd_ptr->count = COLOR_MATRIX_VALUE_COUNT;
        memcpy(buffer, lux19hs_dng_color_matrix1, COLOR_MATRIX_SIZE);
        break;

    case FI_SENSOR_LUX13HS_COLOR:
    case FI_SENSOR_LUX160_COLOR:
    case FI_SENSOR_LUX2100_COLOR:
        break;
    default:
        break;
    }
}

TIFF_TAG_SIZE_FUNCTION(DNG_COLOR_MATRIX2)
{
    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(COLOR_MATRIX_SIZE);
}

TIFF_TAG_READ_FUNCTION(DNG_COLOR_MATRIX2)
{
    (void)tiff;
    (void)fourcc;

    switch (meta->sensor_type)
    {
    case FI_SENSOR_LINCE5M_COLOR:
        ifd_ptr->count = COLOR_MATRIX_VALUE_COUNT;
        memcpy(buffer, lince_dng_color_matrix2, COLOR_MATRIX_SIZE);
        break;
    case FI_SENSOR_LUX19HS_COLOR:
        ifd_ptr->count = COLOR_MATRIX_VALUE_COUNT;
        memcpy(buffer, lux19hs_dng_color_matrix2, COLOR_MATRIX_SIZE);
        break;
    case FI_SENSOR_LUX13HS_COLOR:
    case FI_SENSOR_LUX160_COLOR:
    case FI_SENSOR_LUX2100_COLOR:
    default:
        break;
    }
}

TIFF_TAG_SIZE_FUNCTION(DNG_AS_SHOT_NEUTRAL)
{
    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(3 * TIFF_RATIONAL_SIZE);
}

TIFF_TAG_READ_FUNCTION(DNG_AS_SHOT_NEUTRAL)
{
    uint32_t den;
    int32_t num;

    (void)tiff;
    (void)fourcc;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    den = 1 << AS_SHOT_SHIFT;
    num = static_cast <int>(meta->as_shot_red);

    // We save the AsShot value as a "rational" number, which
    // means we save the numerator followed by the denominator.
    memcpy(buffer, &num, sizeof(uint32_t));
    buffer += sizeof(uint32_t);
    memcpy(buffer, &(den), sizeof(uint32_t));
    buffer += sizeof(uint32_t);

    num = static_cast <int>(meta->as_shot_grn);
    memcpy(buffer, &num, sizeof(uint32_t));
    buffer += sizeof(uint32_t);
    memcpy(buffer, &(den), sizeof(uint32_t));
    buffer += sizeof(uint32_t);

    num = static_cast <int>(meta->as_shot_blu);
    memcpy(buffer, &num, sizeof(uint32_t));
    buffer += sizeof(uint32_t);
    memcpy(buffer, &(den), sizeof(uint32_t));
    //buffer += sizeof(uint32_t);

    ifd_ptr->count = 3;
}

TIFF_TAG_SIZE_FUNCTION(DNG_CALIBRATION_ILLUMINANT1)
{
    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(sizeof(uint16_t));
}

TIFF_TAG_READ_FUNCTION(DNG_CALIBRATION_ILLUMINANT1)
{
    ifd_ptr->count = 1;

    (void)tiff;
    (void)fourcc;
    (void)buffer;

    if (ifd_ptr == nullptr)
        return;
    if (meta == nullptr)
        return;

    switch (meta->sensor_type)
    {
    case FI_SENSOR_LINCE5M_COLOR:
        ifd_ptr->val_offset = LINCE_CALIBRATION_ILLUMINANT1;
        break;
    case FI_SENSOR_LUX19HS_COLOR:
        ifd_ptr->val_offset = LUX19HS_CALIBRATION_ILLUMINANT1;
        break;
    case FI_SENSOR_LUX13HS_COLOR:
    case FI_SENSOR_LUX160_COLOR:
    case FI_SENSOR_LUX2100_COLOR:
    default:
        break;
    }
}


TIFF_TAG_SIZE_FUNCTION(DNG_CALIBRATION_ILLUMINANT2)
{
    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(sizeof(uint16_t));
}

TIFF_TAG_READ_FUNCTION(DNG_CALIBRATION_ILLUMINANT2)
{
    (void)tiff;
    (void)fourcc;
    (void)buffer;

    if (ifd_ptr == nullptr)
        return;
    if (meta == nullptr)
        return;

    switch (meta->sensor_type)
    {
    case FI_SENSOR_LINCE5M_COLOR:
        ifd_ptr->count = 1;
        ifd_ptr->val_offset = LINCE_CALIBRATION_ILLUMINANT2;
        break;
    case FI_SENSOR_LUX19HS_COLOR:
        ifd_ptr->count = 1;
        ifd_ptr->val_offset = LUX19HS_CALIBRATION_ILLUMINANT2;
        break;
    case FI_SENSOR_LUX13HS_COLOR:
    case FI_SENSOR_LUX160_COLOR:
    case FI_SENSOR_LUX2100_COLOR:
    default:
        break;
    }
}

} // End of FI_DNG

// Make things easier for ourselves.
using namespace FI_DNG;
using namespace FAS_VIDEO;

/*!
 * @brief Standard class constructor.
 */
FiDNG::FiDNG() :
    FiTIFF()
{
    FiDNG::init_tag_db();
}

uint16_t FiDNG::get_photometric_interpretation(uint32_t fourcc)
{
    (void)fourcc;
    return DNG_CFA_PHOTOMETRIC_INTERPRETATION;
}

/*!
 * @brief Defines all the tags we will add to a TIFF file.
 * @note The TIFF spec requires that the tags be in the IFD in numerical order.
 */
void FiDNG::init_tag_db()
{
    // Assumption is that the TIFF logic has started the database.
    tag_db.push_back(TAG_DB_ENTRY(DNG_CFA_REPEAT_PATTERN_DIM, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(DNG_EP_CFA_PATTERN, TIFF_TYPE_BYTE));
    tag_db.push_back(TAG_DB_ENTRY(DNG_VERSION, TIFF_TYPE_BYTE));
    tag_db.push_back(TAG_DB_ENTRY(DNG_UNIQUE_CAMERA_MODEL, TIFF_TYPE_ASCII));
    tag_db.push_back(TAG_DB_ENTRY(DNG_CFA_PLANE_COLOR, TIFF_TYPE_BYTE));
    tag_db.push_back(TAG_DB_ENTRY(DNG_CFA_LAYOUT, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(DNG_COLOR_MATRIX1, TIFF_TYPE_SRATIONAL));
    tag_db.push_back(TAG_DB_ENTRY(DNG_COLOR_MATRIX2, TIFF_TYPE_SRATIONAL));
    tag_db.push_back(TAG_DB_ENTRY(DNG_AS_SHOT_NEUTRAL, TIFF_TYPE_RATIONAL));
    tag_db.push_back(TAG_DB_ENTRY(DNG_CALIBRATION_ILLUMINANT1, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(DNG_CALIBRATION_ILLUMINANT2, TIFF_TYPE_SHORT));

    // Make sure the database is in ascending order, per TIFF spec.
    std::sort(tag_db.begin(), tag_db.end());
}

/*!
 * @brief Query the image format supported by this object.
 *
 * @returns The IMAGE_FORMAT value.
 */
FAS_VIDEO::IMAGE_FORMAT FiDNG::GetImageFormat()
{
    //! We are DNGs.
    return FAS_VIDEO::IMAGE_FORMAT::FI_DNG_IMAGE;
}

/*!
 * @brief Test the data in a DNG to verify it and pull out dimensions.
 *
 * @param imageDataToVerify     The image data to verify DNG format.
 * @param imageDataSize         How many uint8_t bytes are in this DNG image.
 *
 * @returns true if DNG data was verified and attach dimension data interpreted.
 */
bool FiDNG::VerifyImageData(const uint8_t *imageDataToVerify, uint32_t imageDataSize)
{
    (void)imageDataToVerify;
    (void)imageDataSize;

    //! There is no implementation for DNG images at this time.
    return true;
}


