/*!
 * @file fi_durations.cpp
 * @brief Statistical support object for tracking duration intervals.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <string>
#include <memory>
#include <chrono>
#include <fmt/printf.h>

#include "fi_durations.h"
#include "log_if.h"

// Make things easier for usage.
using namespace std;
using namespace FAS_VIDEO;

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Static Instances managed in this module.
//--------------------------------------------------------------------

//! Use this to wrap log information.
static std::string fiDurationLogSep = "============================================================"s;

//--------------------------------------------------------------------
// Class implementation of FiDurations.
//--------------------------------------------------------------------

FiDurations::FiDurations()
{
    constructTime = startTime = std::chrono::system_clock::now();
    numberOfIntervals       = 0;
    averageIntervalInMs     = 0;

    smallestIntervalInMS    = 0;
    biggestIntervalInMS     = 0;
    totalIntervalTimeInMS   = 0;
}

FiDurations::~FiDurations() {}

//! Sets a start of interval to mark.
void FiDurations::MarkStart()
{
    startTime = std::chrono::system_clock::now();
}

//! Sets a duration interval in this object.
void FiDurations::MarkInterval()
{
    uint32_t timeInMilliSec = 0;

    numberOfIntervals++;

    auto diff = chrono::system_clock::now() - startTime;
    auto ms = chrono::duration_cast<chrono::milliseconds>(diff);

    // How long in milliseconds was the interval?
    timeInMilliSec = (uint32_t) ms.count();

    // Is this interval the biggest one found so far?
    if (timeInMilliSec > biggestIntervalInMS)
    {
        biggestIntervalInMS = timeInMilliSec;   // Set new high mark for longest interval time.
    }

    if (smallestIntervalInMS == 0)
    {
        smallestIntervalInMS = timeInMilliSec;
    }
    else if (timeInMilliSec < smallestIntervalInMS)
    {
        smallestIntervalInMS = timeInMilliSec;
    }

    // Accumulate a total of all interval times.
    totalIntervalTimeInMS += timeInMilliSec;

    // Calculate a simple average interval time.
    averageIntervalInMs = totalIntervalTimeInMS / numberOfIntervals;
}

std::string FiDurations::DurationToTimeStr(uint32_t msTimeInMS)
{
    std::string timeStr;

    std::chrono::milliseconds ms(msTimeInMS);

    auto hh     = chrono::duration_cast<chrono::hours>(ms);
    auto mm     = chrono::duration_cast<chrono::minutes>(ms % chrono::hours(1));
    auto ss     = chrono::duration_cast<chrono::seconds>(ms % chrono::minutes(1));
    auto mms    = chrono::duration_cast<chrono::milliseconds>(ms % chrono::seconds(1));

    timeStr = fmt::sprintf("%02d:%02d:%02d.%03d", (int)hh.count(), (int)mm.count(), (int)ss.count(), (int)mms.count());
    return timeStr;
}

//! Logs a duration separator mark.
void FiDurations::LogDurationSeparator()
{
    fi_log(LOG_DEBUG, "%s", fiDurationLogSep.c_str());
}

//! Logs duration information.
//!
//! @param logInfoStr - if not null the user wants this to log with this string and time at end.
void FiDurations::LogDurationInfo(const char *logInfoStr)
{
    std::string totalStr = DurationToTimeStr(totalIntervalTimeInMS);
    if (numberOfIntervals > 1)
    {
        if (logInfoStr != nullptr)
        {
            fi_log(LOG_DEBUG, "[TIME %s]%s least[%dms] most[%dms] avr[%dms]",
                   totalStr.c_str(), logInfoStr, smallestIntervalInMS, biggestIntervalInMS, averageIntervalInMs);
        }
        else
        {
            fi_log(LOG_DEBUG, "[TIME %s] least[%dms] most[%dms] avr[%dms]",
                   totalStr.c_str(), smallestIntervalInMS, biggestIntervalInMS, averageIntervalInMs);
        }
    }
    else
    {
        if (logInfoStr != nullptr)
        {
            fi_log(LOG_DEBUG, "[TIME %s]%s", totalStr.c_str(), logInfoStr);
        }
        else
        {
            fi_log(LOG_DEBUG, "[TIME %s]", totalStr.c_str());
        }
    }
}

//! Return the total time of all intervals.
uint32_t FiDurations::GetTotalIntervalTimeInMS()
{
    return totalIntervalTimeInMS;
}
