/*!
 * @file fi_image.h
 * @brief The Fastec Image Class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <memory>

#include "fi_image.h"
#include "fi_image_pkg.h"
#include "fi_bmp.h"
#include "fi_bmp_planar.h"
#include "fi_dng.h"
#include "fi_tiff.h"
#include "fi_jpeg.h"
#include "fas_video.h"
#include "log_if.h"

#ifdef INCLUDE_RDI_PACKED
#include "fi_rdi_packed.h"
#endif /* INCLUDE_RDI_PACKED */

namespace FAS_VIDEO
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

} // End of FAS_VIDEO

using namespace FAS_VIDEO;

//--------------------------------------------------------------------
// FiImage Base Class Implementation
//--------------------------------------------------------------------

/*!
 * @brief FiImage Construction.
 */
FiImage::FiImage() : status(IMAGE_STATUS::IMAGE_NEW),
    headerFourCC(FOURCC_NONE),
    hasKnownImageInfo(false),
    knownImageWidth(0),
    knownImageHeight(0),
    knownBitsPerPixel(8),
    knownRowAlign(1)
{
    // Set the entire headerMetaData to an empty state.
    memset(&headerMetaData, 0, sizeof(pm_metadata));
}

/*!
 * @brief FiImage Destruction.
 */
FiImage::~FiImage()
{
    hasKnownImageInfo = false;
}

/*!
 * @brief Factory for FiImage object
 *
 * @param imageType     One of the FAS_VIDEO::IMAGE_FORMAT values.
 *
 * @returns A std::unique_ptr<FiImage> containing the FiImage object or nullptr if not created.
 */
std::unique_ptr<FAS_VIDEO::FiImage> FiImage::CreateImageObject(FAS_VIDEO::IMAGE_FORMAT imageType)
{
    std::unique_ptr<FAS_VIDEO::FiImage> fiImagePtr;

    // Generate the frame object based on the image format value.
    switch (imageType)
    {
    case FAS_VIDEO::FI_JPEG_IMAGE:
        fiImagePtr = std::unique_ptr<FAS_VIDEO::FiImage>(new FI_JPEG::FiJPEG());
        break;
    case FAS_VIDEO::FI_BMP_IMAGE:
        fiImagePtr = std::unique_ptr<FAS_VIDEO::FiImage>(new FI_BMP::FiBMP());
        break;
    case FAS_VIDEO::FI_BMP_PLANAR_IMAGE:
        fiImagePtr = std::unique_ptr<FAS_VIDEO::FiImage>(new FI_BMP::FiBMPPlanar());
        break;
    case FAS_VIDEO::FI_DNG_IMAGE:
        fiImagePtr = std::unique_ptr<FAS_VIDEO::FiImage>(new FI_DNG::FiDNG());
        break;
    case FAS_VIDEO::FI_TIF_IMAGE:
        fiImagePtr = std::unique_ptr<FAS_VIDEO::FiImage>(new FI_TIFF::FiTIFF());
        break;
#ifdef INCLUDE_RDI_PACKED
    case FAS_VIDEO::FI_RDI_PACKED_IMAGE:
        fiImagePtr = std::unique_ptr<FAS_VIDEO::FiImage>(new FI_RDI_PACKED::FiRDIPacked());
        break;
#endif /* INCLUDE_RDI_PACKED */
    case FAS_VIDEO::FI_INVALID_IMAGE:
    default:
        break;
    }

    return fiImagePtr;
}



/*!
 * @brief Queries the attached state of the image data package.
 *
 * @returns A boolean true if has attached image package, false if not.
 */
bool FiImage::IsImageAttached()
{
    if (status == FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED)
        return true;
    return false;
}

/*!
 * @brief Query the image format supported by this object.
 *
 * @returns The IMAGE_STATUS value.
 */
FAS_VIDEO::IMAGE_STATUS FiImage::GetImageStatus()
{
    return status;
}

/*!
 * @brief Returns the width of the image.
 *
 * @returns 0 or a known image width from previously attached image.
 */
uint32_t FiImage::GetImageWidth()
{
    uint32_t imageWidth = 0;
    if (hasKnownImageInfo)
    {
        imageWidth = knownImageWidth;
    }
    return imageWidth;
}

/*!
 * @brief Returns the height of the image.
 *
 * @returns 0 or a known image height from previously attached image.
 */
uint32_t FiImage::GetImageHeight()
{
    uint32_t imageHeight = 0;
    if (hasKnownImageInfo)
    {
        imageHeight = knownImageHeight;
    }
    return imageHeight;
}

/*!
 * @brief Returns the bit depth of the image.
 *
 * @returns 0 or a known bits depth from previously attached image.
 */
uint32_t FiImage::GetBitsPerPixel()
{
    uint32_t bitsPerPixel = 0;
    if (hasKnownImageInfo)
    {
        bitsPerPixel = knownBitsPerPixel;
    }
    return bitsPerPixel;
}

/*!
 * @brief Returns the row align of the image.
 *
 * @returns 0 or a known row align from previously attached image.
 */
uint32_t FiImage::GetRowAlign()
{
    uint32_t rowAlign = 0;
    if (hasKnownImageInfo)
    {
        rowAlign = knownRowAlign;
    }
    return rowAlign;
}

/*!
 * @brief Returns total number of uint8_t bytes in a frame of data.
 *
 * @returns 0 or a calculated frame size based on known data from
 * previously attached image.
 */
uint32_t FiImage::GetFullFramesize()
{
    uint32_t frameSize = 0;

    if (hasKnownImageInfo)
    {
        uint32_t rowSize = 0;
        uint32_t frameBytesPerPixel = knownBitsPerPixel / 8;
        rowSize = knownImageWidth * frameBytesPerPixel;

        // If rowAlign is set make sure rowSize is byte aligned on specified boundary.
        if (knownRowAlign > 1)
        {
            uint32_t rowAlignMinusOne = knownRowAlign - 1;
            if ((rowSize & rowAlignMinusOne) != 0)
                rowSize = ((rowSize + rowAlignMinusOne) / knownRowAlign) * knownRowAlign;
        }

        frameSize = (uint32_t)(rowSize * knownImageHeight);
    }

    return frameSize;
}

/*!
 * @brief Returns FOURCC value stored for image.
 *
 * @returns a FOURCC code and it could be FOURCC_NONE.
 */
uint32_t FiImage::GetFourCC()
{
    return headerFourCC;
}

/*!
 * @brief Attaches the image data package representing an image to this object.
 *
 * @param imageDataToAttach  The image data package to attach.
 *
 * @returns FAS_VIDEO::IMAGE_STATUS of object.
 */
FAS_VIDEO::IMAGE_STATUS FiImage::AttachImagePkg(std::unique_ptr<FAS_VIDEO::FiImagePkg> imagePkgToAttach)
{
    uint32_t imagePkgSize = 0;
    const uint8_t *imagePkgDataPtr = nullptr;

    // If the status of the FiImage was attached prior reset to empty state.
    // Otherwise leave the status as it is.
    if (status == FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED)
    {
        // Putting the status to empty in this condition tells us that
        // we had a valid image attached at one point.
        status = FAS_VIDEO::IMAGE_STATUS::IMAGE_EMPTY;
        imagePkg.reset();
    }

    // Is the imageDataToAttach a valid unique pointer containing the image data package?
    if (!imagePkgToAttach)
    {
        // Was passed a null unique pointer.
        fi_log(LOG_DEBUG, "%s: imagePkgToAttach was a nullptr.", __PRETTY_FUNCTION__);
        goto exitAttachImagePkg;
    }

    // Make sure there is data in the package.
    imagePkgSize = imagePkgToAttach->GetImageSize();
    imagePkgDataPtr = imagePkgToAttach->GetMemDataPtr();

    if ((imagePkgSize == 0) || (imagePkgDataPtr == nullptr))
    {
        status = FAS_VIDEO::IMAGE_STATUS::BAD_IMAGE_FORMAT;
        goto exitAttachImagePkg;
    }

    // Verify the image package information
    if (!VerifyImageData(imagePkgDataPtr, imagePkgSize))
    {
        status = IMAGE_STATUS::BAD_IMAGE_FORMAT;
        goto exitAttachImagePkg;
    }

    // If make it here the image data is verified and we can attach it
    // officially to this object.
    imagePkg = std::move(imagePkgToAttach);

    // If we have never stored "known" image information query
    // the class hierarchy and save it in case we receive any
    // future placeholder type images.
    if (!hasKnownImageInfo)
    {
        uint32_t theWidth = GetImageWidth();
        uint32_t theHeight = GetImageHeight();
        uint32_t theRowAlign = GetRowAlign();
        uint32_t theBitsPerPixel = GetBitsPerPixel();

        // Do the values make sense?
        if ((theWidth > 0) &&
            (theHeight > 0) &&
            (theRowAlign > 0) &&
            (theBitsPerPixel > 0))
        {
            knownImageWidth = theWidth;
            knownImageHeight = theHeight;
            knownRowAlign = theRowAlign;
            knownBitsPerPixel = theBitsPerPixel;
            hasKnownImageInfo = true;
        }
    }

    // If make it here no error has occurred and image data is now attached
    // and ready to use.
    status = FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED;

exitAttachImagePkg:

    //return the status of the image object.
    return status;
}

/*!
 * @brief Retrieves a read-only frame data pointer to attached image data.
 * data minus any header if one exists.
 *
 * @param attachedFrameSize     Set to length of frame data or zero if none exist.
 * @param frameOrientation      How is the frame supposed to be oriented.
 *
 * @returns read-only pointer to attached frame data or nullptr if none exist.
 */
const uint8_t *FiImage::GetAttachedFrameData(uint32_t& attachedFrameSize,
                                             FAS_VIDEO::FRAME_ORIENTATION frameOrientation)
{
    attachedFrameSize = 0;
    const uint8_t *frameDataPtr = nullptr;

    FAS_VIDEO::FRAME_PLANAR_DATA planarDataAttached;

    // For those calling this function that don't care about Planar data on the attached
    // frame this will pass the local structure to the hierarchy.
    frameDataPtr = GetAttachedFrameData(attachedFrameSize, planarDataAttached, frameOrientation);

    return frameDataPtr;
}

/*!
 * @brief Retrieves a read-only frame data pointer to attached image data.
 * data minus any header if one exists.
 *
 * @param attachedFrameSize     (Reference)Set to length of frame data or zero if none exist.
 * @param planarDataAttached    (Reference)Set to the planar data of attached frame.
 * @param frameOrientation      How is the frame oriented.
 *
 * @returns read-only pointer to attached frame data or nullptr if none exist in addition
 * the references (attachedFrameSize,planarDataAttached) are output settings as well.
 */
const uint8_t *FiImage::GetAttachedFrameData(uint32_t& attachedFrameSize,
                                             FAS_VIDEO::FRAME_PLANAR_DATA& planarDataAttached,
                                             FAS_VIDEO::FRAME_ORIENTATION frameOrientation)
{
    attachedFrameSize = 0;
    const uint8_t *frameDataPtr = nullptr;

    // The frame orientation is not used in the base-class.
    (void) frameOrientation;

    // Initialize the passed in planar record to empty state.
    planarDataAttached = {0, {0}, {nullptr}};

    // A NOTE:
    // If make it here to the base-class implementation then the image package
    // contains all the data necessary for writing to the stream.  If there is a
    // derivation in the class hierarchy (as is currently the case with FiBMP) the
    // data being retrieved in that case will truly be a special headerless frame
    // containing what we call "normalized" bitmap frame data.  In the case of JPEG
    // images the base class will simply return the image data in the package because
    // JPEG data is taken as a whole without any manipulation of the frame data
    // because it is written to the stream in its compressed form for motion jpegs.
    if (imagePkg)
    {
        frameDataPtr = imagePkg->GetMemDataPtr();
        attachedFrameSize = imagePkg->GetImageSize();
    }

    // Fill in the planar data for the caller based on the attached frame data ptr.
    GetAttachedFramePlanarData(planarDataAttached);

    return frameDataPtr;
}

/*!
 * @brief Retrieves a copy of the frame data associated with any attached image
 * data minus any header if one exists.
 *
 * @returns unique pointer to an FiFrame object if one can be created.
 */
std::unique_ptr<FAS_VIDEO::FiFrame> FiImage::CopyAttachedFrameData()
{
    // To Be Implemented.
    std::unique_ptr<FAS_VIDEO::FiFrame> frameData;
    return frameData;
}

/*!
 * @brief If image data attached this returns its size.
 *
 * @returns A zero if no image attached otherwise it returns size of attached image.
 */
uint32_t FiImage::GetAttachedImageSize()
{
    uint32_t retSize = 0;

    // Is there an attached image data package?
    if (imagePkg)
    {
        retSize = imagePkg->GetImageSize();
    }

    return retSize;
}

/*!
 * @brief If image data attached this a read-only pointer to the image data buffer.
 *
 * @returns A nullptr if nothing attached otherwise a read-only image data pointer.
 */
const uint8_t *FiImage::GetAttachedImageData()
{
    const uint8_t *retImagePtr = nullptr;

    // Is there an attached image data package?
    if (imagePkg)
    {
        retImagePtr = imagePkg->GetMemDataPtr();
    }

    return retImagePtr;
}

/*!
 * @brief Retrieves planar information for frame data of attached image.
 *
 * The data returned provides read-only access to plane data if it exists.
 *
 * @param planarDataToSet  The planar information requested.
 *
 * @returns true if planar data is populated with at least one plane or false if not.
 */
bool FiImage::GetAttachedFramePlanarData(FRAME_PLANAR_DATA& planarDataToSet)
{
    // If fall to here in the base class there is no planar data.
    bool onePlaneExists = false;

    planarDataToSet = {0, {0}, {nullptr}};

    return onePlaneExists;
}

/*!
 * @brief Does this image object contain "known" image information?
 *
 * @returns A boolean true if it "knows" image information, false if not.
 */
bool FiImage::HasKnownImageInfo()
{
    // Return the state of the "known" data.
    return hasKnownImageInfo;
}
