/*!
 * @file fi_stream.cpp
 * @brief The stream object of the Fastec Video library.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <mutex>
#include <fmt/printf.h>

#include "fas_video.h"
#include "fi_stream.h"
#include "log_if.h"

namespace FAS_VIDEO
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------
//! If you want to turn on verbose logging for this object set this to 1
//static const bool VERBOSE_SAVE_LOG = true;
static const bool VERBOSE_SAVE_LOG = false;

static std::string incompleteVideoFileNameString = ".incomplete.txt";

} // End of FAS_VIDEO

using namespace FAS_VIDEO;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

/*!
 * @brief Video stream construction.
 */
FiVideoStream::FiVideoStream() : streamType(FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_NONE),
    streamStatus(FAS_VIDEO::STREAM_STATUS::STREAM_BUILT),
    frameRate(FI_DEFAULT_VIDEO_FRAME_RATE),
    frameCount(0),
    lastFrameAdded(0),
    frameWidth(0),
    frameHeight(0),
    fullFrameSize(0),
    firstAddedImageFormat(FAS_VIDEO::IMAGE_FORMAT::FI_INVALID_IMAGE),
    processingVideoFrame(false)
{
    if (VERBOSE_SAVE_LOG)
    {
        fi_log(LOG_DEBUG, "%s: Constructing FiVideoStream", __PRETTY_FUNCTION__);
    }
}

/*!
 * @brief Video stream destruction.
 */
FiVideoStream::~FiVideoStream()
{
    if (VERBOSE_SAVE_LOG)
    {
        fi_log(LOG_INFO, "%s: Destructing FiVideoStream", __PRETTY_FUNCTION__);
    }

    // Give stream hierarchy chance to close down if it set as closed.
    if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_CLOSED)
    {
        fi_log(LOG_DEBUG, "%s: The streamStatus was not set to FAS_VIDEO::STREAM_STATUS::STREAM_CLOSED so video is not complete.", __PRETTY_FUNCTION__);
    }
}

/*!
 * @brief Translates an FAS_VIDEO::IMAGE_FORMAT into readable string.
 *
 * @param FiImageFmtId   The FAS_VIDEO::IMAGE_FORMAT code.
 *
 * @returns The result of the translation.
 */
const char *FiVideoStream::GetStringForFiImageID(FAS_VIDEO::IMAGE_FORMAT FiImageFmtId)
{
    static std::string fiImageFmtIDStr = fmt::sprintf("FiImageType[%d] Unrecognized.", (int)FiImageFmtId);

    // These are the "known" encoder_ids that are supported in this container.
    switch (FiImageFmtId)
    {
    case FAS_VIDEO::IMAGE_FORMAT::FI_JPEG_IMAGE:
        fiImageFmtIDStr = "FI_JPEG_IMAGE";
        break;
    case FAS_VIDEO::IMAGE_FORMAT::FI_BMP_IMAGE:
        fiImageFmtIDStr = "FI_BMP_IMAGE";
        break;
    case FAS_VIDEO::IMAGE_FORMAT::FI_DNG_IMAGE:
        fiImageFmtIDStr = "FI_DNG_IMAGE";
        break;
    case FAS_VIDEO::IMAGE_FORMAT::FI_TIF_IMAGE:
        fiImageFmtIDStr = "FI_TIF_IMAGE";
        break;
    case FAS_VIDEO::IMAGE_FORMAT::FI_BMP_PLANAR_IMAGE:
        fiImageFmtIDStr = "FI_BMP_PLANAR_IMAGE";
        break;
    default:
        break;
    }

    return fiImageFmtIDStr.c_str();
}

/*!
 * @brief Returns the string we append to video filenames
 * when the video is incomplete.
 *
 * @returns A constant reference to the string.
 */
const std::string& FiVideoStream::GetIncompleteVideoFileString()
{
    return incompleteVideoFileNameString;
}

/*!
 * @brief Closes the video stream and releases any allocated resources.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoStream::CloseStream()
{
    // Just set the stream status of this object as being closed.
    streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_CLOSED;
    return streamStatus;
}

/*!
 * @brief Queries the ready status of the stream.
 *
 * @returns A boolean true if ready or false if not.
 */
bool FiVideoStream::IsStreamReady()
{
    if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_READY)
        return true;
    return false;
}

/*!
 * @brief Queries the closed status of the stream.
 *
 * @returns A boolean true if closed or false if not.
 */
bool FiVideoStream::IsStreamClosed()
{
    if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_CLOSED)
        return true;
    return false;
}

/*!
 * @brief Log the stream status for this object into string reference
 * and to the fi_log if requested.
 *
 * @param logMessageStr - reference to hold the log message.
 * @param sendToFiLog - True if user wants this output to our log file.
 *
 */
void FiVideoStream::LogStreamStatus(std::string& logMessageStr, bool sendToFiLog)
{
    // Create a string with readable stream operational state.
    std::string streamOperation = "streamType::";

    // Create a string with readable status information.
    std::string statusInfo = "streamStatus::";

    switch (streamType)
    {
    case FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_READER:
        streamOperation += "READER";
        break;
    case FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER:
        streamOperation += "WRITER";
        break;
    case FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_NONE:
    default:
        streamOperation += "NO-OP";
        break;
    }

    switch (streamStatus)
    {
    case FAS_VIDEO::STREAM_STATUS::STREAM_BUILT:
        statusInfo += "STREAM_BUILT";
        break;
    case FAS_VIDEO::STREAM_STATUS::STREAM_READY:
        statusInfo += "STREAM_READY";
        break;
    case FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED:
        statusInfo += "STREAM_COMPLETED";
        break;
    case FAS_VIDEO::STREAM_STATUS::STREAM_CLOSED:
        statusInfo += "STREAM_CLOSED";
        break;
    case FAS_VIDEO::STREAM_STATUS::STREAM_INCOMPLETE:
        statusInfo += "STREAM_INCOMPLETE";
        break;
    case FAS_VIDEO::STREAM_STATUS::STREAM_TYPE_ERROR:
        statusInfo += "STREAM_TYPE_ERROR";
        break;
    case FAS_VIDEO::STREAM_STATUS::PLACEHOLDER_WAIT:
        statusInfo += "PLACEHOLDER_WAIT";
        break;
    case FAS_VIDEO::STREAM_STATUS::INVALID_CODEC:
        statusInfo += "CODEC TYPE UNKNOWN";
        break;
    case FAS_VIDEO::STREAM_STATUS::BAD_RANGE:
        statusInfo += "INVALID FRAME RANGE";
        break;
    case FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM:
        statusInfo += "MISSING IO STREAM OBJECT";
        break;
    case FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE:
        statusInfo += "MISSING STREAM IMAGE";
        break;
    case FAS_VIDEO::STREAM_STATUS::NO_FILE_FOUND:
        statusInfo += "FILE NOT FOUND";
        break;
    case FAS_VIDEO::STREAM_STATUS::REQUEST_DENIED:
        statusInfo += "INVALID REQUEST";
        break;
    case FAS_VIDEO::STREAM_STATUS::STREAM_REENTRY_ERROR:
        statusInfo += "STREAM RE-ENTRY ERROR";
        break;
    case FAS_VIDEO::STREAM_STATUS::ALLOCATION_FAILED:
        statusInfo += "ALLOCATION FAILED";
        break;
    case FAS_VIDEO::STREAM_STATUS::STREAM_READ_ERROR:
        statusInfo += "STREAM READ REQUEST FAILED";
        break;
    case FAS_VIDEO::STREAM_STATUS::STREAM_EOF:
        statusInfo += "STREAM EOF CONDITION";
        break;
    case FAS_VIDEO::STREAM_STATUS::CODEC_SETUP_ERROR:
        statusInfo += "CODEC_SETUP_ERROR";
        break;
    case FAS_VIDEO::STREAM_STATUS::CODEC_ERROR:
        statusInfo += "CODEC ERROR";
        break;
    case FAS_VIDEO::STREAM_STATUS::EXCEPTION_CAUGHT:
        statusInfo += "EXCEPTION CAUGHT";
        break;
    case FAS_VIDEO::STREAM_STATUS::FFMPEG_ERR:
        statusInfo += "FFMPEG ERROR";
        break;
    case FAS_VIDEO::STREAM_STATUS::STREAM_EMPTY:
    default:
        statusInfo += "NONE";
        break;
    }

    // Build up the log message based on the stream status.
    logMessageStr = "[FI_STREAMER][" + streamOperation + "][" + statusInfo + "]";

    // Is there any additional information available?
    if (additionalStatusInfo.size() > 0)
    {
        logMessageStr += additionalStatusInfo;
    }

    // Sending to the log output too?
    if (sendToFiLog)
        fi_log(LOG_DEBUG, "%s: %s.", __PRETTY_FUNCTION__, logMessageStr.c_str());
}

/*!
 * @brief Log the stream status for this object.
 *
 */
void FiVideoStream::LogStreamStatus()
{
    // Create a message string.
    std::string myLogMessageStr;

    LogStreamStatus(myLogMessageStr, true);
}

/*!
 * @brief Returns the frame count for video stream.
 *
 */
uint32_t FiVideoStream::GetFrameCount()
{
    return frameCount;
}

/*!
 * @brief Returns the frame rate for video stream.
 *
 */
uint32_t FiVideoStream::GetFrameRate()
{
    return frameRate;
}

/*!
 * @brief Return the video name of stream.
 *
 */
const std::string FiVideoStream::GetVideoName()
{
    return videoNameOfStream;
}

/*!
 * @brief Return the last frame number added count.
 *
 */
uint32_t FiVideoStream::GetLastFrameAdded()
{
    return lastFrameAdded;
}

/*!
 * @brief Return the last frame number added count as encoded count as base class default.
 *
 */
uint32_t FiVideoStream::GetFramesEncodedCount()
{
    return lastFrameAdded;
}

/*!
 * @brief Common stream logic to attach the current image frame to stream object.
 *
 * Assumes any derivation has validated the context for which the image provides
 * and this is only available to derived classes in this FiStream hierarchy.
 *
 * @param theImageToAttach  contains the FiImage to use for video stream writer.
 *
 * @returns A boolean true if attached or false if not.
 */
bool FiVideoStream::attachVideoFrameToWrite(std::shared_ptr<FiImage> theImageToAttach)
{
    bool attachedImage = false;

    // Reset the videoFrameToWrite to empty state.
    videoFrameToWrite.reset();

    // Must be a stream writer to attach a video frame.
    if (streamType != FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER)
    {
        fi_log(LOG_DEBUG, "%s: Video frames are only attached for stream writers.", __PRETTY_FUNCTION__);
        goto exitAttachVideoFrameToWrite;
    }

    // Sanity check: Verify there is a valid pointer to the image data.
    if (!theImageToAttach)
    {
        fi_log(LOG_DEBUG, "%s: Empty video frame received for stream writer!", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitAttachVideoFrameToWrite;
    }

    // Sanity check:  Make sure the image to add has attached image data.
    if (!theImageToAttach->IsImageAttached())
    {
        fi_log(LOG_DEBUG, "%s: no attached image in video frame for stream writer.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitAttachVideoFrameToWrite;
    }

    // If no frame has ever been added set the image format for the first added frame.
    if (firstAddedImageFormat == FAS_VIDEO::IMAGE_FORMAT::FI_INVALID_IMAGE)
    {
        firstAddedImageFormat = theImageToAttach->GetImageFormat();
    }
    // Sanity check:  Make sure the frame being added matches the firstAddedImageFormat.
    else if (firstAddedImageFormat != theImageToAttach->GetImageFormat())
    {
        fi_log(LOG_DEBUG, "%s: incompatible video frame received with image type[%d] was expecting[%d].",
               __PRETTY_FUNCTION__, (int)theImageToAttach->GetImageFormat(), (int)firstAddedImageFormat);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitAttachVideoFrameToWrite;
    }

    // It has passed minimal common checks to attach video frame for stream writing.
    videoFrameToWrite = theImageToAttach;

    // Image data was attached for stream writing.
    attachedImage = true;
exitAttachVideoFrameToWrite:
    return attachedImage;
}

