/*!
 * @file fas_stream_avi.cpp
 * @brief The AVI stream object of the Fastec Video library.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include <inttypes.h>

#include <iostream>
#include <fstream>

#include "fas_video.h"
#include "fi_stream_avi.h"
#include "fi_image_pkg.h"
#include "fi_image.h"
#include "fi_bmp.h"
#include "fi_jpeg.h"
#include "log_if.h"


namespace FAS_AVI
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------
//! If you want to turn on verbose logging for this object set this to 1
//static const bool VERBOSE_SAVE_LOG = true;
static const bool VERBOSE_SAVE_LOG = false;

// Defines the frame clock rate used by AVI to calculate frame rate
const uint32_t  AVI_FRAME_CLOCK_RATE = 1000000;

// Target a 1 G max size for each RIFF section
const uint32_t TARGET_RIFF_MAX  = 0x40000000;

// Index Type codes for OpenDML (AVI 2.0) style indexes

// Used when each entry in the Index
// array points to an index chunk
const uint32_t AVI_INDEX_OF_INDEXES = 0x00;

// Used when each entry in the Index
// array points to a chunk in thefile
const uint32_t AVI_INDEX_OF_CHUNKS = 0x01;

//
// AVI tags
//

/// Converts 4 character bytes into a "Four Character Code" (FCC)
#ifndef BYTES_TO_FCC
#define BYTES_TO_FCC( c0, c1, c2, c3 )              \
    ( (uint32_t)(uint8_t)(c0) | ( (uint32_t)(uint8_t)(c1) << 8 ) |  \
    ( (uint32_t)(uint8_t)(c2) << 16 ) | ( (uint32_t)(uint8_t)(c3) << 24 ) )
#endif

// Defines the chunk id to use for the video stream.
#define VIDEO_CHUNK_ID BYTES_TO_FCC('0','0','d','c')

#define AVIH_TAG BYTES_TO_FCC('a','v','i','h')
#define RIFF_TAG BYTES_TO_FCC('R','I','F','F')
#define AVI_TAG  BYTES_TO_FCC('A','V','I',' ')
#define AVIX_TAG BYTES_TO_FCC('A','V','I','X')
#define LIST_TAG BYTES_TO_FCC('L','I','S','T')
#define HDRL_TAG BYTES_TO_FCC('h','d','r','l')
#define ODML_TAG BYTES_TO_FCC('o','d','m','l')
#define DMLH_TAG BYTES_TO_FCC('d','m','l','h')
#define STRL_TAG BYTES_TO_FCC('s','t','r','l')
#define STRH_TAG BYTES_TO_FCC('s','t','r','h')
#define VIDS_TAG BYTES_TO_FCC('v','i','d','s')
#define DIB_TAG  BYTES_TO_FCC('D','I','B',' ')
#define MJPG_TAG BYTES_TO_FCC('M','J','P','G')
#define STRF_TAG BYTES_TO_FCC('s','t','r','f')
#define INDX_TAG BYTES_TO_FCC('i','n','d','x')
#define MOVI_TAG BYTES_TO_FCC('m','o','v','i')
#define IX00_TAG BYTES_TO_FCC('i','x','0','0')
#define IDX1_TAG BYTES_TO_FCC('i','d','x','1')

//
// AVI header flags
//

/// Indicates that the AVI file has an index
#define AVIF_HAS_INDEX      0x10

/// Indicates that the streams are properly interleaved
#define AVIF_IS_INTERLEAVED 0x100

/// Indicates that the keyframe flags in the index are reliable.
#define AVIF_TRUSTCKTYPE    0x800

/// Legacy flags for old (AVI 1.0) index structures
#define AVIIF_KEYFRAME      0x10

#pragma pack(1)
// Defines the head of an AVI list or chunk
typedef struct _fcc_head
{
    /// Specifies the FCC (Four Character Code) for this section
    uint32_t fcc;

    /// Specifies the size of this section
    uint32_t size;
} fcc_head;


// An AVI file is filled with riff_lists containing other lists or chunks.
// The important header and index chunks are defined in the structures below.
typedef struct _riff_list
{
    /// Contains "LIST" and the size of this list.  In the case of the first list
    /// in a riff, may be "riff" instead of "LIST"
    fcc_head h;

    /// Contains the FCC for the list "hdrl", "odml", "strl", etc...
    uint32_t list_id;
} riff_list;


typedef struct _avih
{
    /// Specifies the time for each frame in AVI_FRAME_CLOCK_RATE clock ticks
    uint32_t frame_rate;

    /// Specifies the maximum transmit bit rate for the video.  For uncompressed
    // video, this is FPS * frame size
    uint32_t max_bps;

    /// Specifies the required padding alignment to expect for each riff section
    uint32_t pading_alignment;

    /// AVI flags
    uint32_t flags;

    /// In contrast to the field name, this actually holds only
    /// the count of frames in the first RIFF
    uint32_t total_frames;

    /// Only applies to audio.  Allows for an offset between the start of the video
    /// stream and the start of the audio stream
    uint32_t initial_frames;

    /// Count of streams in the AVI
    uint32_t streams;

    /// Used as a hint for memory allocation by the player so it knows how big
    /// a buffer will be needed to decode the video
    uint32_t suggested_buffer_size;

    /// The width of the video
    uint32_t width;

    /// The height of the video
    uint32_t height;

    /// Unused
    uint32_t reserved[4];
} avih;

typedef struct _avih_chunk
{
    /// Contains "avih" and the size of the chunk
    fcc_head h;

    /// Contains the avih chunk data
    avih data;
} avih_chunk;

typedef struct _strh
{
    /// Should be 'vids'
    uint32_t fcc_type;

    /// Defines the codec 'DIB ' is used for raw uncompressed frames
    uint32_t fcc_handler;

    /// No Flag values apply to us.
    uint32_t flags;

    /// Set priority to 0
    uint16_t priority;

    /// Set language to 0
    uint16_t language;

    /// Set initial_frames to 0
    uint32_t initial_frames;

    /// scale/rate = frame rate
    uint32_t scale;

    /// scale/rate = frame rate
    uint32_t rate;

    /// Set start to 0
    uint32_t start;

    /// Length is the count of frames in the video
    uint32_t length;

    /// Set to the frame size
    uint32_t suggested_buffer_size;

    /// Set to 0
    uint32_t quality;

    /// Set to 0
    uint32_t sample_size;

    /// Set to 0
    uint16_t left;

    /// Set to 0
    uint16_t top;

    /// Set to video width
    uint16_t right;

    /// Set to video height
    uint16_t bottom;

} strh;

typedef struct _strh_chunk
{
    /// Contains the chunk type ('strh') and size
    fcc_head h;

    /// Contains the actual video and codec information
    strh data;
} strh_chunk;


typedef struct _indx_entry
{
    /// Specifies the position in the file of the ix00 chunk pointed to by this entry
    uint64_t offset;

    /// Specifies the size of the ix00 chunk pointed to by this entry
    uint32_t size;

    /// Specifies the count of frames indexed in the ix00 chunk pointed to by this entry
    uint32_t duration;
} indx_entry;

typedef struct _indx
{
    /// Defines the number of longs per entry.  In the case of the indx chunk
    /// this is set to 4
    uint16_t longs_per_entry;

    /// Defines the AVI 2.0 index subtype.  Set to 0 for indx
    uint8_t sub_type;

    /// Defines the AVI 2.0 index type.  Set to AVI_INDEX_OF_INDEXES for this
    /// indx structure
    uint8_t type;

    /// Defines the count of entries in use.  In our implementation, this
    /// matches the number of riff sections since we put an ix00 in each riff.
    uint32_t entries_in_use;

    /// Defines the chunk ID (FCC) for the stream indexed by this table
    uint32_t chunk_id;
    uint32_t reserved[3];
    // This is followed by an array of indx_entry structures
    // note that the bytes in that array are included in the size reported
    // in the fcc_head

} indx;

typedef struct _indx_chunk
{
    /// Contains "indx" and the size of the chunk
    fcc_head h;

    indx i;
} indx_chunk;

typedef struct _idx1_entry
{
    /// Defines the chunk id (FCC) for the stream indexed by this table.
    uint32_t chunk_id;

    /// Flags for each entry.  In an uncompressed video, all frames are keyframes
    /// so this is set to AVIIF_KEYFRAME for each entry.
    uint32_t flags;

    ///  Offset to the header for the chunk pointed to by this entry.  Can be either
    /// absolute or relative to the first byte of the 'movi' chunk.
    /// We use relative offsets.
    uint32_t offset;

    /// Specifies the size of the chunk.
    uint32_t size;
} idx1_entry;


typedef struct _ixxx_entry
{
    /// Defines the offset (relative to the ixxx_chunk base_offset) for the data
    /// section (NOT the chunk header) of the video chunk pointed to by this entry.
    /// Therefore, video data exists at ixxx_chunk.base_offset + ixxx_entry.offset
    uint32_t offset;

    /// Specifies the size of the video chunk pointed to by this entry.
    uint32_t size;
} ixxx_entry;

typedef struct _ixxx
{
    /// Count of longs per index entry.  For "ix00" structure it is set to 2
    uint16_t longs_per_entry;

    /// Defines the AVI 2.0 index subtype.  Set to 0 for ix00
    uint8_t sub_type;

    /// Defines the AVI 2.0 index type.  Set to AVI_INDEX_OF_CHUNKS for this
    /// ix00 structure
    uint8_t type;

    /// Defines the count of entries in use.  In our implementation, this
    /// matches the number of frames in this riff-avix section.
    uint32_t entries_in_use;

    /// Defines the chunk ID (FCC) for the stream indexed by this table
    uint32_t chunk_id;

    /// Defines the base offset for all entries in this table.
    /// Offset of the data portion of a video frame is at base_offset + ixxx_entry.offset
    uint64_t base_offset;

    /// Unused
    uint32_t reserved;
    // This is followed by an array of ixxx_entry structures
    // note that the bytes in that array are included in the size reported
    // in the fcc_head
} ixxx;

typedef struct
{
    /// Contains "ix00" and size of chunk
    fcc_head h;

    ixxx data;
} ixxx_chunk;

typedef struct
{
    /// Contains "strf" and size of chunk
    fcc_head h;

    FI_BMP::DIB_HEADER dib;

    // May have a color table right after if image is mono.  This color table
    // is included in the chunk length.
} strf_chunk;


typedef struct _strl_list
{
    /// List header
    riff_list l;

    /// Contains codec and other video information
    strh_chunk strh_data;

    /// The strf chunk contains a DIB header which is the same as the DIB
    /// in a BMP file.
    strf_chunk strf;

} strl_list;

typedef struct _dmlh
{
    /// Contrary to the name, this contains only the count of frames in the first
    /// riff-avi section
    uint32_t total_number_frames;
} dmlh;

typedef struct _dmlh_chunk
{
    /// Contains "dmlh" and the size
    fcc_head h;

    /// Contains the dmlh data
    dmlh dmlh_data;
} dmlh_chunk;


typedef struct _odml_list
{
    /// Contains "LIST" the "odml" FCC and the size
    riff_list l;
    dmlh_chunk data;
} odml_list;

typedef struct _hdrl_list
{
    /// Contains "LIST" or "riff" and the list FCC and the list size
    riff_list l;

    /// Contains the avih chunk
    avih_chunk avih_data;

    /// contains the odml list
    odml_list odml_data;

    /// contains the strl list
    strl_list strl_data;
} hdrl_list;

// indx_mem_entry MUST be the same size as indx_entry
typedef struct
{
    // Specifies the base offset for all entries in the offsets array
    uint64_t base_offset;

    // An array of frame offsets.  Each video frame is at base_offset + offsets[i].offset
    uint32_t offset_rec;        // ixxx_entry *offsets;

    /// Specifies the count of entries in the offsets array.
    uint32_t frames;

} indx_mem_entry;


typedef union
{
    // NOTE: this union may not be 64 bit processor safe.
    indx_entry disk;
    indx_mem_entry mem;
} indx_union;

#pragma pack()

typedef struct OFFSET_REC_LIST
{
    struct OFFSET_REC_LIST  *next;
    ixxx_entry  offsets[1];             // There's likely to be more than one...but this will do as an overlay.
} OFFSET_REC_LIST;

/// Defines the AVI container context.  This data is necessary to keep
/// track of portions of the file while frames are being added to the video.
typedef struct
{
    uint64_t riff_offset;
    uint64_t movi_offset;
    uint64_t indx_offset;
    uint32_t current_riff;
    uint32_t frames_per_riff;
    uint32_t current_frame;
    uint32_t riff_count;
    uint32_t frame_size;
    //fi_pix_fmt raw_format;
    uint32_t raw_format;

    OFFSET_REC_LIST *rec_head;

    // Stuff used when
    avih *avi_header;
    strh *avi_strh;
    FI_BMP::DIB_HEADER *avi_strf;
    dmlh *avi_dmlh;
    indx *avi_indx;
    indx_union *frame_index;
} AVI_VIDEO_CONTEXT;

/*
 * @brief Private AVI Context Class
 */
class FAS_AVI::FiVideoAVIStream::AVIContextClass
{
public:
    AVIContextClass();
    virtual ~AVIContextClass();

    uint32_t alloc_offset_record(size_t rec_size);
    ixxx_entry *get_offset_rec(uint32_t rec_index);

    AVI_VIDEO_CONTEXT *aviCtxPtr;       //< The AVI context record.
};

FAS_AVI::FiVideoAVIStream::AVIContextClass::AVIContextClass() :
    aviCtxPtr(nullptr)
{
}

FAS_AVI::FiVideoAVIStream::AVIContextClass::~AVIContextClass()
{
    // There is an entire free function that exists to use...

}

uint32_t FAS_AVI::FiVideoAVIStream::AVIContextClass::alloc_offset_record(size_t rec_size)
{
    size_t          malloc_size;
    uint32_t        entry_counter;
    OFFSET_REC_LIST *new_entry;
    OFFSET_REC_LIST *list_walker;
    OFFSET_REC_LIST *prev_rec;

    // If there is no allocated avi context record pointer can't continue.
    if (aviCtxPtr == nullptr)
        return ((uint32_t) -1);

    // To make things easier, we're going to over allocate the memory by a pointer's worth.
    malloc_size = sizeof(OFFSET_REC_LIST) + rec_size;

    // Allocate our entry.
    new_entry = (OFFSET_REC_LIST *)malloc(malloc_size);
    if (new_entry == nullptr)
        return ((uint32_t) -1);

    // Clear our entry.
    memset(new_entry, 0, malloc_size);

    // Find out where to put this.
    prev_rec = list_walker = aviCtxPtr->rec_head;
    for (entry_counter = 0; list_walker != nullptr; ++entry_counter)
    {
        // Save this record location.
        prev_rec = list_walker;

        // Move to the next.
        list_walker = list_walker->next;
    }

    // Is this the head?
    if (prev_rec == nullptr)
        aviCtxPtr->rec_head = new_entry;
    else
        prev_rec->next = new_entry;

    // Return the record index.
    return entry_counter;
}

ixxx_entry *FAS_AVI::FiVideoAVIStream::AVIContextClass::get_offset_rec(uint32_t rec_index)
{
    uint32_t    index;
    OFFSET_REC_LIST *rec_ptr;

    // Make sure the context pointer is valid.
    if (aviCtxPtr == nullptr)
        return nullptr;

    // Sanity check.
    if (aviCtxPtr->rec_head == nullptr)
    {
        // No record have been allocated!
        fi_log(LOG_DEBUG, "%s: No rec_head!", __PRETTY_FUNCTION__);
        return nullptr;
    }

    // Find the record.
    rec_ptr = aviCtxPtr->rec_head;
    for (index = 0; (index < rec_index) && (rec_ptr != nullptr); ++index)
    {
        // Move to the next record.
        rec_ptr = rec_ptr->next;
    }

    // Sanity check.
    if (index != rec_index)
    {
        // Didn't find it!
        fi_log(LOG_DEBUG, "%s: index: %d rec_index: %d", __PRETTY_FUNCTION__, index, rec_index);
        return nullptr;
    }

    // Return the record contents.
    return rec_ptr->offsets;
}


} // End of FAS_AVI

using namespace FAS_VIDEO;
using namespace FAS_AVI;
using namespace std;

//--------------------------------------------------------------------
// Implementation
//--------------------------------------------------------------------

//#define TEST_PLACEHOLDER_FRAME 1
#ifdef TEST_PLACEHOLDER_FRAME
// For testing the place holder logic internally.
static uint32_t phTestFrameNbr = 11;
static std::unique_ptr<FAS_AVI::PLACE_HOLDER_FRAME> phTestFrame;
#endif

/*!
 * @brief Video AVI stream construction.
 */
FiVideoAVIStream::FiVideoAVIStream() : headerProcessed(false),
    runningFrameCount(0),
    curFileOffset(0)
{
    fi_log(LOG_DEBUG, "%s: Constructing FiVideoAVIStream", __PRETTY_FUNCTION__);
}

/*!
 * @brief Video AVI stream destruction.
 */
FiVideoAVIStream::~FiVideoAVIStream()
{
    fi_log(LOG_DEBUG, "%s: Destructing FiVideoAVIStream", __PRETTY_FUNCTION__);

    // If the stream status isn't set to closed - close it.
    if (!IsStreamClosed() && streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_READY)
        FiVideoAVIStream::CloseStream();
}

/*!
 * @brief Set up the video stream object for reading.
 *
 * @param videoNameToOpen         Name of the video file to open for reading.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoAVIStream::PrepareForReading(const std::string& videoNameToOpen)
{
    FAS_VIDEO::STREAM_STATUS retStatus = FAS_VIDEO::STREAM_STATUS::STREAM_EMPTY;

    // The stream should be in a status of construction in order to prepare for reading.
    if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_BUILT)
    {
        // This method is expecting a newly constructed video stream object.
        // Deny the request and leave the stream status set to its state.
        retStatus = FAS_VIDEO::STREAM_STATUS::REQUEST_DENIED;
        fi_log(LOG_ERR, "%s: streamStatus != STREAM_BUILT but set to [%d].", __PRETTY_FUNCTION__, (uint32_t)streamStatus);
        goto exitPrepareForReading;
    }

    // Create the input file stream object using standard input file streaming.
    ioFileStreamPtr = std::unique_ptr<std::fstream>(new std::fstream(videoNameToOpen.c_str(), ios::in | ios::binary));

    // If the ioFileStream object was not created mark error.
    if (!ioFileStreamPtr)
    {
        retStatus = streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
        fi_log(LOG_ERR, "%s: Unable to create fstream for [%s].", __PRETTY_FUNCTION__, videoNameToOpen.c_str());
        goto exitPrepareForReading;
    }
    // Check the status of the output file stream object.
    else if (!ioFileStreamPtr->is_open())
    {
        // Unable to open this file for binary output.
        retStatus = streamStatus = FAS_VIDEO::STREAM_STATUS::NO_FILE_FOUND;
        fi_log(LOG_ERR, "%s: fstream.is_open() failed for [%s].", __PRETTY_FUNCTION__, videoNameToOpen.c_str());
        goto exitPrepareForReading;
    }

    // Create the AVI context for the writer.
    // NOTE: that the internal allocated aviCtxPtr is initialized to
    // nullptr in the class constructor.
    aviContext = std::unique_ptr<FAS_AVI::FiVideoAVIStream::AVIContextClass>(new AVIContextClass());

    // Make sure it was created.
    if (!aviContext)
    {
        streamStatus = FAS_VIDEO::STREAM_STATUS::ALLOCATION_FAILED;
        fi_log(LOG_DEBUG, "%s: Unable to create the AVI Context Record.", __PRETTY_FUNCTION__);
        goto exitPrepareForReading;
    }

    // Need to allocate the internal avi context data structure.
    aviContext->aviCtxPtr = (AVI_VIDEO_CONTEXT *)malloc(sizeof(AVI_VIDEO_CONTEXT));

    // Did the allocation create the data record?
    if (aviContext->aviCtxPtr == nullptr)
    {
        streamStatus = FAS_VIDEO::STREAM_STATUS::ALLOCATION_FAILED;
        fi_log(LOG_DEBUG, "%s: Unable to create the AVI Context Record Internal Context Buffer.", __PRETTY_FUNCTION__);
        goto exitPrepareForReading;
    }

    // Set the record to all zeroes.
    memset(aviContext->aviCtxPtr, 0, sizeof(AVI_VIDEO_CONTEXT));

    // If made it here and the status is still STREAM_BUILT no errors have occurred.
    if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_BUILT)
    {
        streamType = FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_READER;
        streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_READY;
        fi_log(LOG_DEBUG, "%s: STREAM_BUILT for [%s] now verifying the AVI header.", __PRETTY_FUNCTION__, videoNameToOpen.c_str());
    }

    // Scan the AVI file preparing the AVI context information for reading frames.
    if (!ScanHeader())
    {
        // Unable to verify the AVI data.
        fi_log(LOG_ERR, "%s: Unable to ScanHeader for AVI [%s].", __PRETTY_FUNCTION__, videoNameToOpen.c_str());
        retStatus = streamStatus;   // Mark the return status equal to the stream status.
    }
    else
    {
        // Save the video name presented by the caller since it is a valid stream.
        videoNameOfStream = videoNameToOpen;

        // Mark the return status as a stream-ready state.
        retStatus = FAS_VIDEO::STREAM_STATUS::STREAM_READY;
    }
exitPrepareForReading:
    return retStatus;
}

/*!
 * @brief Set up the AVI video stream object for writing.
 *
 * @param videoNameToCreate What to name the created video file.
 * @param frameRateToUse    The number of frames to render per second of video.
 * @param frameWidthToUse   The number of pixel columns across a row.
 * @param frameHeightToUse  The number of rows in a video frame.
 * @param frameCountToUse   The number of frames for this video.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoAVIStream::PrepareForWriting(const std::string& videoNameToCreate,
                                                             uint32_t frameRateToUse,
                                                             uint32_t frameWidthToUse,
                                                             uint32_t frameHeightToUse,
                                                             uint32_t frameCountToUse)
{
    // The stream should be in a status of construction in order to prepare for writing.
    if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_BUILT)
    {
        // This method is expecting a newly constructed video stream object.
        // Deny the request and leave the stream status set to its state.
        return FAS_VIDEO::STREAM_STATUS::REQUEST_DENIED;
    }

    // Until a video frame is added we don't really know what kind of image
    // frames will be streamed.
    // Frame data will be Device Independent Bitmaps create the outFileFrame object
    // With the type of image data expected.
    // outFileImage = std::unique_ptr<FAS_VIDEO::FiImage>(new FI_BMP::FiBMP());
    // TODO_JLC Other Sanity Checks.

    frameCount = frameCountToUse;

    // Is there a range of frames to stream?
    // NOTE_JLC - is there a minimum number of frames to accept for a video stream?
    if (frameCount == 0)
    {
        streamStatus = FAS_VIDEO::STREAM_STATUS::BAD_RANGE;
        fi_log(LOG_DEBUG, "%s: frameCount set to zero.", __PRETTY_FUNCTION__);
        goto exitPrepareForAVIWriting;
    }

    // NOTE:  Sometimes the frame dimensions are unknown at the time
    // of the stream writer preparation.  It will be up to the first
    // add frame to look at the setting of headerProcessed to see
    // if any "housekeeping" is needed and waiting on the first image
    // so that the additional information is provided from the image
    // itself.

    // Save the frame dimension information.
    frameWidth = frameWidthToUse;
    frameHeight = frameHeightToUse;

    // NOTE_JLC is there a minimum frame rate allowed?  Should we flag something under say 10?  TBD.
    frameRate = frameRateToUse;

    // Creating an output file stream using the standard output file streaming object.
    ioFileStreamPtr = std::unique_ptr<std::fstream>(new std::fstream(videoNameToCreate.c_str(), ios::out | ios::binary | ios::trunc));

    // If the outFileStream object was not created mark error.
    if (!ioFileStreamPtr)
    {
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
        goto exitPrepareForAVIWriting;
    }
    // Check the status of the output file stream object.
    else if (!ioFileStreamPtr->is_open())
    {
        // Unable to open this file for binary output.
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_FILE_FOUND;
        goto exitPrepareForAVIWriting;
    }

    // Create the AVI context for the writer.
    // NOTE: that the internal allocated aviCtxPtr is initialized to
    // nullptr in the class constructor.
    aviContext = std::unique_ptr<FAS_AVI::FiVideoAVIStream::AVIContextClass>(new AVIContextClass());

    // Make sure it was created.
    if (!aviContext)
    {
        streamStatus = FAS_VIDEO::STREAM_STATUS::ALLOCATION_FAILED;
        fi_log(LOG_DEBUG, "%s: Unable to create the AVI Context Record.", __PRETTY_FUNCTION__);
        goto exitPrepareForAVIWriting;
    }

    streamType = FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER;
    streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_READY;
    videoNameOfStream = videoNameToCreate;

exitPrepareForAVIWriting:

    // Set the headerProcessed to false.  For writers headerProcessed is
    // set to true when first image attached and any missing header info
    // is processed, for readers headerProcessed is set to true when the
    // headers are scanned and validated.
    headerProcessed = false;

    // Reset the running frame count.
    runningFrameCount = 0;

    // Log the stream status in a debug scenario.
    LogStreamStatus();

    // Return the status of this stream object as being prepared.
    return streamStatus;
}

/*!
 * @brief Common test to determine if the AVI stream is able to read.
 *
 * @param whoIsAsking       The function that called this one for logging.
 * @param addContextCheck   Does this test check internal context pointers.
 *
 * @returns A boolean true if it is ok to write or false if not.  Logs
 * any error condition.
 */
bool FiVideoAVIStream::OkToRead(const char *whoIsAsking, bool addContextCheck)
{
    bool okToRead = false;

    // Is this AVI video stream set up for writing?
    if (streamType != FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_READER)
    {
        fi_log(LOG_DEBUG, "%s: AVI stream not set for reading.", whoIsAsking);
        LogStreamStatus();
        goto exitOkToRead;
    }
    // Is streamStatus in ready state?
    if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_READY)
    {
        fi_log(LOG_DEBUG, "%s: streamStatus is not set to STREAM_READY.", whoIsAsking);
        LogStreamStatus();
        goto exitOkToRead;
    }

    // Make sure there is an AVI context object keeping track of the internal
    // file information.
    if (!aviContext)
    {
        fi_log(LOG_ERR, "%s: missing AVI context object.", whoIsAsking);
        LogStreamStatus();
        goto exitOkToRead;
    }

    // Check to see if caller wants to verify the internal context allocations.
    if (addContextCheck)
    {
        // Check the internal AVI_VIDEO_CONTEXT* data pointer.
        if (aviContext->aviCtxPtr == nullptr)
        {
            fi_log(LOG_ERR, "%s: missing AVI context internal data.", whoIsAsking);
            LogStreamStatus();
            goto exitOkToRead;
        }
    }

    // If made it here the tests were passed.
    okToRead = true;

exitOkToRead:
    return (okToRead);
}


/*!
 * @brief Common test to determine if the AVI stream is able to write.
 *
 * @param whoIsAsking       The function that called this one for logging.
 * @param addContextCheck   Does this test check internal context pointers.
 * @param addImageAttachCheck   Does this test check to see if image is attached.
 *
 * @returns A boolean true if it is ok to write or false if not.  Logs
 * any error condition.
 */
bool FiVideoAVIStream::OkToWrite(const char *whoIsAsking, bool addContextCheck, bool addImageAttachCheck)
{
    bool okToWrite = false;

    // Is this AVI video stream set up for writing?
    if (streamType != FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER)
    {
        fi_log(LOG_DEBUG, "%s: AVI stream not set for writing.", whoIsAsking);
        LogStreamStatus();
        return okToWrite;
    }
    // Is streamStatus in ready state?
    if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_READY)
    {
        fi_log(LOG_DEBUG, "%s: streamStatus is not set to STREAM_READY.", whoIsAsking);
        LogStreamStatus();
        return okToWrite;
    }

    // Make sure there is an AVI context object keeping track of the internal
    // file information.
    if (!aviContext)
    {
        fi_log(LOG_ERR, "%s: missing AVI context object.", whoIsAsking);
        LogStreamStatus();
        return okToWrite;
    }

    // Check to see if caller wants to verify the internal context allocations.
    if (addContextCheck)
    {
        // Check the internal AVI_VIDEO_CONTEXT* data pointer.
        if (aviContext->aviCtxPtr == nullptr)
        {
            fi_log(LOG_ERR, "%s: missing AVI context internal data.", whoIsAsking);
            LogStreamStatus();
            return okToWrite;
        }
    }

    // Does caller want to check for an attached image?
    if (addImageAttachCheck)
    {
        // Make sure there is a unique pointer for the image object.
        if (!videoFrameToWrite)
        {
            fi_log(LOG_DEBUG, "%s: missing videoFrameToWrite for stream writer.", __PRETTY_FUNCTION__);
            streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
            LogStreamStatus();
            return okToWrite;
        }
        // Is there a videoFrameToWrite and if not the header cannot be written.
        if (!videoFrameToWrite->IsImageAttached())
        {
            fi_log(LOG_DEBUG, "%s: missing videoFrameToWrite for stream writer.", __PRETTY_FUNCTION__);
            streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
            LogStreamStatus();
            return okToWrite;
        }
    }

    // If made it here the tests were passed.
    okToWrite = true;

    return (okToWrite);
}

/*!
 * @brief Initialize AVI header building it up in memory in preparation
 * to receive video frames.
 *
 * IMPORTANT:
 * Expects the stream data curVideoFrameToAdd to be valid with an attached image.
 *
 * Note that we need the first frame in cases where the input format is BMP or JPEG
 * so that we can figure out the image dimensions from the image data.
 *
 * The first RIFF in an avi video contains the following
 *      riff-avi list
 *          list-hdrl (more stuff in this)
 *              avih chunk
 *              list-odml
 *              list-strl
 *                  strh chunk
 *                  strf chunk
 *                  indx chunk
 *          list movi
 *              image chunks
 *              ixxx_chunk
 *              idx1 chunk (not written, TODO: Do I need this?)
 *
 * The riff_avi header contains this
 *         riff-avi header
 *             list hdrl header
 *                 avih chunk
 *                 list odml header
 *                     dmlh chunk
 *                 list strl header
 *                     strh chunk
 *                     strf chunk
 *                     indx chunk
 *
 *
 * @return Boolean true if successful or false if error occurred in which case check streamStatus.
 */
bool FiVideoAVIStream::InitializeHeader()
{
    bool initializedHeader = false;
    bool addContextCheck = false;
    bool addImageAttachCheck = true;

    uint32_t fullFrameSize;
    uint32_t frames_per_riff;       // Stores the calculated frames per riff
    uint32_t riff_count;            // Stores the calculated count of riffs

    uint32_t i;
    uint32_t idx_buffer_size = 0;
    uint8_t *idx_buffer;

    // Perform the common status checks.  If they don't pass cannot continue.
    if (!OkToWrite(__PRETTY_FUNCTION__, addContextCheck, addImageAttachCheck))
        return initializedHeader;

    // Calculate frames per RIFF and count of RIFFs in file.
    // Assume uncompressed frames.  No harm in having shorter riffs
    // if frames are compressed.

    // The attached image knows its framesize.
    fullFrameSize = videoFrameToWrite->GetFullFramesize();

    if (fullFrameSize != 0)
        frames_per_riff = TARGET_RIFF_MAX / fullFrameSize;
    else
        frames_per_riff = frameCount;

    if (frames_per_riff >= frameCount)
    {
        frames_per_riff = frameCount;
        riff_count = 1;
    }
    else
    {
        riff_count = frameCount / frames_per_riff;

        // Add one more riff if there are remaining frames.
        if ((riff_count * frames_per_riff) < frameCount)
            riff_count++;
    }

    // Allocate context and initialize pointers
    aviContext->aviCtxPtr = (AVI_VIDEO_CONTEXT *)calloc(1, sizeof(AVI_VIDEO_CONTEXT) + (sizeof(ixxx_entry) * frames_per_riff));
    if (aviContext->aviCtxPtr == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: unable to create AVI context data.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::ALLOCATION_FAILED;
        LogStreamStatus();
        return initializedHeader;
    }

    // Store calculated context information
    aviContext->aviCtxPtr->frames_per_riff = frames_per_riff;
    aviContext->aviCtxPtr->frame_size = fullFrameSize;
    aviContext->aviCtxPtr->riff_count = riff_count;
    //NOTE_JLC come back to this next statement - this was the old PIX FMT data.
    uint32_t theBPP = videoFrameToWrite->GetBitsPerPixel();
    fi_log(LOG_DEBUG, "%s: JPEG_BITS_PER_PIXEL SETTING for aviCtxPtr->raw_format assigned: %d.", __PRETTY_FUNCTION__, theBPP);
    aviContext->aviCtxPtr->raw_format = theBPP; // Should be 8 or 24 for Gray or RGB respectively...

    idx_buffer_size = sizeof(indx) + (aviContext->aviCtxPtr->riff_count * sizeof(indx_union));
    idx_buffer = (uint8_t *)calloc(1, idx_buffer_size);
    if (idx_buffer == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: unable to create AVI context data.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::ALLOCATION_FAILED;
        LogStreamStatus();
        return initializedHeader;
    }

    aviContext->aviCtxPtr->avi_indx = (indx *)idx_buffer;
    aviContext->aviCtxPtr->frame_index = (indx_union *)(idx_buffer + sizeof(indx));
    aviContext->aviCtxPtr->avi_indx->entries_in_use = aviContext->aviCtxPtr->riff_count;

#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "(%s) frame_count: %d, riff_count: %d, frames/riff: %d.", __PRETTY_FUNCTION__,
           frameCount, aviContext->aviCtxPtr->riff_count, aviContext->aviCtxPtr->frames_per_riff);
#endif

    for (i = 0; i < aviContext->aviCtxPtr->riff_count; i++)
    {
        aviContext->aviCtxPtr->frame_index[i].mem.frames = aviContext->aviCtxPtr->frames_per_riff;
        aviContext->aviCtxPtr->frame_index[i].mem.offset_rec = aviContext->alloc_offset_record((aviContext->aviCtxPtr->frames_per_riff * sizeof(ixxx_entry)));
        if (aviContext->aviCtxPtr->frame_index[i].mem.offset_rec == (uint32_t) -1)
        {
            fi_log(LOG_ERR, "(%s) Unable to allocate frame offset index table", __PRETTY_FUNCTION__);
            streamStatus = FAS_VIDEO::STREAM_STATUS::ALLOCATION_FAILED;
            LogStreamStatus();
            AVICleanup();
            return initializedHeader;
        }
    }

    // Start the first RIFF (containing the avi headers).
    if (!AVIStartRiff())
    {
        AVICleanup();
        return initializedHeader;
    }

    // If made it here all is initialized and ready to proceed.
    initializedHeader = true;
    return initializedHeader;
}

/*!
 * @brief Scan AVI header to validate and build index context information.
 *
 * @returns A boolean true if it is ok to write or false if not.  Logs
 * any error condition.
 */
bool FiVideoAVIStream::ScanHeader()
{
    fcc_head h;
    uint32_t bytes;

    bool retCode = false;
    bool addContextCheck = true;
    streampos seekPos = (off64_t)0;

    // Perform the common status checks.  If they don't pass cannot continue.
    if (!OkToRead(__PRETTY_FUNCTION__, addContextCheck))
    {
        goto exitScanHeader;
    }

    // Set the seek position in expected AVI file to first byte.
    if (!SetSeekPos((off64_t) seekPos, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET))
    {
        fi_log(LOG_ERR, "(%s) Unable to set seek position to start of file.", __PRETTY_FUNCTION__);
        goto exitScanHeader;
    }

    // Read the first RIFF header.
    if (!ReadData(&h, sizeof(fcc_head)))
    {
        fi_log(LOG_ERR, "(%s) Unable to read first RIFF header.", __PRETTY_FUNCTION__);
        goto exitScanHeader;
    }

    // Validate that the RIFF tag exists in the data read.
    if (h.fcc != RIFF_TAG)
    {
        fi_log(LOG_ERR, "(%s) First header missing RIFF tag, is this an AVI file?", __PRETTY_FUNCTION__);
        goto exitScanHeader;
    }

    while (h.fcc == RIFF_TAG)
    {
        bytes = AVIReadList(h.size);
        if (bytes != h.size)
        {
            fi_log(LOG_ERR, "(%s) Failed to read list.", __PRETTY_FUNCTION__);
            goto exitScanHeader;
        }

        memset(&h, 0, sizeof(fcc_head));
        if (!ReadData(&h, sizeof(fcc_head)))
        {
            // If the EOF reached on this read attempt all is ok - just at end of file.
            // NOTE_JLC - Really?!  The answer is yes.  And we will set a state of EOF
            // for the stream as this is important for future seek operations that it
            // understand the eof condition reached.  I want to come back to this idea
            // and see if maybe we should reset the bit right here.
            if (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_EOF)
            {
                fi_log(LOG_ERR, "(%s) Unable to read header of next RIFF_TAG.", __PRETTY_FUNCTION__);
                goto exitScanHeader;
            }
            else
            {
                // We have exhausted all data for the header.  EOF means the file is ok and
                // we will at this point set the stream status as ready to read images.
                streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_READY;
            }
        }
    }

    if (aviContext->aviCtxPtr->avi_header == nullptr)
    {
        fi_log(LOG_ERR, "(%s) Missing avi_header.", __PRETTY_FUNCTION__);
        goto exitScanHeader;
    }
    if (aviContext->aviCtxPtr->avi_indx == nullptr)
    {
        fi_log(LOG_ERR, "(%s) Missing avi_indx.", __PRETTY_FUNCTION__);
        goto exitScanHeader;
    }
    if (aviContext->aviCtxPtr->avi_strf == nullptr)
    {
        fi_log(LOG_ERR, "(%s) Missing avi_strf.", __PRETTY_FUNCTION__);
        goto exitScanHeader;
    }
    if (aviContext->aviCtxPtr->avi_strh == nullptr)
    {
        fi_log(LOG_ERR, "(%s) Missing avi_strh.", __PRETTY_FUNCTION__);
        goto exitScanHeader;
    }
    if (aviContext->aviCtxPtr->frame_index == nullptr)
    {
        fi_log(LOG_ERR, "(%s) Missing frame_index.", __PRETTY_FUNCTION__);
        goto exitScanHeader;
    }

    // What type of images are we expecting?
    if (aviContext->aviCtxPtr->avi_strh->fcc_handler == DIB_TAG)
    {
        firstAddedImageFormat = FAS_VIDEO::IMAGE_FORMAT::FI_BMP_IMAGE;
    }
    else if (aviContext->aviCtxPtr->avi_strh->fcc_handler == MJPG_TAG)
    {
        firstAddedImageFormat = FAS_VIDEO::IMAGE_FORMAT::FI_JPEG_IMAGE;
    }
    else
    {
        fi_log(LOG_ERR, "(%s) Invalid Codec specified in avi_strh->fcc_handler[%4.4s].",
               __PRETTY_FUNCTION__, (char *) & (aviContext->aviCtxPtr->avi_strh->fcc_handler));
        goto exitScanHeader;
    }

    frameRate = aviContext->aviCtxPtr->avi_strh->rate;
    frameCount = aviContext->aviCtxPtr->avi_strh->length;
    frameHeight = (uint32_t)aviContext->aviCtxPtr->avi_strf->height;
    frameWidth = (uint32_t)aviContext->aviCtxPtr->avi_strf->width;

    // The header was scanned without any errors.  Assume all is good to go.
    retCode = true;

exitScanHeader:
    return retCode;
}

/*!
 * @brief Scan AVI Index tag and verify contents.
 *
 * @returns A boolean true if index was scanned and is valid false if error.
 */
bool FiVideoAVIStream::ScanIndex()
{
    bool retCode = false;

    off64_t current_offset, index_offset;
    uint32_t size;
    uint32_t i;
    ixxx_chunk ixxx_info;
    fcc_head h;
    uint32_t idx;

    bool addContextCheck = true;

    // Perform the common status checks.  If they don't pass cannot continue.
    if (!OkToRead(__PRETTY_FUNCTION__, addContextCheck))
    {
        goto exitScanIndex;
    }

    // Get current offset
    current_offset = GetSeekPos();

#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "(%s) ScanIndex Current Offset[%" PRIu64 "].",
           __PRETTY_FUNCTION__, current_offset);
#endif

    // If the current offset is not valid this is error condition.
    if (current_offset == -1)
    {
        fi_log(LOG_ERR, "%s: unable to retrieve current seek offset.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
        goto exitScanIndex;
    }

    // Traverse the entries in use in index record.
    for (i = 0; i < aviContext->aviCtxPtr->avi_indx->entries_in_use; i++)
    {
        ixxx_entry  *offset_rec;

        index_offset = (off64_t)aviContext->aviCtxPtr->frame_index[i].disk.offset;
        if (!SetSeekPos(index_offset, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET))
        {
            fi_log(LOG_ERR, "(%s) Unable to set seek position to index_offset[%lu].", __PRETTY_FUNCTION__, index_offset);
            goto exitScanIndex;
        }

        size = aviContext->aviCtxPtr->frame_index[i].disk.size;

        if (!ReadData(&ixxx_info, sizeof(ixxx_chunk)))
        {
            fi_log(LOG_ERR, "(%s) error reading index of frames for index_offset[%lu].", __PRETTY_FUNCTION__, index_offset);
            goto restoreScanIndex;
        }
        if (ixxx_info.h.fcc != IX00_TAG)
        {
            fi_log(LOG_ERR, "(%s) invalid fcc %4.4s, expecting 'ix00' for index_offset[%lu].", __PRETTY_FUNCTION__, (char *) & (ixxx_info.h.fcc), index_offset);
            goto restoreScanIndex;
        }
        size -= sizeof(ixxx_chunk);
        aviContext->aviCtxPtr->frame_index[i].mem.base_offset = ixxx_info.data.base_offset;

        // Allocate an offsets record.
        aviContext->aviCtxPtr->frame_index[i].mem.offset_rec = aviContext->alloc_offset_record(size);
        if (aviContext->aviCtxPtr->frame_index[i].mem.offset_rec == (uint32_t) -1)
        {
            fi_log(LOG_ERR, "(%s) Unable to allocate index %d bytes for fcc %4.4s, in index_offset[%lu].", __PRETTY_FUNCTION__,
                   size, (char *) & (ixxx_info.h.fcc), index_offset);
            goto restoreScanIndex;
        }

        // Get a pointer to the offset record.
        offset_rec = aviContext->get_offset_rec(aviContext->aviCtxPtr->frame_index[i].mem.offset_rec);
        if (offset_rec == nullptr)
        {
            fi_log(LOG_ERR, "(%s) Unable to get_offset_rec for index %d bytes for fcc %4.4s, in index_offset[%lu] i=[%d].", __PRETTY_FUNCTION__,
                   size, (char *) & (ixxx_info.h.fcc), index_offset, i);
            goto restoreScanIndex;
        }
        // Read the data and populate the offset record.
        if (!ReadData(offset_rec, size))
        {
            fi_log(LOG_ERR, "(%s) Failed to read index entries for index %d bytes for fcc %4.4s, in index_offset[%lu] i=[%d].", __PRETTY_FUNCTION__,
                   size, (char *) & (ixxx_info.h.fcc), index_offset, i);
            goto restoreScanIndex;
        }

        // Validate first video chunk in this index list
        idx = 0;
        index_offset = (off64_t)((aviContext->aviCtxPtr->frame_index[i].mem.base_offset + offset_rec[idx].offset) - sizeof(fcc_head));
        if (!SetSeekPos(index_offset, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET))
        {
            fi_log(LOG_ERR, "(%s) Unable to set seek position to index_offset[%lu] i=[%d].", __PRETTY_FUNCTION__, index_offset, i);
            goto restoreScanIndex;
        }

        // Read the fcc header.
        if (!ReadData(&h, sizeof(fcc_head)))
        {
            fi_log(LOG_ERR, "(%s) Failed to read frame chunk header for index_offset[%lu] i=[%d].",
                   __PRETTY_FUNCTION__, index_offset, i);
            goto restoreScanIndex;
        }
        if ((h.fcc != ixxx_info.data.chunk_id) || (h.size != offset_rec[idx].size))
        {
            fi_log(LOG_ERR, "(%s) Invalid size %d in frame chunk header, expected %d for index_offset[%lu] i=[%d].",
                   __PRETTY_FUNCTION__, h.size, offset_rec[idx].size, index_offset, i);
            goto restoreScanIndex;
        }

        // Validate last video chunk in this index list.
        idx = aviContext->aviCtxPtr->frame_index[i].mem.frames - 1;
        index_offset = (off64_t)((aviContext->aviCtxPtr->frame_index[i].mem.base_offset + offset_rec[idx].offset) - sizeof(fcc_head));
        if (!SetSeekPos(index_offset, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET))
        {
            fi_log(LOG_ERR, "(%s) Unable to set seek position to index_offset[%lu] i=[%d] idx=[%d].",
                   __PRETTY_FUNCTION__, index_offset, i, idx);
            goto restoreScanIndex;
        }
        // Read the data for the fcc code.
        if (!ReadData(&h, sizeof(fcc_head)))
        {
            fi_log(LOG_ERR, "(%s) Failed to read frame chunk header for index_offset[%lu] i=[%d] idx=[%d].",
                   __PRETTY_FUNCTION__, index_offset, i, idx);
            goto restoreScanIndex;
        }
        if ((h.fcc != ixxx_info.data.chunk_id) || (h.size != offset_rec[idx].size))
        {
            fi_log(LOG_ERR, "(%s) Invalid size %d in frame chunk header, expected %d for index_offset[%lu] i=[%d] idx=[%d].",
                   __PRETTY_FUNCTION__, h.size, offset_rec[idx].size, index_offset, i, idx);
            goto restoreScanIndex;
        }
    }

restoreScanIndex:
    // Restore the seek position to where it was at when this started.
    if (!SetSeekPos(current_offset, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET))
    {
        fi_log(LOG_ERR, "(%s) Unable to reset seek position to current_offset[%lu].",
               __PRETTY_FUNCTION__, current_offset);
        goto exitScanIndex;
    }

    // If make it here all is ok and the index has been scanned and interpreted.
    retCode = true;

exitScanIndex:

    return retCode;
}

/*!
 * @brief Adds an image or placeholder (fake) image to the video stream.
 *
 * This should only be called by AddVideoFrame and AddPlaceholderFrame methods and
 * it is assumed that any validation has taken place prior to this call to handle
 * AVI context updates for an attached frame or a placeholder frame.
 *
 * @param addPlaceholder    If adding a fake placeholder frame this is true if not
 * then the curVideoFrameToAdd must be set to the current image frame to add.
 *
 * @returns The boolean true if successful, false if error condition.
 */
bool FiVideoAVIStream::AVIAddFrame(bool addPlaceholderFrame)
{
    fcc_head    video_chunk;
    ixxx_entry  *ix00_e;
    ixxx_entry  *offset_rec;
    uint32_t    current_riff;
    uint32_t    current_riff_frame;
    uint32_t    curFrame = 0;

    bool placeHolder = false;       // If writing placeholder frame this will be set true.
    bool addedFrame = false;        // Will determine if the frame was added successfully.
    char placeHolderStr[] = "placeholder-";  // Used for logging.

    // This will be manipulated if writing a bitmap frame.
    uint32_t    imageLenToWrite = 0;

    // This will contain the frame data minus any header information.
    const uint8_t *memDataPtr = nullptr;

    // Pull the current frame value from object.
    curFrame = aviContext->aviCtxPtr->current_frame;

    // This will contain vertical transposition (flipped top to bottom)
    // if writing bitmap frame.
    std::unique_ptr<FAS_VIDEO::FiFrame> normalizedFrame;

    // Frame is placeholder if the imageData is null.
    if (addPlaceholderFrame)
    {
        std::lock_guard<std::mutex> guard(placeHolderFramesMutex);
        placeHolderFrames.push_back(curFrame);

        // Remember we're adding a place holder frame.
        placeHolder = true;

        // Mark the placeholder size at least as big as one full framesize.
        imageLenToWrite = fullFrameSize;
        fi_log(LOG_DEBUG, "%s: ADDED FRAME %d TO placeHolderFrames size=%d.", __PRETTY_FUNCTION__, curFrame, placeHolderFrames.size());

    }
    // Frame to add is not a placeholder so it must be in curVideoFrameToAdd.
    else
    {
        // For logging messages empty this out if it wasn't a placeholder frame.
        placeHolderStr[0] = '\0';

        // Valid data is expected and oriented bottom-up is needed for AVI stream.
        memDataPtr = videoFrameToWrite->GetAttachedFrameData(imageLenToWrite, FAS_VIDEO::FRAME_ORIENTATION::BOTTOM_UP);
    }

    // Sanity check:  Make sure there is a valid image length.
    if (imageLenToWrite == 0)
    {
        fi_log(LOG_DEBUG, "%s: image data framesize is zero or image data pointer is null.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitAVIAddFrame;
    }
    // Sanity check:  Make sure there is a valid image data pointer.
    if (!addPlaceholderFrame && (memDataPtr == nullptr))
    {
        fi_log(LOG_DEBUG, "%s: image data pointer is null.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitAVIAddFrame;
    }

#ifdef TEST_PLACEHOLDER_FRAME
    // To test out the "placeholder" internally use static globals.
    if (curFrame == phTestFrameNbr)
    {
        fi_log(LOG_DEBUG, "%s: AVI %sframe Faking Lost Frame %u.", __PRETTY_FUNCTION__, placeHolderStr, phTestFrameNbr);
        phTestFrame = std::unique_ptr<FAS_AVI::PLACE_HOLDER_FRAME>(new FAS_AVI::PLACE_HOLDER_FRAME(static_cast<uint64_t>(imageLenToWrite), 0));

        uint8_t *posToWrite;      // Write position in buffer.

        // Get our pointer.
        posToWrite = phTestFrame->data();

        // Save the image normalized or not image data.
        memcpy(posToWrite, memDataPtr, imageLenToWrite);

        // Now set the image pointer to null to fake out a "lost" frame.
        memDataPtr = nullptr;
        imageLenToWrite = fullFrameSize;
        placeHolder = true;
    }
#endif

    // Make length an even number.  This seems to make players happier
    imageLenToWrite += (imageLenToWrite & 1);
    video_chunk.fcc = VIDEO_CHUNK_ID;
    video_chunk.size = imageLenToWrite;

    // Write the chunk header
    if (!WriteData(&video_chunk, sizeof(video_chunk)))
    {
        fi_log(LOG_ERR, "%s: Unable to write VIDEO_CHUNK_ID.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitAVIAddFrame;
    }

    // Update the index (ix00) entry
    current_riff = aviContext->aviCtxPtr->current_riff;
    current_riff_frame = aviContext->aviCtxPtr->current_frame % aviContext->aviCtxPtr->frames_per_riff;
    offset_rec = aviContext->get_offset_rec(aviContext->aviCtxPtr->frame_index[ current_riff ].mem.offset_rec);
    if (offset_rec == nullptr)
    {
        fi_log(LOG_ERR, "(%s) no offset record for current_riff: %d, current_riff_frame: %d", __PRETTY_FUNCTION__, current_riff, current_riff_frame);
        streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
        goto exitAVIAddFrame;
    }
    ix00_e = &offset_rec[ current_riff_frame ];
    if (ix00_e == nullptr)
    {
        fi_log(LOG_ERR, "(%s) No ix00_e for current_riff: %d, current_riff_frame: %d", __PRETTY_FUNCTION__, current_riff, current_riff_frame);
        streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
        goto exitAVIAddFrame;
    }

    // I set the ixxx_chunk base offset to the first byte of "movi" tag
    // Therefore the offset of the frame data (not frame chunk header) is...
    ix00_e->offset = (uint32_t)(curFileOffset - aviContext->aviCtxPtr->movi_offset);
    ix00_e->size = imageLenToWrite;

    // Write out the frame data if not a placeholder.
    if (!placeHolder)
    {
        if (!WriteData((const void *) memDataPtr, imageLenToWrite))
        {
            fi_log(LOG_ERR, "(%s) Unable to write image data frame to AVI stream.", __PRETTY_FUNCTION__);
            goto exitAVIAddFrame;
        }
    }
    else
    {
        // Placeholder frame simply moves and records the seek position by the
        // length of the future image.
        // Make sure that there is an output file I/O stream to write to.
        if (!ioFileStreamPtr)
        {
            fi_log(LOG_DEBUG, "%s: AVI stream missing ofstream object.", __PRETTY_FUNCTION__);
            streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
            goto exitAVIAddFrame;
        }
        // Prepare to move the seek position from current place to end of placeholder length.
        streampos seekBefore = ioFileStreamPtr->tellp();
        streampos seekPos = seekBefore;
        seekPos += (off64_t)imageLenToWrite;

        // Place seek position to end of placeholder frame.
        if (!SetSeekPos((off64_t) seekPos, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET))
        {
            fi_log(LOG_DEBUG, "%s: AVI stream unable to set seek position for placeholder frame.", __PRETTY_FUNCTION__);
            streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
            goto exitAVIAddFrame;
        }

        streampos seekCurPos = ioFileStreamPtr->tellp();

        // The seek delta to test becomes the current seek position
        // stored in seekCurPos minus the position before seek request
        // which is in seekBefore.  The delta should equal imageLenToWrite.
        long long seekDelta = seekCurPos - seekBefore;
        if ((uint32_t) seekDelta != imageLenToWrite)
        {
            fi_log(LOG_ERR, "%s: ofstream ok but seek len requested %d != seekDelta %d.", __PRETTY_FUNCTION__, imageLenToWrite, (uint32_t) seekDelta);
            streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
            goto exitAVIAddFrame;
        }

        // Update the file offset to the current seek position.
        // Note that in the WriteData case curFileOffset is updated too.
        curFileOffset = (uint64_t) seekCurPos;

        // Record the placeholder data in the debug logging stream.
        fi_log(LOG_DEBUG, "%s: Place held for frame=%u seekBefore=%lu seekAfter=%lu imgLen=%lu.", __PRETTY_FUNCTION__, curFrame, (uint32_t) seekBefore, (uint32_t) seekCurPos, imageLenToWrite);
    }

    aviContext->aviCtxPtr->current_frame++;

    // Check if we are done with this riff
    if (((aviContext->aviCtxPtr->current_frame % aviContext->aviCtxPtr->frames_per_riff) == 0) ||
        (aviContext->aviCtxPtr->current_frame == frameCount))
    {
        if (!AVIEndRiff())
        {
            fi_log(LOG_ERR, "(%s) AVIEndRiff Failed.", __PRETTY_FUNCTION__);
            streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
            goto exitAVIAddFrame;
        }

        // Check if we are done with the whole video, if not, start a new riff.
        if ((aviContext->aviCtxPtr->current_riff * aviContext->aviCtxPtr->frames_per_riff) < frameCount)
        {
            aviContext->aviCtxPtr->riff_offset = curFileOffset;

            if (!AVIStartRiff())
            {
                fi_log(LOG_ERR, "(%s) AVIStartRiff Failed.", __PRETTY_FUNCTION__);
                streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
                goto exitAVIAddFrame;
            }
        }
    }

    // If make it here in the function the frame was added.
    addedFrame = true;
#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "%s: %sframe (%u) was added to stream.", __PRETTY_FUNCTION__, placeHolderStr, curFrame);
#endif

#ifdef TEST_PLACEHOLDER_FRAME
    // If testing the placeholder frame and have written out all frames - use the saved place holder frame
    // and attempt to edit it...
    if (curFrame == frameCount - 1)
    {
        // Is there a lost frame?
        if (phTestFrame)
        {
            /*const uint8_t *phTestFramePtr = */phTestFrame->data();
            uint32_t phTestFrameLen = (uint32_t) phTestFrame->size();
            fi_log(LOG_DEBUG, "%s: AVI %sframe calling EditVideoFrame for frame=%u imglen=%u.", __PRETTY_FUNCTION__, placeHolderStr, phTestFrameNbr, phTestFrameLen);

            std::unique_ptr<FAS_VIDEO::FiImage> AVIEditFrame;           // Our reusable image data object managing stream images.
            if (firstAddedImageFormat == FAS_VIDEO::IMAGE_FORMAT::FI_BMP_IMAGE)
            {
                AVIEditFrame = std::unique_ptr<FAS_VIDEO::FiImage>(new FI_BMP::FiBMP());
            }
            else
            {
                AVIEditFrame = std::unique_ptr<FAS_VIDEO::FiImage>(new FI_JPEG::FiJPEG());
            }

            // Use our current frame image to attach this fake frame because we want as close to a
            // system frame as possible - even though we have saved data in phTestFrame.
            FAS_VIDEO::IMAGE_STATUS imageResult =
                AVIEditFrame->AttachImagePtr(curVideoFrameToAdd->GetAttachedImageData(), curVideoFrameToAdd->GetAttachedImageSize());
            fi_log(LOG_DEBUG, "%s: AVI %sframe imageAttachResult for frame=%u imglen=%u.  imageResult=%u", __PRETTY_FUNCTION__, placeHolderStr, phTestFrameNbr, phTestFrameLen, imageResult);

            /*FAS_VIDEO::STREAM_STATUS streamResult = */EditVideoFrame(phTestFrameNbr, AVIEditFrame);
            LogStreamStatus();
        }
    }
#endif

exitAVIAddFrame:

    // If the frame was not added report the stream status.
    if (!addedFrame)
    {
        fi_log(LOG_DEBUG, "%s: AVI %sframe was not added to stream.  %s.", __PRETTY_FUNCTION__, placeHolderStr);
        LogStreamStatus();
    }
    else
    {
        // Without regard to the placeholder state keep a running total of
        // frames added to the AVI index.
        lastFrameAdded++;
    }
    return addedFrame;
}

/*!
 * @brief Adds an image to the video stream.
 *
 * Note:  Inside of the imageToAdd right now is attached image data frame
 * in the form of uint8_t* and a length.  This external data pointer is managed
 * by an external component of the system currently.  In the future this may be
 * attached data that we can own and manage or that is a shared_ptr - The point
 * is that it shouldn't matter to this object as the interface is expecting one
 * of our fas_video library objects FiImage and it always should just be usable
 * within the scope of this functionality.
 *
 * @param imageToAdd    The FiImage containing an attached image data frame.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoAVIStream::AddVideoFrame(std::shared_ptr<FAS_VIDEO::FiImage> imageToAdd)
{
    bool addContextCheck = headerProcessed; // Once header processed need to check AVI context state too.
    bool addImageAttachedCheck = false;     // We are adding so don't check for attached image at top.
    bool addPlaceholderFrame = false;       // This is not a placeholder frame.
    bool addedFrame = false;                // Will only be true if make it all way through without error.

    if( VERBOSE_SAVE_LOG )
    {
        fi_log(LOG_INFO, "%s: runningFrameCount:%d lastFrameAdded:%d.", __PRETTY_FUNCTION__, runningFrameCount, lastFrameAdded);
    }

    // If the processingVideoFrame is true this is an invalid condition that should not occur!
    if (processingVideoFrame)
    {
        fi_log(LOG_ERR, "%s: RE-ENTRY occurred in stream writer!", __PRETTY_FUNCTION__);

        // Go ahead and leave the streamer in its current stream status state
        // and return the re-entry error condition right away.
        return FAS_VIDEO::STREAM_STATUS::STREAM_REENTRY_ERROR;
    }

    // Mark the condition of processing a video frame.
    processingVideoFrame = true;

    // Verify that data can be written to the stream.
    if (!OkToWrite(__PRETTY_FUNCTION__, addContextCheck, addImageAttachedCheck))
        goto exitAddVideoFrame;

    // Assign the imageToAdd to our local smart pointer.
    videoFrameToWrite = imageToAdd;

    // Sanity check: Verify there is a valid pointer to the image data.
    if (!videoFrameToWrite)
    {
        fi_log(LOG_DEBUG, "%s: empty FiImage for stream writer.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitAddVideoFrame;
    }

    // Sanity check:  Make sure the image to add has attached image data.
    if (videoFrameToWrite->GetImageStatus() != FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED)
    {
        fi_log(LOG_DEBUG, "%s: no image data was attached for AVI Stream Writer.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitAddVideoFrame;
    }

    // Sanity check:  Make sure the image to add has a valid image format for an AVI file.
    if ((videoFrameToWrite->GetImageFormat() != FAS_VIDEO::IMAGE_FORMAT::FI_BMP_IMAGE) &&
        (videoFrameToWrite->GetImageFormat() != FAS_VIDEO::IMAGE_FORMAT::FI_JPEG_IMAGE))
    {
        fi_log(LOG_DEBUG, "%s: image data in unsupported image format for AVI Stream Writer.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitAddVideoFrame;
    }

    // If no frame has ever been added set the image format for the first added frame.
    if (firstAddedImageFormat == FAS_VIDEO::IMAGE_FORMAT::FI_INVALID_IMAGE)
    {
        firstAddedImageFormat = videoFrameToWrite->GetImageFormat();
    }
    // Sanity check:  Make sure the frame being added matches the firstAddedImageFormat.
    else if (firstAddedImageFormat != videoFrameToWrite->GetImageFormat())
    {
        fi_log(LOG_DEBUG, "%s: AVI Stream writer received incompatible image type[%d] was expecting[%d].",
               __PRETTY_FUNCTION__, (uint32_t)videoFrameToWrite->GetImageFormat(), (uint32_t)firstAddedImageFormat);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitAddVideoFrame;
    }

    // If make it here without error return has the header been written?
    if (!headerProcessed)
    {
        // Make sure the stream's frame information matches the data
        // in the image.  The stream creation sometimes doesn't know
        // the real information for the frames coming to it.
        //
        // The opportunity to determine real information is now
        // available to reuse for additional AddVideoFrame calls.
        //
        // The header is being initialized right now means that:
        //
        // A) The first frame has been received, attached, and verified.
        // B) The frame information should be valid and reliable.
        //
        uint32_t imageFrameWidth = videoFrameToWrite->GetImageWidth();
        uint32_t imageFrameHeight = videoFrameToWrite->GetImageHeight();
        fullFrameSize = videoFrameToWrite->GetFullFramesize();

        // Make sure that the image has extent information.
        // Note that when an image is "attached" to the FiImage object
        // the frame information will be populated by a query on
        // data in the image file itself.
        if ((imageFrameWidth == 0) || (imageFrameHeight == 0))
        {
            fi_log(LOG_ERR, "%s: Attached image size error width[%lu] height[%lu].",
                   __PRETTY_FUNCTION__, imageFrameWidth, imageFrameHeight);
            streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
            goto exitAddVideoFrame;
        }

        if (frameWidth != imageFrameWidth)
        {
            // What if the frameWidth stored isn't zero?
            // NOTE_JLC - should this be considered an error?
            if (frameWidth != 0)
            {
                fi_log(LOG_DEBUG, "%s: !headerProcessed and frameWidth[%lu] not zero, imageFrameWidth[%lu]",
                       __PRETTY_FUNCTION__, frameWidth, imageFrameWidth);
            }
            frameWidth = imageFrameWidth;
        }
        if (frameHeight != imageFrameHeight)
        {
            // What if the frameHeight stored isn't zero?
            // NOTE_JLC - should this be considered an error?
            if (frameHeight != 0)
            {
                fi_log(LOG_DEBUG, "%s: !headerProcessed and frameHeight[%lu] not zero, imageFrameHeight[%lu]",
                       __PRETTY_FUNCTION__, frameHeight, imageFrameHeight);
            }
            frameHeight = imageFrameHeight;
        }

        // Initialize the AVI header information preparing for AVI file creation.
        if (!InitializeHeader())
        {
            fi_log(LOG_DEBUG, "%s: Unable to initialize AVI header.", __PRETTY_FUNCTION__);
            goto exitAddVideoFrame;
        }
        headerProcessed = true;
    }

    // If our running frame total isn't in sync with our last frame processed we have missing
    // frame(s). Add an AVI frame for each frame missing to sync up and create a valid AVI
    // placement map in preparation for any retry pass later when EditVideoFrame is called.
    if( runningFrameCount > lastFrameAdded )
    {
        for(uint32_t frameId = lastFrameAdded; frameId < runningFrameCount; frameId++)
        {
            fi_log(LOG_DEBUG, "%s: CATCH-UP runningFrameCount:%d lastFrameAdded:%d.", __PRETTY_FUNCTION__, runningFrameCount, lastFrameAdded);
            // Insert a real placeholder frame here!
            if( !InsertPlaceholderFrame() )
            {
                fi_log(LOG_DEBUG, "%s: Unable to add placeholder frame for frameId:%d.", __PRETTY_FUNCTION__, frameId);
                goto exitAddVideoFrame;
            }
        }
    }

    // Call the actual AVI add frame logic to insert this image data frame into avi stream.
    // We know that we have stored the curVideoFrameToAdd so the AVI add function can use it.
    if (!AVIAddFrame(addPlaceholderFrame))
    {
        fi_log(LOG_DEBUG, "%s: Unable to add AVI frame to stream, lastFrameAdded:%d.", __PRETTY_FUNCTION__, lastFrameAdded);
        goto exitAddVideoFrame;
    }

    // This real image frame counts in our running frame count too.
    runningFrameCount++;

    // Make here and we can safely assume the frame has been added successfully to the stream.
    addedFrame = true;

exitAddVideoFrame:
    // If we did not successfully add the frame then make sure to reset it to a nullptr.
    if (!addedFrame)
        videoFrameToWrite = nullptr;
    else
    {
        // Is this stream complete?
        if (frameCount == lastFrameAdded)
        {
            // Is this stream complete?
            if (placeHolderFrames.empty())
            {
                // Are we done writing this stream?
                if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_READY)
                {
                    streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED;
                }
            }
        }
    }

    // Reset the re-entry boolean so that we don't false flag a re-entry condition.
    processingVideoFrame = false;

    return streamStatus;
}

/*!
 * @brief As long as we have received a valid frame this inserts a placeholder
 * spot in the AVI file being created. It will set stream status to error condition
 * if there is an issue.
 */
bool FiVideoAVIStream::InsertPlaceholderFrame()
{
    bool addedEmptyFrame = false;

    if (firstAddedImageFormat != FAS_VIDEO::IMAGE_FORMAT::FI_INVALID_IMAGE)
    {
        // Set the size of a placeholder frame.  We can assume if make it here
        // in the scope one frame was successfully added and we have a full framesize
        // value to lock into place for the placeholder frame.
        uint32_t frameSize = fullFrameSize;

        // If the size is zero will not be able to insert a frame.
        if (frameSize == 0)
        {
            fi_log(LOG_WARNING, "%s: unable to calculate frameSize for placeholder frame.", __PRETTY_FUNCTION__);
            streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
            return addedEmptyFrame;
        }

        // This is an empty addition to our stream.
        addedEmptyFrame = AVIAddFrame(true);
    }
    return addedEmptyFrame;
}

/*!
 * @brief Inserts an empty frame into AVI header and advances seek pointer
 * for an image frame to prepare for a AddRetryFrame call when frame retry
 * is attempted by the user of this stream object. If no valid frames have
 * been received yet it just increments our runningFrameCount and moves on.
 * The first added frame with real data checks to see if it needs to add any
 * placeholder frames to catch up.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoAVIStream::AddPlaceholderFrame()
{
    bool addContextCheck = headerProcessed;     // Once header processed need to check AVI context state too.
    bool addImageAttachedCheck  = false;        // No image being added nothing to check.

    fi_log(LOG_DEBUG, "%s: runningFrameCount:%d lastFrameAdded:%d.", __PRETTY_FUNCTION__, runningFrameCount, lastFrameAdded);

    // Verify that data can be written to the stream.
    if (!OkToWrite(__PRETTY_FUNCTION__, addContextCheck, addImageAttachedCheck))
        goto exitAddPlaceholderFrame;

    // Increase our running frame count as this is the next valid frame in our stream.
    runningFrameCount++;

    // Attempt to add the frame - it won't add it yet if no valid images have been received for the stream.
    InsertPlaceholderFrame();

exitAddPlaceholderFrame:
    return streamStatus;
}

/*!
 * @brief Insert the real frame of an image into the created AVI stream at a
 * specified frameNumber representing a placeholder frame location.
 *
 * Note:  Inside of the imageDataFrame right now is attached image data frame
 * in the form of uint8_t* and a length.  This external data pointer is managed
 * by an external component of the system currently.  In the future this may be
 * attached data that we can own and manage or that is a shared_ptr - The point
 * is that it shouldn't matter to this object as the interface is expecting one
 * of our fas_video library objects FiImage and it always should just be usable
 * within the scope of this functionality.
 *
 * @param frameNumber:  Zero-based frame number for the placeholder frame to edit.
 * @param imageDataFrame:  The image frame data to edit in place.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoAVIStream::EditVideoFrame(uint32_t frameNumber, std::shared_ptr<FAS_VIDEO::FiImage> imageDataFrame)
{
    off64_t offset;
    size_t size;
    uint32_t frameSize;

    bool addContextCheck = headerProcessed;     // Once header processed need to check AVI context state too.
    bool addImageAttachedCheck = false;         // Not checking attachment of normal add frame.

    // This will be manipulated if writing a bitmap frame.
    uint32_t imageLenToWrite = 0;

    // This will contain the frame data minus any header information.
    const uint8_t *memDataPtr = nullptr;

    // This will contain vertical transposition (flipped top to bottom)
    // if writing bitmap frame.
    std::unique_ptr<FAS_VIDEO::FiFrame> normalizedFrame;

    // If the processingVideoFrame is true this is an invalid condition that should not occur!
    if (processingVideoFrame)
    {
        fi_log(LOG_ERR, "%s: RE-ENTRY occurred in stream writer!", __PRETTY_FUNCTION__);

        // Go ahead and leave the streamer in its current stream status state
        // and return the re-entry error condition right away.
        return FAS_VIDEO::STREAM_STATUS::STREAM_REENTRY_ERROR;
    }

    // Set the processingVideoFrame true.
    processingVideoFrame = true;

    // Verify that data can be written to the stream.
    if (!OkToWrite(__PRETTY_FUNCTION__, addContextCheck, addImageAttachedCheck))
        goto exitEditVideoFrame;

    // Sanity check:  imageDataFrame should not be nullptr.
    if (!imageDataFrame)
    {
        fi_log(LOG_DEBUG, "%s: missing FiImage for editing frame[%u] in AVI stream writer.",
               __PRETTY_FUNCTION__, frameNumber);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitEditVideoFrame;
    }

    // Sanity check:  imageDataFrame should have attached image data.
    if (imageDataFrame->GetImageStatus() != FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED)
    {
        fi_log(LOG_DEBUG, "%s: missing attached image data in FiImage for editing frame[%u] in AVI stream writer.",
               __PRETTY_FUNCTION__, frameNumber);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitEditVideoFrame;
    }

    // Pull the framesize from the image.
    frameSize = imageDataFrame->GetFullFramesize();
    if (frameSize == 0)
    {
        fi_log(LOG_DEBUG, "%s: attached image returned frameSize=0 for frame_num %d.", __PRETTY_FUNCTION__, frameNumber);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitEditVideoFrame;
    }

    offset = AVILookupImageOffset(frameNumber, &size);
    if (offset == 0)
    {
        fi_log(LOG_ERR, "%s: Unable to calculate frame offset for frame: %d.", __PRETTY_FUNCTION__, frameNumber);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitEditVideoFrame;
    }

    // Pull out the image data frame information.
    memDataPtr = imageDataFrame->GetAttachedFrameData(imageLenToWrite, FAS_VIDEO::FRAME_ORIENTATION::BOTTOM_UP);

    // Sanity check:  Make sure image length and pointer are valid.
    if ((imageLenToWrite == 0) || (memDataPtr == nullptr))
    {
        fi_log(LOG_DEBUG, "%s: invalid image frame data provided to edit retry for frame_num %d.",
               __PRETTY_FUNCTION__, frameNumber);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitEditVideoFrame;
    }

    // Sanity check:  Was there enough room saved during the placeholder add?
    if (size < imageLenToWrite)
    {
        fi_log(LOG_ERR, "%s: lookup for frame %d, size=%lu but length=%u.",
               __PRETTY_FUNCTION__, frameNumber, size, imageLenToWrite);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitEditVideoFrame;
    }

    // Record where this is going in the file.
#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "%s: frame %d, offset=%lu avail=%lu imglen=%lu.",
           __PRETTY_FUNCTION__, frameNumber, offset, size, imageLenToWrite);
#endif

    // Seek to placeholder frame position.
    if (!SetSeekPos(offset, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET))
    {
        fi_log(LOG_ERR, "%s: unable to set the seek offset to %u for frame %d.",
               __PRETTY_FUNCTION__, offset, frameNumber);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
        goto exitEditVideoFrame;
    }

    // Overwrite the frame data into the placeholder position.
    if (!WriteData((const void *)memDataPtr, imageLenToWrite))
    {
        fi_log(LOG_ERR, "(%s) Unable to write frame to AVI file.", __PRETTY_FUNCTION__);
    }

    // Reseek to the end of the file.
    if (!SetSeekPos(0, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_END))
    {
        fi_log(LOG_ERR, "%s: unable to set the seek offset to eof for frame %d.",
               __PRETTY_FUNCTION__, offset, frameNumber);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_STREAM_IMAGE;
    }

exitEditVideoFrame:

    // Remove the placeholder frame from our stored list.
    {
        std::lock_guard<std::mutex> guard(placeHolderFramesMutex);
        uint32_t theSize = (uint32_t)placeHolderFrames.size();

        // Run through the number of entries that are in the list.
        for (uint32_t theIndex = 0; theIndex < theSize; theIndex++)
        {
            uint32_t frameId = placeHolderFrames.front();
            placeHolderFrames.pop_front();
            if (frameId == frameNumber)
            {
                fi_log(LOG_DEBUG, "%s frameId:%d POPPED placeHolderFrames.size()=%d",
                       __PRETTY_FUNCTION__, frameId, placeHolderFrames.size());
                break;
            }
            else
            {
                placeHolderFrames.push_back(frameId);
                fi_log(LOG_DEBUG, "%s frameId:%d RESTORED placeHolderFrames.size()=%d",
                       __PRETTY_FUNCTION__, frameId, placeHolderFrames.size());
            }
        }
    }

    // Is this stream complete?
    if (placeHolderFrames.empty())
    {
        // Are we done writing this stream?
        if (frameCount == lastFrameAdded)
        {
            if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_READY)
            {
                streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED;
            }
        }
    }

    // Reset the processingVideoFrame to false.
    processingVideoFrame = false;

    return streamStatus;
}

/*!
 * @brief Queries complete state of this stream.
 *
 * @returns The status of the stream writer.  If it is complete
 * without error the status should be FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED.
 */
FAS_VIDEO::STREAM_STATUS FiVideoAVIStream::IsStreamWriterComplete()
{
    // Make sure this is a stream writer otherwise this makes no sense.
    if (streamType != FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER)
    {
        // Note that this is only a status being returned not the
        // status set in the stream object.
        return FAS_VIDEO::STREAM_STATUS::STREAM_TYPE_ERROR;
    }

    // For the AVI stream once the encoded frames count and frame count are equal
    // test for placeholder frames in case they exist.
    if (GetFrameCount() == GetFramesEncodedCount())
    {
        // If still in a ready state check for placeholder frames.
        if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_READY)
        {
            // Are we waiting on placeholder frames?  The only reason for this
            // to not be empty is a stream waiting for more placeholder frames.
            if (placeHolderFrames.empty())
            {
                // We are complete.
                streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED;
            }
            else
            {
                // Note that this is only a status being returned not the
                // status set in the stream object.
                return FAS_VIDEO::STREAM_STATUS::PLACEHOLDER_WAIT;
            }
        }
    }

    // At this point return the actual stream status.
    return streamStatus;
}

/*!
 * @brief Loads a frame from AVI stream into internal frame object.
 *
 * @param frameNum:  Which frame is requested.
 *
 * @returns The video stream status setting of the object.
 */
std::unique_ptr<FAS_VIDEO::FiImage> FiVideoAVIStream::LoadVideoFrame(uint32_t frameNum)
{
    bool addContextCheck = true; // Should have AVI video context record.
    uint32_t headerSize = 0;
    off64_t frameOffset = -1;
    size_t imageSize = 0;
    uint32_t imageDataSize = 0;

    std::unique_ptr<FAS_VIDEO::IMAGE_DATA> memDataPtr;        // The vector containing the IMAGE_DATA read from stream.
    std::unique_ptr<FAS_VIDEO::FiImagePkg> imageDataPkg;        // The image data package wrapping the memDataPtr.
    std::unique_ptr<FAS_VIDEO::FiImage> imageLoaded;            // The loaded image object.

    // Create a pointer to navigate through IMAGE_DATA vector array.
    uint8_t *destPtr;

    FI_BMP::BMP_HEADER bmp_head;

    // Verify that stream is readable.
    if (!OkToRead(__PRETTY_FUNCTION__, addContextCheck))
    {
        goto exitLoadVideoFrame;
    }

    // Does the AVI file contain DIB bitmap images?
    if (firstAddedImageFormat == FAS_VIDEO::IMAGE_FORMAT::FI_BMP_IMAGE)
    {
        // Build up the bitmap header information.
        headerSize = (uint32_t)(sizeof(FI_BMP::BMP_HEADER) + strlen(FI_BMP::BMP_MAGIC));

        if (aviContext->aviCtxPtr->avi_strf->depth == 8)
            headerSize += FI_BMP::DIB_COLOR_TABLE_SIZE;

        bmp_head.data_offset = headerSize;
        bmp_head.reserved[0] = 0;
        bmp_head.reserved[1] = 0;
        bmp_head.reserved[2] = 0;
        bmp_head.reserved[3] = 0;
        bmp_head.length = aviContext->aviCtxPtr->avi_strf->data_size + headerSize;

        imageLoaded = std::unique_ptr<FAS_VIDEO::FiImage>(new FI_BMP::FiBMP());
#ifdef USE_VERBOSE_INFO_DEBUG
        fi_log(LOG_DEBUG, "%s: Loaded FI_BMP Image for frame[%u].", __PRETTY_FUNCTION__, frameNum);
#endif
    }
    if (firstAddedImageFormat == FAS_VIDEO::IMAGE_FORMAT::FI_JPEG_IMAGE)
    {
        imageLoaded = std::unique_ptr<FAS_VIDEO::FiImage>(new FI_JPEG::FiJPEG());
#ifdef USE_VERBOSE_INFO_DEBUG
        fi_log(LOG_DEBUG, "%s: Loaded FI_JPEG Image for frame[%u].", __PRETTY_FUNCTION__, frameNum);
#endif
    }

    // Must have a curVideoFrameRead object to hold the loaded image data.
    if (!imageLoaded)
    {
        fi_log(LOG_ERR, "%s: Unable to create FiImage for frame[%u].", __PRETTY_FUNCTION__, frameNum);
        streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
        goto exitLoadVideoFrame;
    }

    // Find location of the image in the AVI file stream.
    frameOffset = FiVideoAVIStream::AVILookupImageOffset(frameNum, &imageSize);
    if (frameOffset == 0)
    {
        fi_log(LOG_ERR, "%s: Unable to locate frameOffset for frame[%u].", __PRETTY_FUNCTION__, frameNum);
        streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
        goto exitLoadVideoFrame;
    }

    // Create an image data unique pointer with a vector array containing
    // the header and the image data together to attach to stream for use.
    imageDataSize = (uint32_t)(imageSize + headerSize);
    memDataPtr = std::unique_ptr<FAS_VIDEO::IMAGE_DATA>(new FAS_VIDEO::IMAGE_DATA(static_cast<uint64_t>(imageDataSize), 0));

    // If the memDataPtr could not be created report the error.
    if (!memDataPtr)
    {
        fi_log(LOG_ERR, "%s: Unable to create memDataPtr for frame[%u].", __PRETTY_FUNCTION__, frameNum);
        streamStatus = FAS_VIDEO::STREAM_STATUS::ALLOCATION_FAILED;
        goto exitLoadVideoFrame;
    }

    // Provide a destination pointer to move through vector data.
    destPtr = memDataPtr->data();

    // If has a BMP header populate that first.
    if ((headerSize > 0) && firstAddedImageFormat == FAS_VIDEO::IMAGE_FORMAT::FI_BMP_IMAGE)
    {
        memcpy(destPtr, FI_BMP::BMP_MAGIC, strlen(FI_BMP::BMP_MAGIC));
        destPtr += strlen(FI_BMP::BMP_MAGIC);
        headerSize -= strlen(FI_BMP::BMP_MAGIC);

        memcpy(destPtr, &bmp_head, sizeof(FI_BMP::BMP_HEADER) - sizeof(FI_BMP::DIB_HEADER));
        destPtr += sizeof(FI_BMP::BMP_HEADER) - sizeof(FI_BMP::DIB_HEADER);
        headerSize -= sizeof(FI_BMP::BMP_HEADER) - sizeof(FI_BMP::DIB_HEADER);

        memcpy(destPtr, aviContext->aviCtxPtr->avi_strf, headerSize);
        destPtr += headerSize;
    }

    // Seek to the position in the file for the frame.
    if (!SetSeekPos(frameOffset, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET))
    {
        fi_log(LOG_ERR, "%s: Unable to set seek position[%ld] for frame[%u].", __PRETTY_FUNCTION__, frameOffset, frameNum);
        streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
        goto exitLoadVideoFrame;
    }

    // Read the frame data into the destination pointer.
    if (!ReadData(destPtr, (uint32_t)imageSize))
    {
        fi_log(LOG_ERR, "%s: Unable to read position[%ld] for imageSize[%u] for frame[%u].",
               __PRETTY_FUNCTION__, frameOffset, imageSize, frameNum);
        streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
        goto exitLoadVideoFrame;
    }

    // Create an image data package wrapping the image data just read.
    imageDataPkg = FiImagePkg::Create(std::move(memDataPtr));

    // Attach the image data package to the image object.
    if (imageLoaded->AttachImagePkg(std::move(imageDataPkg)) != FAS_VIDEO::IMAGE_STATUS::IMAGE_ATTACHED)
    {
        fi_log(LOG_ERR, "%s: Unable to attach imageDataPkg for frame[%u].",
               __PRETTY_FUNCTION__, frameNum);
        streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
        goto exitLoadVideoFrame;
    }

exitLoadVideoFrame:
    return imageLoaded;
}

/*!
 * @brief Closes the video stream and releases any allocated resources.
 *
 * @returns The video stream status setting of the object.
 */
FAS_VIDEO::STREAM_STATUS FiVideoAVIStream::CloseStream()
{
    std::string myStreamStatusString;
    if (VERBOSE_SAVE_LOG)
    {
        LogStreamStatus(myStreamStatusString, false);
        fi_log(LOG_DEBUG, "%s: %s Before AVICleanup()", __PRETTY_FUNCTION__, myStreamStatusString);
    }

    // Just call the cleanup routine to release any memory allocated
    // the old fashioned way for the AVI writer.
    // NOTE_JLC - It would be nice to make a pass through the old AVI
    // C code and cleanup the traditional C memory allocations replacing
    // with C++ vector objects as unique pointers.
    AVICleanup();

    // Close the std::fstream AVI File I/O stream object.
    if (ioFileStreamPtr)
        ioFileStreamPtr->close();

    if (VERBOSE_SAVE_LOG)
    {
        LogStreamStatus(myStreamStatusString, false);
        fi_log(LOG_DEBUG, "%s: %s After AVICleanup()", __PRETTY_FUNCTION__, myStreamStatusString);
    }

    // If this is a stream writer check for completion.
    if (streamType == FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER)
    {
        // Did this stream complete?
        bool looksLikeWeAreCompleted = false;
        if (VERBOSE_SAVE_LOG)
        {
            fi_log(LOG_DEBUG, "%s: frameCount[%d] lastFrameAdded[%d] placeHolderFrames.size()[%d].",
                   __PRETTY_FUNCTION__, frameCount, lastFrameAdded, placeHolderFrames.size());
        }

        // Determine if we actually completed the AVI file.
        if (frameCount == lastFrameAdded)
        {
            // Is this stream complete?
            if (placeHolderFrames.empty())
            {
                if (VERBOSE_SAVE_LOG)
                {
                    LogStreamStatus(myStreamStatusString, false);
                    fi_log(LOG_DEBUG, "%s: %s AVI Video should be complete.", __PRETTY_FUNCTION__, myStreamStatusString);
                }
                // Make sure our status reflects our complete state.
                if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_READY ||
                    streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED)
                {
                    // Should not happen but if it does catch it.
                    if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_READY)
                    {
                        streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED;
                    }
                    looksLikeWeAreCompleted = true;
                }
            }
        }

        // Handle marking this video as incomplete if necessary.
        if (!looksLikeWeAreCompleted)
        {
            // Make sure this stream is marked incomplete if still ready.
            if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_READY)
            {
                streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_INCOMPLETE;
            }

            std::string videoNameWithErrorTag = videoNameOfStream + FiVideoStream::GetIncompleteVideoFileString();
            std::string statusLog;
            LogStreamStatus(statusLog, false);

            fi_log(LOG_ERR, "%s: AVI streamStatus is not marked complete.  Writing %s to document incomplete status.", __PRETTY_FUNCTION__, videoNameWithErrorTag.c_str());
            // using ofstream constructors.
            std::ofstream outfile(videoNameWithErrorTag.c_str());
            outfile << "AVI video creation is incomplete and was not fully processed." << std::endl;
            outfile << "Please review log data for possible error conditions." << std::endl;
            outfile << "Status:" << statusLog << std::endl;
            outfile.close();
        }
    }

    // Call the base class' CloseStream to set the stream status closed.
    FiVideoStream::CloseStream();

    // The stream status of the object should be FAS_VIDEO::STREAM_STATUS::STREAM_CLOSED
    return streamStatus;
}

/*!
 * @brief Closing the video stream for AVI file either by error or by completion.
 *
 * Close the file and release any resources.  Make sure not to use video_context
 * for anything after calling this function because the pointer will be free'd.
 *
 * NOTE_JLC - Would like to make all this allocated memory become "smart" and
 * "unique" so it can clean itself up and protect against overruns etc.
 *
 * @returns The boolean true value if cleanup was successful, false if error.
 */
bool FiVideoAVIStream::AVICleanup()
{
    bool cleanupStatus = false;
    OFFSET_REC_LIST *rec_ptr;

    // If there is an AVI Context object free up any internal allocations.
    if (aviContext)
    {
        // Pop a message in the log about cleaning up and dump the status.
        fi_log(LOG_DEBUG, "%s: Cleaning AVI Context Data.", __PRETTY_FUNCTION__);
        LogStreamStatus();

        // Make sure the internally allocated aviCtxPtr exists.
        if (aviContext->aviCtxPtr != nullptr)
        {
            // Walk through the record list and delete all the entries.
            rec_ptr = aviContext->aviCtxPtr->rec_head;
            while (rec_ptr != nullptr)
            {
                OFFSET_REC_LIST *next_ptr;

                // Remember where we going.
                next_ptr = rec_ptr->next;

                // Release this record.
                free(rec_ptr);

                // Move on.
                rec_ptr = next_ptr;
            }

            // Terminate the head.
            aviContext->aviCtxPtr->rec_head = nullptr;

            if (aviContext->aviCtxPtr->avi_dmlh != nullptr)
            {
                free(aviContext->aviCtxPtr->avi_dmlh);
                aviContext->aviCtxPtr->avi_dmlh = nullptr;
            }

            if (aviContext->aviCtxPtr->avi_header != nullptr)
            {
                free(aviContext->aviCtxPtr->avi_header);
                aviContext->aviCtxPtr->avi_header = nullptr;
            }

            if (aviContext->aviCtxPtr->avi_strf != nullptr)
            {
                free(aviContext->aviCtxPtr->avi_strf);
                aviContext->aviCtxPtr->avi_strf = nullptr;
            }

            if (aviContext->aviCtxPtr->avi_strh != nullptr)
            {
                free(aviContext->aviCtxPtr->avi_strh);
                aviContext->aviCtxPtr->avi_strh = nullptr;
            }

            if (aviContext->aviCtxPtr->avi_indx != nullptr)
            {
                aviContext->aviCtxPtr->frame_index = nullptr;
                free(aviContext->aviCtxPtr->avi_indx);
                aviContext->aviCtxPtr->avi_indx = nullptr;
            }

            // Release the aviCtxPtr itself.
            free(aviContext->aviCtxPtr);
            aviContext->aviCtxPtr = nullptr;
        }

        // Finally, release the avi context object.
        aviContext.release();
        cleanupStatus = true;
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: AVI Context Data Empty.", __PRETTY_FUNCTION__);
    }
    return cleanupStatus;
}

/*!
 * @brief Read and intrepret an AVI List tag in AVI header.
 *
 * Recursive support to scan and intrepret an AVI List tag.
 *
 * @param sizeToRead: Number of uint8_t bytes to read.
 *
 * @return Number of uint8_t bytes read in this method.
 */
uint32_t FiVideoAVIStream::AVIReadList(uint32_t sizeToRead)
{
    uint32_t retTotalBytes = 0;

    fcc_head h;
    size_t bytesRead = 0;
    uint32_t list_id;

    bool addContextCheck = true;

#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "(%s) AVIReadList sizeToRead[%u].",
           __PRETTY_FUNCTION__, sizeToRead);
#endif

    if (!OkToRead(__PRETTY_FUNCTION__, addContextCheck))
    {
        goto exitAVIReadList;
    }

    if (!ReadData(&list_id, sizeof(list_id)))
    {
        fi_log(LOG_ERR, "(%s) Unable to read AVI List Id Value.", __PRETTY_FUNCTION__);
        goto exitAVIReadList;
    }

    // Number of bytes in the list ID.
    bytesRead = sizeof(list_id);

    // Update the total amount read.
    retTotalBytes += bytesRead;

    // Work through the list.
    while (bytesRead && (retTotalBytes < sizeToRead))
    {
        // Read the fcc_head data from AVI file.
        if (!ReadData(&h, sizeof(fcc_head)))
        {
            fi_log(LOG_WARNING, "(%s) Expected more data in AVI list failed to read fcc_head.", __PRETTY_FUNCTION__);
            break;
        }

        // Update the total amount read.
        retTotalBytes += sizeof(fcc_head);

        // Verify that the list tag is in the AVI header data.
        if (h.fcc == LIST_TAG)
        {
            // Recurse to read the next list data tag.
            bytesRead = AVIReadList(h.size);
        }
        else
        {
            // Not a list so read the chunk data.
            bytesRead = AVIReadChunk(h.fcc, h.size);
        }

        // Update the total amount read.
        retTotalBytes += bytesRead;

        // Verify we read what we expect to read.
        if (bytesRead != h.size)
        {
            fi_log(LOG_WARNING, "(%s) Expected [%d] bytes in %4.4s, bytesRead=[%d].",
                   __PRETTY_FUNCTION__, (int)h.size, (char *) & (h.fcc), (int)bytesRead);
            goto exitAVIReadList;
        }
    }

exitAVIReadList:
    return retTotalBytes;
}

/*!
 * @brief Read and intrepret an AVI Chunk of Index data in AVI header.
 *
 * Read support to scan and intrepret an AVI index data.
 *
 * @param fccId: The FourCC Tag Id to interpret.
 * @param sizeToRead: Number of uint8_t bytes to read.
 *
 * @return Number of uint8_t bytes read in this method.
 */
uint32_t FiVideoAVIStream::AVIReadChunk(uint32_t fccId, uint32_t sizeToRead)
{
    uint8_t *buffer = nullptr;
    uint32_t bytesInChunk = 0;

#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "(%s) AVIReadChunk fccId[%4.4s] sizeToRead[%u].",
           __PRETTY_FUNCTION__, (char *) & (fccId), sizeToRead);
#endif

    bool addContextCheck = true;
    if (!OkToRead(__PRETTY_FUNCTION__, addContextCheck))
    {
        goto exitAVIReadChunk;
    }

    switch (fccId)
    {
    case AVIH_TAG:
        buffer = (uint8_t *)malloc(sizeToRead);
        aviContext->aviCtxPtr->avi_header = (avih *)buffer;
        break;
    case STRH_TAG:
        buffer = (uint8_t *)malloc(sizeToRead);
        aviContext->aviCtxPtr->avi_strh = (strh *)buffer;
        break;
    case STRF_TAG:
        buffer = (uint8_t *)malloc(sizeToRead);
        aviContext->aviCtxPtr->avi_strf = (FI_BMP::DIB_HEADER *)buffer;
        break;
    case INDX_TAG:
        buffer = (uint8_t *)malloc(sizeToRead);
        aviContext->aviCtxPtr->avi_indx = (indx *)buffer;
        aviContext->aviCtxPtr->frame_index = (indx_union *)(buffer + sizeof(indx));
        break;
    case DMLH_TAG:
        buffer = (uint8_t *)malloc(sizeToRead);
        aviContext->aviCtxPtr->avi_dmlh = (dmlh *)buffer;
        break;
    case IDX1_TAG:
        if (aviContext->aviCtxPtr->avi_indx == nullptr)
        {
            // TODO: build 2.0 index from old IDX1 index.
            fi_log(LOG_ERR, "(%s) AVI 1.0 index in use!", __PRETTY_FUNCTION__);
        }
        break;
    default:
#ifdef USE_VERBOSE_INFO_DEBUG
        fi_log(LOG_DEBUG, "(%s) Ignoring fccId[%4.4s]!", __PRETTY_FUNCTION__, (char *) & (fccId));
#endif
        break;
    }

    // read the data, or skip it
    if (buffer != nullptr)
    {
        if (!ReadData(buffer, sizeToRead))
        {
            fi_log(LOG_ERR, "(%s) Failed to read %4.4s chunk.", __PRETTY_FUNCTION__, (char *) & (fccId));
        }
        else
        {
            // If come across another index scan it.
            if (fccId == INDX_TAG)
            {
                // Scan index entries (it internally maintains current seek postition).
                if (!ScanIndex())
                {
                    fi_log(LOG_ERR, "(%s) Failed to scan avi 2.0 index.", __PRETTY_FUNCTION__);
                }
            }
        }
    }
    else
    {
        if (!SetSeekPos(sizeToRead, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_CUR))
        {
            fi_log(LOG_ERR, "(%s) Unable to set seek position past %4.4s chunk.", __PRETTY_FUNCTION__, (char *) & (fccId));
            goto exitAVIReadChunk;
        }
    }

    // Update our return value.
    bytesInChunk = sizeToRead;

exitAVIReadChunk:
    return bytesInChunk;
}

/*!
 * @brief Begin a new RIFF-LIST in an avi file.
 *
 * An AVI file contains a sequence of RIFF-LISTs.  This function starts a new
 * RIFF list and keeps track of the staring point so that it's length can be
 * updated when we are done.
 *
 * @param video_context: a video context created from fi_avi_default_context()
 *
 * @return Boolean true if successfully started Riff, false if error.
 */
bool FiVideoAVIStream::AVIStartRiff()
{
    bool startedRiff = false;
    bool addContextCheck = true;
    bool addImageAttachedCheck = true;
    uint32_t frameSize = 0;

    uint32_t i;                 // General purpose loop variable
    uint8_t color_entry[4];     // Used for writing a gray scale color table
    uint32_t indx_size;     // Used to calculate the size of the indx structure plus it's entries
    riff_list lst;          // riff list to be filled and written when adding a list.
    hdrl_list hdrl;         // The hdrl avi header structure
    indx_chunk indx;        // Used for writing the indx chunk
    indx_entry indx_e;      // Used to write empty data as index entries to be filled in later

    // Set to 0 later if we determine we don't need a color table.
    uint32_t color_table_size = FI_BMP::DIB_COLOR_TABLE_SIZE;

    // Make sure the avi stream is ok for writing.
    if (!OkToWrite(__PRETTY_FUNCTION__, addContextCheck, addImageAttachedCheck))
        goto exitAVIStartRiff;

    // Pop a message in the log about cleaning up and dump the status.
#ifdef USE_VERBOSE_INFO_DEBUG
    fi_log(LOG_DEBUG, "%s: Starting a Riff record.", __PRETTY_FUNCTION__);
#endif

    // Pull a local copy of the framesize.
    frameSize = videoFrameToWrite->GetFullFramesize();

    // Write the initial RIFF-AVI or subsequent RIFF-AVIX lists
    lst.h.fcc = RIFF_TAG;
    lst.h.size = 0;
    aviContext->aviCtxPtr->riff_offset = curFileOffset;

    // The take we use depends on if this is the first RIFF
    // or subsequent RIFFs. The first one is special because
    // it is supported by legacy apps - AVI 1.0 compliant.
    if (aviContext->aviCtxPtr->current_riff == 0)
        lst.list_id = AVI_TAG;
    else
        lst.list_id = AVIX_TAG;

    // Write the riff list identifier
    if (!WriteData(&lst, sizeof(lst)))
        goto exitAVIStartRiff;

    // If we are starting the first riff, then we need the AVI header info.
    if (aviContext->aviCtxPtr->current_riff == 0)
    {
        // No color table for RGB input pixel formats because they
        // contain their color values within the frame data.
        if (aviContext->aviCtxPtr->raw_format == 24)
        {
            color_table_size = 0;
            fi_log(LOG_DEBUG, "%s: A color table will NOT be written as bpp=24.", __PRETTY_FUNCTION__);
        }
        else
        {
            fi_log(LOG_DEBUG, "%s: A color table WILL be written as bpp=[%u].", __PRETTY_FUNCTION__, aviContext->aviCtxPtr->raw_format);
        }


        // The INDX chunk has an index entry for each RIFF.
        indx_size = sizeof(indx_chunk) + (sizeof(indx_entry) * aviContext->aviCtxPtr->riff_count);

        // Setup the AVI header info.
        memset(&hdrl, 0, sizeof(hdrl));

        hdrl.l.h.fcc = LIST_TAG;
        hdrl.l.h.size = (sizeof(hdrl) - sizeof(fcc_head)) + color_table_size + indx_size;
        hdrl.l.list_id = HDRL_TAG;

        // Now fill in AVIH chunk of header
        hdrl.avih_data.h.fcc = AVIH_TAG;
        hdrl.avih_data.h.size = sizeof(avih_chunk) - sizeof(fcc_head);
        hdrl.avih_data.data.frame_rate = AVI_FRAME_CLOCK_RATE / frameRate;/**/
        hdrl.avih_data.data.max_bps = 0;
        hdrl.avih_data.data.total_frames = frameCount;

        hdrl.avih_data.data.streams = 1;
        hdrl.avih_data.data.width = frameWidth;
        hdrl.avih_data.data.height = frameHeight;
        hdrl.avih_data.data.suggested_buffer_size = 0;
        hdrl.avih_data.data.flags = AVIF_HAS_INDEX;

        // Now fill in OMDL list
        hdrl.odml_data.l.h.fcc = LIST_TAG;
        hdrl.odml_data.l.h.size = sizeof(odml_list) - sizeof(fcc_head);
        hdrl.odml_data.l.list_id = ODML_TAG;
        hdrl.odml_data.data.h.fcc = DMLH_TAG;
        hdrl.odml_data.data.h.size = sizeof(dmlh_chunk) - sizeof(fcc_head);
        hdrl.odml_data.data.dmlh_data.total_number_frames = hdrl.avih_data.data.total_frames;

        // Now fill in STRL list
        hdrl.strl_data.l.h.fcc = LIST_TAG;
        hdrl.strl_data.l.h.size = (sizeof(strl_list) - sizeof(fcc_head)) + color_table_size + indx_size;
        hdrl.strl_data.l.list_id = STRL_TAG;

        // fill in STRH chunk
        hdrl.strl_data.strh_data.h.fcc = STRH_TAG;
        hdrl.strl_data.strh_data.h.size = sizeof(strh_chunk) - sizeof(fcc_head);
        hdrl.strl_data.strh_data.data.fcc_type = VIDS_TAG;

        // Fill in some codec information
        if (firstAddedImageFormat == FAS_VIDEO::IMAGE_FORMAT::FI_BMP_IMAGE)
        {
            hdrl.strl_data.strh_data.data.fcc_handler = DIB_TAG;
        }
        else
        {
            hdrl.strl_data.strh_data.data.fcc_handler = MJPG_TAG;
            hdrl.strl_data.strf.dib.compression = MJPG_TAG;
        }

        // Some video parameters
        hdrl.strl_data.strh_data.data.scale = (uint32_t) 1;
        hdrl.strl_data.strh_data.data.rate = (uint32_t) frameRate;
        hdrl.strl_data.strh_data.data.length = (uint32_t)frameCount;
        hdrl.strl_data.strh_data.data.suggested_buffer_size = (uint32_t) frameSize;
        hdrl.strl_data.strh_data.data.right = (uint16_t) frameWidth;
        hdrl.strl_data.strh_data.data.bottom = (uint16_t) frameHeight;
        hdrl.strl_data.strh_data.data.quality = (uint32_t) -1U;

        // Fill in strf chunk, more video parameters and sizes.
        hdrl.strl_data.strf.h.fcc = STRF_TAG;
        hdrl.strl_data.strf.dib.data_header_size = sizeof(strf_chunk) - sizeof(fcc_head);
        hdrl.strl_data.strf.h.size = hdrl.strl_data.strf.dib.data_header_size + color_table_size;
        hdrl.strl_data.strf.dib.width = (uint16_t)frameWidth;
        hdrl.strl_data.strf.dib.height = (uint16_t)frameHeight;
        hdrl.strl_data.strf.dib.color_planes = 1;
        hdrl.strl_data.strf.dib.data_size = frameSize;
        if (color_table_size)
        {
            hdrl.strl_data.strf.dib.depth = 8;
            hdrl.strl_data.strf.dib.actual_colors_used = 256;
        }
        else
        {
            hdrl.strl_data.strf.dib.depth = 24;
            hdrl.strl_data.strf.dib.actual_colors_used = 0;
        }

        // Write the AVI header
        if (!WriteData(&hdrl, sizeof(hdrl)))
            goto exitAVIStartRiff;

        // If we need a color table, the write it out now.
        if (color_table_size)
        {
            for (i = 0; i < (FI_BMP::DIB_COLOR_TABLE_SIZE / sizeof(uint32_t)); i++)
            {
                color_entry[0] = (uint8_t)i;
                color_entry[1] = (uint8_t)i;
                color_entry[2] = (uint8_t)i;
                color_entry[3] = 0;

                if (!WriteData(color_entry, 4))
                    goto exitAVIStartRiff;
            }
        }

        // Save start of INDX for updating later.

        aviContext->aviCtxPtr->indx_offset = curFileOffset;

        // Now, fill in the INDX section, this is still in the STRL.
        memset(&indx, 0, sizeof(indx));

        indx.h.fcc = INDX_TAG;
        indx.h.size = indx_size - sizeof(fcc_head);
        indx.i.longs_per_entry = 4;
        indx.i.sub_type = 0;
        indx.i.type = AVI_INDEX_OF_INDEXES;
        indx.i.entries_in_use = aviContext->aviCtxPtr->riff_count;
        indx.i.chunk_id = VIDEO_CHUNK_ID;

        if (!WriteData(&indx, sizeof(indx)))
            goto exitAVIStartRiff;

        memset(&indx_e, 0, sizeof(indx_e));

        // Save space for empty INDX entries, one per RIFF.
        for (i = 0 ; i < aviContext->aviCtxPtr->riff_count; i++)
        {
            // True values will be updated later.
            if (!WriteData(&(indx_e), sizeof(indx_entry)))
                goto exitAVIStartRiff;
        }
    }

    // Write the MOVI-LIST.
    lst.h.fcc = LIST_TAG;
    lst.h.size = 0;
    lst.list_id = MOVI_TAG;

    // Save the offset so we can update the file later.
    aviContext->aviCtxPtr->movi_offset = curFileOffset;

    // Write the movi LIST identifier.
    if (!WriteData(&lst, sizeof(lst)))
        goto exitAVIStartRiff;

    // Now we are ready to write video data.
    startedRiff = true;

exitAVIStartRiff:
    return startedRiff;
}


/*
 * @brief end or close out the current RIFF list by updating sizes, etc...
 *
 * @param video_context: a video context created from fi_avi_default_context()
 *
 * @return 1 on success, 0 on error and errno is set.
 */
bool FiVideoAVIStream::AVIEndRiff()
{
    bool addContextCheck = true;
    bool addImageAttachedCheck = true;
    bool endRiffResult = false;

    ixxx_chunk ix00;                // Contains the ix00 chunk header written at the end of each movi-LIST
    uint32_t frames_in_this_riff;   // Contains a count of the frames in the RIFF
    uint32_t riff_size;             // Contains the total size of the riff
    uint32_t current_riff;          // The number of riff sections completed so far
    uint32_t movi_size;             // Contains the size of the movi
    indx_entry indx_e;              // Storage for an index entry to be written in AVI header.
    idx1_entry idx1_e;              // A legacy idx1 structure to be written for each frame in the first RIFF.
    fcc_head h;                     // Used to write riff headers
    uint64_t offset;
    ixxx_entry  *offset_rec;


    // Make sure the avi stream is ok for writing.
    if (!OkToWrite(__PRETTY_FUNCTION__, addContextCheck, addImageAttachedCheck))
        return endRiffResult;

    // Calculate the frames in this riff.
    frames_in_this_riff = aviContext->aviCtxPtr->current_frame % aviContext->aviCtxPtr->frames_per_riff;/**/

    // If the calculation above produced 0, then we have a full movi-LIST
    if (frames_in_this_riff == 0)
        frames_in_this_riff = aviContext->aviCtxPtr->frames_per_riff;/**/

    // Create an indx entry to point to this ix00 chunk, later we will
    // seek back to the avi header and write this in it's proper place.
    indx_e.offset = curFileOffset;/**/
    indx_e.duration = frames_in_this_riff;
    indx_e.size = sizeof(ix00) + sizeof(ixxx_entry) * frames_in_this_riff;

    // save the current riff for later
    current_riff = aviContext->aviCtxPtr->current_riff;

    // Write the ix00_chunk and entries

    // Fill in the ix00_chunk
    ix00.h.fcc = IX00_TAG;
    ix00.h.size = indx_e.size - sizeof(fcc_head);

    ix00.data.base_offset = aviContext->aviCtxPtr->movi_offset;/**/
    ix00.data.chunk_id = VIDEO_CHUNK_ID;
    ix00.data.entries_in_use = frames_in_this_riff;
    ix00.data.longs_per_entry = 2;
    ix00.data.sub_type = 0;
    ix00.data.type = AVI_INDEX_OF_CHUNKS;

    // Save the ix_00 data for later frame edits
    aviContext->aviCtxPtr->frame_index[current_riff].mem.base_offset = aviContext->aviCtxPtr->movi_offset;
    aviContext->aviCtxPtr->frame_index[current_riff].mem.frames = frames_in_this_riff;

    // Write ixxx header
    if (!WriteData(&ix00, sizeof(ix00)))
    {
        return endRiffResult;
    }

    // Now write ixxx_entries
    offset_rec = aviContext->get_offset_rec(aviContext->aviCtxPtr->frame_index[ current_riff ].mem.offset_rec);
    if (offset_rec == nullptr)
        return endRiffResult;

    if (!WriteData(offset_rec, sizeof(ixxx_entry) * frames_in_this_riff))
    {
        return endRiffResult;
    }

    // Calculate the movi size (to be written later)
    movi_size = (uint32_t)(curFileOffset - aviContext->aviCtxPtr->movi_offset) - sizeof(fcc_head);

    // Write IDX1 after MOVI size calculated, but only on first RIFF.
    if (current_riff == 0)
    {
        // Build up our IDX1 header.
        h.fcc = IDX1_TAG;
        h.size = sizeof(idx1_e) * frames_in_this_riff;

        // Write IDX1 header.
        if (!WriteData(&h, sizeof(h)))
            return endRiffResult;

        // Write one idx1 entry for each frame in the first riff, based on saved
        // ix00 entries.
        for (uint32_t i = 0; i < frames_in_this_riff ; i++)
        {
            idx1_e.chunk_id = VIDEO_CHUNK_ID;
            idx1_e.flags = AVIIF_KEYFRAME;

            // ix00 offset is to frame data, idx1 offset is to chunk header
            // Back up from the start of the frame (found at frame_index[].mem.offsets[i].offset)
            // to the start of the "chunk" containing the frame.
            idx1_e.offset = offset_rec[ i ].offset - sizeof(fcc_head);

            // The offset we want does NOT include the MOVI chunk header, so back that out too.
            idx1_e.offset -= sizeof(fcc_head);

            // Need the length of this frame.
            idx1_e.size = offset_rec[ i ].size;

            // Write IDX1 entry
            if (!WriteData(&idx1_e, sizeof(idx1_e)))
                return endRiffResult;
        }
    }

    // RIFF size includes idx1.
    riff_size = (uint32_t)(curFileOffset - aviContext->aviCtxPtr->riff_offset) - sizeof(fcc_head);

    // Seek to position and update RIFF length.
    offset = aviContext->aviCtxPtr->riff_offset;
    offset += sizeof(uint32_t);

    // Position the output stream to write the data in the proper place.
    if (!SetSeekPos((off64_t)offset, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET))
    {
        return endRiffResult;
    }

    if (!WriteData(&riff_size, sizeof(uint32_t)))
    {
        return endRiffResult;
    }

    // Seek to position and update indx entry for this riff
    offset = aviContext->aviCtxPtr->indx_offset;
    offset += sizeof(indx_chunk);
    offset += (sizeof(indx_entry) * (uint64_t)aviContext->aviCtxPtr->current_riff);

    // Position the output stream to write the data in the proper place.
    if (!SetSeekPos((off64_t)offset, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET))
    {
        return endRiffResult;
    }

    if (!WriteData(&indx_e, sizeof(indx_entry)))
    {
        return endRiffResult;
    }

    // Seek to position and update movi length
    offset = aviContext->aviCtxPtr->movi_offset; /**/
    offset += sizeof(uint32_t);

    if (!SetSeekPos((off64_t)offset, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET))
    {
        return endRiffResult;
    }

    if (!WriteData(&movi_size, sizeof(uint32_t)))
    {
        return endRiffResult;
    }

    if (!SetSeekPos(0, FAS_AVI::AVI_SEEK_POS::AVI_SEEK_END))
    {
        return endRiffResult;
    }

    aviContext->aviCtxPtr->current_riff++;/**/

    // If reach here the end riff was created and seek position set for next one.
    endRiffResult = true;

    // Now we are ready to start a new riff if necessary
    return endRiffResult;
}

/*!
 * @brief Read data from an avi file
 *
 * Read data from the avi file, log any errors, and update current context.
 * Updates the curFileOffset to match the seek position in file.
 *
 * @param buffer: the data buffer
 * @param length: the amount of data to write
 *
 * @return A boolean true if successful, false if error condition.
 */
bool FiVideoAVIStream::ReadData(void *buffer, uint32_t length)
{
    bool readData = false;
    bool addContextCheck = true;
    streampos seekBefore = 0;
    streampos seekLater = 0;
    long long seekDelta = 0;

    if (!OkToRead(__PRETTY_FUNCTION__, addContextCheck))
    {
        goto exitReadData;
    }

    // Make sure that there is an output file I/O stream to write to.
    if (!ioFileStreamPtr)
    {
        fi_log(LOG_DEBUG, "%s: AVI stream missing io stream object.", __PRETTY_FUNCTION__);
        LogStreamStatus();
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
        goto exitReadData;
    }

    // Validate the buffer and length requested for writing.
    if ((buffer == nullptr) || (length == 0))
    {
        fi_log(LOG_ERR, "(%s) check failed, either null buffer or zero length.  Length[%u].", __PRETTY_FUNCTION__, length);
        goto exitReadData;
    }

    // Write the data to the output stream.
    seekBefore = ioFileStreamPtr->tellp();
    ioFileStreamPtr->read((char *)buffer, length);

    // Is the output stream IO object in good state?
    if (!ioFileStreamPtr->good())
    {
        // Is it due to an EOF condition?  If so set appropriately.
        if (ioFileStreamPtr->eof())
        {
#ifdef USE_VERBOSE_INFO_DEBUG
            fi_log(LOG_DEBUG, "%s: io stream eof during read request.", __PRETTY_FUNCTION__);
#endif
            streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_EOF;
        }
        else
        {
            fi_log(LOG_ERR, "%s: io stream read failure.", __PRETTY_FUNCTION__);
            streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_READ_ERROR;
        }
        goto exitReadData;
    }
    seekLater = ioFileStreamPtr->tellp();

    // The number of bytes should be equal to the length of the requested write.
    seekDelta = seekLater - seekBefore;
    if ((uint32_t) seekDelta != length)
    {
        fi_log(LOG_ERR, "%s: io stream read ok but len requested %d != len written %d.", __PRETTY_FUNCTION__, length, (uint32_t) seekDelta);
        LogStreamStatus();
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
        goto exitReadData;
    }

    // Update the file offset to the current seek position.
    curFileOffset = (uint64_t) seekLater;

    // Read was successful.
    readData = true;

exitReadData:
    return readData;
}

/*!
 * @brief Write data to an avi file
 *
 * Write data to the avi file, log any errors, and update current context.
 *
 * NOTE_JLC - does this belong in base stream class?  The reason
 * it is not there is because AVI files are the only streams where
 * we control the writes.  In the FFmpeg stream situations the library
 * will be handling the writing.  I am being inconsistent here because
 * the actual ofStream class in a member of the base class and not here.
 * Come back to this issue when FFmpeg architecture is defined.
 *
 * @param video_context: a video context created from avi_default_context()
 * @param buffer: the data buffer
 * @param length: the amount of data to write
 *
 * @return A boolean true if successful, false if error condition.
 */
bool FiVideoAVIStream::WriteData(const void *buffer, uint32_t length)
{
    bool wroteData = false;
    bool addContextCheck = true;
    bool addImageAttachedCheck = false;     // Don't force image attachment because can receive placeholder frames.
    streampos seekBefore = 0;
    streampos seekLater = 0;
    long long seekDelta = 0;

    if (!OkToWrite(__PRETTY_FUNCTION__, addContextCheck, addImageAttachedCheck))
    {
        goto exitWriteData;
    }

    // Make sure that there is an output file I/O stream to write to.
    if (!ioFileStreamPtr)
    {
        fi_log(LOG_DEBUG, "%s: AVI stream missing io stream object.", __PRETTY_FUNCTION__);
        LogStreamStatus();
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
        goto exitWriteData;
    }

    // Validate the buffer and length requested for writing.
    if ((buffer == nullptr) || (length == 0))
    {
        fi_log(LOG_ERR, "(%s) check failed, either null buffer or zero length.  Length[%u].", __PRETTY_FUNCTION__, length);
        goto exitWriteData;
    }

    // Write the data to the output stream.
    seekBefore = ioFileStreamPtr->tellp();
    ioFileStreamPtr->write((const char *)buffer, length);

    // Is the output stream IO object in good state?
    if (!ioFileStreamPtr->good())
    {
        fi_log(LOG_ERR, "%s: io stream write failed.", __PRETTY_FUNCTION__);
        LogStreamStatus();
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
        goto exitWriteData;
    }
    seekLater = ioFileStreamPtr->tellp();

    // The number of bytes should be equal to the length of the requested write.
    seekDelta = seekLater - seekBefore;
    if ((uint32_t) seekDelta != length)
    {
        fi_log(LOG_ERR, "%s: io stream write ok but len requested %d != len written %d.", __PRETTY_FUNCTION__, length, (uint32_t) seekDelta);
        LogStreamStatus();
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
        goto exitWriteData;
    }

    // Update the file offset to the current seek position.
    curFileOffset = (uint64_t) seekLater;

    // Write was successful.
    wroteData = true;

exitWriteData:
    return wroteData;
}


/*!
 * @brief Get seek position in the io stream object.
 *
 * @return off64_t current seek position or -1 if unable to query.
 */
off64_t FiVideoAVIStream::GetSeekPos()
{
    streampos curSeekPos;
    off64_t retSeekPos = -1;

    // Make sure that there is an output file I/O stream object to query.
    if (!ioFileStreamPtr)
    {
        fi_log(LOG_DEBUG, "%s: no io file stream object exists.", __PRETTY_FUNCTION__);
        LogStreamStatus();
        goto exitGetSeekPos;
    }

    // Get the current seek position in the io file stream.
    curSeekPos = ioFileStreamPtr->tellp();
    retSeekPos = (off64_t)curSeekPos;

exitGetSeekPos:
    return retSeekPos;
}

/*!
 * @brief Set seek position in the io stream object.
 *
 * @param seekPos:      Where in the stream file to set seek position.
 * @param howToSeek:    How to set the AVI seek position.
 *
 * @return A boolean true if successful, false if error condition.
 */
//NOTE_JLC this should become a base-class implementation.
bool FiVideoAVIStream::SetSeekPos(off64_t seekPos, FAS_AVI::AVI_SEEK_POS howToSeek)
{
    bool seekWasSet = false;

    streampos seekBeforeSet = 0;
    streampos seekToSet = seekPos;
    streampos seekAfterSet = 0;

    // Need to check the validity of the avi context data.
    bool addContextCheck = true;

    // It is ok to seek if there is not an attached image.
    bool addImageAttachedCheck = false;

    // Validate the seekPos.
    if (seekPos < 0)
    {
        fi_log(LOG_ERR, "(%s) invalid seek position (%ld) received.", __PRETTY_FUNCTION__, seekPos);
        goto exitSetSeekPos;
    }

    if (streamType == FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_READER)
    {
        if (!OkToRead(__PRETTY_FUNCTION__, addContextCheck))
            goto exitSetSeekPos;

        // Make sure that there is an output file I/O stream object for writer.
        if (!ioFileStreamPtr)
        {
            fi_log(LOG_DEBUG, "%s: AVI stream reader missing fstream object.", __PRETTY_FUNCTION__);
            LogStreamStatus();
            streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
            goto exitSetSeekPos;
        }
    }
    else if (streamType == FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER)
    {
        if (!OkToWrite(__PRETTY_FUNCTION__, addContextCheck, addImageAttachedCheck))
            goto exitSetSeekPos;

        // Make sure that there is an output file I/O stream object for writer.
        if (!ioFileStreamPtr)
        {
            fi_log(LOG_DEBUG, "%s: AVI stream writer missing fstream object.", __PRETTY_FUNCTION__);
            LogStreamStatus();
            streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
            goto exitSetSeekPos;
        }
    }
    else
    {
        // The I/O stream is not setup properly.
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
        goto exitSetSeekPos;
    }

    // If the eof bit was set in the stream ptr attempt to clear, reset to zero
    // and retest the seekBeforeSet.
    if (ioFileStreamPtr->eof())
    {
        ioFileStreamPtr->clear();
        ioFileStreamPtr->seekp(0, ios_base::beg);
    }

    // Set the requested seek position in the output stream.
    seekBeforeSet = ioFileStreamPtr->tellp();

    switch (howToSeek)
    {
    case FAS_AVI::AVI_SEEK_POS::AVI_SEEK_SET:
        ioFileStreamPtr->seekp(seekToSet, ios_base::beg);
        break;
    case FAS_AVI::AVI_SEEK_POS::AVI_SEEK_END:
        ioFileStreamPtr->seekp(seekToSet, ios_base::end);
        break;
    case FAS_AVI::AVI_SEEK_POS::AVI_SEEK_CUR:
        ioFileStreamPtr->seekp(seekToSet, ios_base::cur);
        break;
    default:
        fi_log(LOG_ERR, "(%s) invalid seek request (%d) received.", __PRETTY_FUNCTION__, (int)howToSeek);
        goto exitSetSeekPos;
    }

    // Is the output stream IO object in good state?
    if (!ioFileStreamPtr->good())
    {
        fi_log(LOG_ERR, "%s: fstream failure on seekp() call for seekPos[%" PRId64 "].  Seek before[%" PRId64 "]", __PRETTY_FUNCTION__, seekPos, (off64_t)seekBeforeSet);
        LogStreamStatus();
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
        goto exitSetSeekPos;
    }
    seekAfterSet = ioFileStreamPtr->tellp();

    // Update the file offset to the current seek position.
    curFileOffset = (uint64_t) seekAfterSet;

    // Seek position was successful.
    seekWasSet = true;

exitSetSeekPos:
    return seekWasSet;
}

/*!
 * @brief Determine where in the file to write a lost frame.
 *
 *
 * @param frame_num:        Which frame was lost.
 * @param size:             How large is the new frame for the placeholder sanity check.
 *
 * @return A valid offset or 0 if unable to find one.
 *
 */
off64_t FiVideoAVIStream::AVILookupImageOffset(uint32_t frame_num, size_t *size)
{
    uint32_t i;
    off64_t offset = 0;
    uint64_t tempOffset = 0;
    uint32_t total_frames = 0;
    uint32_t index_frame = 0;
    ixxx_entry  *offset_rec;

    bool offsetFound = false;

    // Need to check the validity of the avi context data.
    bool addContextCheck = true;

    // It is ok to seek if there is not an attached image.
    bool addImageAttachedCheck = false;

    if (streamType == FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_READER)
    {
        if (!OkToRead(__PRETTY_FUNCTION__, addContextCheck))
            goto exitAVILookupImageOffset;
    }
    else if (streamType == FAS_VIDEO::STREAM_IO_TYPE::FI_STREAM_WRITER)
    {
        if (!OkToWrite(__PRETTY_FUNCTION__, addContextCheck, addImageAttachedCheck))
            goto exitAVILookupImageOffset;
    }
    else
    {
        fi_log(LOG_DEBUG, "%s: AVI stream type set to none.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
        goto exitAVILookupImageOffset;
    }

    // Make sure that there is an output file I/O stream to write to.
    if (!ioFileStreamPtr)
    {
        fi_log(LOG_DEBUG, "%s: AVI stream missing ofstream object.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::NO_IO_STREAM;
        goto exitAVILookupImageOffset;
    }

    for (i = 0; i < aviContext->aviCtxPtr->avi_indx->entries_in_use; i++)
    {
        if (total_frames + aviContext->aviCtxPtr->frame_index[i].mem.frames > frame_num)
        {
            break;
        }
        total_frames += aviContext->aviCtxPtr->frame_index[i].mem.frames;
    }
    if (i >= aviContext->aviCtxPtr->avi_indx->entries_in_use)
    {
        fi_log(LOG_DEBUG, "%s: index[%d] exceeded entries in use[%d].", __PRETTY_FUNCTION__, i, aviContext->aviCtxPtr->avi_indx->entries_in_use);
        streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
        goto exitAVILookupImageOffset;
    }

    index_frame = frame_num - total_frames;

    offset_rec = aviContext->get_offset_rec(aviContext->aviCtxPtr->frame_index[i].mem.offset_rec);
    if (offset_rec == nullptr)
    {
        fi_log(LOG_DEBUG, "%s: offset_rec was nullptr.", __PRETTY_FUNCTION__);
        streamStatus = FAS_VIDEO::STREAM_STATUS::CODEC_ERROR;
        goto exitAVILookupImageOffset;
    }

    tempOffset = aviContext->aviCtxPtr->frame_index[i].mem.base_offset;

    // Make sure to cast the right size it was casting to a uint32_t.
    offset = (off64_t)tempOffset;
    offset += offset_rec[index_frame].offset;

    //NOTE_JLC revisit this and should there be an else and error condition?
    if (size != nullptr)
        *size = offset_rec[index_frame].size;

    offsetFound = true;

exitAVILookupImageOffset:

    // If some kind of error condition return zero offset.
    if (!offsetFound)
        offset = 0;

    return offset;
}

