/*!
 * @file fi_tiff.cpp
 * @brief Contains the implementation of the FiTIFF class.
 *
 * TIFF specific terms
 * - IFD: Image File Directory
 * - TIFF: Tag Image File Format
 *
 * For help with the meaning of the TIFF tags, refer to
 * https://www.awaresystems.be/imaging/tiff/tifftags/baseline.html.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <string.h>
#include <stdint.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>

#include <algorithm>

#include "fi_tiff.h"
#include "log_if.h"

namespace FI_TIFF
{
//--------------------------------------------------------------------
// Constants, defines, Typedefs and Macros
//--------------------------------------------------------------------

// Bits per sample.  For mono, 8 bit raw and color, we use 8 bps per sample
// (3 samples for color).  For raw 10 bit, we use 16 bits per sample.
const uint16_t BPS_8         = 8;
const uint16_t BPS_16        = 16;

// If a tag is shorter than MAX_EMBEDDED_TAG_SIZE, then it is stored directly
// in the IFD in place of the offset.
const uint32_t TIF_MAX_EMBEDDED_TAG_SIZE    = 4;

// TIFF magic constants
const uint16_t TIFF_LE_BYTE_ORDER           = 0x4949;
//const uint16_t TIFF_BE_BYTE_ORDER           = 0x4d4d;     // Currently not used by library but left for posterity.

// Defines that the data is in "Chunky" format.
const uint16_t TIFF_PLANAR_CONFIG_VALUE     = 1;

//
// TIFF Tags (https://www.awaresystems.be/imaging/tiff/tifftags.html)
//
#define TIFF_NEW_SUBFILE_TYPE           254
#define TIFF_SUBFILE_TYPE               255
#define TIFF_IMAGE_WIDTH                256
#define TIFF_IMAGE_LENGTH               257
#define TIFF_BITS_PER_SAMPLE            258
#define TIFF_COMPRESSION                259
#define TIFF_PHOTOMETRIC_INTERPRETATION 262
#define TIFF_FILL_ORDER                 266
#define TIFF_IMAGE_DESCRIPTION          270
#define TIFF_MAKE                       271
#define TIFF_MODEL                      272
#define TIFF_STRIP_OFFSETS              273
#define TIFF_ORIENTATION                274
#define TIFF_SAMPLES_PER_PIXEL          277
#define TIFF_ROWS_PER_STRIP             278
#define TIFF_STRIP_BYTE_COUNTS          279
#define TIFF_X_RESOLUTION               282
#define TIFF_Y_RESOLUTION               283
#define TIFF_PLANAR_CONFIG              284
#define TIFF_RESOLUTION_UNIT            296

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Function Definitions
//--------------------------------------------------------------------

TIFF_TAG_SIZE_FUNCTION(TIFF_NEW_SUBFILE_TYPE)
{
    uint32_t ret_size = sizeof(uint32_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief A general indication of the kind of data contained in this subfile.
 *
 * This field is made up of a set of 32 flag bits.
 * - Bit-0: '1' if the image is a reduced-resolution version of another image in
 *          this TIFF file, otherwise '0'.
 * - Bit-1: '1' if the image is a single page of a multi-page image, otherwise '0'.
 * - Bit-2: '1' if the image defines a transparency mask for another image in this TIFF file.
 */
TIFF_TAG_READ_FUNCTION(TIFF_NEW_SUBFILE_TYPE)
{
    uint32_t val = 0;

    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint32_t));
}


TIFF_TAG_SIZE_FUNCTION(TIFF_SUBFILE_TYPE)
{
    uint32_t ret_size = sizeof(uint16_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief A general indication of the kind of data contained in this subfile.
 *
 * The specification defines these values:
 * - 1 = full-resolution image data.
 * - 2 = reduced-resolution image data.
 * - 3 = a single page of a multi-page image.
 */
TIFF_TAG_READ_FUNCTION(TIFF_SUBFILE_TYPE)
{
    uint16_t val = 1;

    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint16_t));
}


TIFF_TAG_SIZE_FUNCTION(TIFF_IMAGE_WIDTH)
{
    uint32_t ret_size = sizeof(uint16_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The number of columns in the images.
 */
TIFF_TAG_READ_FUNCTION(TIFF_IMAGE_WIDTH)
{
    uint16_t val;

    (void)tiff;
    (void)fourcc;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    ifd_ptr->count = 1;
    val = meta->roi_width;
    memcpy(buffer, &val, sizeof(uint16_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_IMAGE_LENGTH)
{
    uint32_t ret_size = sizeof(uint16_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The number of rows of pixels in the image.
 */
TIFF_TAG_READ_FUNCTION(TIFF_IMAGE_LENGTH)
{
    uint16_t val;

    (void)tiff;
    (void)fourcc;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    ifd_ptr->count = 1;
    val = meta->roi_height;
    memcpy(buffer, &val, sizeof(uint16_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_BITS_PER_SAMPLE)
{
    uint32_t ret_size;

    (void)tiff;
    (void)meta;

    if ((fourcc == FAS_VIDEO::FOURCC_ARGB) || (fourcc == FAS_VIDEO::FOURCC_RGB))
        ret_size = 3 * sizeof(uint16_t);
    else
        ret_size = sizeof(uint16_t);

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief Number of bits per component.
 */
TIFF_TAG_READ_FUNCTION(TIFF_BITS_PER_SAMPLE)
{
    uint16_t val;

    (void)tiff;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    if ((fourcc == FAS_VIDEO::FOURCC_ARGB) || (fourcc == FAS_VIDEO::FOURCC_RGB))
    {
        ifd_ptr->count = 3;
        val = BPS_8;

        memcpy(buffer, &val, sizeof(uint16_t));
        memcpy(buffer + sizeof(uint16_t), &val, sizeof(uint16_t));
        memcpy(buffer + (2 * sizeof(uint16_t)), &val, sizeof(uint16_t));
    }
    else
    {
        ifd_ptr->count = 1;
        if ((fourcc == FAS_VIDEO::FOURCC_BYR2) || (fourcc == FAS_VIDEO::FOURCC_Y16))
            val = BPS_16;
        else
            val = BPS_8;

        memcpy(buffer, &val, sizeof(uint16_t));
    }
}

TIFF_TAG_SIZE_FUNCTION(TIFF_COMPRESSION)
{
    uint32_t ret_size = sizeof(uint16_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief Compression scheme used on the image data.
 *
 * The specification defines these values to be baseline:
 * - 1 = no compression.
 * - 2 = CCITT modified Huffman RLE.
 * - 32773 = PackBits compression, aka Macintosh RLE.
 */
TIFF_TAG_READ_FUNCTION(TIFF_COMPRESSION)
{
    uint16_t val = 1;   // No compression.

    (void)tiff;
    (void)fourcc;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint16_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_PHOTOMETRIC_INTERPRETATION)
{
    uint32_t ret_size = sizeof(uint16_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The color space of the image data.
 *
 * The specification considers these values baseline:
 * - 0 = WhiteIsZero. For grayscale images, 0 is imaged as white.
 * - 1 = BlackIsZero. For grayscale images, 0 is imaged as black.
 * - 2 = RGB. RGB value of (0, 0, 0) represents black, and (255, 255, 255)
 *       represents white.
 * - 3 = Palette color.
 */
TIFF_TAG_READ_FUNCTION(TIFF_PHOTOMETRIC_INTERPRETATION)
{
    uint16_t val;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    val = ((FiTIFF *)tiff)->get_photometric_interpretation(fourcc);
    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint16_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_FILL_ORDER)
{
    uint32_t ret_size = sizeof(uint16_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The logical order of bits within a byte.
 *
 * The specification defines these values:
 * - 1 [FILLORDER_MSB2LSB]: pixels with lower column values are stored in
 *     the higher-order bits of the byte.
 * - 2 [FILLORDER_LSB2MSB]: pixels with lower column values are stored in
 *     the lower-order bits of the byte.
 */
TIFF_TAG_READ_FUNCTION(TIFF_FILL_ORDER)
{
    uint16_t val = 1;   // FILLORDER_MSB2LSB

    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint16_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_IMAGE_DESCRIPTION)
{
    uint32_t ret_size;

    (void)tiff;
    (void)fourcc;

    if (meta == nullptr)
    {
        return 0;
    }

    if (meta->comment[0] == 0)
        return 0;

    ret_size = static_cast<uint32_t>(strlen((const char *)meta->comment) + 1);
    return TIFF_ALIGNED_SIZE(ret_size);
}

TIFF_TAG_READ_FUNCTION(TIFF_IMAGE_DESCRIPTION)
{
    (void)tiff;
    (void)fourcc;

    if (ifd_ptr == nullptr)
    {
        return;
    }
    /* set count to zero in case early return for null pointers */
    ifd_ptr->count = 0;

    if (buffer == nullptr)
    {
        return;
    }
    if (meta == nullptr)
    {
        return;
    }
    if (meta->comment[0] == 0)
    {
        ifd_ptr->count = 0;
    }
    else
    {
        //strcpy(buffer, (const char *)meta->comment);/**/
        strncpy(buffer, (const char *)meta->comment, sizeof(meta->comment) - 1);

        ifd_ptr->count = static_cast<uint32_t>(strlen((const char *) meta->comment) + 1);
    }
}

TIFF_TAG_SIZE_FUNCTION(TIFF_MAKE)
{
    uint32_t ret_size;

    (void)tiff;
    (void)fourcc;

    if (meta == nullptr)
        return 0;

    ret_size = static_cast<uint32_t>(strlen(meta->man_name) + 1);
    if (ret_size > GIGE_DSC_ACK_VENDOR_B_LEN)
        ret_size = GIGE_DSC_ACK_VENDOR_B_LEN;
    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The manufacturer.
 */
TIFF_TAG_READ_FUNCTION(TIFF_MAKE)
{
    uint32_t ret_size;

    (void)tiff;
    (void)fourcc;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    ret_size = static_cast<uint32_t>(strlen(meta->man_name) + 1);
    if (ret_size > GIGE_DSC_ACK_VENDOR_B_LEN)
        ret_size = GIGE_DSC_ACK_VENDOR_B_LEN;

    ifd_ptr->count = ret_size;
    strncpy(buffer, meta->man_name, ret_size);
}

TIFF_TAG_SIZE_FUNCTION(TIFF_MODEL)
{
    uint32_t ret_size;

    (void)tiff;
    (void)fourcc;

    if (meta == nullptr)
        return 0;

    ret_size = static_cast<uint32_t>(strlen(meta->model_name) + 1);
    if (ret_size > GIGE_DSC_ACK_MODEL_B_LEN)
        ret_size = GIGE_DSC_ACK_MODEL_B_LEN;
    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The model name or number.
 */
TIFF_TAG_READ_FUNCTION(TIFF_MODEL)
{
    uint32_t ret_size;

    (void)tiff;
    (void)fourcc;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    ret_size = static_cast<uint32_t>(strlen(meta->model_name) + 1);
    if (ret_size > GIGE_DSC_ACK_MODEL_B_LEN)
        ret_size = GIGE_DSC_ACK_MODEL_B_LEN;

    ifd_ptr->count = ret_size;
    strncpy(buffer, meta->model_name, ret_size);
}

TIFF_TAG_SIZE_FUNCTION(TIFF_STRIP_OFFSETS)
{
    uint32_t ret_size = sizeof(uint32_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief For each strip, the byte offset of that strip.
 *
 * We only have one strip, so the byte offset will be where the
 * image data starts within the file. I.e., immediately following
 * the header.
 */
TIFF_TAG_READ_FUNCTION(TIFF_STRIP_OFFSETS)
{
    uint32_t offset;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    offset = ((FiTIFF *)tiff)->get_header_size(fourcc, meta);
    ifd_ptr->count = 1;
    memcpy(buffer, &offset, sizeof(uint32_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_ORIENTATION)
{
    uint32_t ret_size = sizeof(uint16_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The orientation of the image with respect to
 *        the rows and columns.
 *
 * - 1 = The 0th row represents the visual top of the image, and the 0th column represents the visual left-hand side.
 * - 2 = The 0th row represents the visual top of the image, and the 0th column represents the visual right-hand side.
 * - 3 = The 0th row represents the visual bottom of the image, and the 0th column represents the visual right-hand side.
 * - 4 = The 0th row represents the visual bottom of the image, and the 0th column represents the visual left-hand side.
 * - 5 = The 0th row represents the visual left-hand side of the image, and the 0th column represents the visual top.
 * - 6 = The 0th row represents the visual right-hand side of the image, and the 0th column represents the visual top.
 * - 7 = The 0th row represents the visual right-hand side of the image, and the 0th column represents the visual bottom.
 * - 8 = The 0th row represents the visual left-hand side of the image, and the 0th column represents the visual bottom.
 *
 * Default is 1, which is convenient because that's what we want.
 */
TIFF_TAG_READ_FUNCTION(TIFF_ORIENTATION)
{
    uint16_t val = 1;   // ORIENTATION_TOPLEFT


    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint16_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_SAMPLES_PER_PIXEL)
{
    uint32_t ret_size = sizeof(uint16_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The number of components per pixel.
 *
 * SamplesPerPixel is usually 1 for grayscale and palette-color images. It
 * is usually 3 for RGB images.
 */
TIFF_TAG_READ_FUNCTION(TIFF_SAMPLES_PER_PIXEL)
{
    uint16_t val;

    (void)tiff;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    if ((fourcc == FAS_VIDEO::FOURCC_ARGB) || (fourcc == FAS_VIDEO::FOURCC_RGB))
        val = 3;  // Indicates 3 samples per pixel R,G,B
    else
        val = 1;  // Gray scale is 1 sample per pixel

    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint16_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_ROWS_PER_STRIP)
{
    uint32_t ret_size = sizeof(uint32_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The number of rows per strip.
 *
 * We only have one strip, so this is the number of rows (aka "height") in the image.
 */
TIFF_TAG_READ_FUNCTION(TIFF_ROWS_PER_STRIP)
{
    uint32_t val;

    (void)tiff;
    (void)fourcc;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    val = meta->roi_height;
    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint32_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_STRIP_BYTE_COUNTS)
{
    uint32_t ret_size = sizeof(uint32_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief For each strip, the number of bytes in the strip. For us,
 *        this is the size of the image.
 */
TIFF_TAG_READ_FUNCTION(TIFF_STRIP_BYTE_COUNTS)
{
    uint32_t val;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;
    if (meta == nullptr)
        return;

    val = ((FiTIFF *)tiff)->get_image_size(fourcc, meta);
    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint32_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_X_RESOLUTION)
{
    uint32_t ret_size = 2 * sizeof(uint32_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The number of pixels per "resolution unit" in the width direction.
 *
 * This is a rational number (numerator/denominator), and we've picked 100 pixels
 * per "unit" (maybe inch? centimeter? whatever).
 *
 * @sa TIFF_RESOLUTION_UNIT
 */
TIFF_TAG_READ_FUNCTION(TIFF_X_RESOLUTION)
{
    uint32_t numerator = 100, denominator = 1;

    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    ifd_ptr->count = 1;
    memcpy(buffer, &numerator, sizeof(uint32_t));
    memcpy(buffer + sizeof(uint32_t), &denominator, sizeof(uint32_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_Y_RESOLUTION)
{
    uint32_t ret_size = 2 * sizeof(uint32_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The number of pixels per "resolution unit" in the height direction.
 *
 * This is a rational number (numerator/denominator), and we've picked 100 pixels
 * per "unit" (maybe inch? centimeter? whatever).
 *
 * @sa TIFF_RESOLUTION_UNIT
 */
TIFF_TAG_READ_FUNCTION(TIFF_Y_RESOLUTION)
{
    uint32_t numerator = 100, denominator = 1;

    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    ifd_ptr->count = 1;
    memcpy(buffer, &numerator, sizeof(uint32_t));
    memcpy(buffer + sizeof(uint32_t), &denominator, sizeof(uint32_t));
}

TIFF_TAG_SIZE_FUNCTION(TIFF_PLANAR_CONFIG)
{
    uint32_t ret_size = sizeof(uint16_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief Specifies how the components of each pixel are stored.
 *
 * The specification defines these values:
 * - 1: Chunky format. The component values for each pixel are stored contiguously.
 *      For example, for RGB data, the data is stored as RGBRGBRGB.
 * - 2: Planar format. The components are stored in separate component planes.
 *
 * The Fastec Video library uses the "chunky" format.
 */
TIFF_TAG_READ_FUNCTION(TIFF_PLANAR_CONFIG)
{
    uint16_t val = TIFF_PLANAR_CONFIG_VALUE;

    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint16_t));
}


TIFF_TAG_SIZE_FUNCTION(TIFF_RESOLUTION_UNIT)
{
    uint32_t ret_size = sizeof(uint16_t);

    (void)tiff;
    (void)fourcc;
    (void)meta;

    return TIFF_ALIGNED_SIZE(ret_size);
}

/*!
 * @brief The unit of measurement for XResolution and YResolution.
 *
 * The specification defines these values:
 * - 1 = No absolute unit of measurement.
 * - 2 = Inch.
 * - 3 = Centimeter.
 */
TIFF_TAG_READ_FUNCTION(TIFF_RESOLUTION_UNIT)
{
    uint16_t val = 2;   // We'll use inches.

    (void)tiff;
    (void)fourcc;
    (void)meta;

    if (ifd_ptr == nullptr)
        return;
    if (buffer == nullptr)
        return;

    ifd_ptr->count = 1;
    memcpy(buffer, &val, sizeof(uint16_t));
}

} // End of FI_TIFF

// Make things easier for ourselves.
using namespace FI_TIFF;
using namespace FAS_VIDEO;

tagEntry::tagEntry(void *parent, uint16_t tag, uint16_t type, TAG_SIZE_FN size_fn, TAG_READ_FN read_fn)
{
    tiff = (FiTIFF *)parent;
    tag_id = tag;
    tag_type = type;
    tag_size = size_fn;
    tag_read = read_fn;
}

uint64_t tagEntry::size(uint32_t fourcc, const pm_metadata *meta)
{
    uint64_t rc(0);

    // Sanity check.
    if (tag_size != nullptr)
        rc = (*tag_size)(tiff, fourcc, meta);
    return rc;
}

void tagEntry::read(uint32_t fourcc, const pm_metadata *meta, IFD_ENTRY *ifd_ptr, char *buffer)
{
    // Sanity check.
    if (tag_read != nullptr)
        tag_read(tiff, fourcc, meta, ifd_ptr, buffer);
}

/*!
 * @brief Standard class constructor.
 */
FiTIFF::FiTIFF() : FiImage()
{
    // Make sure we initialize our members.
    header.clear();

    // Initialize our database of tags we're interested in.
    FiTIFF::init_tag_db();
}

uint16_t FiTIFF::get_photometric_interpretation(uint32_t fourcc)
{
    uint16_t val;

    if ((fourcc == FAS_VIDEO::FOURCC_ARGB) || (fourcc == FAS_VIDEO::FOURCC_RGB))
        val = 2;    // Indicates RGB
    else
        val = 1;    // Indicates black is 0, set to 0 if black is 255.

    return val;
}

/*!
 * @brief Defines all the tags we will add to a TIFF file.
 * @note The TIFF spec requires that the tags be in the IFD in numerical order.
 */
void FiTIFF::init_tag_db()
{
    // Make sure we're starting with a clean slate.
    tag_db.clear();

    // Add all the tags we support.
    tag_db.push_back(TAG_DB_ENTRY(TIFF_NEW_SUBFILE_TYPE, TIFF_TYPE_LONG));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_SUBFILE_TYPE, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_IMAGE_WIDTH, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_IMAGE_LENGTH, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_BITS_PER_SAMPLE, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_COMPRESSION, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_PHOTOMETRIC_INTERPRETATION, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_FILL_ORDER, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_IMAGE_DESCRIPTION, TIFF_TYPE_ASCII));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_MAKE, TIFF_TYPE_ASCII));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_MODEL, TIFF_TYPE_ASCII));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_STRIP_OFFSETS, TIFF_TYPE_LONG));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_ORIENTATION, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_SAMPLES_PER_PIXEL, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_ROWS_PER_STRIP, TIFF_TYPE_LONG));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_STRIP_BYTE_COUNTS, TIFF_TYPE_LONG));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_X_RESOLUTION, TIFF_TYPE_RATIONAL));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_Y_RESOLUTION, TIFF_TYPE_RATIONAL));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_PLANAR_CONFIG, TIFF_TYPE_SHORT));
    tag_db.push_back(TAG_DB_ENTRY(TIFF_RESOLUTION_UNIT, TIFF_TYPE_SHORT));

    // Make sure the database is in ascending order, per TIFF spec.
    std::sort(tag_db.begin(), tag_db.end());
}

/*!
 * @brief Calculate the size of the IFD data.
 *
 * @param fourcc    One of the supported FAS_VIDEO FourCC codes.
 * @param meta      The metadata associated with the image.
 *
 * @return Total size of the TAG data (outside of the IFD space).
 */
uint32_t FiTIFF::get_ifd_data_size(uint32_t fourcc, const pm_metadata *meta)
{
    uint32_t i;                 // Loop iteration variable.
    uint32_t total_size = 0;    // Total size calculated.
    uint32_t current_size;      // Size of current tag entry.

    // Sanity check.
    if (tag_db.empty())
    {
        errno = EINVAL;
        return 0;
    }

    // Iterate the tag array.
    for (i = 0; i < tag_db.size(); i++)
    {
        // Get the size of the current tag data.
        current_size = static_cast<uint32_t>(tag_db[i].size(fourcc, meta));

        // Only add to the size if the tag size will not fit in IFD, since TIFF
        // states that data elements smaller than this value should be stored
        // in the IFD in place of the offset field.
        if (current_size > TIF_MAX_EMBEDDED_TAG_SIZE)
            total_size += current_size;
    }

    // Return the computed total size.
    return total_size;
}

/*!
 * @brief Calculate the TIFF header total size.
 *
 * @param fourcc    One of the supported FAS_VIDEO FourCC codes.
 * @param meta      The metadata associated with the image.
 *
 * @return Size of the TIFF header.
 */
uint32_t FiTIFF::get_header_size(uint32_t fourcc, const pm_metadata *meta)
{
    uint32_t header_size;

    // Header starts with TIFF_HEADER.
    header_size = TIFF_ALIGNED_SIZE(sizeof(TIFF_HEADER));

    // Add and align the Image File Directory (IFD) data.
    header_size += get_ifd_data_size(fourcc, meta);
    header_size = TIFF_ALIGNED_SIZE(header_size);

    // Add the size of the IFD (which points to the IFD data).
    // - First 2 bytes contains the number of tags to follow.
    // - The next chunk is the IFD (which may or may not contain the tag data).
    // - A "null" offset to the next IFD (there may be multiple, but not in ours).
    header_size += sizeof(uint16_t) + (tag_db.size() * sizeof(IFD_ENTRY)) + sizeof(uint32_t);
    header_size = TIFF_ALIGNED_SIZE(header_size);

    return header_size;
}

/*!
 * @brief Calculate the TIFF image size.
 *
 * @param fourcc    One of the supported FAS_VIDEO FourCC codes.
 * @param meta      The metadata associated with the image.
 *
 * @return Size of the TIFF image.
 */
uint32_t FiTIFF::get_image_size(uint32_t fourcc, const pm_metadata *meta)
{
    uint32_t image_size;        // The image size to return.

    // Start out with the width of the image.
    image_size = meta->roi_width;
    if ((fourcc == FOURCC_ARGB) || (fourcc == FOURCC_RGB))
    {
        // For RGB images, we have 3 bytes per pixel.
        image_size *= 3;
    }
    else if ((fourcc == FOURCC_BYR2) || (fourcc == FOURCC_Y16))
    {
        // These formats use 2 bytes per pixel.
        image_size *= 2;
    }

    // NOTE, no row alignment with TIFF!
    image_size *= meta->roi_height;

    // All done.
    return image_size;
}

/*!
 * @brief Add tags to the TIFF image buffer.
 *
 * @param buffer: Pointer to the buffer structure.
 * @param ifd_offset: offset of the IFD in the buffer.
 * @param data_offset: position of the IFD data in the buffer/file
 * @param tag_info: definitions of the tags to add.
 *
 * @return True on success, false on error.
 */
bool FiTIFF::add_tags(uint32_t fourcc, const pm_metadata *meta, uint32_t ifd_offset, uint32_t data_offset)
{
    uint16_t current_tag_count;         // The current tag count in the IFD.
    uint32_t file_offset;               // The file position.
    uint16_t used_count = 0;            // Count of tags USED from tag_info array.
    uint8_t *buffer_position;           // Position in buffer for current tag data.
    uint16_t i;                         // Loop index.
    IFD_ENTRY *ifd_entry;               // Current ifd entry.

    // Sanity checking
    if (tag_db.empty())
    {
        fi_log(LOG_ERR, "%s: empty tag database", __PRETTY_FUNCTION__);
        errno = EINVAL;
        return false;
    }

    // In this implementation, the file_offset IS the data_offset.
    file_offset = data_offset;

    // Find the position of the tag data in the buffer.
    buffer_position = header.data() + data_offset;

    // Get the current count of tags.
    memcpy(&current_tag_count, header.data() + ifd_offset, sizeof(uint16_t));

    // Get pointer to first IFD entry.
    ifd_entry = (IFD_ENTRY *)(header.data() + ifd_offset + sizeof(uint16_t));

    // Jump pointer to what will be the next IFD entry.
    ifd_entry = ifd_entry + current_tag_count;

    // Iterate the tag array
    used_count = 0;
    for (i = 0; i < tag_db.size(); i++)
    {
        uint32_t temp_bytes;    // Temporary size of current tag data.

        // Get the tag ID and type for this entry.
        ifd_entry->tag = tag_db[i].get_id();
        ifd_entry->type = tag_db[i].get_type();

        // Get the size of the tag data.
        temp_bytes = static_cast<uint32_t>(tag_db[i].size(fourcc, meta));

        // Tags not used for this format will return a length of 0.
        if (temp_bytes == 0)
            continue;

        // Increment the number of tags we've used.
        used_count++;

        // Will the tag data fit in the IFD entry offset field?
        if (temp_bytes <= TIF_MAX_EMBEDDED_TAG_SIZE)
        {
            // Copy tag data to image buffer. It is put in the ifd_entry offset field.
            tag_db[i].read(fourcc, meta, ifd_entry, (char *) & (ifd_entry->val_offset));
        }
        else
        {
            // Copy tag data to the image buffer.
            ifd_entry->val_offset = file_offset;
            tag_db[i].read(fourcc, meta, ifd_entry, (char *)buffer_position);

            // Update our positions and offsets.
            buffer_position += temp_bytes;
            file_offset += temp_bytes;
        }

        // Bump the IFD entry so we write the next tag to the right place.
        ifd_entry++;
    }

    // New count is sum of old + new.
    current_tag_count += used_count;

    // Record the new count of IFD entries in the IFD.
    memcpy(header.data() + ifd_offset, &current_tag_count, sizeof(uint16_t));

    // Put a 0 at the end to indicate that there is no "next" IFD.
    memset(ifd_entry, 0, sizeof(uint32_t));

    return true;
}

/*!
 * @brief Create a TIFF header based on the specified parameters.
 *
 * @param fourcc    One of the supported FAS_VIDEO @ref fv_fourcc "FourCC" codes.
 * @param meta      Pointer to the image's metadata.
 *
 * @returns The number of bytes the header uses, or a negative number
 *          indicating an error.
 */
int FiTIFF::CreateHeader(uint32_t fourcc, const pm_metadata *meta)
{
    uint32_t ifd_data_size;             // Size of the IFD data (outside IFD).
    uint32_t current_file_offset;       // Offset in file of current tag data.
    TIFF_HEADER *hd_ptr;                // TIFF header.
    int header_size;                    // Calculated size of our total header.

    // Save the data in base class properties so it persists.
    headerFourCC = fourcc;
    memcpy(&headerMetaData, meta, sizeof(pm_metadata));

    // Figure out what we will need in terms of buffer space.
    header_size = static_cast<int>(get_header_size(fourcc, meta));

    // Create our array.
    header.resize(static_cast<uint64_t>(header_size), 0);

    // Calculate IFD data size for positioning of IFD.
    ifd_data_size = get_ifd_data_size(fourcc, meta);

    // Prepare the TIFF header
    hd_ptr = (TIFF_HEADER *)header.data();
    hd_ptr->byte_order = TIFF_LE_BYTE_ORDER;
    hd_ptr->tif_magic = TIFF_MAGIC;

    // Current file offset keeps track of the position in the TIFF file for tag data.
    current_file_offset = sizeof(TIFF_HEADER);

    // Align tag data
    current_file_offset = TIFF_ALIGNED_SIZE(current_file_offset);

    // Each ifd data tag should report it's size as a multiple of
    // TIFF_IFD_DATA_ALIGNMENT.  Therefore, this should already be aligned.
    hd_ptr->ifd_offset = current_file_offset + ifd_data_size;

    // Add the tags to the image buffer.
    if (!add_tags(fourcc, meta, hd_ptr->ifd_offset, current_file_offset))
    {
        errno = EIO;
        return -1;
    }

    // Return the size of the header.
    return header_size;
}

/*!
 * @brief Get a unique pointer to an array of bytes containing
 *    the TIFF header.
 *
 * The returned data is ready to be prepended onto an image to
 * create a TIFF file.
 *
 * @returns A std::unique_ptr<char> containing the TIFF header.
 */
std::unique_ptr<FAS_VIDEO::HEADER_DATA> FiTIFF::GetHeader()
{
    uint8_t *current_position;      // Write position in buffer.
    std::unique_ptr<FAS_VIDEO::HEADER_DATA> hdr_ptr = std::unique_ptr<FAS_VIDEO::HEADER_DATA>(new FAS_VIDEO::HEADER_DATA(header.size(), 0));

    // Get our pointer.
    current_position = hdr_ptr.get()->data();

    // Copy the header into the buffer. We could use the "swap" method, but I want
    // to keep the "header" around...just in case.
    memcpy(current_position, ((const char *)header.data()), header.size());

    // Return the pointer.
    return hdr_ptr;
}

/*!
 * @brief Query the image format supported by this object.
 *
 * @returns The IMAGE_FORMAT value.
 */
FAS_VIDEO::IMAGE_FORMAT FiTIFF::GetImageFormat()
{
    // We are TIFFs.
    return FAS_VIDEO::IMAGE_FORMAT::FI_TIF_IMAGE;
}

/*!
 * @brief Test the data in a TIFF to verify it and pull out dimensions.
 *
 * @param imageDataToVerify     The Image data to verify TIFF format.
 * @param imageDataSize         How many uint8_t bytes are in this TIFF image.
 *
 * @returns true if TIFF data was verified and attach dimension data interpreted.
 */
bool FiTIFF::VerifyImageData(const uint8_t *imageDataToVerify, uint32_t imageDataSize)
{
    (void)imageDataToVerify;
    (void)imageDataSize;

    // There is no implementation for TIFF images at this time.
    return true;
}

